//
// Copyright(C) 2005-2014 Simon Howard
// Copyright(C) 2014-2019 Fabian Greffrath
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
//
// Parses [PARS] sections in BEX files
//

#ifndef DEH_BEXPARS_H
#define DEH_BEXPARS_H

extern int bex_pars[7][10];
extern int bex_cpars[32];

#endif /* #ifndef DEH_BEXPARS_H */
