//
// Copyright(C) 1993-1996 Id Software, Inc.
// Copyright(C) 1993-2008 Raven Software
// Copyright(C) 2005-2014 Simon Howard
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// generated by stateco

typedef enum
{
    SPR_MAN1,
    SPR_ACLO,
    SPR_TLGL,
    SPR_FBL1,
    SPR_XPL1,
    SPR_ARRW,
    SPR_DART,
    SPR_RIPP,
    SPR_CFCF,
    SPR_BLAD,
    SPR_SHRD,
    SPR_FFSM,
    SPR_FFLG,
    SPR_PTN1,
    SPR_PTN2,
    SPR_SOAR,
    SPR_INVU,
    SPR_SUMN,
    SPR_TSPK,
    SPR_TELO,
    SPR_TRNG,
    SPR_ROCK,
    SPR_FOGS,
    SPR_FOGM,
    SPR_FOGL,
    SPR_SGSA,
    SPR_SGSB,
    SPR_PORK,
    SPR_EGGM,
    SPR_FHFX,
    SPR_SPHL,
    SPR_STWN,
    SPR_GMPD,
    SPR_ASKU,
    SPR_ABGM,
    SPR_AGMR,
    SPR_AGMG,
    SPR_AGG2,
    SPR_AGMB,
    SPR_AGB2,
    SPR_ABK1,
    SPR_ABK2,
    SPR_ASK2,
    SPR_AFWP,
    SPR_ACWP,
    SPR_AMWP,
    SPR_AGER,
    SPR_AGR2,
    SPR_AGR3,
    SPR_AGR4,
    SPR_TRCH,
    SPR_PSBG,
    SPR_ATLP,
    SPR_THRW,
    SPR_SPED,
    SPR_BMAN,
    SPR_BRAC,
    SPR_BLST,
    SPR_HRAD,
    SPR_SPSH,
    SPR_LVAS,
    SPR_SLDG,
    SPR_STTW,
    SPR_RCK1,
    SPR_RCK2,
    SPR_RCK3,
    SPR_RCK4,
    SPR_CDLR,
    SPR_TRE1,
    SPR_TRDT,
    SPR_TRE2,
    SPR_TRE3,
    SPR_STM1,
    SPR_STM2,
    SPR_STM3,
    SPR_STM4,
    SPR_MSH1,
    SPR_MSH2,
    SPR_MSH3,
    SPR_MSH4,
    SPR_MSH5,
    SPR_MSH6,
    SPR_MSH7,
    SPR_MSH8,
    SPR_SGMP,
    SPR_SGM1,
    SPR_SGM2,
    SPR_SGM3,
    SPR_SLC1,
    SPR_SLC2,
    SPR_SLC3,
    SPR_MSS1,
    SPR_MSS2,
    SPR_SWMV,
    SPR_CPS1,
    SPR_CPS2,
    SPR_TMS1,
    SPR_TMS2,
    SPR_TMS3,
    SPR_TMS4,
    SPR_TMS5,
    SPR_TMS6,
    SPR_TMS7,
    SPR_CPS3,
    SPR_STT2,
    SPR_STT3,
    SPR_STT4,
    SPR_STT5,
    SPR_GAR1,
    SPR_GAR2,
    SPR_GAR3,
    SPR_GAR4,
    SPR_GAR5,
    SPR_GAR6,
    SPR_GAR7,
    SPR_GAR8,
    SPR_GAR9,
    SPR_BNR1,
    SPR_TRE4,
    SPR_TRE5,
    SPR_TRE6,
    SPR_TRE7,
    SPR_LOGG,
    SPR_ICT1,
    SPR_ICT2,
    SPR_ICT3,
    SPR_ICT4,
    SPR_ICM1,
    SPR_ICM2,
    SPR_ICM3,
    SPR_ICM4,
    SPR_RKBL,
    SPR_RKBS,
    SPR_RKBK,
    SPR_RBL1,
    SPR_RBL2,
    SPR_RBL3,
    SPR_VASE,
    SPR_POT1,
    SPR_POT2,
    SPR_POT3,
    SPR_PBIT,
    SPR_CPS4,
    SPR_CPS5,
    SPR_CPS6,
    SPR_CPB1,
    SPR_CPB2,
    SPR_CPB3,
    SPR_CPB4,
    SPR_BDRP,
    SPR_BDSH,
    SPR_BDPL,
    SPR_CNDL,
    SPR_LEF1,
    SPR_LEF3,
    SPR_LEF2,
    SPR_TWTR,
    SPR_WLTR,
    SPR_BARL,
    SPR_SHB1,
    SPR_SHB2,
    SPR_BCKT,
    SPR_SHRM,
    SPR_FBUL,
    SPR_FSKL,
    SPR_BRTR,
    SPR_SUIT,
    SPR_BBLL,
    SPR_CAND,
    SPR_IRON,
    SPR_XMAS,
    SPR_CDRN,
    SPR_CHNS,
    SPR_TST1,
    SPR_TST2,
    SPR_TST3,
    SPR_TST4,
    SPR_TST5,
    SPR_TST6,
    SPR_TST7,
    SPR_TST8,
    SPR_TST9,
    SPR_TST0,
    SPR_TELE,
    SPR_TSMK,
    SPR_FPCH,
    SPR_WFAX,
    SPR_FAXE,
    SPR_WFHM,
    SPR_FHMR,
    SPR_FSRD,
    SPR_FSFX,
    SPR_CMCE,
    SPR_WCSS,
    SPR_CSSF,
    SPR_WCFM,
    SPR_CFLM,
    SPR_CFFX,
    SPR_CHLY,
    SPR_SPIR,
    SPR_MWND,
    SPR_WMLG,
    SPR_MLNG,
    SPR_MLFX,
    SPR_MLF2,
    SPR_MSTF,
    SPR_MSP1,
    SPR_MSP2,
    SPR_WFR1,
    SPR_WFR2,
    SPR_WFR3,
    SPR_WCH1,
    SPR_WCH2,
    SPR_WCH3,
    SPR_WMS1,
    SPR_WMS2,
    SPR_WMS3,
    SPR_WPIG,
    SPR_WMCS,
    SPR_CONE,
    SPR_SHEX,
    SPR_BLOD,
    SPR_GIBS,
    SPR_PLAY,
    SPR_FDTH,
    SPR_BSKL,
    SPR_ICEC,
    SPR_CLER,
    SPR_MAGE,
    SPR_PIGY,
    SPR_CENT,
    SPR_CTXD,
    SPR_CTFX,
    SPR_CTDP,
    SPR_DEMN,
    SPR_DEMA,
    SPR_DEMB,
    SPR_DEMC,
    SPR_DEMD,
    SPR_DEME,
    SPR_DMFX,
    SPR_DEM2,
    SPR_DMBA,
    SPR_DMBB,
    SPR_DMBC,
    SPR_DMBD,
    SPR_DMBE,
    SPR_D2FX,
    SPR_WRTH,
    SPR_WRT2,
    SPR_WRBL,
    SPR_MNTR,
    SPR_FX12,
    SPR_FX13,
    SPR_MNSM,
    SPR_SSPT,
    SPR_SSDV,
    SPR_SSXD,
    SPR_SSFX,
    SPR_BISH,
    SPR_BPFX,
    SPR_DRAG,
    SPR_DRFX,
    SPR_ARM1,
    SPR_ARM2,
    SPR_ARM3,
    SPR_ARM4,
    SPR_MAN2,
    SPR_MAN3,
    SPR_KEY1,
    SPR_KEY2,
    SPR_KEY3,
    SPR_KEY4,
    SPR_KEY5,
    SPR_KEY6,
    SPR_KEY7,
    SPR_KEY8,
    SPR_KEY9,
    SPR_KEYA,
    SPR_KEYB,
    SPR_ETTN,
    SPR_ETTB,
    SPR_FDMN,
    SPR_FDMB,
    SPR_ICEY,
    SPR_ICPR,
    SPR_ICWS,
    SPR_SORC,
    SPR_SBMP,
    SPR_SBS4,
    SPR_SBMB,
    SPR_SBS3,
    SPR_SBMG,
    SPR_SBS1,
    SPR_SBS2,
    SPR_SBFX,
    SPR_RADE,
    SPR_WATR,
    SPR_KORX,
    SPR_ABAT,
    NUMSPRITES
} spritenum_t;

typedef enum
{
    S_NULL,
    S_FREETARGMOBJ,
    S_MAPSPOT,
    S_FIREBALL1_1,
    S_FIREBALL1_2,
    S_FIREBALL1_X1,
    S_FIREBALL1_X2,
    S_FIREBALL1_X3,
    S_FIREBALL1_X4,
    S_FIREBALL1_X5,
    S_FIREBALL1_X6,
    S_ARROW_1,
    S_ARROW_X1,
    S_DART_1,
    S_DART_X1,
    S_POISONDART_1,
    S_POISONDART_X1,
    S_RIPPERBALL_1,
    S_RIPPERBALL_2,
    S_RIPPERBALL_3,
    S_RIPPERBALL_X1,
    S_RIPPERBALL_X2,
    S_RIPPERBALL_X3,
    S_RIPPERBALL_X4,
    S_RIPPERBALL_X5,
    S_RIPPERBALL_X6,
    S_RIPPERBALL_X7,
    S_RIPPERBALL_X8,
    S_RIPPERBALL_X9,
    S_RIPPERBALL_X10,
    S_PRJ_BLADE1,
    S_PRJ_BLADE_X1,
    S_ICESHARD1,
    S_ICESHARD2,
    S_ICESHARD3,
    S_FLAME_TSMALL1,
    S_FLAME_TSMALL2,
    S_FLAME_TSMALL3,
    S_FLAME_TSMALL4,
    S_FLAME_TSMALL5,
    S_FLAME_TSMALL6,
    S_FLAME_TLARGE1,
    S_FLAME_TLARGE2,
    S_FLAME_TLARGE3,
    S_FLAME_TLARGE4,
    S_FLAME_TLARGE5,
    S_FLAME_TLARGE6,
    S_FLAME_TLARGE7,
    S_FLAME_TLARGE8,
    S_FLAME_TLARGE9,
    S_FLAME_TLARGE10,
    S_FLAME_TLARGE11,
    S_FLAME_TLARGE12,
    S_FLAME_TLARGE13,
    S_FLAME_TLARGE14,
    S_FLAME_TLARGE15,
    S_FLAME_TLARGE16,
    S_FLAME_SDORM1,
    S_FLAME_SDORM2,
    S_FLAME_SDORM3,
    S_FLAME_SMALL1,
    S_FLAME_SMALL2,
    S_FLAME_SMALL3,
    S_FLAME_SMALL4,
    S_FLAME_SMALL5,
    S_FLAME_SMALL6,
    S_FLAME_SMALL7,
    S_FLAME_LDORM1,
    S_FLAME_LDORM2,
    S_FLAME_LDORM3,
    S_FLAME_LDORM4,
    S_FLAME_LDORM5,
    S_FLAME_LARGE1,
    S_FLAME_LARGE2,
    S_FLAME_LARGE3,
    S_FLAME_LARGE4,
    S_FLAME_LARGE5,
    S_FLAME_LARGE6,
    S_FLAME_LARGE7,
    S_FLAME_LARGE8,
    S_FLAME_LARGE9,
    S_FLAME_LARGE10,
    S_FLAME_LARGE11,
    S_FLAME_LARGE12,
    S_FLAME_LARGE13,
    S_FLAME_LARGE14,
    S_FLAME_LARGE15,
    S_FLAME_LARGE16,
    S_FLAME_LARGE17,
    S_FLAME_LARGE18,
    S_ITEM_PTN1_1,
    S_ITEM_PTN1_2,
    S_ITEM_PTN1_3,
    S_HIDESPECIAL1,
    S_HIDESPECIAL2,
    S_HIDESPECIAL3,
    S_HIDESPECIAL4,
    S_HIDESPECIAL5,
    S_HIDESPECIAL6,
    S_HIDESPECIAL7,
    S_HIDESPECIAL8,
    S_HIDESPECIAL9,
    S_HIDESPECIAL10,
    S_HIDESPECIAL11,
    S_DORMANTARTI1_1,
    S_DORMANTARTI1_2,
    S_DORMANTARTI1_3,
    S_DORMANTARTI1_4,
    S_DORMANTARTI1_5,
    S_DORMANTARTI1_6,
    S_DORMANTARTI1_7,
    S_DORMANTARTI1_8,
    S_DORMANTARTI1_9,
    S_DORMANTARTI1_10,
    S_DORMANTARTI1_11,
    S_DORMANTARTI1_12,
    S_DORMANTARTI1_13,
    S_DORMANTARTI1_14,
    S_DORMANTARTI1_15,
    S_DORMANTARTI1_16,
    S_DORMANTARTI1_17,
    S_DORMANTARTI1_18,
    S_DORMANTARTI1_19,
    S_DORMANTARTI1_20,
    S_DORMANTARTI1_21,
    S_DORMANTARTI2_1,
    S_DORMANTARTI2_2,
    S_DORMANTARTI2_3,
    S_DORMANTARTI2_4,
    S_DORMANTARTI2_5,
    S_DORMANTARTI2_6,
    S_DORMANTARTI2_7,
    S_DORMANTARTI2_8,
    S_DORMANTARTI2_9,
    S_DORMANTARTI2_10,
    S_DORMANTARTI2_11,
    S_DORMANTARTI2_12,
    S_DORMANTARTI2_13,
    S_DORMANTARTI2_14,
    S_DORMANTARTI2_15,
    S_DORMANTARTI2_16,
    S_DORMANTARTI2_17,
    S_DORMANTARTI2_18,
    S_DORMANTARTI2_19,
    S_DORMANTARTI2_20,
    S_DORMANTARTI2_21,
    S_DORMANTARTI3_1,
    S_DORMANTARTI3_2,
    S_DORMANTARTI3_3,
    S_DORMANTARTI3_4,
    S_DORMANTARTI3_5,
    S_DORMANTARTI3_6,
    S_DORMANTARTI3_7,
    S_DORMANTARTI3_8,
    S_DORMANTARTI3_9,
    S_DORMANTARTI3_10,
    S_DORMANTARTI3_11,
    S_DORMANTARTI3_12,
    S_DORMANTARTI3_13,
    S_DORMANTARTI3_14,
    S_DORMANTARTI3_15,
    S_DORMANTARTI3_16,
    S_DORMANTARTI3_17,
    S_DORMANTARTI3_18,
    S_DORMANTARTI3_19,
    S_DORMANTARTI3_20,
    S_DORMANTARTI3_21,
    S_DEADARTI1,
    S_DEADARTI2,
    S_DEADARTI3,
    S_DEADARTI4,
    S_DEADARTI5,
    S_DEADARTI6,
    S_DEADARTI7,
    S_DEADARTI8,
    S_DEADARTI9,
    S_DEADARTI10,
    S_ARTI_PTN2_1,
    S_ARTI_PTN2_2,
    S_ARTI_PTN2_3,
    S_ARTI_SOAR1,
    S_ARTI_SOAR2,
    S_ARTI_SOAR3,
    S_ARTI_SOAR4,
    S_ARTI_INVU1,
    S_ARTI_INVU2,
    S_ARTI_INVU3,
    S_ARTI_INVU4,
    S_ARTI_SUMMON,
    S_SUMMON_FX1_1,
    S_SUMMON_FX2_1,
    S_SUMMON_FX2_2,
    S_SUMMON_FX2_3,
    S_THRUSTINIT2_1,
    S_THRUSTINIT2_2,
    S_BTHRUSTINIT2_1,
    S_BTHRUSTINIT2_2,
    S_THRUSTINIT1_1,
    S_THRUSTINIT1_2,
    S_BTHRUSTINIT1_1,
    S_BTHRUSTINIT1_2,
    S_THRUSTRAISE1,
    S_THRUSTRAISE2,
    S_THRUSTRAISE3,
    S_THRUSTRAISE4,
    S_BTHRUSTRAISE1,
    S_BTHRUSTRAISE2,
    S_BTHRUSTRAISE3,
    S_BTHRUSTRAISE4,
    S_THRUSTIMPALE,
    S_BTHRUSTIMPALE,
    S_THRUSTRAISE,
    S_BTHRUSTRAISE,
    S_THRUSTBLOCK,
    S_BTHRUSTBLOCK,
    S_THRUSTLOWER,
    S_BTHRUSTLOWER,
    S_THRUSTSTAY,
    S_BTHRUSTSTAY,
    S_ARTI_TELOTHER1,
    S_ARTI_TELOTHER2,
    S_ARTI_TELOTHER3,
    S_ARTI_TELOTHER4,
    S_TELO_FX1,
    S_TELO_FX2,
    S_TELO_FX3,
    S_TELO_FX4,
    S_TELO_FX5,
    S_TELO_FX6,
    S_TELO_FX7,
    S_TELO_FX8,
    S_TELO_FX9,
    S_TELO_FX2_1,
    S_TELO_FX2_2,
    S_TELO_FX2_3,
    S_TELO_FX2_4,
    S_TELO_FX2_5,
    S_TELO_FX2_6,
    S_TELO_FX3_1,
    S_TELO_FX3_2,
    S_TELO_FX3_3,
    S_TELO_FX3_4,
    S_TELO_FX3_5,
    S_TELO_FX3_6,
    S_TELO_FX4_1,
    S_TELO_FX4_2,
    S_TELO_FX4_3,
    S_TELO_FX4_4,
    S_TELO_FX4_5,
    S_TELO_FX4_6,
    S_TELO_FX5_1,
    S_TELO_FX5_2,
    S_TELO_FX5_3,
    S_TELO_FX5_4,
    S_TELO_FX5_5,
    S_TELO_FX5_6,
    S_DIRT1_1,
    S_DIRT1_D,
    S_DIRT2_1,
    S_DIRT2_D,
    S_DIRT3_1,
    S_DIRT3_D,
    S_DIRT4_1,
    S_DIRT4_D,
    S_DIRT5_1,
    S_DIRT5_D,
    S_DIRT6_1,
    S_DIRT6_D,
    S_DIRTCLUMP1,
    S_ROCK1_1,
    S_ROCK1_D,
    S_ROCK2_1,
    S_ROCK2_D,
    S_ROCK3_1,
    S_ROCK3_D,
    S_SPAWNFOG1,
    S_FOGPATCHS1,
    S_FOGPATCHS2,
    S_FOGPATCHS3,
    S_FOGPATCHS4,
    S_FOGPATCHS5,
    S_FOGPATCHS0,
    S_FOGPATCHM1,
    S_FOGPATCHM2,
    S_FOGPATCHM3,
    S_FOGPATCHM4,
    S_FOGPATCHM5,
    S_FOGPATCHM0,
    S_FOGPATCHMA,
    S_FOGPATCHMB,
    S_FOGPATCHMC,
    S_FOGPATCHMD,
    S_FOGPATCHL1,
    S_FOGPATCHL2,
    S_FOGPATCHL3,
    S_FOGPATCHL4,
    S_FOGPATCHL5,
    S_FOGPATCHL0,
    S_FOGPATCHLA,
    S_FOGPATCHLB,
    S_FOGPATCHLC,
    S_FOGPATCHLD,
    S_QUAKE_ACTIVE1,
    S_QUAKE_ACTIVE2,
    S_QUAKE_ACTIVE3,
    S_QUAKE_ACTIVE4,
    S_QUAKE_ACTIVE5,
    S_QUAKE_ACTIVE6,
    S_QUAKE_ACTIVE7,
    S_QUAKE_ACTIVE8,
    S_QUAKE_ACTIVE9,
    S_QUAKE_ACTIVE0,
    S_QUAKE_ACTIVEA,
    S_QUAKE_ACTIVEB,
    S_QUAKE_ACTIVEC,
    S_QUAKE_ACTIVED,
    S_QUAKE_ACTIVEE,
    S_QUAKE_ACTIVEF,
    S_QUAKE_ACTIVEG,
    S_QUAKE_ACTIVEH,
    S_QUAKE_ACTIVEI,
    S_QUAKE_ACTIVEJ,
    S_QUAKE_ACTIVEK,
    S_QUAKE_ACTIVEL,
    S_QUAKE_ACTIVEM,
    S_QUAKE_ACTIVEN,
    S_QUAKE_ACTIVEO,
    S_QUAKE_ACTIVEP,
    S_QUAKE_ACTIVEQ,
    S_QUAKE_ACTIVER,
    S_QUAKE_ACTIVES,
    S_QUAKE_ACTIVET,
    S_QUAKE_ACTIVEU,
    S_QUAKE_ACTIVEV,
    S_QUAKE_ACTIVEW,
    S_QUAKE_ACTIVEX,
    S_QUAKE_ACTIVEY,
    S_QUAKE_ACTIVEZ,
    S_QUAKE_ACT1,
    S_QUAKE_ACT2,
    S_QUAKE_ACT3,
    S_QUAKE_ACT4,
    S_QUAKE_ACT5,
    S_QUAKE_ACT6,
    S_QUAKE_ACT7,
    S_QUAKE_ACT8,
    S_QUAKE_ACT9,
    S_QUAKE_ACT0,
    S_SGSHARD1_1,
    S_SGSHARD1_2,
    S_SGSHARD1_3,
    S_SGSHARD1_4,
    S_SGSHARD1_5,
    S_SGSHARD1_D,
    S_SGSHARD2_1,
    S_SGSHARD2_2,
    S_SGSHARD2_3,
    S_SGSHARD2_4,
    S_SGSHARD2_5,
    S_SGSHARD2_D,
    S_SGSHARD3_1,
    S_SGSHARD3_2,
    S_SGSHARD3_3,
    S_SGSHARD3_4,
    S_SGSHARD3_5,
    S_SGSHARD3_D,
    S_SGSHARD4_1,
    S_SGSHARD4_2,
    S_SGSHARD4_3,
    S_SGSHARD4_4,
    S_SGSHARD4_5,
    S_SGSHARD4_D,
    S_SGSHARD5_1,
    S_SGSHARD5_2,
    S_SGSHARD5_3,
    S_SGSHARD5_4,
    S_SGSHARD5_5,
    S_SGSHARD5_D,
    S_SGSHARD6_1,
    S_SGSHARD6_D,
    S_SGSHARD7_1,
    S_SGSHARD7_D,
    S_SGSHARD8_1,
    S_SGSHARD8_D,
    S_SGSHARD9_1,
    S_SGSHARD9_D,
    S_SGSHARD0_1,
    S_SGSHARD0_D,
    S_ARTI_EGGC1,
    S_ARTI_EGGC2,
    S_ARTI_EGGC3,
    S_ARTI_EGGC4,
    S_ARTI_EGGC5,
    S_ARTI_EGGC6,
    S_ARTI_EGGC7,
    S_ARTI_EGGC8,
    S_EGGFX1,
    S_EGGFX2,
    S_EGGFX3,
    S_EGGFX4,
    S_EGGFX5,
    S_EGGFXI1_1,
    S_EGGFXI1_2,
    S_EGGFXI1_3,
    S_EGGFXI1_4,
    S_ARTI_SPHL1,
    S_ZWINGEDSTATUENOSKULL,
    S_ZWINGEDSTATUENOSKULL2,
    S_ZGEMPEDESTAL1,
    S_ZGEMPEDESTAL2,
    S_ARTIPUZZSKULL,
    S_ARTIPUZZGEMBIG,
    S_ARTIPUZZGEMRED,
    S_ARTIPUZZGEMGREEN1,
    S_ARTIPUZZGEMGREEN2,
    S_ARTIPUZZGEMBLUE1,
    S_ARTIPUZZGEMBLUE2,
    S_ARTIPUZZBOOK1,
    S_ARTIPUZZBOOK2,
    S_ARTIPUZZSKULL2,
    S_ARTIPUZZFWEAPON,
    S_ARTIPUZZCWEAPON,
    S_ARTIPUZZMWEAPON,
    S_ARTIPUZZGEAR_1,
    S_ARTIPUZZGEAR_2,
    S_ARTIPUZZGEAR_3,
    S_ARTIPUZZGEAR_4,
    S_ARTIPUZZGEAR_5,
    S_ARTIPUZZGEAR_6,
    S_ARTIPUZZGEAR_7,
    S_ARTIPUZZGEAR_8,
    S_ARTIPUZZGEAR2_1,
    S_ARTIPUZZGEAR2_2,
    S_ARTIPUZZGEAR2_3,
    S_ARTIPUZZGEAR2_4,
    S_ARTIPUZZGEAR2_5,
    S_ARTIPUZZGEAR2_6,
    S_ARTIPUZZGEAR2_7,
    S_ARTIPUZZGEAR2_8,
    S_ARTIPUZZGEAR3_1,
    S_ARTIPUZZGEAR3_2,
    S_ARTIPUZZGEAR3_3,
    S_ARTIPUZZGEAR3_4,
    S_ARTIPUZZGEAR3_5,
    S_ARTIPUZZGEAR3_6,
    S_ARTIPUZZGEAR3_7,
    S_ARTIPUZZGEAR3_8,
    S_ARTIPUZZGEAR4_1,
    S_ARTIPUZZGEAR4_2,
    S_ARTIPUZZGEAR4_3,
    S_ARTIPUZZGEAR4_4,
    S_ARTIPUZZGEAR4_5,
    S_ARTIPUZZGEAR4_6,
    S_ARTIPUZZGEAR4_7,
    S_ARTIPUZZGEAR4_8,
    S_ARTI_TRCH1,
    S_ARTI_TRCH2,
    S_ARTI_TRCH3,
    S_FIREBOMB1,
    S_FIREBOMB2,
    S_FIREBOMB3,
    S_FIREBOMB4,
    S_FIREBOMB5,
    S_FIREBOMB6,
    S_FIREBOMB7,
    S_FIREBOMB8,
    S_FIREBOMB9,
    S_FIREBOMB10,
    S_FIREBOMB11,
    S_ARTI_ATLP1,
    S_ARTI_ATLP2,
    S_ARTI_ATLP3,
    S_ARTI_ATLP4,
    S_ARTI_PSBG1,
    S_POISONBAG1,
    S_POISONBAG2,
    S_POISONBAG3,
    S_POISONBAG4,
    S_POISONCLOUD1,
    S_POISONCLOUD2,
    S_POISONCLOUD3,
    S_POISONCLOUD4,
    S_POISONCLOUD5,
    S_POISONCLOUD6,
    S_POISONCLOUD7,
    S_POISONCLOUD8,
    S_POISONCLOUD9,
    S_POISONCLOUD10,
    S_POISONCLOUD11,
    S_POISONCLOUD12,
    S_POISONCLOUD13,
    S_POISONCLOUD14,
    S_POISONCLOUD15,
    S_POISONCLOUD16,
    S_POISONCLOUD17,
    S_POISONCLOUD18,
    S_POISONCLOUD_X1,
    S_POISONCLOUD_X2,
    S_POISONCLOUD_X3,
    S_POISONCLOUD_X4,
    S_THROWINGBOMB1,
    S_THROWINGBOMB2,
    S_THROWINGBOMB3,
    S_THROWINGBOMB4,
    S_THROWINGBOMB5,
    S_THROWINGBOMB6,
    S_THROWINGBOMB7,
    S_THROWINGBOMB8,
    S_THROWINGBOMB9,
    S_THROWINGBOMB10,
    S_THROWINGBOMB11,
    S_THROWINGBOMB12,
    S_THROWINGBOMB_X1,
    S_THROWINGBOMB_X2,
    S_THROWINGBOMB_X3,
    S_THROWINGBOMB_X4,
    S_THROWINGBOMB_X5,
    S_THROWINGBOMB_X6,
    S_THROWINGBOMB_X7,
    S_THROWINGBOMB_X8,
    S_ARTI_BOOTS1,
    S_ARTI_BOOTS2,
    S_ARTI_BOOTS3,
    S_ARTI_BOOTS4,
    S_ARTI_BOOTS5,
    S_ARTI_BOOTS6,
    S_ARTI_BOOTS7,
    S_ARTI_BOOTS8,
    S_ARTI_MANA,
    S_ARTI_ARMOR1,
    S_ARTI_ARMOR2,
    S_ARTI_ARMOR3,
    S_ARTI_ARMOR4,
    S_ARTI_ARMOR5,
    S_ARTI_ARMOR6,
    S_ARTI_ARMOR7,
    S_ARTI_ARMOR8,
    S_ARTI_BLAST1,
    S_ARTI_BLAST2,
    S_ARTI_BLAST3,
    S_ARTI_BLAST4,
    S_ARTI_BLAST5,
    S_ARTI_BLAST6,
    S_ARTI_BLAST7,
    S_ARTI_BLAST8,
    S_ARTI_HEALRAD1,
    S_ARTI_HEALRAD2,
    S_ARTI_HEALRAD3,
    S_ARTI_HEALRAD4,
    S_ARTI_HEALRAD5,
    S_ARTI_HEALRAD6,
    S_ARTI_HEALRAD7,
    S_ARTI_HEALRAD8,
    S_ARTI_HEALRAD9,
    S_ARTI_HEALRAD0,
    S_ARTI_HEALRADA,
    S_ARTI_HEALRADB,
    S_ARTI_HEALRADC,
    S_ARTI_HEALRADD,
    S_ARTI_HEALRADE,
    S_ARTI_HEALRADF,
    S_SPLASH1,
    S_SPLASH2,
    S_SPLASH3,
    S_SPLASH4,
    S_SPLASHX,
    S_SPLASHBASE1,
    S_SPLASHBASE2,
    S_SPLASHBASE3,
    S_SPLASHBASE4,
    S_SPLASHBASE5,
    S_SPLASHBASE6,
    S_SPLASHBASE7,
    S_LAVASPLASH1,
    S_LAVASPLASH2,
    S_LAVASPLASH3,
    S_LAVASPLASH4,
    S_LAVASPLASH5,
    S_LAVASPLASH6,
    S_LAVASMOKE1,
    S_LAVASMOKE2,
    S_LAVASMOKE3,
    S_LAVASMOKE4,
    S_LAVASMOKE5,
    S_SLUDGECHUNK1,
    S_SLUDGECHUNK2,
    S_SLUDGECHUNK3,
    S_SLUDGECHUNK4,
    S_SLUDGECHUNKX,
    S_SLUDGESPLASH1,
    S_SLUDGESPLASH2,
    S_SLUDGESPLASH3,
    S_SLUDGESPLASH4,
    S_ZWINGEDSTATUE1,
    S_ZROCK1_1,
    S_ZROCK2_1,
    S_ZROCK3_1,
    S_ZROCK4_1,
    S_ZCHANDELIER1,
    S_ZCHANDELIER2,
    S_ZCHANDELIER3,
    S_ZCHANDELIER_U,
    S_ZTREEDEAD1,
    S_ZTREE,
    S_ZTREEDESTRUCTIBLE1,
    S_ZTREEDES_D1,
    S_ZTREEDES_D2,
    S_ZTREEDES_D3,
    S_ZTREEDES_D4,
    S_ZTREEDES_D5,
    S_ZTREEDES_D6,
    S_ZTREEDES_X1,
    S_ZTREEDES_X2,
    S_ZTREEDES_X3,
    S_ZTREEDES_X4,
    S_ZTREEDES_X5,
    S_ZTREEDES_X6,
    S_ZTREEDES_X7,
    S_ZTREEDES_X8,
    S_ZTREEDES_X9,
    S_ZTREEDES_X10,
    S_ZTREESWAMP182_1,
    S_ZTREESWAMP172_1,
    S_ZSTUMPBURNED1,
    S_ZSTUMPBARE1,
    S_ZSTUMPSWAMP1_1,
    S_ZSTUMPSWAMP2_1,
    S_ZSHROOMLARGE1_1,
    S_ZSHROOMLARGE2_1,
    S_ZSHROOMLARGE3_1,
    S_ZSHROOMSMALL1_1,
    S_ZSHROOMSMALL2_1,
    S_ZSHROOMSMALL3_1,
    S_ZSHROOMSMALL4_1,
    S_ZSHROOMSMALL5_1,
    S_ZSTALAGMITEPILLAR1,
    S_ZSTALAGMITELARGE1,
    S_ZSTALAGMITEMEDIUM1,
    S_ZSTALAGMITESMALL1,
    S_ZSTALACTITELARGE1,
    S_ZSTALACTITEMEDIUM1,
    S_ZSTALACTITESMALL1,
    S_ZMOSSCEILING1_1,
    S_ZMOSSCEILING2_1,
    S_ZSWAMPVINE1,
    S_ZCORPSEKABOB1,
    S_ZCORPSESLEEPING1,
    S_ZTOMBSTONERIP1,
    S_ZTOMBSTONESHANE1,
    S_ZTOMBSTONEBIGCROSS1,
    S_ZTOMBSTONEBRIANR1,
    S_ZTOMBSTONECROSSCIRCLE1,
    S_ZTOMBSTONESMALLCROSS1,
    S_ZTOMBSTONEBRIANP1,
    S_CORPSEHANGING_1,
    S_ZSTATUEGARGOYLEGREENTALL_1,
    S_ZSTATUEGARGOYLEBLUETALL_1,
    S_ZSTATUEGARGOYLEGREENSHORT_1,
    S_ZSTATUEGARGOYLEBLUESHORT_1,
    S_ZSTATUEGARGOYLESTRIPETALL_1,
    S_ZSTATUEGARGOYLEDARKREDTALL_1,
    S_ZSTATUEGARGOYLEREDTALL_1,
    S_ZSTATUEGARGOYLETANTALL_1,
    S_ZSTATUEGARGOYLERUSTTALL_1,
    S_ZSTATUEGARGOYLEDARKREDSHORT_1,
    S_ZSTATUEGARGOYLEREDSHORT_1,
    S_ZSTATUEGARGOYLETANSHORT_1,
    S_ZSTATUEGARGOYLERUSTSHORT_1,
    S_ZBANNERTATTERED_1,
    S_ZTREELARGE1,
    S_ZTREELARGE2,
    S_ZTREEGNARLED1,
    S_ZTREEGNARLED2,
    S_ZLOG,
    S_ZSTALACTITEICELARGE,
    S_ZSTALACTITEICEMEDIUM,
    S_ZSTALACTITEICESMALL,
    S_ZSTALACTITEICETINY,
    S_ZSTALAGMITEICELARGE,
    S_ZSTALAGMITEICEMEDIUM,
    S_ZSTALAGMITEICESMALL,
    S_ZSTALAGMITEICETINY,
    S_ZROCKBROWN1,
    S_ZROCKBROWN2,
    S_ZROCKBLACK,
    S_ZRUBBLE1,
    S_ZRUBBLE2,
    S_ZRUBBLE3,
    S_ZVASEPILLAR,
    S_ZPOTTERY1,
    S_ZPOTTERY2,
    S_ZPOTTERY3,
    S_ZPOTTERY_EXPLODE,
    S_POTTERYBIT_1,
    S_POTTERYBIT_2,
    S_POTTERYBIT_3,
    S_POTTERYBIT_4,
    S_POTTERYBIT_5,
    S_POTTERYBIT_EX0,
    S_POTTERYBIT_EX1,
    S_POTTERYBIT_EX1_2,
    S_POTTERYBIT_EX2,
    S_POTTERYBIT_EX2_2,
    S_POTTERYBIT_EX3,
    S_POTTERYBIT_EX3_2,
    S_POTTERYBIT_EX4,
    S_POTTERYBIT_EX4_2,
    S_POTTERYBIT_EX5,
    S_POTTERYBIT_EX5_2,
    S_ZCORPSELYNCHED1,
    S_ZCORPSELYNCHED2,
    S_ZCORPSESITTING,
    S_ZCORPSESITTING_X,
    S_CORPSEBIT_1,
    S_CORPSEBIT_2,
    S_CORPSEBIT_3,
    S_CORPSEBIT_4,
    S_CORPSEBLOODDRIP,
    S_CORPSEBLOODDRIP_X1,
    S_CORPSEBLOODDRIP_X2,
    S_CORPSEBLOODDRIP_X3,
    S_CORPSEBLOODDRIP_X4,
    S_BLOODPOOL,
    S_ZCANDLE1,
    S_ZCANDLE2,
    S_ZCANDLE3,
    S_ZLEAFSPAWNER,
    S_LEAF1_1,
    S_LEAF1_2,
    S_LEAF1_3,
    S_LEAF1_4,
    S_LEAF1_5,
    S_LEAF1_6,
    S_LEAF1_7,
    S_LEAF1_8,
    S_LEAF1_9,
    S_LEAF1_10,
    S_LEAF1_11,
    S_LEAF1_12,
    S_LEAF1_13,
    S_LEAF1_14,
    S_LEAF1_15,
    S_LEAF1_16,
    S_LEAF1_17,
    S_LEAF1_18,
    S_LEAF_X1,
    S_LEAF2_1,
    S_LEAF2_2,
    S_LEAF2_3,
    S_LEAF2_4,
    S_LEAF2_5,
    S_LEAF2_6,
    S_LEAF2_7,
    S_LEAF2_8,
    S_LEAF2_9,
    S_LEAF2_10,
    S_LEAF2_11,
    S_LEAF2_12,
    S_LEAF2_13,
    S_LEAF2_14,
    S_LEAF2_15,
    S_LEAF2_16,
    S_LEAF2_17,
    S_LEAF2_18,
    S_ZTWINEDTORCH_1,
    S_ZTWINEDTORCH_2,
    S_ZTWINEDTORCH_3,
    S_ZTWINEDTORCH_4,
    S_ZTWINEDTORCH_5,
    S_ZTWINEDTORCH_6,
    S_ZTWINEDTORCH_7,
    S_ZTWINEDTORCH_8,
    S_ZTWINEDTORCH_UNLIT,
    S_BRIDGE1,
    S_BRIDGE2,
    S_BRIDGE3,
    S_FREE_BRIDGE1,
    S_FREE_BRIDGE2,
    S_BBALL1,
    S_BBALL2,
    S_ZWALLTORCH1,
    S_ZWALLTORCH2,
    S_ZWALLTORCH3,
    S_ZWALLTORCH4,
    S_ZWALLTORCH5,
    S_ZWALLTORCH6,
    S_ZWALLTORCH7,
    S_ZWALLTORCH8,
    S_ZWALLTORCH_U,
    S_ZBARREL1,
    S_ZSHRUB1,
    S_ZSHRUB1_DIE,
    S_ZSHRUB1_X1,
    S_ZSHRUB1_X2,
    S_ZSHRUB1_X3,
    S_ZSHRUB2,
    S_ZSHRUB2_DIE,
    S_ZSHRUB2_X1,
    S_ZSHRUB2_X2,
    S_ZSHRUB2_X3,
    S_ZSHRUB2_X4,
    S_ZBUCKET1,
    S_ZPOISONSHROOM1,
    S_ZPOISONSHROOM_P1,
    S_ZPOISONSHROOM_P2,
    S_ZPOISONSHROOM_X1,
    S_ZPOISONSHROOM_X2,
    S_ZPOISONSHROOM_X3,
    S_ZPOISONSHROOM_X4,
    S_ZFIREBULL1,
    S_ZFIREBULL2,
    S_ZFIREBULL3,
    S_ZFIREBULL4,
    S_ZFIREBULL5,
    S_ZFIREBULL6,
    S_ZFIREBULL7,
    S_ZFIREBULL_DEATH,
    S_ZFIREBULL_DEATH2,
    S_ZFIREBULL_U,
    S_ZFIREBULL_BIRTH,
    S_ZFIREBULL_BIRTH2,
    S_ZFIRETHING1,
    S_ZFIRETHING2,
    S_ZFIRETHING3,
    S_ZFIRETHING4,
    S_ZFIRETHING5,
    S_ZFIRETHING6,
    S_ZFIRETHING7,
    S_ZFIRETHING8,
    S_ZFIRETHING9,
    S_ZBRASSTORCH1,
    S_ZBRASSTORCH2,
    S_ZBRASSTORCH3,
    S_ZBRASSTORCH4,
    S_ZBRASSTORCH5,
    S_ZBRASSTORCH6,
    S_ZBRASSTORCH7,
    S_ZBRASSTORCH8,
    S_ZBRASSTORCH9,
    S_ZBRASSTORCH10,
    S_ZBRASSTORCH11,
    S_ZBRASSTORCH12,
    S_ZBRASSTORCH13,
    S_ZSUITOFARMOR,
    S_ZSUITOFARMOR_X1,
    S_ZARMORCHUNK1,
    S_ZARMORCHUNK2,
    S_ZARMORCHUNK3,
    S_ZARMORCHUNK4,
    S_ZARMORCHUNK5,
    S_ZARMORCHUNK6,
    S_ZARMORCHUNK7,
    S_ZARMORCHUNK8,
    S_ZARMORCHUNK9,
    S_ZARMORCHUNK10,
    S_ZBELL,
    S_ZBELL_X1,
    S_ZBELL_X2,
    S_ZBELL_X3,
    S_ZBELL_X4,
    S_ZBELL_X5,
    S_ZBELL_X6,
    S_ZBELL_X7,
    S_ZBELL_X8,
    S_ZBELL_X9,
    S_ZBELL_X10,
    S_ZBELL_X11,
    S_ZBELL_X12,
    S_ZBELL_X13,
    S_ZBELL_X14,
    S_ZBELL_X15,
    S_ZBELL_X16,
    S_ZBELL_X17,
    S_ZBELL_X18,
    S_ZBELL_X19,
    S_ZBELL_X20,
    S_ZBELL_X21,
    S_ZBELL_X22,
    S_ZBELL_X23,
    S_ZBELL_X24,
    S_ZBELL_X25,
    S_ZBELL_X26,
    S_ZBELL_X27,
    S_ZBELL_X28,
    S_ZBELL_X29,
    S_ZBELL_X30,
    S_ZBELL_X31,
    S_ZBELL_X32,
    S_ZBELL_X33,
    S_ZBELL_X34,
    S_ZBELL_X35,
    S_ZBELL_X36,
    S_ZBELL_X37,
    S_ZBELL_X38,
    S_ZBELL_X39,
    S_ZBELL_X40,
    S_ZBELL_X41,
    S_ZBELL_X42,
    S_ZBELL_X43,
    S_ZBELL_X44,
    S_ZBELL_X45,
    S_ZBELL_X46,
    S_ZBELL_X47,
    S_ZBLUE_CANDLE1,
    S_ZBLUE_CANDLE2,
    S_ZBLUE_CANDLE3,
    S_ZBLUE_CANDLE4,
    S_ZBLUE_CANDLE5,
    S_ZIRON_MAIDEN,
    S_ZXMAS_TREE,
    S_ZXMAS_TREE_DIE,
    S_ZXMAS_TREE_X1,
    S_ZXMAS_TREE_X2,
    S_ZXMAS_TREE_X3,
    S_ZXMAS_TREE_X4,
    S_ZXMAS_TREE_X5,
    S_ZXMAS_TREE_X6,
    S_ZXMAS_TREE_X7,
    S_ZXMAS_TREE_X8,
    S_ZXMAS_TREE_X9,
    S_ZXMAS_TREE_X10,
    S_ZCAULDRON1,
    S_ZCAULDRON2,
    S_ZCAULDRON3,
    S_ZCAULDRON4,
    S_ZCAULDRON5,
    S_ZCAULDRON6,
    S_ZCAULDRON7,
    S_ZCAULDRON_U,
    S_ZCHAINBIT32,
    S_ZCHAINBIT64,
    S_ZCHAINEND_HEART,
    S_ZCHAINEND_HOOK1,
    S_ZCHAINEND_HOOK2,
    S_ZCHAINEND_SPIKE,
    S_ZCHAINEND_SKULL,
    S_TABLE_SHIT1,
    S_TABLE_SHIT2,
    S_TABLE_SHIT3,
    S_TABLE_SHIT4,
    S_TABLE_SHIT5,
    S_TABLE_SHIT6,
    S_TABLE_SHIT7,
    S_TABLE_SHIT8,
    S_TABLE_SHIT9,
    S_TABLE_SHIT10,
    S_TFOG1,
    S_TFOG2,
    S_TFOG3,
    S_TFOG4,
    S_TFOG5,
    S_TFOG6,
    S_TFOG7,
    S_TFOG8,
    S_TFOG9,
    S_TFOG10,
    S_TFOG11,
    S_TFOG12,
    S_TFOG13,
    S_TELESMOKE1,
    S_TELESMOKE2,
    S_TELESMOKE3,
    S_TELESMOKE4,
    S_TELESMOKE5,
    S_TELESMOKE6,
    S_TELESMOKE7,
    S_TELESMOKE8,
    S_TELESMOKE9,
    S_TELESMOKE10,
    S_TELESMOKE11,
    S_TELESMOKE12,
    S_TELESMOKE13,
    S_TELESMOKE14,
    S_TELESMOKE15,
    S_TELESMOKE16,
    S_TELESMOKE17,
    S_TELESMOKE18,
    S_TELESMOKE19,
    S_TELESMOKE20,
    S_TELESMOKE21,
    S_TELESMOKE22,
    S_TELESMOKE23,
    S_TELESMOKE24,
    S_TELESMOKE25,
    S_TELESMOKE26,
    S_LIGHTDONE,
    S_PUNCHREADY,
    S_PUNCHDOWN,
    S_PUNCHUP,
    S_PUNCHATK1_1,
    S_PUNCHATK1_2,
    S_PUNCHATK1_3,
    S_PUNCHATK1_4,
    S_PUNCHATK1_5,
    S_PUNCHATK2_1,
    S_PUNCHATK2_2,
    S_PUNCHATK2_3,
    S_PUNCHATK2_4,
    S_PUNCHATK2_5,
    S_PUNCHATK2_6,
    S_PUNCHATK2_7,
    S_PUNCHATK2_8,
    S_PUNCHATK2_9,
    S_PUNCHPUFF1,
    S_PUNCHPUFF2,
    S_PUNCHPUFF3,
    S_PUNCHPUFF4,
    S_PUNCHPUFF5,
    S_AXE,
    S_FAXEREADY,
    S_FAXEDOWN,
    S_FAXEUP,
    S_FAXEATK_1,
    S_FAXEATK_2,
    S_FAXEATK_3,
    S_FAXEATK_4,
    S_FAXEATK_5,
    S_FAXEATK_6,
    S_FAXEATK_7,
    S_FAXEATK_8,
    S_FAXEATK_9,
    S_FAXEATK_10,
    S_FAXEATK_11,
    S_FAXEATK_12,
    S_FAXEATK_13,
    S_FAXEREADY_G,
    S_FAXEREADY_G1,
    S_FAXEREADY_G2,
    S_FAXEREADY_G3,
    S_FAXEREADY_G4,
    S_FAXEREADY_G5,
    S_FAXEDOWN_G,
    S_FAXEUP_G,
    S_FAXEATK_G1,
    S_FAXEATK_G2,
    S_FAXEATK_G3,
    S_FAXEATK_G4,
    S_FAXEATK_G5,
    S_FAXEATK_G6,
    S_FAXEATK_G7,
    S_FAXEATK_G8,
    S_FAXEATK_G9,
    S_FAXEATK_G10,
    S_FAXEATK_G11,
    S_FAXEATK_G12,
    S_FAXEATK_G13,
    S_AXEPUFF_GLOW1,
    S_AXEPUFF_GLOW2,
    S_AXEPUFF_GLOW3,
    S_AXEPUFF_GLOW4,
    S_AXEPUFF_GLOW5,
    S_AXEPUFF_GLOW6,
    S_AXEPUFF_GLOW7,
    S_AXEBLOOD1,
    S_AXEBLOOD2,
    S_AXEBLOOD3,
    S_AXEBLOOD4,
    S_AXEBLOOD5,
    S_AXEBLOOD6,
    S_HAMM,
    S_FHAMMERREADY,
    S_FHAMMERDOWN,
    S_FHAMMERUP,
    S_FHAMMERATK_1,
    S_FHAMMERATK_2,
    S_FHAMMERATK_3,
    S_FHAMMERATK_4,
    S_FHAMMERATK_5,
    S_FHAMMERATK_6,
    S_FHAMMERATK_7,
    S_FHAMMERATK_8,
    S_FHAMMERATK_9,
    S_FHAMMERATK_10,
    S_FHAMMERATK_11,
    S_FHAMMERATK_12,
    S_HAMMER_MISSILE_1,
    S_HAMMER_MISSILE_2,
    S_HAMMER_MISSILE_3,
    S_HAMMER_MISSILE_4,
    S_HAMMER_MISSILE_5,
    S_HAMMER_MISSILE_6,
    S_HAMMER_MISSILE_7,
    S_HAMMER_MISSILE_8,
    S_HAMMER_MISSILE_X1,
    S_HAMMER_MISSILE_X2,
    S_HAMMER_MISSILE_X3,
    S_HAMMER_MISSILE_X4,
    S_HAMMER_MISSILE_X5,
    S_HAMMER_MISSILE_X6,
    S_HAMMER_MISSILE_X7,
    S_HAMMER_MISSILE_X8,
    S_HAMMER_MISSILE_X9,
    S_HAMMER_MISSILE_X10,
    S_HAMMERPUFF1,
    S_HAMMERPUFF2,
    S_HAMMERPUFF3,
    S_HAMMERPUFF4,
    S_HAMMERPUFF5,
    S_FSWORDREADY,
    S_FSWORDREADY1,
    S_FSWORDREADY2,
    S_FSWORDREADY3,
    S_FSWORDREADY4,
    S_FSWORDREADY5,
    S_FSWORDREADY6,
    S_FSWORDREADY7,
    S_FSWORDREADY8,
    S_FSWORDREADY9,
    S_FSWORDREADY10,
    S_FSWORDREADY11,
    S_FSWORDDOWN,
    S_FSWORDUP,
    S_FSWORDATK_1,
    S_FSWORDATK_2,
    S_FSWORDATK_3,
    S_FSWORDATK_4,
    S_FSWORDATK_5,
    S_FSWORDATK_6,
    S_FSWORDATK_7,
    S_FSWORDATK_8,
    S_FSWORDATK_9,
    S_FSWORDATK_10,
    S_FSWORDATK_11,
    S_FSWORDATK_12,
    S_FSWORD_MISSILE1,
    S_FSWORD_MISSILE2,
    S_FSWORD_MISSILE3,
    S_FSWORD_MISSILE_X1,
    S_FSWORD_MISSILE_X2,
    S_FSWORD_MISSILE_X3,
    S_FSWORD_MISSILE_X4,
    S_FSWORD_MISSILE_X5,
    S_FSWORD_MISSILE_X6,
    S_FSWORD_MISSILE_X7,
    S_FSWORD_MISSILE_X8,
    S_FSWORD_MISSILE_X9,
    S_FSWORD_MISSILE_X10,
    S_FSWORD_FLAME1,
    S_FSWORD_FLAME2,
    S_FSWORD_FLAME3,
    S_FSWORD_FLAME4,
    S_FSWORD_FLAME5,
    S_FSWORD_FLAME6,
    S_FSWORD_FLAME7,
    S_FSWORD_FLAME8,
    S_FSWORD_FLAME9,
    S_FSWORD_FLAME10,
    S_CMACEREADY,
    S_CMACEDOWN,
    S_CMACEUP,
    S_CMACEATK_1,
    S_CMACEATK_2,
    S_CMACEATK_3,
    S_CMACEATK_4,
    S_CMACEATK_5,
    S_CMACEATK_6,
    S_CMACEATK_7,
    S_CMACEATK_8,
    S_CMACEATK_9,
    S_CMACEATK_10,
    S_CMACEATK_11,
    S_CMACEATK_12,
    S_CMACEATK_13,
    S_CMACEATK_14,
    S_CMACEATK_15,
    S_CMACEATK_16,
    S_CMACEATK_17,
    S_CSTAFF,
    S_CSTAFFREADY,
    S_CSTAFFREADY1,
    S_CSTAFFREADY2,
    S_CSTAFFREADY3,
    S_CSTAFFREADY4,
    S_CSTAFFREADY5,
    S_CSTAFFREADY6,
    S_CSTAFFREADY7,
    S_CSTAFFREADY8,
    S_CSTAFFREADY9,
    S_CSTAFFBLINK1,
    S_CSTAFFBLINK2,
    S_CSTAFFBLINK3,
    S_CSTAFFBLINK4,
    S_CSTAFFBLINK5,
    S_CSTAFFBLINK6,
    S_CSTAFFBLINK7,
    S_CSTAFFBLINK8,
    S_CSTAFFBLINK9,
    S_CSTAFFBLINK10,
    S_CSTAFFBLINK11,
    S_CSTAFFDOWN,
    S_CSTAFFDOWN2,
    S_CSTAFFDOWN3,
    S_CSTAFFUP,
    S_CSTAFFATK_1,
    S_CSTAFFATK_2,
    S_CSTAFFATK_3,
    S_CSTAFFATK_4,
    S_CSTAFFATK_5,
    S_CSTAFFATK_6,
    S_CSTAFFATK2_1,
    S_CSTAFF_MISSILE1,
    S_CSTAFF_MISSILE2,
    S_CSTAFF_MISSILE3,
    S_CSTAFF_MISSILE4,
    S_CSTAFF_MISSILE_X1,
    S_CSTAFF_MISSILE_X2,
    S_CSTAFF_MISSILE_X3,
    S_CSTAFF_MISSILE_X4,
    S_CSTAFFPUFF1,
    S_CSTAFFPUFF2,
    S_CSTAFFPUFF3,
    S_CSTAFFPUFF4,
    S_CSTAFFPUFF5,
    S_CFLAME1,
    S_CFLAME2,
    S_CFLAME3,
    S_CFLAME4,
    S_CFLAME5,
    S_CFLAME6,
    S_CFLAME7,
    S_CFLAME8,
    S_CFLAMEREADY1,
    S_CFLAMEREADY2,
    S_CFLAMEREADY3,
    S_CFLAMEREADY4,
    S_CFLAMEREADY5,
    S_CFLAMEREADY6,
    S_CFLAMEREADY7,
    S_CFLAMEREADY8,
    S_CFLAMEREADY9,
    S_CFLAMEREADY10,
    S_CFLAMEREADY11,
    S_CFLAMEREADY12,
    S_CFLAMEDOWN,
    S_CFLAMEUP,
    S_CFLAMEATK_1,
    S_CFLAMEATK_2,
    S_CFLAMEATK_3,
    S_CFLAMEATK_4,
    S_CFLAMEATK_5,
    S_CFLAMEATK_6,
    S_CFLAMEATK_7,
    S_CFLAMEATK_8,
    S_CFLAMEFLOOR1,
    S_CFLAMEFLOOR2,
    S_CFLAMEFLOOR3,
    S_FLAMEPUFF1,
    S_FLAMEPUFF2,
    S_FLAMEPUFF3,
    S_FLAMEPUFF4,
    S_FLAMEPUFF5,
    S_FLAMEPUFF6,
    S_FLAMEPUFF7,
    S_FLAMEPUFF8,
    S_FLAMEPUFF9,
    S_FLAMEPUFF10,
    S_FLAMEPUFF11,
    S_FLAMEPUFF12,
    S_FLAMEPUFF13,
    S_FLAMEPUFF2_1,
    S_FLAMEPUFF2_2,
    S_FLAMEPUFF2_3,
    S_FLAMEPUFF2_4,
    S_FLAMEPUFF2_5,
    S_FLAMEPUFF2_6,
    S_FLAMEPUFF2_7,
    S_FLAMEPUFF2_8,
    S_FLAMEPUFF2_9,
    S_FLAMEPUFF2_10,
    S_FLAMEPUFF2_11,
    S_FLAMEPUFF2_12,
    S_FLAMEPUFF2_13,
    S_FLAMEPUFF2_14,
    S_FLAMEPUFF2_15,
    S_FLAMEPUFF2_16,
    S_FLAMEPUFF2_17,
    S_FLAMEPUFF2_18,
    S_FLAMEPUFF2_19,
    S_FLAMEPUFF2_20,
    S_CIRCLE_FLAME1,
    S_CIRCLE_FLAME2,
    S_CIRCLE_FLAME3,
    S_CIRCLE_FLAME4,
    S_CIRCLE_FLAME5,
    S_CIRCLE_FLAME6,
    S_CIRCLE_FLAME7,
    S_CIRCLE_FLAME8,
    S_CIRCLE_FLAME9,
    S_CIRCLE_FLAME10,
    S_CIRCLE_FLAME11,
    S_CIRCLE_FLAME12,
    S_CIRCLE_FLAME13,
    S_CIRCLE_FLAME14,
    S_CIRCLE_FLAME15,
    S_CIRCLE_FLAME16,
    S_CIRCLE_FLAME_X1,
    S_CIRCLE_FLAME_X2,
    S_CIRCLE_FLAME_X3,
    S_CIRCLE_FLAME_X4,
    S_CIRCLE_FLAME_X5,
    S_CIRCLE_FLAME_X6,
    S_CIRCLE_FLAME_X7,
    S_CIRCLE_FLAME_X8,
    S_CIRCLE_FLAME_X9,
    S_CIRCLE_FLAME_X10,
    S_CFLAME_MISSILE1,
    S_CFLAME_MISSILE2,
    S_CFLAME_MISSILE_X,
    S_CHOLYREADY,
    S_CHOLYDOWN,
    S_CHOLYUP,
    S_CHOLYATK_1,
    S_CHOLYATK_2,
    S_CHOLYATK_3,
    S_CHOLYATK_4,
    S_CHOLYATK_5,
    S_CHOLYATK_6,
    S_CHOLYATK_7,
    S_CHOLYATK_8,
    S_CHOLYATK_9,
    S_HOLY_FX1,
    S_HOLY_FX2,
    S_HOLY_FX3,
    S_HOLY_FX4,
    S_HOLY_FX_X1,
    S_HOLY_FX_X2,
    S_HOLY_FX_X3,
    S_HOLY_FX_X4,
    S_HOLY_FX_X5,
    S_HOLY_FX_X6,
    S_HOLY_TAIL1,
    S_HOLY_TAIL2,
    S_HOLY_PUFF1,
    S_HOLY_PUFF2,
    S_HOLY_PUFF3,
    S_HOLY_PUFF4,
    S_HOLY_PUFF5,
    S_HOLY_MISSILE1,
    S_HOLY_MISSILE2,
    S_HOLY_MISSILE3,
    S_HOLY_MISSILE4,
    S_HOLY_MISSILE_X,
    S_HOLY_MISSILE_P1,
    S_HOLY_MISSILE_P2,
    S_HOLY_MISSILE_P3,
    S_HOLY_MISSILE_P4,
    S_HOLY_MISSILE_P5,
    S_MWANDREADY,
    S_MWANDDOWN,
    S_MWANDUP,
    S_MWANDATK_1,
    S_MWANDATK_2,
    S_MWANDATK_3,
    S_MWANDATK_4,
    S_MWANDPUFF1,
    S_MWANDPUFF2,
    S_MWANDPUFF3,
    S_MWANDPUFF4,
    S_MWANDPUFF5,
    S_MWANDSMOKE1,
    S_MWANDSMOKE2,
    S_MWANDSMOKE3,
    S_MWANDSMOKE4,
    S_MWAND_MISSILE1,
    S_MWAND_MISSILE2,
    S_MW_LIGHTNING1,
    S_MW_LIGHTNING2,
    S_MW_LIGHTNING3,
    S_MW_LIGHTNING4,
    S_MW_LIGHTNING5,
    S_MW_LIGHTNING6,
    S_MW_LIGHTNING7,
    S_MW_LIGHTNING8,
    S_MLIGHTNINGREADY,
    S_MLIGHTNINGREADY2,
    S_MLIGHTNINGREADY3,
    S_MLIGHTNINGREADY4,
    S_MLIGHTNINGREADY5,
    S_MLIGHTNINGREADY6,
    S_MLIGHTNINGREADY7,
    S_MLIGHTNINGREADY8,
    S_MLIGHTNINGREADY9,
    S_MLIGHTNINGREADY10,
    S_MLIGHTNINGREADY11,
    S_MLIGHTNINGREADY12,
    S_MLIGHTNINGREADY13,
    S_MLIGHTNINGREADY14,
    S_MLIGHTNINGREADY15,
    S_MLIGHTNINGREADY16,
    S_MLIGHTNINGREADY17,
    S_MLIGHTNINGREADY18,
    S_MLIGHTNINGREADY19,
    S_MLIGHTNINGREADY20,
    S_MLIGHTNINGREADY21,
    S_MLIGHTNINGREADY22,
    S_MLIGHTNINGREADY23,
    S_MLIGHTNINGREADY24,
    S_MLIGHTNINGDOWN,
    S_MLIGHTNINGUP,
    S_MLIGHTNINGATK_1,
    S_MLIGHTNINGATK_2,
    S_MLIGHTNINGATK_3,
    S_MLIGHTNINGATK_4,
    S_MLIGHTNINGATK_5,
    S_MLIGHTNINGATK_6,
    S_MLIGHTNINGATK_7,
    S_MLIGHTNINGATK_8,
    S_MLIGHTNINGATK_9,
    S_MLIGHTNINGATK_10,
    S_MLIGHTNINGATK_11,
    S_LIGHTNING_CEILING1,
    S_LIGHTNING_CEILING2,
    S_LIGHTNING_CEILING3,
    S_LIGHTNING_CEILING4,
    S_LIGHTNING_C_X1,
    S_LIGHTNING_C_X2,
    S_LIGHTNING_C_X3,
    S_LIGHTNING_C_X4,
    S_LIGHTNING_C_X5,
    S_LIGHTNING_C_X6,
    S_LIGHTNING_C_X7,
    S_LIGHTNING_C_X8,
    S_LIGHTNING_C_X9,
    S_LIGHTNING_C_X10,
    S_LIGHTNING_C_X11,
    S_LIGHTNING_C_X12,
    S_LIGHTNING_C_X13,
    S_LIGHTNING_C_X14,
    S_LIGHTNING_C_X15,
    S_LIGHTNING_C_X16,
    S_LIGHTNING_C_X17,
    S_LIGHTNING_C_X18,
    S_LIGHTNING_C_X19,
    S_LIGHTNING_FLOOR1,
    S_LIGHTNING_FLOOR2,
    S_LIGHTNING_FLOOR3,
    S_LIGHTNING_FLOOR4,
    S_LIGHTNING_F_X1,
    S_LIGHTNING_F_X2,
    S_LIGHTNING_F_X3,
    S_LIGHTNING_F_X4,
    S_LIGHTNING_F_X5,
    S_LIGHTNING_F_X6,
    S_LIGHTNING_F_X7,
    S_LIGHTNING_F_X8,
    S_LIGHTNING_F_X9,
    S_LIGHTNING_F_X10,
    S_LIGHTNING_F_X11,
    S_LIGHTNING_F_X12,
    S_LIGHTNING_F_X13,
    S_LIGHTNING_F_X14,
    S_LIGHTNING_F_X15,
    S_LIGHTNING_F_X16,
    S_LIGHTNING_F_X17,
    S_LIGHTNING_F_X18,
    S_LIGHTNING_F_X19,
    S_LIGHTNING_ZAP1,
    S_LIGHTNING_ZAP2,
    S_LIGHTNING_ZAP3,
    S_LIGHTNING_ZAP4,
    S_LIGHTNING_ZAP5,
    S_LIGHTNING_ZAP_X1,
    S_LIGHTNING_ZAP_X2,
    S_LIGHTNING_ZAP_X3,
    S_LIGHTNING_ZAP_X4,
    S_LIGHTNING_ZAP_X5,
    S_LIGHTNING_ZAP_X6,
    S_LIGHTNING_ZAP_X7,
    S_LIGHTNING_ZAP_X8,
    S_MSTAFFREADY,
    S_MSTAFFREADY2,
    S_MSTAFFREADY3,
    S_MSTAFFREADY4,
    S_MSTAFFREADY5,
    S_MSTAFFREADY6,
    S_MSTAFFREADY7,
    S_MSTAFFREADY8,
    S_MSTAFFREADY9,
    S_MSTAFFREADY10,
    S_MSTAFFREADY11,
    S_MSTAFFREADY12,
    S_MSTAFFREADY13,
    S_MSTAFFREADY14,
    S_MSTAFFREADY15,
    S_MSTAFFREADY16,
    S_MSTAFFREADY17,
    S_MSTAFFREADY18,
    S_MSTAFFREADY19,
    S_MSTAFFREADY20,
    S_MSTAFFREADY21,
    S_MSTAFFREADY22,
    S_MSTAFFREADY23,
    S_MSTAFFREADY24,
    S_MSTAFFREADY25,
    S_MSTAFFREADY26,
    S_MSTAFFREADY27,
    S_MSTAFFREADY28,
    S_MSTAFFREADY29,
    S_MSTAFFREADY30,
    S_MSTAFFREADY31,
    S_MSTAFFREADY32,
    S_MSTAFFREADY33,
    S_MSTAFFREADY34,
    S_MSTAFFREADY35,
    S_MSTAFFDOWN,
    S_MSTAFFUP,
    S_MSTAFFATK_1,
    S_MSTAFFATK_2,
    S_MSTAFFATK_3,
    S_MSTAFFATK_4,
    S_MSTAFFATK_5,
    S_MSTAFFATK_6,
    S_MSTAFFATK_7,
    S_MSTAFF_FX1_1,
    S_MSTAFF_FX1_2,
    S_MSTAFF_FX1_3,
    S_MSTAFF_FX1_4,
    S_MSTAFF_FX1_5,
    S_MSTAFF_FX1_6,
    S_MSTAFF_FX_X1,
    S_MSTAFF_FX_X2,
    S_MSTAFF_FX_X3,
    S_MSTAFF_FX_X4,
    S_MSTAFF_FX_X5,
    S_MSTAFF_FX_X6,
    S_MSTAFF_FX_X7,
    S_MSTAFF_FX_X8,
    S_MSTAFF_FX_X9,
    S_MSTAFF_FX_X10,
    S_MSTAFF_FX2_1,
    S_MSTAFF_FX2_2,
    S_MSTAFF_FX2_3,
    S_MSTAFF_FX2_4,
    S_MSTAFF_FX2_X1,
    S_MSTAFF_FX2_X2,
    S_MSTAFF_FX2_X3,
    S_MSTAFF_FX2_X4,
    S_MSTAFF_FX2_X5,
    S_FSWORD1,
    S_FSWORD2,
    S_FSWORD3,
    S_CHOLY1,
    S_CHOLY2,
    S_CHOLY3,
    S_MSTAFF1,
    S_MSTAFF2,
    S_MSTAFF3,
    S_SNOUTREADY,
    S_SNOUTDOWN,
    S_SNOUTUP,
    S_SNOUTATK1,
    S_SNOUTATK2,
    S_COS1,
    S_COS2,
    S_COS3,
    S_CONEREADY,
    S_CONEDOWN,
    S_CONEUP,
    S_CONEATK1_1,
    S_CONEATK1_2,
    S_CONEATK1_3,
    S_CONEATK1_4,
    S_CONEATK1_5,
    S_CONEATK1_6,
    S_CONEATK1_7,
    S_CONEATK1_8,
    S_SHARDFX1_1,
    S_SHARDFX1_2,
    S_SHARDFX1_3,
    S_SHARDFX1_4,
    S_SHARDFXE1_1,
    S_SHARDFXE1_2,
    S_SHARDFXE1_3,
    S_SHARDFXE1_4,
    S_SHARDFXE1_5,
    S_BLOOD1,
    S_BLOOD2,
    S_BLOOD3,
    S_BLOODSPLATTER1,
    S_BLOODSPLATTER2,
    S_BLOODSPLATTER3,
    S_BLOODSPLATTERX,
    S_GIBS1,
    S_FPLAY,
    S_FPLAY_RUN1,
    S_FPLAY_RUN2,
    S_FPLAY_RUN3,
    S_FPLAY_RUN4,
    S_FPLAY_ATK1,
    S_FPLAY_ATK2,
    S_FPLAY_PAIN,
    S_FPLAY_PAIN2,
    S_FPLAY_DIE1,
    S_FPLAY_DIE2,
    S_FPLAY_DIE3,
    S_FPLAY_DIE4,
    S_FPLAY_DIE5,
    S_FPLAY_DIE6,
    S_FPLAY_DIE7,
    S_FPLAY_XDIE1,
    S_FPLAY_XDIE2,
    S_FPLAY_XDIE3,
    S_FPLAY_XDIE4,
    S_FPLAY_XDIE5,
    S_FPLAY_XDIE6,
    S_FPLAY_XDIE7,
    S_FPLAY_XDIE8,
    S_FPLAY_ICE,
    S_FPLAY_ICE2,
    S_PLAY_F_FDTH1,
    S_PLAY_F_FDTH2,
    S_PLAY_C_FDTH1,
    S_PLAY_C_FDTH2,
    S_PLAY_M_FDTH1,
    S_PLAY_M_FDTH2,
    S_PLAY_FDTH3,
    S_PLAY_FDTH4,
    S_PLAY_FDTH5,
    S_PLAY_FDTH6,
    S_PLAY_FDTH7,
    S_PLAY_FDTH8,
    S_PLAY_FDTH9,
    S_PLAY_FDTH10,
    S_PLAY_FDTH11,
    S_PLAY_FDTH12,
    S_PLAY_FDTH13,
    S_PLAY_FDTH14,
    S_PLAY_FDTH15,
    S_PLAY_FDTH16,
    S_PLAY_FDTH17,
    S_PLAY_FDTH18,
    S_PLAY_FDTH19,
    S_PLAY_FDTH20,
    S_BLOODYSKULL1,
    S_BLOODYSKULL2,
    S_BLOODYSKULL3,
    S_BLOODYSKULL4,
    S_BLOODYSKULL5,
    S_BLOODYSKULL6,
    S_BLOODYSKULL7,
    S_BLOODYSKULLX1,
    S_BLOODYSKULLX2,
    S_PLAYER_SPEED1,
    S_PLAYER_SPEED2,
    S_ICECHUNK1,
    S_ICECHUNK2,
    S_ICECHUNK3,
    S_ICECHUNK4,
    S_ICECHUNK_HEAD,
    S_ICECHUNK_HEAD2,
    S_CPLAY,
    S_CPLAY_RUN1,
    S_CPLAY_RUN2,
    S_CPLAY_RUN3,
    S_CPLAY_RUN4,
    S_CPLAY_ATK1,
    S_CPLAY_ATK2,
    S_CPLAY_ATK3,
    S_CPLAY_PAIN,
    S_CPLAY_PAIN2,
    S_CPLAY_DIE1,
    S_CPLAY_DIE2,
    S_CPLAY_DIE3,
    S_CPLAY_DIE4,
    S_CPLAY_DIE5,
    S_CPLAY_DIE6,
    S_CPLAY_DIE7,
    S_CPLAY_DIE8,
    S_CPLAY_DIE9,
    S_CPLAY_XDIE1,
    S_CPLAY_XDIE2,
    S_CPLAY_XDIE3,
    S_CPLAY_XDIE4,
    S_CPLAY_XDIE5,
    S_CPLAY_XDIE6,
    S_CPLAY_XDIE7,
    S_CPLAY_XDIE8,
    S_CPLAY_XDIE9,
    S_CPLAY_XDIE10,
    S_CPLAY_ICE,
    S_CPLAY_ICE2,
    S_MPLAY,
    S_MPLAY_RUN1,
    S_MPLAY_RUN2,
    S_MPLAY_RUN3,
    S_MPLAY_RUN4,
    S_MPLAY_ATK1,
    S_MPLAY_ATK2,
    S_MPLAY_PAIN,
    S_MPLAY_PAIN2,
    S_MPLAY_DIE1,
    S_MPLAY_DIE2,
    S_MPLAY_DIE3,
    S_MPLAY_DIE4,
    S_MPLAY_DIE5,
    S_MPLAY_DIE6,
    S_MPLAY_DIE7,
    S_MPLAY_XDIE1,
    S_MPLAY_XDIE2,
    S_MPLAY_XDIE3,
    S_MPLAY_XDIE4,
    S_MPLAY_XDIE5,
    S_MPLAY_XDIE6,
    S_MPLAY_XDIE7,
    S_MPLAY_XDIE8,
    S_MPLAY_XDIE9,
    S_MPLAY_ICE,
    S_MPLAY_ICE2,
    S_PIGPLAY,
    S_PIGPLAY_RUN1,
    S_PIGPLAY_RUN2,
    S_PIGPLAY_RUN3,
    S_PIGPLAY_RUN4,
    S_PIGPLAY_ATK1,
    S_PIGPLAY_PAIN,
    S_PIG_LOOK1,
    S_PIG_WALK1,
    S_PIG_WALK2,
    S_PIG_WALK3,
    S_PIG_WALK4,
    S_PIG_PAIN,
    S_PIG_ATK1,
    S_PIG_ATK2,
    S_PIG_DIE1,
    S_PIG_DIE2,
    S_PIG_DIE3,
    S_PIG_DIE4,
    S_PIG_DIE5,
    S_PIG_DIE6,
    S_PIG_DIE7,
    S_PIG_DIE8,
    S_PIG_ICE,
    S_PIG_ICE2,
    S_CENTAUR_LOOK1,
    S_CENTAUR_LOOK2,
    S_CENTAUR_WALK1,
    S_CENTAUR_WALK2,
    S_CENTAUR_WALK3,
    S_CENTAUR_WALK4,
    S_CENTAUR_ATK1,
    S_CENTAUR_ATK2,
    S_CENTAUR_ATK3,
    S_CENTAUR_MISSILE1,
    S_CENTAUR_MISSILE2,
    S_CENTAUR_MISSILE3,
    S_CENTAUR_MISSILE4,
    S_CENTAUR_PAIN1,
    S_CENTAUR_PAIN2,
    S_CENTAUR_PAIN3,
    S_CENTAUR_PAIN4,
    S_CENTAUR_PAIN5,
    S_CENTAUR_PAIN6,
    S_CENTAUR_DEATH1,
    S_CENTAUR_DEATH2,
    S_CENTAUR_DEATH3,
    S_CENTAUR_DEATH4,
    S_CENTAUR_DEATH5,
    S_CENTAUR_DEATH6,
    S_CENTAUR_DEATH7,
    S_CENTAUR_DEATH8,
    S_CENTAUR_DEATH9,
    S_CENTAUR_DEATH0,
    S_CENTAUR_DEATH_X1,
    S_CENTAUR_DEATH_X2,
    S_CENTAUR_DEATH_X3,
    S_CENTAUR_DEATH_X4,
    S_CENTAUR_DEATH_X5,
    S_CENTAUR_DEATH_X6,
    S_CENTAUR_DEATH_X7,
    S_CENTAUR_DEATH_X8,
    S_CENTAUR_DEATH_X9,
    S_CENTAUR_DEATH_X10,
    S_CENTAUR_DEATH_X11,
    S_CENTAUR_ICE,
    S_CENTAUR_ICE2,
    S_CENTAUR_FX1,
    S_CENTAUR_FX_X1,
    S_CENTAUR_FX_X2,
    S_CENTAUR_FX_X3,
    S_CENTAUR_FX_X4,
    S_CENTAUR_FX_X5,
    S_CENTAUR_SHIELD1,
    S_CENTAUR_SHIELD2,
    S_CENTAUR_SHIELD3,
    S_CENTAUR_SHIELD4,
    S_CENTAUR_SHIELD5,
    S_CENTAUR_SHIELD6,
    S_CENTAUR_SHIELD_X1,
    S_CENTAUR_SHIELD_X2,
    S_CENTAUR_SHIELD_X3,
    S_CENTAUR_SHIELD_X4,
    S_CENTAUR_SWORD1,
    S_CENTAUR_SWORD2,
    S_CENTAUR_SWORD3,
    S_CENTAUR_SWORD4,
    S_CENTAUR_SWORD5,
    S_CENTAUR_SWORD6,
    S_CENTAUR_SWORD7,
    S_CENTAUR_SWORD_X1,
    S_CENTAUR_SWORD_X2,
    S_CENTAUR_SWORD_X3,
    S_DEMN_LOOK1,
    S_DEMN_LOOK2,
    S_DEMN_CHASE1,
    S_DEMN_CHASE2,
    S_DEMN_CHASE3,
    S_DEMN_CHASE4,
    S_DEMN_ATK1_1,
    S_DEMN_ATK1_2,
    S_DEMN_ATK1_3,
    S_DEMN_ATK2_1,
    S_DEMN_ATK2_2,
    S_DEMN_ATK2_3,
    S_DEMN_PAIN1,
    S_DEMN_PAIN2,
    S_DEMN_DEATH1,
    S_DEMN_DEATH2,
    S_DEMN_DEATH3,
    S_DEMN_DEATH4,
    S_DEMN_DEATH5,
    S_DEMN_DEATH6,
    S_DEMN_DEATH7,
    S_DEMN_DEATH8,
    S_DEMN_DEATH9,
    S_DEMN_XDEATH1,
    S_DEMN_XDEATH2,
    S_DEMN_XDEATH3,
    S_DEMN_XDEATH4,
    S_DEMN_XDEATH5,
    S_DEMN_XDEATH6,
    S_DEMN_XDEATH7,
    S_DEMN_XDEATH8,
    S_DEMN_XDEATH9,
    S_DEMON_ICE,
    S_DEMON_ICE2,
    S_DEMONCHUNK1_1,
    S_DEMONCHUNK1_2,
    S_DEMONCHUNK1_3,
    S_DEMONCHUNK1_4,
    S_DEMONCHUNK2_1,
    S_DEMONCHUNK2_2,
    S_DEMONCHUNK2_3,
    S_DEMONCHUNK2_4,
    S_DEMONCHUNK3_1,
    S_DEMONCHUNK3_2,
    S_DEMONCHUNK3_3,
    S_DEMONCHUNK3_4,
    S_DEMONCHUNK4_1,
    S_DEMONCHUNK4_2,
    S_DEMONCHUNK4_3,
    S_DEMONCHUNK4_4,
    S_DEMONCHUNK5_1,
    S_DEMONCHUNK5_2,
    S_DEMONCHUNK5_3,
    S_DEMONCHUNK5_4,
    S_DEMONFX_MOVE1,
    S_DEMONFX_MOVE2,
    S_DEMONFX_MOVE3,
    S_DEMONFX_BOOM1,
    S_DEMONFX_BOOM2,
    S_DEMONFX_BOOM3,
    S_DEMONFX_BOOM4,
    S_DEMONFX_BOOM5,
    S_DEMN2_LOOK1,
    S_DEMN2_LOOK2,
    S_DEMN2_CHASE1,
    S_DEMN2_CHASE2,
    S_DEMN2_CHASE3,
    S_DEMN2_CHASE4,
    S_DEMN2_ATK1_1,
    S_DEMN2_ATK1_2,
    S_DEMN2_ATK1_3,
    S_DEMN2_ATK2_1,
    S_DEMN2_ATK2_2,
    S_DEMN2_ATK2_3,
    S_DEMN2_PAIN1,
    S_DEMN2_PAIN2,
    S_DEMN2_DEATH1,
    S_DEMN2_DEATH2,
    S_DEMN2_DEATH3,
    S_DEMN2_DEATH4,
    S_DEMN2_DEATH5,
    S_DEMN2_DEATH6,
    S_DEMN2_DEATH7,
    S_DEMN2_DEATH8,
    S_DEMN2_DEATH9,
    S_DEMN2_XDEATH1,
    S_DEMN2_XDEATH2,
    S_DEMN2_XDEATH3,
    S_DEMN2_XDEATH4,
    S_DEMN2_XDEATH5,
    S_DEMN2_XDEATH6,
    S_DEMN2_XDEATH7,
    S_DEMN2_XDEATH8,
    S_DEMN2_XDEATH9,
    S_DEMON2CHUNK1_1,
    S_DEMON2CHUNK1_2,
    S_DEMON2CHUNK1_3,
    S_DEMON2CHUNK1_4,
    S_DEMON2CHUNK2_1,
    S_DEMON2CHUNK2_2,
    S_DEMON2CHUNK2_3,
    S_DEMON2CHUNK2_4,
    S_DEMON2CHUNK3_1,
    S_DEMON2CHUNK3_2,
    S_DEMON2CHUNK3_3,
    S_DEMON2CHUNK3_4,
    S_DEMON2CHUNK4_1,
    S_DEMON2CHUNK4_2,
    S_DEMON2CHUNK4_3,
    S_DEMON2CHUNK4_4,
    S_DEMON2CHUNK5_1,
    S_DEMON2CHUNK5_2,
    S_DEMON2CHUNK5_3,
    S_DEMON2CHUNK5_4,
    S_DEMON2FX_MOVE1,
    S_DEMON2FX_MOVE2,
    S_DEMON2FX_MOVE3,
    S_DEMON2FX_MOVE4,
    S_DEMON2FX_MOVE5,
    S_DEMON2FX_MOVE6,
    S_DEMON2FX_BOOM1,
    S_DEMON2FX_BOOM2,
    S_DEMON2FX_BOOM3,
    S_DEMON2FX_BOOM4,
    S_DEMON2FX_BOOM5,
    S_DEMON2FX_BOOM6,
    S_WRAITH_RAISE1,
    S_WRAITH_RAISE2,
    S_WRAITH_RAISE3,
    S_WRAITH_RAISE4,
    S_WRAITH_RAISE5,
    S_WRAITH_INIT1,
    S_WRAITH_INIT2,
    S_WRAITH_LOOK1,
    S_WRAITH_LOOK2,
    S_WRAITH_CHASE1,
    S_WRAITH_CHASE2,
    S_WRAITH_CHASE3,
    S_WRAITH_CHASE4,
    S_WRAITH_ATK1_1,
    S_WRAITH_ATK1_2,
    S_WRAITH_ATK1_3,
    S_WRAITH_ATK2_1,
    S_WRAITH_ATK2_2,
    S_WRAITH_ATK2_3,
    S_WRAITH_PAIN1,
    S_WRAITH_PAIN2,
    S_WRAITH_DEATH1_1,
    S_WRAITH_DEATH1_2,
    S_WRAITH_DEATH1_3,
    S_WRAITH_DEATH1_4,
    S_WRAITH_DEATH1_5,
    S_WRAITH_DEATH1_6,
    S_WRAITH_DEATH1_7,
    S_WRAITH_DEATH1_8,
    S_WRAITH_DEATH1_9,
    S_WRAITH_DEATH1_0,
    S_WRAITH_DEATH2_1,
    S_WRAITH_DEATH2_2,
    S_WRAITH_DEATH2_3,
    S_WRAITH_DEATH2_4,
    S_WRAITH_DEATH2_5,
    S_WRAITH_DEATH2_6,
    S_WRAITH_DEATH2_7,
    S_WRAITH_DEATH2_8,
    S_WRAITH_ICE,
    S_WRAITH_ICE2,
    S_WRTHFX_MOVE1,
    S_WRTHFX_MOVE2,
    S_WRTHFX_MOVE3,
    S_WRTHFX_BOOM1,
    S_WRTHFX_BOOM2,
    S_WRTHFX_BOOM3,
    S_WRTHFX_BOOM4,
    S_WRTHFX_BOOM5,
    S_WRTHFX_BOOM6,
    S_WRTHFX_SIZZLE1,
    S_WRTHFX_SIZZLE2,
    S_WRTHFX_SIZZLE3,
    S_WRTHFX_SIZZLE4,
    S_WRTHFX_SIZZLE5,
    S_WRTHFX_SIZZLE6,
    S_WRTHFX_SIZZLE7,
    S_WRTHFX_DROP1,
    S_WRTHFX_DROP2,
    S_WRTHFX_DROP3,
    S_WRTHFX_DEAD1,
    S_WRTHFX_ADROP1,
    S_WRTHFX_ADROP2,
    S_WRTHFX_ADROP3,
    S_WRTHFX_ADROP4,
    S_WRTHFX_ADEAD1,
    S_WRTHFX_BDROP1,
    S_WRTHFX_BDROP2,
    S_WRTHFX_BDROP3,
    S_WRTHFX_BDEAD1,
    S_MNTR_SPAWN1,
    S_MNTR_SPAWN2,
    S_MNTR_SPAWN3,
    S_MNTR_LOOK1,
    S_MNTR_LOOK2,
    S_MNTR_WALK1,
    S_MNTR_WALK2,
    S_MNTR_WALK3,
    S_MNTR_WALK4,
    S_MNTR_ROAM1,
    S_MNTR_ROAM2,
    S_MNTR_ROAM3,
    S_MNTR_ROAM4,
    S_MNTR_ATK1_1,
    S_MNTR_ATK1_2,
    S_MNTR_ATK1_3,
    S_MNTR_ATK2_1,
    S_MNTR_ATK2_2,
    S_MNTR_ATK2_3,
    S_MNTR_ATK3_1,
    S_MNTR_ATK3_2,
    S_MNTR_ATK3_3,
    S_MNTR_ATK3_4,
    S_MNTR_ATK4_1,
    S_MNTR_PAIN1,
    S_MNTR_PAIN2,
    S_MNTR_DIE1,
    S_MNTR_DIE2,
    S_MNTR_DIE3,
    S_MNTR_DIE4,
    S_MNTR_DIE5,
    S_MNTR_DIE6,
    S_MNTR_DIE7,
    S_MNTR_DIE8,
    S_MNTR_DIE9,
    S_MNTRFX1_1,
    S_MNTRFX1_2,
    S_MNTRFXI1_1,
    S_MNTRFXI1_2,
    S_MNTRFXI1_3,
    S_MNTRFXI1_4,
    S_MNTRFXI1_5,
    S_MNTRFXI1_6,
    S_MNTRFX2_1,
    S_MNTRFXI2_1,
    S_MNTRFXI2_2,
    S_MNTRFXI2_3,
    S_MNTRFXI2_4,
    S_MNTRFXI2_5,
    S_MNTRFX3_1,
    S_MNTRFX3_2,
    S_MNTRFX3_3,
    S_MNTRFX3_4,
    S_MNTRFX3_5,
    S_MNTRFX3_6,
    S_MNTRFX3_7,
    S_MNTRFX3_8,
    S_MNTRFX3_9,
    S_MINOSMOKE1,
    S_MINOSMOKE2,
    S_MINOSMOKE3,
    S_MINOSMOKE4,
    S_MINOSMOKE5,
    S_MINOSMOKE6,
    S_MINOSMOKE7,
    S_MINOSMOKE8,
    S_MINOSMOKE9,
    S_MINOSMOKE0,
    S_MINOSMOKEA,
    S_MINOSMOKEB,
    S_MINOSMOKEC,
    S_MINOSMOKED,
    S_MINOSMOKEE,
    S_MINOSMOKEF,
    S_MINOSMOKEG,
    S_MINOSMOKEX1,
    S_MINOSMOKEX2,
    S_MINOSMOKEX3,
    S_MINOSMOKEX4,
    S_MINOSMOKEX5,
    S_MINOSMOKEX6,
    S_MINOSMOKEX7,
    S_MINOSMOKEX8,
    S_MINOSMOKEX9,
    S_MINOSMOKEX0,
    S_MINOSMOKEXA,
    S_MINOSMOKEXB,
    S_MINOSMOKEXC,
    S_MINOSMOKEXD,
    S_MINOSMOKEXE,
    S_MINOSMOKEXF,
    S_MINOSMOKEXG,
    S_MINOSMOKEXH,
    S_MINOSMOKEXI,
    S_SERPENT_LOOK1,
    S_SERPENT_SWIM1,
    S_SERPENT_SWIM2,
    S_SERPENT_SWIM3,
    S_SERPENT_HUMP1,
    S_SERPENT_HUMP2,
    S_SERPENT_HUMP3,
    S_SERPENT_HUMP4,
    S_SERPENT_HUMP5,
    S_SERPENT_HUMP6,
    S_SERPENT_HUMP7,
    S_SERPENT_HUMP8,
    S_SERPENT_HUMP9,
    S_SERPENT_HUMP10,
    S_SERPENT_HUMP11,
    S_SERPENT_HUMP12,
    S_SERPENT_HUMP13,
    S_SERPENT_HUMP14,
    S_SERPENT_HUMP15,
    S_SERPENT_SURFACE1,
    S_SERPENT_SURFACE2,
    S_SERPENT_SURFACE3,
    S_SERPENT_SURFACE4,
    S_SERPENT_SURFACE5,
    S_SERPENT_DIVE1,
    S_SERPENT_DIVE2,
    S_SERPENT_DIVE3,
    S_SERPENT_DIVE4,
    S_SERPENT_DIVE5,
    S_SERPENT_DIVE6,
    S_SERPENT_DIVE7,
    S_SERPENT_DIVE8,
    S_SERPENT_DIVE9,
    S_SERPENT_DIVE10,
    S_SERPENT_WALK1,
    S_SERPENT_WALK2,
    S_SERPENT_WALK3,
    S_SERPENT_WALK4,
    S_SERPENT_PAIN1,
    S_SERPENT_PAIN2,
    S_SERPENT_ATK1,
    S_SERPENT_ATK2,
    S_SERPENT_MELEE1,
    S_SERPENT_MISSILE1,
    S_SERPENT_DIE1,
    S_SERPENT_DIE2,
    S_SERPENT_DIE3,
    S_SERPENT_DIE4,
    S_SERPENT_DIE5,
    S_SERPENT_DIE6,
    S_SERPENT_DIE7,
    S_SERPENT_DIE8,
    S_SERPENT_DIE9,
    S_SERPENT_DIE10,
    S_SERPENT_DIE11,
    S_SERPENT_DIE12,
    S_SERPENT_XDIE1,
    S_SERPENT_XDIE2,
    S_SERPENT_XDIE3,
    S_SERPENT_XDIE4,
    S_SERPENT_XDIE5,
    S_SERPENT_XDIE6,
    S_SERPENT_XDIE7,
    S_SERPENT_XDIE8,
    S_SERPENT_ICE,
    S_SERPENT_ICE2,
    S_SERPENT_FX1,
    S_SERPENT_FX2,
    S_SERPENT_FX3,
    S_SERPENT_FX4,
    S_SERPENT_FX_X1,
    S_SERPENT_FX_X2,
    S_SERPENT_FX_X3,
    S_SERPENT_FX_X4,
    S_SERPENT_FX_X5,
    S_SERPENT_FX_X6,
    S_SERPENT_HEAD1,
    S_SERPENT_HEAD2,
    S_SERPENT_HEAD3,
    S_SERPENT_HEAD4,
    S_SERPENT_HEAD5,
    S_SERPENT_HEAD6,
    S_SERPENT_HEAD7,
    S_SERPENT_HEAD8,
    S_SERPENT_HEAD_X1,
    S_SERPENT_GIB1_1,
    S_SERPENT_GIB1_2,
    S_SERPENT_GIB1_3,
    S_SERPENT_GIB1_4,
    S_SERPENT_GIB1_5,
    S_SERPENT_GIB1_6,
    S_SERPENT_GIB1_7,
    S_SERPENT_GIB1_8,
    S_SERPENT_GIB1_9,
    S_SERPENT_GIB1_10,
    S_SERPENT_GIB1_11,
    S_SERPENT_GIB1_12,
    S_SERPENT_GIB2_1,
    S_SERPENT_GIB2_2,
    S_SERPENT_GIB2_3,
    S_SERPENT_GIB2_4,
    S_SERPENT_GIB2_5,
    S_SERPENT_GIB2_6,
    S_SERPENT_GIB2_7,
    S_SERPENT_GIB2_8,
    S_SERPENT_GIB2_9,
    S_SERPENT_GIB2_10,
    S_SERPENT_GIB2_11,
    S_SERPENT_GIB2_12,
    S_SERPENT_GIB3_1,
    S_SERPENT_GIB3_2,
    S_SERPENT_GIB3_3,
    S_SERPENT_GIB3_4,
    S_SERPENT_GIB3_5,
    S_SERPENT_GIB3_6,
    S_SERPENT_GIB3_7,
    S_SERPENT_GIB3_8,
    S_SERPENT_GIB3_9,
    S_SERPENT_GIB3_10,
    S_SERPENT_GIB3_11,
    S_SERPENT_GIB3_12,
    S_BISHOP_LOOK1,
    S_BISHOP_DECIDE,
    S_BISHOP_BLUR1,
    S_BISHOP_BLUR2,
    S_BISHOP_WALK1,
    S_BISHOP_WALK2,
    S_BISHOP_WALK3,
    S_BISHOP_WALK4,
    S_BISHOP_WALK5,
    S_BISHOP_WALK6,
    S_BISHOP_ATK1,
    S_BISHOP_ATK2,
    S_BISHOP_ATK3,
    S_BISHOP_ATK4,
    S_BISHOP_ATK5,
    S_BISHOP_PAIN1,
    S_BISHOP_PAIN2,
    S_BISHOP_PAIN3,
    S_BISHOP_PAIN4,
    S_BISHOP_PAIN5,
    S_BISHOP_DEATH1,
    S_BISHOP_DEATH2,
    S_BISHOP_DEATH3,
    S_BISHOP_DEATH4,
    S_BISHOP_DEATH5,
    S_BISHOP_DEATH6,
    S_BISHOP_DEATH7,
    S_BISHOP_DEATH8,
    S_BISHOP_DEATH9,
    S_BISHOP_DEATH10,
    S_BISHOP_ICE,
    S_BISHOP_ICE2,
    S_BISHOP_PUFF1,
    S_BISHOP_PUFF2,
    S_BISHOP_PUFF3,
    S_BISHOP_PUFF4,
    S_BISHOP_PUFF5,
    S_BISHOP_PUFF6,
    S_BISHOP_PUFF7,
    S_BISHOPBLUR1,
    S_BISHOPBLUR2,
    S_BISHOPPAINBLUR1,
    S_BISHFX1_1,
    S_BISHFX1_2,
    S_BISHFX1_3,
    S_BISHFX1_4,
    S_BISHFX1_5,
    S_BISHFXI1_1,
    S_BISHFXI1_2,
    S_BISHFXI1_3,
    S_BISHFXI1_4,
    S_BISHFXI1_5,
    S_BISHFXI1_6,
    S_DRAGON_LOOK1,
    S_DRAGON_INIT,
    S_DRAGON_INIT2,
    S_DRAGON_INIT3,
    S_DRAGON_WALK1,
    S_DRAGON_WALK2,
    S_DRAGON_WALK3,
    S_DRAGON_WALK4,
    S_DRAGON_WALK5,
    S_DRAGON_WALK6,
    S_DRAGON_WALK7,
    S_DRAGON_WALK8,
    S_DRAGON_WALK9,
    S_DRAGON_WALK10,
    S_DRAGON_WALK11,
    S_DRAGON_WALK12,
    S_DRAGON_ATK1,
    S_DRAGON_PAIN1,
    S_DRAGON_DEATH1,
    S_DRAGON_DEATH2,
    S_DRAGON_DEATH3,
    S_DRAGON_DEATH4,
    S_DRAGON_CRASH1,
    S_DRAGON_CRASH2,
    S_DRAGON_CRASH3,
    S_DRAGON_FX1_1,
    S_DRAGON_FX1_2,
    S_DRAGON_FX1_3,
    S_DRAGON_FX1_4,
    S_DRAGON_FX1_5,
    S_DRAGON_FX1_6,
    S_DRAGON_FX1_X1,
    S_DRAGON_FX1_X2,
    S_DRAGON_FX1_X3,
    S_DRAGON_FX1_X4,
    S_DRAGON_FX1_X5,
    S_DRAGON_FX1_X6,
    S_DRAGON_FX2_1,
    S_DRAGON_FX2_2,
    S_DRAGON_FX2_3,
    S_DRAGON_FX2_4,
    S_DRAGON_FX2_5,
    S_DRAGON_FX2_6,
    S_DRAGON_FX2_7,
    S_DRAGON_FX2_8,
    S_DRAGON_FX2_9,
    S_DRAGON_FX2_10,
    S_DRAGON_FX2_11,
    S_ARMOR_1,
    S_ARMOR_2,
    S_ARMOR_3,
    S_ARMOR_4,
    S_MANA1_1,
    S_MANA1_2,
    S_MANA1_3,
    S_MANA1_4,
    S_MANA1_5,
    S_MANA1_6,
    S_MANA1_7,
    S_MANA1_8,
    S_MANA1_9,
    S_MANA2_1,
    S_MANA2_2,
    S_MANA2_3,
    S_MANA2_4,
    S_MANA2_5,
    S_MANA2_6,
    S_MANA2_7,
    S_MANA2_8,
    S_MANA2_9,
    S_MANA2_10,
    S_MANA2_11,
    S_MANA2_12,
    S_MANA2_13,
    S_MANA2_14,
    S_MANA2_15,
    S_MANA2_16,
    S_MANA3_1,
    S_MANA3_2,
    S_MANA3_3,
    S_MANA3_4,
    S_MANA3_5,
    S_MANA3_6,
    S_MANA3_7,
    S_MANA3_8,
    S_MANA3_9,
    S_MANA3_10,
    S_MANA3_11,
    S_MANA3_12,
    S_MANA3_13,
    S_MANA3_14,
    S_MANA3_15,
    S_MANA3_16,
    S_KEY1,
    S_KEY2,
    S_KEY3,
    S_KEY4,
    S_KEY5,
    S_KEY6,
    S_KEY7,
    S_KEY8,
    S_KEY9,
    S_KEYA,
    S_KEYB,
    S_SND_WIND1,
    S_SND_WIND2,
    S_SND_WATERFALL,
    S_ETTIN_LOOK1,
    S_ETTIN_LOOK2,
    S_ETTIN_CHASE1,
    S_ETTIN_CHASE2,
    S_ETTIN_CHASE3,
    S_ETTIN_CHASE4,
    S_ETTIN_PAIN1,
    S_ETTIN_ATK1_1,
    S_ETTIN_ATK1_2,
    S_ETTIN_ATK1_3,
    S_ETTIN_DEATH1_1,
    S_ETTIN_DEATH1_2,
    S_ETTIN_DEATH1_3,
    S_ETTIN_DEATH1_4,
    S_ETTIN_DEATH1_5,
    S_ETTIN_DEATH1_6,
    S_ETTIN_DEATH1_7,
    S_ETTIN_DEATH1_8,
    S_ETTIN_DEATH1_9,
    S_ETTIN_DEATH2_1,
    S_ETTIN_DEATH2_2,
    S_ETTIN_DEATH2_3,
    S_ETTIN_DEATH2_4,
    S_ETTIN_DEATH2_5,
    S_ETTIN_DEATH2_6,
    S_ETTIN_DEATH2_7,
    S_ETTIN_DEATH2_8,
    S_ETTIN_DEATH2_9,
    S_ETTIN_DEATH2_0,
    S_ETTIN_DEATH2_A,
    S_ETTIN_DEATH2_B,
    S_ETTIN_ICE1,
    S_ETTIN_ICE2,
    S_ETTIN_MACE1,
    S_ETTIN_MACE2,
    S_ETTIN_MACE3,
    S_ETTIN_MACE4,
    S_ETTIN_MACE5,
    S_ETTIN_MACE6,
    S_ETTIN_MACE7,
    S_FIRED_SPAWN1,
    S_FIRED_LOOK1,
    S_FIRED_LOOK2,
    S_FIRED_LOOK3,
    S_FIRED_LOOK4,
    S_FIRED_LOOK5,
    S_FIRED_LOOK6,
    S_FIRED_LOOK7,
    S_FIRED_LOOK8,
    S_FIRED_LOOK9,
    S_FIRED_LOOK0,
    S_FIRED_LOOKA,
    S_FIRED_LOOKB,
    S_FIRED_WALK1,
    S_FIRED_WALK2,
    S_FIRED_WALK3,
    S_FIRED_PAIN1,
    S_FIRED_ATTACK1,
    S_FIRED_ATTACK2,
    S_FIRED_ATTACK3,
    S_FIRED_ATTACK4,
    S_FIRED_DEATH1,
    S_FIRED_DEATH2,
    S_FIRED_DEATH3,
    S_FIRED_DEATH4,
    S_FIRED_XDEATH1,
    S_FIRED_XDEATH2,
    S_FIRED_XDEATH3,
    S_FIRED_ICE1,
    S_FIRED_ICE2,
    S_FIRED_CORPSE1,
    S_FIRED_CORPSE2,
    S_FIRED_CORPSE3,
    S_FIRED_CORPSE4,
    S_FIRED_CORPSE5,
    S_FIRED_CORPSE6,
    S_FIRED_RDROP1,
    S_FIRED_RDEAD1_1,
    S_FIRED_RDEAD1_2,
    S_FIRED_RDROP2,
    S_FIRED_RDEAD2_1,
    S_FIRED_RDEAD2_2,
    S_FIRED_RDROP3,
    S_FIRED_RDEAD3_1,
    S_FIRED_RDEAD3_2,
    S_FIRED_RDROP4,
    S_FIRED_RDEAD4_1,
    S_FIRED_RDEAD4_2,
    S_FIRED_RDROP5,
    S_FIRED_RDEAD5_1,
    S_FIRED_RDEAD5_2,
    S_FIRED_FX6_1,
    S_FIRED_FX6_2,
    S_FIRED_FX6_3,
    S_FIRED_FX6_4,
    S_FIRED_FX6_5,
    S_ICEGUY_LOOK,
    S_ICEGUY_DORMANT,
    S_ICEGUY_WALK1,
    S_ICEGUY_WALK2,
    S_ICEGUY_WALK3,
    S_ICEGUY_WALK4,
    S_ICEGUY_ATK1,
    S_ICEGUY_ATK2,
    S_ICEGUY_ATK3,
    S_ICEGUY_ATK4,
    S_ICEGUY_PAIN1,
    S_ICEGUY_DEATH,
    S_ICEGUY_FX1,
    S_ICEGUY_FX2,
    S_ICEGUY_FX3,
    S_ICEGUY_FX_X1,
    S_ICEGUY_FX_X2,
    S_ICEGUY_FX_X3,
    S_ICEGUY_FX_X4,
    S_ICEGUY_FX_X5,
    S_ICEFX_PUFF1,
    S_ICEFX_PUFF2,
    S_ICEFX_PUFF3,
    S_ICEFX_PUFF4,
    S_ICEFX_PUFF5,
    S_ICEGUY_FX2_1,
    S_ICEGUY_FX2_2,
    S_ICEGUY_FX2_3,
    S_ICEGUY_BIT1,
    S_ICEGUY_BIT2,
    S_ICEGUY_WISP1_1,
    S_ICEGUY_WISP1_2,
    S_ICEGUY_WISP1_3,
    S_ICEGUY_WISP1_4,
    S_ICEGUY_WISP1_5,
    S_ICEGUY_WISP1_6,
    S_ICEGUY_WISP1_7,
    S_ICEGUY_WISP1_8,
    S_ICEGUY_WISP1_9,
    S_ICEGUY_WISP2_1,
    S_ICEGUY_WISP2_2,
    S_ICEGUY_WISP2_3,
    S_ICEGUY_WISP2_4,
    S_ICEGUY_WISP2_5,
    S_ICEGUY_WISP2_6,
    S_ICEGUY_WISP2_7,
    S_ICEGUY_WISP2_8,
    S_ICEGUY_WISP2_9,
    S_FIGHTER,
    S_FIGHTER2,
    S_FIGHTERLOOK,
    S_FIGHTER_RUN1,
    S_FIGHTER_RUN2,
    S_FIGHTER_RUN3,
    S_FIGHTER_RUN4,
    S_FIGHTER_ATK1,
    S_FIGHTER_ATK2,
    S_FIGHTER_PAIN,
    S_FIGHTER_PAIN2,
    S_FIGHTER_DIE1,
    S_FIGHTER_DIE2,
    S_FIGHTER_DIE3,
    S_FIGHTER_DIE4,
    S_FIGHTER_DIE5,
    S_FIGHTER_DIE6,
    S_FIGHTER_DIE7,
    S_FIGHTER_XDIE1,
    S_FIGHTER_XDIE2,
    S_FIGHTER_XDIE3,
    S_FIGHTER_XDIE4,
    S_FIGHTER_XDIE5,
    S_FIGHTER_XDIE6,
    S_FIGHTER_XDIE7,
    S_FIGHTER_XDIE8,
    S_FIGHTER_ICE,
    S_FIGHTER_ICE2,
    S_CLERIC,
    S_CLERIC2,
    S_CLERICLOOK,
    S_CLERIC_RUN1,
    S_CLERIC_RUN2,
    S_CLERIC_RUN3,
    S_CLERIC_RUN4,
    S_CLERIC_ATK1,
    S_CLERIC_ATK2,
    S_CLERIC_ATK3,
    S_CLERIC_PAIN,
    S_CLERIC_PAIN2,
    S_CLERIC_DIE1,
    S_CLERIC_DIE2,
    S_CLERIC_DIE3,
    S_CLERIC_DIE4,
    S_CLERIC_DIE5,
    S_CLERIC_DIE6,
    S_CLERIC_DIE7,
    S_CLERIC_DIE8,
    S_CLERIC_DIE9,
    S_CLERIC_XDIE1,
    S_CLERIC_XDIE2,
    S_CLERIC_XDIE3,
    S_CLERIC_XDIE4,
    S_CLERIC_XDIE5,
    S_CLERIC_XDIE6,
    S_CLERIC_XDIE7,
    S_CLERIC_XDIE8,
    S_CLERIC_XDIE9,
    S_CLERIC_XDIE10,
    S_CLERIC_ICE,
    S_CLERIC_ICE2,
    S_MAGE,
    S_MAGE2,
    S_MAGELOOK,
    S_MAGE_RUN1,
    S_MAGE_RUN2,
    S_MAGE_RUN3,
    S_MAGE_RUN4,
    S_MAGE_ATK1,
    S_MAGE_ATK2,
    S_MAGE_PAIN,
    S_MAGE_PAIN2,
    S_MAGE_DIE1,
    S_MAGE_DIE2,
    S_MAGE_DIE3,
    S_MAGE_DIE4,
    S_MAGE_DIE5,
    S_MAGE_DIE6,
    S_MAGE_DIE7,
    S_MAGE_XDIE1,
    S_MAGE_XDIE2,
    S_MAGE_XDIE3,
    S_MAGE_XDIE4,
    S_MAGE_XDIE5,
    S_MAGE_XDIE6,
    S_MAGE_XDIE7,
    S_MAGE_XDIE8,
    S_MAGE_XDIE9,
    S_MAGE_ICE,
    S_MAGE_ICE2,
    S_SORC_SPAWN1,
    S_SORC_SPAWN2,
    S_SORC_LOOK1,
    S_SORC_WALK1,
    S_SORC_WALK2,
    S_SORC_WALK3,
    S_SORC_WALK4,
    S_SORC_PAIN1,
    S_SORC_PAIN2,
    S_SORC_ATK2_1,
    S_SORC_ATK2_2,
    S_SORC_ATK2_3,
    S_SORC_ATTACK1,
    S_SORC_ATTACK2,
    S_SORC_ATTACK3,
    S_SORC_ATTACK4,
    S_SORC_ATTACK5,
    S_SORC_DIE1,
    S_SORC_DIE2,
    S_SORC_DIE3,
    S_SORC_DIE4,
    S_SORC_DIE5,
    S_SORC_DIE6,
    S_SORC_DIE7,
    S_SORC_DIE8,
    S_SORC_DIE9,
    S_SORC_DIE0,
    S_SORC_DIEA,
    S_SORC_DIEB,
    S_SORC_DIEC,
    S_SORC_DIED,
    S_SORC_DIEE,
    S_SORC_DIEF,
    S_SORC_DIEG,
    S_SORC_DIEH,
    S_SORC_DIEI,
    S_SORCBALL1_1,
    S_SORCBALL1_2,
    S_SORCBALL1_3,
    S_SORCBALL1_4,
    S_SORCBALL1_5,
    S_SORCBALL1_6,
    S_SORCBALL1_7,
    S_SORCBALL1_8,
    S_SORCBALL1_9,
    S_SORCBALL1_0,
    S_SORCBALL1_A,
    S_SORCBALL1_B,
    S_SORCBALL1_C,
    S_SORCBALL1_D,
    S_SORCBALL1_E,
    S_SORCBALL1_F,
    S_SORCBALL1_D1,
    S_SORCBALL1_D2,
    S_SORCBALL1_D5,
    S_SORCBALL1_D6,
    S_SORCBALL1_D7,
    S_SORCBALL1_D8,
    S_SORCBALL1_D9,
    S_SORCBALL2_1,
    S_SORCBALL2_2,
    S_SORCBALL2_3,
    S_SORCBALL2_4,
    S_SORCBALL2_5,
    S_SORCBALL2_6,
    S_SORCBALL2_7,
    S_SORCBALL2_8,
    S_SORCBALL2_9,
    S_SORCBALL2_0,
    S_SORCBALL2_A,
    S_SORCBALL2_B,
    S_SORCBALL2_C,
    S_SORCBALL2_D,
    S_SORCBALL2_E,
    S_SORCBALL2_F,
    S_SORCBALL2_D1,
    S_SORCBALL2_D2,
    S_SORCBALL2_D5,
    S_SORCBALL2_D6,
    S_SORCBALL2_D7,
    S_SORCBALL2_D8,
    S_SORCBALL2_D9,
    S_SORCBALL3_1,
    S_SORCBALL3_2,
    S_SORCBALL3_3,
    S_SORCBALL3_4,
    S_SORCBALL3_5,
    S_SORCBALL3_6,
    S_SORCBALL3_7,
    S_SORCBALL3_8,
    S_SORCBALL3_9,
    S_SORCBALL3_0,
    S_SORCBALL3_A,
    S_SORCBALL3_B,
    S_SORCBALL3_C,
    S_SORCBALL3_D,
    S_SORCBALL3_E,
    S_SORCBALL3_F,
    S_SORCBALL3_D1,
    S_SORCBALL3_D2,
    S_SORCBALL3_D5,
    S_SORCBALL3_D6,
    S_SORCBALL3_D7,
    S_SORCBALL3_D8,
    S_SORCBALL3_D9,
    S_SORCFX1_1,
    S_SORCFX1_2,
    S_SORCFX1_3,
    S_SORCFX1_4,
    S_SORCFX1_D1,
    S_SORCFX1_D2,
    S_SORCFX1_D3,
    S_SORCFX2_SPLIT1,
    S_SORCFX2_ORBIT1,
    S_SORCFX2_ORBIT2,
    S_SORCFX2_ORBIT3,
    S_SORCFX2_ORBIT4,
    S_SORCFX2_ORBIT5,
    S_SORCFX2_ORBIT6,
    S_SORCFX2_ORBIT7,
    S_SORCFX2_ORBIT8,
    S_SORCFX2_ORBIT9,
    S_SORCFX2_ORBIT0,
    S_SORCFX2_ORBITA,
    S_SORCFX2_ORBITB,
    S_SORCFX2_ORBITC,
    S_SORCFX2_ORBITD,
    S_SORCFX2_ORBITE,
    S_SORCFX2_ORBITF,
    S_SORCFX2T1,
    S_SORCFX3_1,
    S_SORCFX3_2,
    S_SORCFX3_3,
    S_BISHMORPH1,
    S_BISHMORPHA,
    S_BISHMORPHB,
    S_BISHMORPHC,
    S_BISHMORPHD,
    S_BISHMORPHE,
    S_BISHMORPHF,
    S_BISHMORPHG,
    S_BISHMORPHH,
    S_BISHMORPHI,
    S_BISHMORPHJ,
    S_SORCFX3_EXP1,
    S_SORCFX3_EXP2,
    S_SORCFX3_EXP3,
    S_SORCFX3_EXP4,
    S_SORCFX3_EXP5,
    S_SORCFX4_1,
    S_SORCFX4_2,
    S_SORCFX4_3,
    S_SORCFX4_D1,
    S_SORCFX4_D2,
    S_SORCFX4_D3,
    S_SORCFX4_D4,
    S_SORCFX4_D5,
    S_SORCSPARK1,
    S_SORCSPARK2,
    S_SORCSPARK3,
    S_SORCSPARK4,
    S_SORCSPARK5,
    S_SORCSPARK6,
    S_SORCSPARK7,
    S_BLASTEFFECT1,
    S_BLASTEFFECT2,
    S_BLASTEFFECT3,
    S_BLASTEFFECT4,
    S_BLASTEFFECT5,
    S_BLASTEFFECT6,
    S_BLASTEFFECT7,
    S_BLASTEFFECT8,
    S_BLASTEFFECT9,
    S_WATERDRIP1,
    S_KORAX_LOOK1,
    S_KORAX_CHASE1,
    S_KORAX_CHASE2,
    S_KORAX_CHASE3,
    S_KORAX_CHASE4,
    S_KORAX_CHASE5,
    S_KORAX_CHASE6,
    S_KORAX_CHASE7,
    S_KORAX_CHASE8,
    S_KORAX_CHASE9,
    S_KORAX_CHASE0,
    S_KORAX_CHASEA,
    S_KORAX_CHASEB,
    S_KORAX_CHASEC,
    S_KORAX_CHASED,
    S_KORAX_CHASEE,
    S_KORAX_CHASEF,
    S_KORAX_PAIN1,
    S_KORAX_PAIN2,
    S_KORAX_ATTACK1,
    S_KORAX_ATTACK2,
    S_KORAX_MISSILE1,
    S_KORAX_MISSILE2,
    S_KORAX_MISSILE3,
    S_KORAX_COMMAND1,
    S_KORAX_COMMAND2,
    S_KORAX_COMMAND3,
    S_KORAX_COMMAND4,
    S_KORAX_COMMAND5,
    S_KORAX_DEATH1,
    S_KORAX_DEATH2,
    S_KORAX_DEATH3,
    S_KORAX_DEATH4,
    S_KORAX_DEATH5,
    S_KORAX_DEATH6,
    S_KORAX_DEATH7,
    S_KORAX_DEATH8,
    S_KORAX_DEATH9,
    S_KORAX_DEATH0,
    S_KORAX_DEATHA,
    S_KORAX_DEATHB,
    S_KORAX_DEATHC,
    S_KORAX_DEATHD,
    S_KSPIRIT_ROAM1,
    S_KSPIRIT_ROAM2,
    S_KSPIRIT_DEATH1,
    S_KSPIRIT_DEATH2,
    S_KSPIRIT_DEATH3,
    S_KSPIRIT_DEATH4,
    S_KSPIRIT_DEATH5,
    S_KSPIRIT_DEATH6,
    S_KBOLT1,
    S_KBOLT2,
    S_KBOLT3,
    S_KBOLT4,
    S_KBOLT5,
    S_KBOLT6,
    S_KBOLT7,
    S_SPAWNBATS1,
    S_SPAWNBATS2,
    S_SPAWNBATS3,
    S_SPAWNBATS_OFF,
    S_BAT1,
    S_BAT2,
    S_BAT3,
    S_BAT_DEATH,
    NUMSTATES
} statenum_t;


struct mobj_s;
struct player_s;
struct pspdef_s;


typedef struct
{
    spritenum_t sprite;
    int frame;
    int tics;
    void (*action) (struct mobj_s *, struct player_s *, struct pspdef_s *);
    statenum_t nextstate;
    int misc1, misc2;
} state_t;

extern state_t states[NUMSTATES];
extern const char *sprnames[];



typedef enum
{
    MT_MAPSPOT,
    MT_MAPSPOTGRAVITY,
    MT_FIREBALL1,
    MT_ARROW,
    MT_DART,
    MT_POISONDART,
    MT_RIPPERBALL,
    MT_PROJECTILE_BLADE,
    MT_ICESHARD,
    MT_FLAME_SMALL_TEMP,
    MT_FLAME_LARGE_TEMP,
    MT_FLAME_SMALL,
    MT_FLAME_LARGE,
    MT_HEALINGBOTTLE,
    MT_HEALTHFLASK,
    MT_ARTIFLY,
    MT_ARTIINVULNERABILITY,
    MT_SUMMONMAULATOR,
    MT_SUMMON_FX,
    MT_THRUSTFLOOR_UP,
    MT_THRUSTFLOOR_DOWN,
    MT_TELEPORTOTHER,
    MT_TELOTHER_FX1,
    MT_TELOTHER_FX2,
    MT_TELOTHER_FX3,
    MT_TELOTHER_FX4,
    MT_TELOTHER_FX5,
    MT_DIRT1,
    MT_DIRT2,
    MT_DIRT3,
    MT_DIRT4,
    MT_DIRT5,
    MT_DIRT6,
    MT_DIRTCLUMP,
    MT_ROCK1,
    MT_ROCK2,
    MT_ROCK3,
    MT_FOGSPAWNER,
    MT_FOGPATCHS,
    MT_FOGPATCHM,
    MT_FOGPATCHL,
    MT_QUAKE_FOCUS,
    MT_SGSHARD1,
    MT_SGSHARD2,
    MT_SGSHARD3,
    MT_SGSHARD4,
    MT_SGSHARD5,
    MT_SGSHARD6,
    MT_SGSHARD7,
    MT_SGSHARD8,
    MT_SGSHARD9,
    MT_SGSHARD0,
    MT_ARTIEGG,
    MT_EGGFX,
    MT_ARTISUPERHEAL,
    MT_ZWINGEDSTATUENOSKULL,
    MT_ZGEMPEDESTAL,
    MT_ARTIPUZZSKULL,
    MT_ARTIPUZZGEMBIG,
    MT_ARTIPUZZGEMRED,
    MT_ARTIPUZZGEMGREEN1,
    MT_ARTIPUZZGEMGREEN2,
    MT_ARTIPUZZGEMBLUE1,
    MT_ARTIPUZZGEMBLUE2,
    MT_ARTIPUZZBOOK1,
    MT_ARTIPUZZBOOK2,
    MT_ARTIPUZZSKULL2,
    MT_ARTIPUZZFWEAPON,
    MT_ARTIPUZZCWEAPON,
    MT_ARTIPUZZMWEAPON,
    MT_ARTIPUZZGEAR,
    MT_ARTIPUZZGEAR2,
    MT_ARTIPUZZGEAR3,
    MT_ARTIPUZZGEAR4,
    MT_ARTITORCH,
    MT_FIREBOMB,
    MT_ARTITELEPORT,
    MT_ARTIPOISONBAG,
    MT_POISONBAG,
    MT_POISONCLOUD,
    MT_THROWINGBOMB,
    MT_SPEEDBOOTS,
    MT_BOOSTMANA,
    MT_BOOSTARMOR,
    MT_BLASTRADIUS,
    MT_HEALRADIUS,
    MT_SPLASH,
    MT_SPLASHBASE,
    MT_LAVASPLASH,
    MT_LAVASMOKE,
    MT_SLUDGECHUNK,
    MT_SLUDGESPLASH,
    MT_MISC0,
    MT_MISC1,
    MT_MISC2,
    MT_MISC3,
    MT_MISC4,
    MT_MISC5,
    MT_MISC6,
    MT_MISC7,
    MT_MISC8,
    MT_TREEDESTRUCTIBLE,
    MT_MISC9,
    MT_MISC10,
    MT_MISC11,
    MT_MISC12,
    MT_MISC13,
    MT_MISC14,
    MT_MISC15,
    MT_MISC16,
    MT_MISC17,
    MT_MISC18,
    MT_MISC19,
    MT_MISC20,
    MT_MISC21,
    MT_MISC22,
    MT_MISC23,
    MT_MISC24,
    MT_MISC25,
    MT_MISC26,
    MT_MISC27,
    MT_MISC28,
    MT_MISC29,
    MT_MISC30,
    MT_MISC31,
    MT_MISC32,
    MT_MISC33,
    MT_MISC34,
    MT_MISC35,
    MT_MISC36,
    MT_MISC37,
    MT_MISC38,
    MT_MISC39,
    MT_MISC40,
    MT_MISC41,
    MT_MISC42,
    MT_MISC43,
    MT_MISC44,
    MT_MISC45,
    MT_MISC46,
    MT_MISC47,
    MT_MISC48,
    MT_MISC49,
    MT_MISC50,
    MT_MISC51,
    MT_MISC52,
    MT_MISC53,
    MT_MISC54,
    MT_MISC55,
    MT_MISC56,
    MT_MISC57,
    MT_MISC58,
    MT_MISC59,
    MT_MISC60,
    MT_MISC61,
    MT_MISC62,
    MT_MISC63,
    MT_MISC64,
    MT_MISC65,
    MT_MISC66,
    MT_MISC67,
    MT_MISC68,
    MT_MISC69,
    MT_MISC70,
    MT_MISC71,
    MT_MISC72,
    MT_MISC73,
    MT_MISC74,
    MT_MISC75,
    MT_MISC76,
    MT_POTTERY1,
    MT_POTTERY2,
    MT_POTTERY3,
    MT_POTTERYBIT1,
    MT_MISC77,
    MT_ZLYNCHED_NOHEART,
    MT_MISC78,
    MT_CORPSEBIT,
    MT_CORPSEBLOODDRIP,
    MT_BLOODPOOL,
    MT_MISC79,
    MT_MISC80,
    MT_LEAF1,
    MT_LEAF2,
    MT_ZTWINEDTORCH,
    MT_ZTWINEDTORCH_UNLIT,
    MT_BRIDGE,
    MT_BRIDGEBALL,
    MT_ZWALLTORCH,
    MT_ZWALLTORCH_UNLIT,
    MT_ZBARREL,
    MT_ZSHRUB1,
    MT_ZSHRUB2,
    MT_ZBUCKET,
    MT_ZPOISONSHROOM,
    MT_ZFIREBULL,
    MT_ZFIREBULL_UNLIT,
    MT_FIRETHING,
    MT_BRASSTORCH,
    MT_ZSUITOFARMOR,
    MT_ZARMORCHUNK,
    MT_ZBELL,
    MT_ZBLUE_CANDLE,
    MT_ZIRON_MAIDEN,
    MT_ZXMAS_TREE,
    MT_ZCAULDRON,
    MT_ZCAULDRON_UNLIT,
    MT_ZCHAINBIT32,
    MT_ZCHAINBIT64,
    MT_ZCHAINEND_HEART,
    MT_ZCHAINEND_HOOK1,
    MT_ZCHAINEND_HOOK2,
    MT_ZCHAINEND_SPIKE,
    MT_ZCHAINEND_SKULL,
    MT_TABLE_SHIT1,
    MT_TABLE_SHIT2,
    MT_TABLE_SHIT3,
    MT_TABLE_SHIT4,
    MT_TABLE_SHIT5,
    MT_TABLE_SHIT6,
    MT_TABLE_SHIT7,
    MT_TABLE_SHIT8,
    MT_TABLE_SHIT9,
    MT_TABLE_SHIT10,
    MT_TFOG,
    MT_MISC81,
    MT_TELEPORTMAN,
    MT_PUNCHPUFF,
    MT_FW_AXE,
    MT_AXEPUFF,
    MT_AXEPUFF_GLOW,
    MT_AXEBLOOD,
    MT_FW_HAMMER,
    MT_HAMMER_MISSILE,
    MT_HAMMERPUFF,
    MT_FSWORD_MISSILE,
    MT_FSWORD_FLAME,
    MT_CW_SERPSTAFF,
    MT_CSTAFF_MISSILE,
    MT_CSTAFFPUFF,
    MT_CW_FLAME,
    MT_CFLAMEFLOOR,
    MT_FLAMEPUFF,
    MT_FLAMEPUFF2,
    MT_CIRCLEFLAME,
    MT_CFLAME_MISSILE,
    MT_HOLY_FX,
    MT_HOLY_TAIL,
    MT_HOLY_PUFF,
    MT_HOLY_MISSILE,
    MT_HOLY_MISSILE_PUFF,
    MT_MWANDPUFF,
    MT_MWANDSMOKE,
    MT_MWAND_MISSILE,
    MT_MW_LIGHTNING,
    MT_LIGHTNING_CEILING,
    MT_LIGHTNING_FLOOR,
    MT_LIGHTNING_ZAP,
    MT_MSTAFF_FX,
    MT_MSTAFF_FX2,
    MT_FW_SWORD1,
    MT_FW_SWORD2,
    MT_FW_SWORD3,
    MT_CW_HOLY1,
    MT_CW_HOLY2,
    MT_CW_HOLY3,
    MT_MW_STAFF1,
    MT_MW_STAFF2,
    MT_MW_STAFF3,
    MT_SNOUTPUFF,
    MT_MW_CONE,
    MT_SHARDFX1,
    MT_BLOOD,
    MT_BLOODSPLATTER,
    MT_GIBS,
    MT_PLAYER_FIGHTER,
    MT_BLOODYSKULL,
    MT_PLAYER_SPEED,
    MT_ICECHUNK,
    MT_PLAYER_CLERIC,
    MT_PLAYER_MAGE,
    MT_PIGPLAYER,
    MT_PIG,
    MT_CENTAUR,
    MT_CENTAURLEADER,
    MT_CENTAUR_FX,
    MT_CENTAUR_SHIELD,
    MT_CENTAUR_SWORD,
    MT_DEMON,
    MT_DEMONCHUNK1,
    MT_DEMONCHUNK2,
    MT_DEMONCHUNK3,
    MT_DEMONCHUNK4,
    MT_DEMONCHUNK5,
    MT_DEMONFX1,
    MT_DEMON2,
    MT_DEMON2CHUNK1,
    MT_DEMON2CHUNK2,
    MT_DEMON2CHUNK3,
    MT_DEMON2CHUNK4,
    MT_DEMON2CHUNK5,
    MT_DEMON2FX1,
    MT_WRAITHB,
    MT_WRAITH,
    MT_WRAITHFX1,
    MT_WRAITHFX2,
    MT_WRAITHFX3,
    MT_WRAITHFX4,
    MT_WRAITHFX5,
    MT_MINOTAUR,
    MT_MNTRFX1,
    MT_MNTRFX2,
    MT_MNTRFX3,
    MT_MNTRSMOKE,
    MT_MNTRSMOKEEXIT,
    MT_SERPENT,
    MT_SERPENTLEADER,
    MT_SERPENTFX,
    MT_SERPENT_HEAD,
    MT_SERPENT_GIB1,
    MT_SERPENT_GIB2,
    MT_SERPENT_GIB3,
    MT_BISHOP,
    MT_BISHOP_PUFF,
    MT_BISHOPBLUR,
    MT_BISHOPPAINBLUR,
    MT_BISH_FX,
    MT_DRAGON,
    MT_DRAGON_FX,
    MT_DRAGON_FX2,
    MT_ARMOR_1,
    MT_ARMOR_2,
    MT_ARMOR_3,
    MT_ARMOR_4,
    MT_MANA1,
    MT_MANA2,
    MT_MANA3,
    MT_KEY1,
    MT_KEY2,
    MT_KEY3,
    MT_KEY4,
    MT_KEY5,
    MT_KEY6,
    MT_KEY7,
    MT_KEY8,
    MT_KEY9,
    MT_KEYA,
    MT_KEYB,
    MT_SOUNDWIND,
    MT_SOUNDWATERFALL,
    MT_ETTIN,
    MT_ETTIN_MACE,
    MT_FIREDEMON,
    MT_FIREDEMON_SPLOTCH1,
    MT_FIREDEMON_SPLOTCH2,
    MT_FIREDEMON_FX1,
    MT_FIREDEMON_FX2,
    MT_FIREDEMON_FX3,
    MT_FIREDEMON_FX4,
    MT_FIREDEMON_FX5,
    MT_FIREDEMON_FX6,
    MT_ICEGUY,
    MT_ICEGUY_FX,
    MT_ICEFX_PUFF,
    MT_ICEGUY_FX2,
    MT_ICEGUY_BIT,
    MT_ICEGUY_WISP1,
    MT_ICEGUY_WISP2,
    MT_FIGHTER_BOSS,
    MT_CLERIC_BOSS,
    MT_MAGE_BOSS,
    MT_SORCBOSS,
    MT_SORCBALL1,
    MT_SORCBALL2,
    MT_SORCBALL3,
    MT_SORCFX1,
    MT_SORCFX2,
    MT_SORCFX2_T1,
    MT_SORCFX3,
    MT_SORCFX3_EXPLOSION,
    MT_SORCFX4,
    MT_SORCSPARK1,
    MT_BLASTEFFECT,
    MT_WATER_DRIP,
    MT_KORAX,
    MT_KORAX_SPIRIT1,
    MT_KORAX_SPIRIT2,
    MT_KORAX_SPIRIT3,
    MT_KORAX_SPIRIT4,
    MT_KORAX_SPIRIT5,
    MT_KORAX_SPIRIT6,
    MT_DEMON_MASH,
    MT_DEMON2_MASH,
    MT_ETTIN_MASH,
    MT_CENTAUR_MASH,
    MT_KORAX_BOLT,
    MT_BAT_SPAWNER,
    MT_BAT,
    NUMMOBJTYPES
} mobjtype_t;

typedef struct
{
    int doomednum;
    int spawnstate;
    int spawnhealth;
    int seestate;
    int seesound;
    int reactiontime;
    int attacksound;
    int painstate;
    int painchance;
    int painsound;
    int meleestate;
    int missilestate;
    int crashstate;
    int deathstate;
    int xdeathstate;
    int deathsound;
    int speed;
    int radius;
    int height;
    int mass;
    int damage;
    int activesound;
    int flags;
    int flags2;
} mobjinfo_t;

extern mobjinfo_t mobjinfo[NUMMOBJTYPES];
