//
// Copyright(C) 1993-1996 Id Software, Inc.
// Copyright(C) 1993-2008 Raven Software
// Copyright(C) 2005-2014 Simon Howard
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
#include "h2def.h"
#include "i_swap.h"
#include "p_local.h"

// generated by stateco

const char *sprnames[] = {
    "MAN1","ACLO","TLGL","FBL1","XPL1","ARRW","DART","RIPP","CFCF","BLAD",
    "SHRD","FFSM","FFLG","PTN1","PTN2","SOAR","INVU","SUMN","TSPK","TELO",
    "TRNG","ROCK","FOGS","FOGM","FOGL","SGSA","SGSB","PORK","EGGM","FHFX",
    "SPHL","STWN","GMPD","ASKU","ABGM","AGMR","AGMG","AGG2","AGMB","AGB2",
    "ABK1","ABK2","ASK2","AFWP","ACWP","AMWP","AGER","AGR2","AGR3","AGR4",
    "TRCH","PSBG","ATLP","THRW","SPED","BMAN","BRAC","BLST","HRAD","SPSH",
    "LVAS","SLDG","STTW","RCK1","RCK2","RCK3","RCK4","CDLR","TRE1","TRDT",
    "TRE2","TRE3","STM1","STM2","STM3","STM4","MSH1","MSH2","MSH3","MSH4",
    "MSH5","MSH6","MSH7","MSH8","SGMP","SGM1","SGM2","SGM3","SLC1","SLC2",
    "SLC3","MSS1","MSS2","SWMV","CPS1","CPS2","TMS1","TMS2","TMS3","TMS4",
    "TMS5","TMS6","TMS7","CPS3","STT2","STT3","STT4","STT5","GAR1","GAR2",
    "GAR3","GAR4","GAR5","GAR6","GAR7","GAR8","GAR9","BNR1","TRE4","TRE5",
    "TRE6","TRE7","LOGG","ICT1","ICT2","ICT3","ICT4","ICM1","ICM2","ICM3",
    "ICM4","RKBL","RKBS","RKBK","RBL1","RBL2","RBL3","VASE","POT1","POT2",
    "POT3","PBIT","CPS4","CPS5","CPS6","CPB1","CPB2","CPB3","CPB4","BDRP",
    "BDSH","BDPL","CNDL","LEF1","LEF3","LEF2","TWTR","WLTR","BARL","SHB1",
    "SHB2","BCKT","SHRM","FBUL","FSKL","BRTR","SUIT","BBLL","CAND","IRON",
    "XMAS","CDRN","CHNS","TST1","TST2","TST3","TST4","TST5","TST6","TST7",
    "TST8","TST9","TST0","TELE","TSMK","FPCH","WFAX","FAXE","WFHM","FHMR",
    "FSRD","FSFX","CMCE","WCSS","CSSF","WCFM","CFLM","CFFX","CHLY","SPIR",
    "MWND","WMLG","MLNG","MLFX","MLF2","MSTF","MSP1","MSP2","WFR1","WFR2",
    "WFR3","WCH1","WCH2","WCH3","WMS1","WMS2","WMS3","WPIG","WMCS","CONE",
    "SHEX","BLOD","GIBS","PLAY","FDTH","BSKL","ICEC","CLER","MAGE","PIGY",
    "CENT","CTXD","CTFX","CTDP","DEMN","DEMA","DEMB","DEMC","DEMD","DEME",
    "DMFX","DEM2","DMBA","DMBB","DMBC","DMBD","DMBE","D2FX","WRTH","WRT2",
    "WRBL","MNTR","FX12","FX13","MNSM","SSPT","SSDV","SSXD","SSFX","BISH",
    "BPFX","DRAG","DRFX","ARM1","ARM2","ARM3","ARM4","MAN2","MAN3","KEY1",
    "KEY2","KEY3","KEY4","KEY5","KEY6","KEY7","KEY8","KEY9","KEYA","KEYB",
    "ETTN","ETTB","FDMN","FDMB","ICEY","ICPR","ICWS","SORC","SBMP","SBS4",
    "SBMB","SBS3","SBMG","SBS1","SBS2","SBFX","RADE","WATR","KORX","ABAT",
    NULL
};


state_t states[NUMSTATES] = {
    {SPR_MAN1, 0, -1, NULL, S_NULL, 0, 0},      // S_NULL
    {SPR_ACLO, 4, 1050, A_FreeTargMobj, S_NULL, 0, 0},  // S_FREETARGMOBJ
    {SPR_TLGL, 0, -1, NULL, S_NULL, 0, 0},      // S_MAPSPOT
    {SPR_FBL1, 32768, 4, NULL, S_FIREBALL1_2, 0, 0},    // S_FIREBALL1_1
    {SPR_FBL1, 32769, 4, NULL, S_FIREBALL1_1, 0, 0},    // S_FIREBALL1_2
    {SPR_XPL1, 32768, 4, NULL, S_FIREBALL1_X2, 0, 0},   // S_FIREBALL1_X1
    {SPR_XPL1, 32769, 4, NULL, S_FIREBALL1_X3, 0, 0},   // S_FIREBALL1_X2
    {SPR_XPL1, 32770, 4, NULL, S_FIREBALL1_X4, 0, 0},   // S_FIREBALL1_X3
    {SPR_XPL1, 32771, 4, NULL, S_FIREBALL1_X5, 0, 0},   // S_FIREBALL1_X4
    {SPR_XPL1, 32772, 4, NULL, S_FIREBALL1_X6, 0, 0},   // S_FIREBALL1_X5
    {SPR_XPL1, 32773, 4, NULL, S_NULL, 0, 0},   // S_FIREBALL1_X6
    {SPR_ARRW, 0, -1, NULL, S_NULL, 0, 0},      // S_ARROW_1
    {SPR_ARRW, 0, 1, NULL, S_NULL, 0, 0},       // S_ARROW_X1
    {SPR_DART, 0, -1, NULL, S_NULL, 0, 0},      // S_DART_1
    {SPR_DART, 0, 1, NULL, S_NULL, 0, 0},       // S_DART_X1
    {SPR_DART, 0, -1, NULL, S_NULL, 0, 0},      // S_POISONDART_1
    {SPR_DART, 0, 1, NULL, S_NULL, 0, 0},       // S_POISONDART_X1
    {SPR_RIPP, 0, 3, NULL, S_RIPPERBALL_2, 0, 0},       // S_RIPPERBALL_1
    {SPR_RIPP, 1, 3, NULL, S_RIPPERBALL_3, 0, 0},       // S_RIPPERBALL_2
    {SPR_RIPP, 2, 3, NULL, S_RIPPERBALL_1, 0, 0},       // S_RIPPERBALL_3
    {SPR_CFCF, 32784, 4, NULL, S_RIPPERBALL_X2, 0, 0},  // S_RIPPERBALL_X1
    {SPR_CFCF, 32785, 3, NULL, S_RIPPERBALL_X3, 0, 0},  // S_RIPPERBALL_X2
    {SPR_CFCF, 32786, 4, NULL, S_RIPPERBALL_X4, 0, 0},  // S_RIPPERBALL_X3
    {SPR_CFCF, 32787, 3, NULL, S_RIPPERBALL_X5, 0, 0},  // S_RIPPERBALL_X4
    {SPR_CFCF, 32788, 4, NULL, S_RIPPERBALL_X6, 0, 0},  // S_RIPPERBALL_X5
    {SPR_CFCF, 32789, 3, NULL, S_RIPPERBALL_X7, 0, 0},  // S_RIPPERBALL_X6
    {SPR_CFCF, 32790, 4, NULL, S_RIPPERBALL_X8, 0, 0},  // S_RIPPERBALL_X7
    {SPR_CFCF, 32791, 3, NULL, S_RIPPERBALL_X9, 0, 0},  // S_RIPPERBALL_X8
    {SPR_CFCF, 32792, 4, NULL, S_RIPPERBALL_X10, 0, 0}, // S_RIPPERBALL_X9
    {SPR_CFCF, 32793, 3, NULL, S_NULL, 0, 0},   // S_RIPPERBALL_X10
    {SPR_BLAD, 0, -1, NULL, S_NULL, 0, 0},      // S_PRJ_BLADE1
    {SPR_BLAD, 0, 1, NULL, S_NULL, 0, 0},       // S_PRJ_BLADE_X1
    {SPR_SHRD, 32768, 3, NULL, S_ICESHARD2, 0, 0},      // S_ICESHARD1
    {SPR_SHRD, 32769, 3, NULL, S_ICESHARD3, 0, 0},      // S_ICESHARD2
    {SPR_SHRD, 32770, 3, NULL, S_ICESHARD1, 0, 0},      // S_ICESHARD3
    {SPR_FFSM, 32768, 3, NULL, S_FLAME_TSMALL2, 0, 0},  // S_FLAME_TSMALL1
    {SPR_FFSM, 32769, 3, NULL, S_FLAME_TSMALL3, 0, 0},  // S_FLAME_TSMALL2
    {SPR_FFSM, 32770, 2, A_FlameCheck, S_FLAME_TSMALL4, 0, 0},  // S_FLAME_TSMALL3
    {SPR_FFSM, 32770, 2, NULL, S_FLAME_TSMALL5, 0, 0},  // S_FLAME_TSMALL4
    {SPR_FFSM, 32771, 3, NULL, S_FLAME_TSMALL6, 0, 0},  // S_FLAME_TSMALL5
    {SPR_FFSM, 32772, 3, A_FlameCheck, S_FLAME_TSMALL1, 0, 0},  // S_FLAME_TSMALL6
    {SPR_FFLG, 32768, 4, NULL, S_FLAME_TLARGE2, 0, 0},  // S_FLAME_TLARGE1
    {SPR_FFLG, 32769, 4, A_FlameCheck, S_FLAME_TLARGE3, 0, 0},  // S_FLAME_TLARGE2
    {SPR_FFLG, 32770, 4, NULL, S_FLAME_TLARGE4, 0, 0},  // S_FLAME_TLARGE3
    {SPR_FFLG, 32771, 4, A_FlameCheck, S_FLAME_TLARGE5, 0, 0},  // S_FLAME_TLARGE4
    {SPR_FFLG, 32772, 4, NULL, S_FLAME_TLARGE6, 0, 0},  // S_FLAME_TLARGE5
    {SPR_FFLG, 32773, 4, A_FlameCheck, S_FLAME_TLARGE7, 0, 0},  // S_FLAME_TLARGE6
    {SPR_FFLG, 32774, 4, NULL, S_FLAME_TLARGE8, 0, 0},  // S_FLAME_TLARGE7
    {SPR_FFLG, 32775, 4, A_FlameCheck, S_FLAME_TLARGE9, 0, 0},  // S_FLAME_TLARGE8
    {SPR_FFLG, 32776, 4, NULL, S_FLAME_TLARGE10, 0, 0}, // S_FLAME_TLARGE9
    {SPR_FFLG, 32777, 4, A_FlameCheck, S_FLAME_TLARGE11, 0, 0}, // S_FLAME_TLARGE10
    {SPR_FFLG, 32778, 4, NULL, S_FLAME_TLARGE12, 0, 0}, // S_FLAME_TLARGE11
    {SPR_FFLG, 32779, 4, A_FlameCheck, S_FLAME_TLARGE13, 0, 0}, // S_FLAME_TLARGE12
    {SPR_FFLG, 32780, 4, NULL, S_FLAME_TLARGE14, 0, 0}, // S_FLAME_TLARGE13
    {SPR_FFLG, 32781, 4, A_FlameCheck, S_FLAME_TLARGE15, 0, 0}, // S_FLAME_TLARGE14
    {SPR_FFLG, 32782, 4, NULL, S_FLAME_TLARGE16, 0, 0}, // S_FLAME_TLARGE15
    {SPR_FFLG, 32783, 4, A_FlameCheck, S_FLAME_TLARGE5, 0, 0},  // S_FLAME_TLARGE16
    {SPR_FFSM, 0, 2, NULL, S_FLAME_SDORM2, 0, 0},       // S_FLAME_SDORM1
    {SPR_FFSM, 1, 2, A_HideThing, S_FLAME_SDORM3, 0, 0},        // S_FLAME_SDORM2
    {SPR_FFSM, 2, 200, NULL, S_FLAME_SDORM3, 0, 0},     // S_FLAME_SDORM3
    {SPR_FFSM, 32768, 3, NULL, S_FLAME_SMALL2, 0, 0},   // S_FLAME_SMALL1
    {SPR_FFSM, 32768, 3, A_UnHideThing, S_FLAME_SMALL3, 0, 0},  // S_FLAME_SMALL2
    {SPR_FFSM, 32768, 3, NULL, S_FLAME_SMALL4, 0, 0},   // S_FLAME_SMALL3
    {SPR_FFSM, 32769, 3, NULL, S_FLAME_SMALL5, 0, 0},   // S_FLAME_SMALL4
    {SPR_FFSM, 32770, 3, NULL, S_FLAME_SMALL6, 0, 0},   // S_FLAME_SMALL5
    {SPR_FFSM, 32771, 3, NULL, S_FLAME_SMALL7, 0, 0},   // S_FLAME_SMALL6
    {SPR_FFSM, 32772, 3, NULL, S_FLAME_SMALL3, 0, 0},   // S_FLAME_SMALL7
    {SPR_FFLG, 3, 2, NULL, S_FLAME_LDORM2, 0, 0},       // S_FLAME_LDORM1
    {SPR_FFLG, 2, 2, NULL, S_FLAME_LDORM3, 0, 0},       // S_FLAME_LDORM2
    {SPR_FFLG, 1, 2, NULL, S_FLAME_LDORM4, 0, 0},       // S_FLAME_LDORM3
    {SPR_FFLG, 0, 2, A_HideThing, S_FLAME_LDORM5, 0, 0},        // S_FLAME_LDORM4
    {SPR_FFLG, 0, 200, NULL, S_FLAME_LDORM5, 0, 0},     // S_FLAME_LDORM5
    {SPR_FFLG, 32768, 2, NULL, S_FLAME_LARGE2, 0, 0},   // S_FLAME_LARGE1
    {SPR_FFLG, 32768, 2, A_UnHideThing, S_FLAME_LARGE3, 0, 0},  // S_FLAME_LARGE2
    {SPR_FFLG, 32768, 4, NULL, S_FLAME_LARGE4, 0, 0},   // S_FLAME_LARGE3
    {SPR_FFLG, 32769, 4, NULL, S_FLAME_LARGE5, 0, 0},   // S_FLAME_LARGE4
    {SPR_FFLG, 32770, 4, NULL, S_FLAME_LARGE6, 0, 0},   // S_FLAME_LARGE5
    {SPR_FFLG, 32771, 4, NULL, S_FLAME_LARGE7, 0, 0},   // S_FLAME_LARGE6
    {SPR_FFLG, 32772, 4, NULL, S_FLAME_LARGE8, 0, 0},   // S_FLAME_LARGE7
    {SPR_FFLG, 32773, 4, NULL, S_FLAME_LARGE9, 0, 0},   // S_FLAME_LARGE8
    {SPR_FFLG, 32774, 4, NULL, S_FLAME_LARGE10, 0, 0},  // S_FLAME_LARGE9
    {SPR_FFLG, 32775, 4, NULL, S_FLAME_LARGE11, 0, 0},  // S_FLAME_LARGE10
    {SPR_FFLG, 32776, 4, NULL, S_FLAME_LARGE12, 0, 0},  // S_FLAME_LARGE11
    {SPR_FFLG, 32777, 4, NULL, S_FLAME_LARGE13, 0, 0},  // S_FLAME_LARGE12
    {SPR_FFLG, 32778, 4, NULL, S_FLAME_LARGE14, 0, 0},  // S_FLAME_LARGE13
    {SPR_FFLG, 32779, 4, NULL, S_FLAME_LARGE15, 0, 0},  // S_FLAME_LARGE14
    {SPR_FFLG, 32780, 4, NULL, S_FLAME_LARGE16, 0, 0},  // S_FLAME_LARGE15
    {SPR_FFLG, 32781, 4, NULL, S_FLAME_LARGE17, 0, 0},  // S_FLAME_LARGE16
    {SPR_FFLG, 32782, 4, NULL, S_FLAME_LARGE18, 0, 0},  // S_FLAME_LARGE17
    {SPR_FFLG, 32783, 4, NULL, S_FLAME_LARGE7, 0, 0},   // S_FLAME_LARGE18
    {SPR_PTN1, 0, 3, NULL, S_ITEM_PTN1_2, 0, 0},        // S_ITEM_PTN1_1
    {SPR_PTN1, 1, 3, NULL, S_ITEM_PTN1_3, 0, 0},        // S_ITEM_PTN1_2
    {SPR_PTN1, 2, 3, NULL, S_ITEM_PTN1_1, 0, 0},        // S_ITEM_PTN1_3
    {SPR_ACLO, 4, 1400, NULL, S_HIDESPECIAL2, 0, 0},    // S_HIDESPECIAL1
    {SPR_ACLO, 0, 4, A_RestoreSpecialThing1, S_HIDESPECIAL3, 0, 0},     // S_HIDESPECIAL2
    {SPR_ACLO, 1, 4, NULL, S_HIDESPECIAL4, 0, 0},       // S_HIDESPECIAL3
    {SPR_ACLO, 0, 4, NULL, S_HIDESPECIAL5, 0, 0},       // S_HIDESPECIAL4
    {SPR_ACLO, 1, 4, NULL, S_HIDESPECIAL6, 0, 0},       // S_HIDESPECIAL5
    {SPR_ACLO, 2, 4, NULL, S_HIDESPECIAL7, 0, 0},       // S_HIDESPECIAL6
    {SPR_ACLO, 1, 4, NULL, S_HIDESPECIAL8, 0, 0},       // S_HIDESPECIAL7
    {SPR_ACLO, 2, 4, NULL, S_HIDESPECIAL9, 0, 0},       // S_HIDESPECIAL8
    {SPR_ACLO, 3, 4, NULL, S_HIDESPECIAL10, 0, 0},      // S_HIDESPECIAL9
    {SPR_ACLO, 2, 4, NULL, S_HIDESPECIAL11, 0, 0},      // S_HIDESPECIAL10
    {SPR_ACLO, 3, 4, A_RestoreSpecialThing2, S_NULL, 0, 0},     // S_HIDESPECIAL11
    {SPR_ACLO, 3, 3, NULL, S_DORMANTARTI1_2, 0, 0},     // S_DORMANTARTI1_1
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI1_3, 0, 0},     // S_DORMANTARTI1_2
    {SPR_ACLO, 3, 3, NULL, S_DORMANTARTI1_4, 0, 0},     // S_DORMANTARTI1_3
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI1_5, 0, 0},     // S_DORMANTARTI1_4
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI1_6, 0, 0},     // S_DORMANTARTI1_5
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI1_7, 0, 0},     // S_DORMANTARTI1_6
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI1_8, 0, 0},     // S_DORMANTARTI1_7
    {SPR_ACLO, 0, 3, NULL, S_DORMANTARTI1_9, 0, 0},     // S_DORMANTARTI1_8
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI1_10, 0, 0},    // S_DORMANTARTI1_9
    {SPR_ACLO, 0, 3, NULL, S_DORMANTARTI1_11, 0, 0},    // S_DORMANTARTI1_10
    {SPR_ACLO, 0, 1400, A_HideThing, S_DORMANTARTI1_12, 0, 0},  // S_DORMANTARTI1_11
    {SPR_ACLO, 0, 3, A_UnHideThing, S_DORMANTARTI1_13, 0, 0},   // S_DORMANTARTI1_12
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI1_14, 0, 0},    // S_DORMANTARTI1_13
    {SPR_ACLO, 0, 3, NULL, S_DORMANTARTI1_15, 0, 0},    // S_DORMANTARTI1_14
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI1_16, 0, 0},    // S_DORMANTARTI1_15
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI1_17, 0, 0},    // S_DORMANTARTI1_16
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI1_18, 0, 0},    // S_DORMANTARTI1_17
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI1_19, 0, 0},    // S_DORMANTARTI1_18
    {SPR_ACLO, 3, 3, NULL, S_DORMANTARTI1_20, 0, 0},    // S_DORMANTARTI1_19
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI1_21, 0, 0},    // S_DORMANTARTI1_20
    {SPR_ACLO, 3, 3, A_RestoreArtifact, S_NULL, 0, 0},  // S_DORMANTARTI1_21
    {SPR_ACLO, 3, 3, NULL, S_DORMANTARTI2_2, 0, 0},     // S_DORMANTARTI2_1
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI2_3, 0, 0},     // S_DORMANTARTI2_2
    {SPR_ACLO, 3, 3, NULL, S_DORMANTARTI2_4, 0, 0},     // S_DORMANTARTI2_3
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI2_5, 0, 0},     // S_DORMANTARTI2_4
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI2_6, 0, 0},     // S_DORMANTARTI2_5
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI2_7, 0, 0},     // S_DORMANTARTI2_6
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI2_8, 0, 0},     // S_DORMANTARTI2_7
    {SPR_ACLO, 0, 3, NULL, S_DORMANTARTI2_9, 0, 0},     // S_DORMANTARTI2_8
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI2_10, 0, 0},    // S_DORMANTARTI2_9
    {SPR_ACLO, 0, 3, NULL, S_DORMANTARTI2_11, 0, 0},    // S_DORMANTARTI2_10
    {SPR_ACLO, 0, 4200, A_HideThing, S_DORMANTARTI2_12, 0, 0},  // S_DORMANTARTI2_11
    {SPR_ACLO, 0, 3, A_UnHideThing, S_DORMANTARTI2_13, 0, 0},   // S_DORMANTARTI2_12
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI2_14, 0, 0},    // S_DORMANTARTI2_13
    {SPR_ACLO, 0, 3, NULL, S_DORMANTARTI2_15, 0, 0},    // S_DORMANTARTI2_14
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI2_16, 0, 0},    // S_DORMANTARTI2_15
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI2_17, 0, 0},    // S_DORMANTARTI2_16
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI2_18, 0, 0},    // S_DORMANTARTI2_17
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI2_19, 0, 0},    // S_DORMANTARTI2_18
    {SPR_ACLO, 3, 3, NULL, S_DORMANTARTI2_20, 0, 0},    // S_DORMANTARTI2_19
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI2_21, 0, 0},    // S_DORMANTARTI2_20
    {SPR_ACLO, 3, 3, A_RestoreArtifact, S_NULL, 0, 0},  // S_DORMANTARTI2_21
    {SPR_ACLO, 3, 3, NULL, S_DORMANTARTI3_2, 0, 0},     // S_DORMANTARTI3_1
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI3_3, 0, 0},     // S_DORMANTARTI3_2
    {SPR_ACLO, 3, 3, NULL, S_DORMANTARTI3_4, 0, 0},     // S_DORMANTARTI3_3
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI3_5, 0, 0},     // S_DORMANTARTI3_4
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI3_6, 0, 0},     // S_DORMANTARTI3_5
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI3_7, 0, 0},     // S_DORMANTARTI3_6
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI3_8, 0, 0},     // S_DORMANTARTI3_7
    {SPR_ACLO, 0, 3, NULL, S_DORMANTARTI3_9, 0, 0},     // S_DORMANTARTI3_8
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI3_10, 0, 0},    // S_DORMANTARTI3_9
    {SPR_ACLO, 0, 3, NULL, S_DORMANTARTI3_11, 0, 0},    // S_DORMANTARTI3_10
    {SPR_ACLO, 0, 21000, A_HideThing, S_DORMANTARTI3_12, 0, 0}, // S_DORMANTARTI3_11
    {SPR_ACLO, 0, 3, A_UnHideThing, S_DORMANTARTI3_13, 0, 0},   // S_DORMANTARTI3_12
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI3_14, 0, 0},    // S_DORMANTARTI3_13
    {SPR_ACLO, 0, 3, NULL, S_DORMANTARTI3_15, 0, 0},    // S_DORMANTARTI3_14
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI3_16, 0, 0},    // S_DORMANTARTI3_15
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI3_17, 0, 0},    // S_DORMANTARTI3_16
    {SPR_ACLO, 1, 3, NULL, S_DORMANTARTI3_18, 0, 0},    // S_DORMANTARTI3_17
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI3_19, 0, 0},    // S_DORMANTARTI3_18
    {SPR_ACLO, 3, 3, NULL, S_DORMANTARTI3_20, 0, 0},    // S_DORMANTARTI3_19
    {SPR_ACLO, 2, 3, NULL, S_DORMANTARTI3_21, 0, 0},    // S_DORMANTARTI3_20
    {SPR_ACLO, 3, 3, A_RestoreArtifact, S_NULL, 0, 0},  // S_DORMANTARTI3_21
    {SPR_ACLO, 3, 3, NULL, S_DEADARTI2, 0, 0},  // S_DEADARTI1
    {SPR_ACLO, 2, 3, NULL, S_DEADARTI3, 0, 0},  // S_DEADARTI2
    {SPR_ACLO, 3, 3, NULL, S_DEADARTI4, 0, 0},  // S_DEADARTI3
    {SPR_ACLO, 2, 3, NULL, S_DEADARTI5, 0, 0},  // S_DEADARTI4
    {SPR_ACLO, 1, 3, NULL, S_DEADARTI6, 0, 0},  // S_DEADARTI5
    {SPR_ACLO, 2, 3, NULL, S_DEADARTI7, 0, 0},  // S_DEADARTI6
    {SPR_ACLO, 1, 3, NULL, S_DEADARTI8, 0, 0},  // S_DEADARTI7
    {SPR_ACLO, 0, 3, NULL, S_DEADARTI9, 0, 0},  // S_DEADARTI8
    {SPR_ACLO, 1, 3, NULL, S_DEADARTI10, 0, 0}, // S_DEADARTI9
    {SPR_ACLO, 0, 3, NULL, S_NULL, 0, 0},       // S_DEADARTI10
    {SPR_PTN2, 0, 4, NULL, S_ARTI_PTN2_2, 0, 0},        // S_ARTI_PTN2_1
    {SPR_PTN2, 1, 4, NULL, S_ARTI_PTN2_3, 0, 0},        // S_ARTI_PTN2_2
    {SPR_PTN2, 2, 4, NULL, S_ARTI_PTN2_1, 0, 0},        // S_ARTI_PTN2_3
    {SPR_SOAR, 0, 5, NULL, S_ARTI_SOAR2, 0, 0}, // S_ARTI_SOAR1
    {SPR_SOAR, 1, 5, NULL, S_ARTI_SOAR3, 0, 0}, // S_ARTI_SOAR2
    {SPR_SOAR, 2, 5, NULL, S_ARTI_SOAR4, 0, 0}, // S_ARTI_SOAR3
    {SPR_SOAR, 1, 5, NULL, S_ARTI_SOAR1, 0, 0}, // S_ARTI_SOAR4
    {SPR_INVU, 0, 3, NULL, S_ARTI_INVU2, 0, 0}, // S_ARTI_INVU1
    {SPR_INVU, 1, 3, NULL, S_ARTI_INVU3, 0, 0}, // S_ARTI_INVU2
    {SPR_INVU, 2, 3, NULL, S_ARTI_INVU4, 0, 0}, // S_ARTI_INVU3
    {SPR_INVU, 3, 3, NULL, S_ARTI_INVU1, 0, 0}, // S_ARTI_INVU4
    {SPR_SUMN, 0, 350, NULL, S_ARTI_SUMMON, 0, 0},      // S_ARTI_SUMMON
    {SPR_SUMN, 0, 4, NULL, S_SUMMON_FX1_1, 0, 0},       // S_SUMMON_FX1_1
    {SPR_SUMN, 0, 4, NULL, S_SUMMON_FX2_2, 0, 0},       // S_SUMMON_FX2_1
    {SPR_SUMN, 0, 4, NULL, S_SUMMON_FX2_3, 0, 0},       // S_SUMMON_FX2_2
    {SPR_SUMN, 0, 4, A_Summon, S_NULL, 0, 0},   // S_SUMMON_FX2_3
    {SPR_TSPK, 0, 3, NULL, S_THRUSTINIT2_2, 0, 0},      // S_THRUSTINIT2_1
    {SPR_TSPK, 0, 4, A_ThrustInitUp, S_THRUSTBLOCK, 0, 0},      // S_THRUSTINIT2_2
    {SPR_TSPK, 1, 3, NULL, S_BTHRUSTINIT2_2, 0, 0},     // S_BTHRUSTINIT2_1
    {SPR_TSPK, 1, 4, A_ThrustInitUp, S_BTHRUSTBLOCK, 0, 0},     // S_BTHRUSTINIT2_2
    {SPR_TSPK, 0, 3, NULL, S_THRUSTINIT1_2, 0, 0},      // S_THRUSTINIT1_1
    {SPR_TSPK, 0, 4, A_ThrustInitDn, S_THRUSTSTAY, 0, 0},       // S_THRUSTINIT1_2
    {SPR_TSPK, 1, 3, NULL, S_BTHRUSTINIT1_2, 0, 0},     // S_BTHRUSTINIT1_1
    {SPR_TSPK, 1, 4, A_ThrustInitDn, S_BTHRUSTSTAY, 0, 0},      // S_BTHRUSTINIT1_2
    {SPR_TSPK, 0, 8, A_ThrustRaise, S_THRUSTRAISE2, 0, 0},      // S_THRUSTRAISE1
    {SPR_TSPK, 0, 6, A_ThrustRaise, S_THRUSTRAISE3, 0, 0},      // S_THRUSTRAISE2
    {SPR_TSPK, 0, 4, A_ThrustRaise, S_THRUSTRAISE4, 0, 0},      // S_THRUSTRAISE3
    {SPR_TSPK, 0, 3, A_ThrustBlock, S_THRUSTIMPALE, 0, 0},      // S_THRUSTRAISE4
    {SPR_TSPK, 1, 8, A_ThrustRaise, S_BTHRUSTRAISE2, 0, 0},     // S_BTHRUSTRAISE1
    {SPR_TSPK, 1, 6, A_ThrustRaise, S_BTHRUSTRAISE3, 0, 0},     // S_BTHRUSTRAISE2
    {SPR_TSPK, 1, 4, A_ThrustRaise, S_BTHRUSTRAISE4, 0, 0},     // S_BTHRUSTRAISE3
    {SPR_TSPK, 1, 3, A_ThrustBlock, S_BTHRUSTIMPALE, 0, 0},     // S_BTHRUSTRAISE4
    {SPR_TSPK, 0, 2, A_ThrustImpale, S_THRUSTRAISE, 0, 0},      // S_THRUSTIMPALE
    {SPR_TSPK, 1, 2, A_ThrustImpale, S_BTHRUSTRAISE, 0, 0},     // S_BTHRUSTIMPALE
    {SPR_TSPK, 0, 2, A_ThrustRaise, S_THRUSTRAISE, 0, 0},       // S_THRUSTRAISE
    {SPR_TSPK, 1, 2, A_ThrustRaise, S_BTHRUSTRAISE, 0, 0},      // S_BTHRUSTRAISE
    {SPR_TSPK, 0, 10, NULL, S_THRUSTBLOCK, 0, 0},       // S_THRUSTBLOCK
    {SPR_TSPK, 1, 10, NULL, S_BTHRUSTBLOCK, 0, 0},      // S_BTHRUSTBLOCK
    {SPR_TSPK, 0, 2, A_ThrustLower, S_THRUSTLOWER, 0, 0},       // S_THRUSTLOWER
    {SPR_TSPK, 1, 2, A_ThrustLower, S_BTHRUSTLOWER, 0, 0},      // S_BTHRUSTLOWER
    {SPR_TSPK, 0, -1, NULL, S_THRUSTSTAY, 0, 0},        // S_THRUSTSTAY
    {SPR_TSPK, 1, -1, NULL, S_BTHRUSTSTAY, 0, 0},       // S_BTHRUSTSTAY
    {SPR_TELO, 0, 5, NULL, S_ARTI_TELOTHER2, 0, 0},     // S_ARTI_TELOTHER1
    {SPR_TELO, 1, 5, NULL, S_ARTI_TELOTHER3, 0, 0},     // S_ARTI_TELOTHER2
    {SPR_TELO, 2, 5, NULL, S_ARTI_TELOTHER4, 0, 0},     // S_ARTI_TELOTHER3
    {SPR_TELO, 3, 5, NULL, S_ARTI_TELOTHER1, 0, 0},     // S_ARTI_TELOTHER4
    {SPR_TRNG, 32772, 5, NULL, S_TELO_FX2, 0, 0},       // S_TELO_FX1
    {SPR_TRNG, 32771, 4, NULL, S_TELO_FX3, 0, 0},       // S_TELO_FX2
    {SPR_TRNG, 32770, 3, A_TeloSpawnC, S_TELO_FX4, 0, 0},       // S_TELO_FX3
    {SPR_TRNG, 32769, 3, A_TeloSpawnB, S_TELO_FX5, 0, 0},       // S_TELO_FX4
    {SPR_TRNG, 32768, 3, A_TeloSpawnA, S_TELO_FX6, 0, 0},       // S_TELO_FX5
    {SPR_TRNG, 32769, 3, A_TeloSpawnB, S_TELO_FX7, 0, 0},       // S_TELO_FX6
    {SPR_TRNG, 32770, 3, A_TeloSpawnC, S_TELO_FX8, 0, 0},       // S_TELO_FX7
    {SPR_TRNG, 32771, 3, A_TeloSpawnD, S_TELO_FX3, 0, 0},       // S_TELO_FX8
    {SPR_TRNG, 32772, 3, NULL, S_NULL, 0, 0},   // S_TELO_FX9
    {SPR_TRNG, 32769, 4, NULL, S_TELO_FX2_2, 0, 0},     // S_TELO_FX2_1
    {SPR_TRNG, 32770, 4, NULL, S_TELO_FX2_3, 0, 0},     // S_TELO_FX2_2
    {SPR_TRNG, 32771, 4, NULL, S_TELO_FX2_4, 0, 0},     // S_TELO_FX2_3
    {SPR_TRNG, 32770, 4, NULL, S_TELO_FX2_5, 0, 0},     // S_TELO_FX2_4
    {SPR_TRNG, 32769, 4, NULL, S_TELO_FX2_6, 0, 0},     // S_TELO_FX2_5
    {SPR_TRNG, 32768, 4, A_CheckTeleRing, S_TELO_FX2_1, 0, 0},  // S_TELO_FX2_6
    {SPR_TRNG, 32770, 4, NULL, S_TELO_FX3_2, 0, 0},     // S_TELO_FX3_1
    {SPR_TRNG, 32771, 4, NULL, S_TELO_FX3_3, 0, 0},     // S_TELO_FX3_2
    {SPR_TRNG, 32770, 4, NULL, S_TELO_FX3_4, 0, 0},     // S_TELO_FX3_3
    {SPR_TRNG, 32769, 4, NULL, S_TELO_FX3_5, 0, 0},     // S_TELO_FX3_4
    {SPR_TRNG, 32768, 4, NULL, S_TELO_FX3_6, 0, 0},     // S_TELO_FX3_5
    {SPR_TRNG, 32769, 4, A_CheckTeleRing, S_TELO_FX3_1, 0, 0},  // S_TELO_FX3_6
    {SPR_TRNG, 32771, 4, NULL, S_TELO_FX4_2, 0, 0},     // S_TELO_FX4_1
    {SPR_TRNG, 32770, 4, NULL, S_TELO_FX4_3, 0, 0},     // S_TELO_FX4_2
    {SPR_TRNG, 32769, 4, NULL, S_TELO_FX4_4, 0, 0},     // S_TELO_FX4_3
    {SPR_TRNG, 32768, 4, NULL, S_TELO_FX4_5, 0, 0},     // S_TELO_FX4_4
    {SPR_TRNG, 32769, 4, NULL, S_TELO_FX4_6, 0, 0},     // S_TELO_FX4_5
    {SPR_TRNG, 32770, 4, A_CheckTeleRing, S_TELO_FX4_1, 0, 0},  // S_TELO_FX4_6
    {SPR_TRNG, 32770, 4, NULL, S_TELO_FX5_2, 0, 0},     // S_TELO_FX5_1
    {SPR_TRNG, 32769, 4, NULL, S_TELO_FX5_3, 0, 0},     // S_TELO_FX5_2
    {SPR_TRNG, 32768, 4, NULL, S_TELO_FX5_4, 0, 0},     // S_TELO_FX5_3
    {SPR_TRNG, 32769, 4, NULL, S_TELO_FX5_5, 0, 0},     // S_TELO_FX5_4
    {SPR_TRNG, 32770, 4, NULL, S_TELO_FX5_6, 0, 0},     // S_TELO_FX5_5
    {SPR_TRNG, 32771, 4, A_CheckTeleRing, S_TELO_FX5_1, 0, 0},  // S_TELO_FX5_6
    {SPR_ROCK, 3, 20, NULL, S_DIRT1_1, 0, 0},   // S_DIRT1_1
    {SPR_ROCK, 3, 10, NULL, S_NULL, 0, 0},      // S_DIRT1_D
    {SPR_ROCK, 4, 20, NULL, S_DIRT2_1, 0, 0},   // S_DIRT2_1
    {SPR_ROCK, 4, 10, NULL, S_NULL, 0, 0},      // S_DIRT2_D
    {SPR_ROCK, 5, 20, NULL, S_DIRT3_1, 0, 0},   // S_DIRT3_1
    {SPR_ROCK, 5, 10, NULL, S_NULL, 0, 0},      // S_DIRT3_D
    {SPR_ROCK, 6, 20, NULL, S_DIRT4_1, 0, 0},   // S_DIRT4_1
    {SPR_ROCK, 6, 10, NULL, S_NULL, 0, 0},      // S_DIRT4_D
    {SPR_ROCK, 7, 20, NULL, S_DIRT5_1, 0, 0},   // S_DIRT5_1
    {SPR_ROCK, 7, 10, NULL, S_NULL, 0, 0},      // S_DIRT5_D
    {SPR_ROCK, 8, 20, NULL, S_DIRT6_1, 0, 0},   // S_DIRT6_1
    {SPR_ROCK, 8, 10, NULL, S_NULL, 0, 0},      // S_DIRT6_D
    {SPR_TSPK, 2, 20, NULL, S_DIRTCLUMP1, 0, 0},        // S_DIRTCLUMP1
    {SPR_ROCK, 0, 20, NULL, S_ROCK1_1, 0, 0},   // S_ROCK1_1
    {SPR_ROCK, 0, 10, NULL, S_NULL, 0, 0},      // S_ROCK1_D
    {SPR_ROCK, 1, 20, NULL, S_ROCK2_1, 0, 0},   // S_ROCK2_1
    {SPR_ROCK, 1, 10, NULL, S_NULL, 0, 0},      // S_ROCK2_D
    {SPR_ROCK, 2, 20, NULL, S_ROCK3_1, 0, 0},   // S_ROCK3_1
    {SPR_ROCK, 2, 10, NULL, S_NULL, 0, 0},      // S_ROCK3_D
    {SPR_MAN1, 0, 20, A_FogSpawn, S_SPAWNFOG1, 0, 0},   // S_SPAWNFOG1
    {SPR_FOGS, 0, 7, A_FogMove, S_FOGPATCHS2, 0, 0},    // S_FOGPATCHS1
    {SPR_FOGS, 1, 7, A_FogMove, S_FOGPATCHS3, 0, 0},    // S_FOGPATCHS2
    {SPR_FOGS, 2, 7, A_FogMove, S_FOGPATCHS4, 0, 0},    // S_FOGPATCHS3
    {SPR_FOGS, 3, 7, A_FogMove, S_FOGPATCHS5, 0, 0},    // S_FOGPATCHS4
    {SPR_FOGS, 4, 7, A_FogMove, S_FOGPATCHS1, 0, 0},    // S_FOGPATCHS5
    {SPR_FOGS, 4, 5, NULL, S_NULL, 0, 0},       // S_FOGPATCHS0
    {SPR_FOGM, 0, 7, A_FogMove, S_FOGPATCHM2, 0, 0},    // S_FOGPATCHM1
    {SPR_FOGM, 1, 7, A_FogMove, S_FOGPATCHM3, 0, 0},    // S_FOGPATCHM2
    {SPR_FOGM, 2, 7, A_FogMove, S_FOGPATCHM4, 0, 0},    // S_FOGPATCHM3
    {SPR_FOGM, 3, 7, A_FogMove, S_FOGPATCHM5, 0, 0},    // S_FOGPATCHM4
    {SPR_FOGM, 4, 7, A_FogMove, S_FOGPATCHM1, 0, 0},    // S_FOGPATCHM5
    {SPR_FOGS, 0, 5, NULL, S_FOGPATCHMA, 0, 0}, // S_FOGPATCHM0
    {SPR_FOGS, 1, 5, NULL, S_FOGPATCHMB, 0, 0}, // S_FOGPATCHMA
    {SPR_FOGS, 2, 5, NULL, S_FOGPATCHMC, 0, 0}, // S_FOGPATCHMB
    {SPR_FOGS, 3, 5, NULL, S_FOGPATCHMD, 0, 0}, // S_FOGPATCHMC
    {SPR_FOGS, 4, 5, NULL, S_FOGPATCHS0, 0, 0}, // S_FOGPATCHMD
    {SPR_FOGL, 0, 7, A_FogMove, S_FOGPATCHL2, 0, 0},    // S_FOGPATCHL1
    {SPR_FOGL, 1, 7, A_FogMove, S_FOGPATCHL3, 0, 0},    // S_FOGPATCHL2
    {SPR_FOGL, 2, 7, A_FogMove, S_FOGPATCHL4, 0, 0},    // S_FOGPATCHL3
    {SPR_FOGL, 3, 7, A_FogMove, S_FOGPATCHL5, 0, 0},    // S_FOGPATCHL4
    {SPR_FOGL, 4, 7, A_FogMove, S_FOGPATCHL1, 0, 0},    // S_FOGPATCHL5
    {SPR_FOGM, 0, 4, NULL, S_FOGPATCHLA, 0, 0}, // S_FOGPATCHL0
    {SPR_FOGM, 1, 4, NULL, S_FOGPATCHLB, 0, 0}, // S_FOGPATCHLA
    {SPR_FOGM, 2, 4, NULL, S_FOGPATCHLC, 0, 0}, // S_FOGPATCHLB
    {SPR_FOGM, 3, 4, NULL, S_FOGPATCHLD, 0, 0}, // S_FOGPATCHLC
    {SPR_FOGM, 4, 4, NULL, S_FOGPATCHM0, 0, 0}, // S_FOGPATCHLD
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVE2, 0, 0},   // S_QUAKE_ACTIVE1
    {SPR_MAN1, 0, 1, A_ContMobjSound, S_QUAKE_ACTIVE3, 0, 0},   // S_QUAKE_ACTIVE2
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVE4, 0, 0},   // S_QUAKE_ACTIVE3
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVE5, 0, 0},   // S_QUAKE_ACTIVE4
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVE6, 0, 0},   // S_QUAKE_ACTIVE5
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVE7, 0, 0},   // S_QUAKE_ACTIVE6
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVE8, 0, 0},   // S_QUAKE_ACTIVE7
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVE9, 0, 0},   // S_QUAKE_ACTIVE8
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVE0, 0, 0},   // S_QUAKE_ACTIVE9
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEA, 0, 0},   // S_QUAKE_ACTIVE0
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEB, 0, 0},   // S_QUAKE_ACTIVEA
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEC, 0, 0},   // S_QUAKE_ACTIVEB
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVED, 0, 0},   // S_QUAKE_ACTIVEC
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEE, 0, 0},   // S_QUAKE_ACTIVED
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEF, 0, 0},   // S_QUAKE_ACTIVEE
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEG, 0, 0},   // S_QUAKE_ACTIVEF
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEH, 0, 0},   // S_QUAKE_ACTIVEG
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEI, 0, 0},   // S_QUAKE_ACTIVEH
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEJ, 0, 0},   // S_QUAKE_ACTIVEI
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEK, 0, 0},   // S_QUAKE_ACTIVEJ
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEL, 0, 0},   // S_QUAKE_ACTIVEK
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEM, 0, 0},   // S_QUAKE_ACTIVEL
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEN, 0, 0},   // S_QUAKE_ACTIVEM
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEO, 0, 0},   // S_QUAKE_ACTIVEN
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEP, 0, 0},   // S_QUAKE_ACTIVEO
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEQ, 0, 0},   // S_QUAKE_ACTIVEP
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVER, 0, 0},   // S_QUAKE_ACTIVEQ
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVES, 0, 0},   // S_QUAKE_ACTIVER
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVET, 0, 0},   // S_QUAKE_ACTIVES
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEU, 0, 0},   // S_QUAKE_ACTIVET
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEV, 0, 0},   // S_QUAKE_ACTIVEU
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEW, 0, 0},   // S_QUAKE_ACTIVEV
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEX, 0, 0},   // S_QUAKE_ACTIVEW
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEY, 0, 0},   // S_QUAKE_ACTIVEX
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVEZ, 0, 0},   // S_QUAKE_ACTIVEY
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACT1, 0, 0},      // S_QUAKE_ACTIVEZ
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACT2, 0, 0},      // S_QUAKE_ACT1
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACT3, 0, 0},      // S_QUAKE_ACT2
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACT4, 0, 0},      // S_QUAKE_ACT3
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACT5, 0, 0},      // S_QUAKE_ACT4
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACT6, 0, 0},      // S_QUAKE_ACT5
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACT7, 0, 0},      // S_QUAKE_ACT6
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACT8, 0, 0},      // S_QUAKE_ACT7
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACT9, 0, 0},      // S_QUAKE_ACT8
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACT0, 0, 0},      // S_QUAKE_ACT9
    {SPR_MAN1, 0, 2, A_Quake, S_QUAKE_ACTIVE1, 0, 0},   // S_QUAKE_ACT0
    {SPR_SGSA, 0, 4, NULL, S_SGSHARD1_2, 0, 0}, // S_SGSHARD1_1
    {SPR_SGSA, 1, 4, NULL, S_SGSHARD1_3, 0, 0}, // S_SGSHARD1_2
    {SPR_SGSA, 2, 4, NULL, S_SGSHARD1_4, 0, 0}, // S_SGSHARD1_3
    {SPR_SGSA, 3, 4, NULL, S_SGSHARD1_5, 0, 0}, // S_SGSHARD1_4
    {SPR_SGSA, 4, 4, NULL, S_SGSHARD1_1, 0, 0}, // S_SGSHARD1_5
    {SPR_SGSA, 4, 30, NULL, S_NULL, 0, 0},      // S_SGSHARD1_D
    {SPR_SGSA, 5, 4, NULL, S_SGSHARD2_2, 0, 0}, // S_SGSHARD2_1
    {SPR_SGSA, 6, 4, NULL, S_SGSHARD2_3, 0, 0}, // S_SGSHARD2_2
    {SPR_SGSA, 7, 4, NULL, S_SGSHARD2_4, 0, 0}, // S_SGSHARD2_3
    {SPR_SGSA, 8, 4, NULL, S_SGSHARD2_5, 0, 0}, // S_SGSHARD2_4
    {SPR_SGSA, 9, 4, NULL, S_SGSHARD2_1, 0, 0}, // S_SGSHARD2_5
    {SPR_SGSA, 9, 30, NULL, S_NULL, 0, 0},      // S_SGSHARD2_D
    {SPR_SGSA, 10, 4, NULL, S_SGSHARD3_2, 0, 0},        // S_SGSHARD3_1
    {SPR_SGSA, 11, 4, NULL, S_SGSHARD3_3, 0, 0},        // S_SGSHARD3_2
    {SPR_SGSA, 12, 4, NULL, S_SGSHARD3_4, 0, 0},        // S_SGSHARD3_3
    {SPR_SGSA, 13, 4, NULL, S_SGSHARD3_5, 0, 0},        // S_SGSHARD3_4
    {SPR_SGSA, 14, 4, NULL, S_SGSHARD3_1, 0, 0},        // S_SGSHARD3_5
    {SPR_SGSA, 14, 30, NULL, S_NULL, 0, 0},     // S_SGSHARD3_D
    {SPR_SGSA, 15, 4, NULL, S_SGSHARD4_2, 0, 0},        // S_SGSHARD4_1
    {SPR_SGSA, 16, 4, NULL, S_SGSHARD4_3, 0, 0},        // S_SGSHARD4_2
    {SPR_SGSA, 17, 4, NULL, S_SGSHARD4_4, 0, 0},        // S_SGSHARD4_3
    {SPR_SGSA, 18, 4, NULL, S_SGSHARD4_5, 0, 0},        // S_SGSHARD4_4
    {SPR_SGSA, 19, 4, NULL, S_SGSHARD4_1, 0, 0},        // S_SGSHARD4_5
    {SPR_SGSA, 19, 30, NULL, S_NULL, 0, 0},     // S_SGSHARD4_D
    {SPR_SGSA, 20, 4, NULL, S_SGSHARD5_2, 0, 0},        // S_SGSHARD5_1
    {SPR_SGSA, 21, 4, NULL, S_SGSHARD5_3, 0, 0},        // S_SGSHARD5_2
    {SPR_SGSA, 22, 4, NULL, S_SGSHARD5_4, 0, 0},        // S_SGSHARD5_3
    {SPR_SGSA, 23, 4, NULL, S_SGSHARD5_5, 0, 0},        // S_SGSHARD5_4
    {SPR_SGSA, 24, 4, NULL, S_SGSHARD5_1, 0, 0},        // S_SGSHARD5_5
    {SPR_SGSA, 24, 30, NULL, S_NULL, 0, 0},     // S_SGSHARD5_D
    {SPR_SGSB, 0, 4, NULL, S_SGSHARD6_1, 0, 0}, // S_SGSHARD6_1
    {SPR_SGSB, 0, 30, NULL, S_NULL, 0, 0},      // S_SGSHARD6_D
    {SPR_SGSB, 1, 4, NULL, S_SGSHARD7_1, 0, 0}, // S_SGSHARD7_1
    {SPR_SGSB, 1, 30, NULL, S_NULL, 0, 0},      // S_SGSHARD7_D
    {SPR_SGSB, 2, 4, NULL, S_SGSHARD8_1, 0, 0}, // S_SGSHARD8_1
    {SPR_SGSB, 2, 30, NULL, S_NULL, 0, 0},      // S_SGSHARD8_D
    {SPR_SGSB, 3, 4, NULL, S_SGSHARD9_1, 0, 0}, // S_SGSHARD9_1
    {SPR_SGSB, 3, 30, NULL, S_NULL, 0, 0},      // S_SGSHARD9_D
    {SPR_SGSB, 4, 4, NULL, S_SGSHARD0_1, 0, 0}, // S_SGSHARD0_1
    {SPR_SGSB, 4, 30, NULL, S_NULL, 0, 0},      // S_SGSHARD0_D
    {SPR_PORK, 0, 5, NULL, S_ARTI_EGGC2, 0, 0}, // S_ARTI_EGGC1
    {SPR_PORK, 1, 5, NULL, S_ARTI_EGGC3, 0, 0}, // S_ARTI_EGGC2
    {SPR_PORK, 2, 5, NULL, S_ARTI_EGGC4, 0, 0}, // S_ARTI_EGGC3
    {SPR_PORK, 3, 5, NULL, S_ARTI_EGGC5, 0, 0}, // S_ARTI_EGGC4
    {SPR_PORK, 4, 5, NULL, S_ARTI_EGGC6, 0, 0}, // S_ARTI_EGGC5
    {SPR_PORK, 5, 5, NULL, S_ARTI_EGGC7, 0, 0}, // S_ARTI_EGGC6
    {SPR_PORK, 6, 5, NULL, S_ARTI_EGGC8, 0, 0}, // S_ARTI_EGGC7
    {SPR_PORK, 7, 5, NULL, S_ARTI_EGGC1, 0, 0}, // S_ARTI_EGGC8
    {SPR_EGGM, 0, 4, NULL, S_EGGFX2, 0, 0},     // S_EGGFX1
    {SPR_EGGM, 1, 4, NULL, S_EGGFX3, 0, 0},     // S_EGGFX2
    {SPR_EGGM, 2, 4, NULL, S_EGGFX4, 0, 0},     // S_EGGFX3
    {SPR_EGGM, 3, 4, NULL, S_EGGFX5, 0, 0},     // S_EGGFX4
    {SPR_EGGM, 4, 4, NULL, S_EGGFX1, 0, 0},     // S_EGGFX5
    {SPR_FHFX, 32776, 3, NULL, S_EGGFXI1_2, 0, 0},      // S_EGGFXI1_1
    {SPR_FHFX, 32777, 3, NULL, S_EGGFXI1_3, 0, 0},      // S_EGGFXI1_2
    {SPR_FHFX, 32778, 3, NULL, S_EGGFXI1_4, 0, 0},      // S_EGGFXI1_3
    {SPR_FHFX, 32779, 3, NULL, S_NULL, 0, 0},   // S_EGGFXI1_4
    {SPR_SPHL, 0, 350, NULL, S_ARTI_SPHL1, 0, 0},       // S_ARTI_SPHL1
    {SPR_STWN, 0, -1, NULL, S_NULL, 0, 0},      // S_ZWINGEDSTATUENOSKULL
    {SPR_STWN, 1, -1, NULL, S_NULL, 0, 0},      // S_ZWINGEDSTATUENOSKULL2
    {SPR_GMPD, 0, -1, NULL, S_NULL, 0, 0},      // S_ZGEMPEDESTAL1
    {SPR_GMPD, 1, -1, NULL, S_NULL, 0, 0},      // S_ZGEMPEDESTAL2
    {SPR_ASKU, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZSKULL
    {SPR_ABGM, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZGEMBIG
    {SPR_AGMR, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZGEMRED
    {SPR_AGMG, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZGEMGREEN1
    {SPR_AGG2, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZGEMGREEN2
    {SPR_AGMB, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZGEMBLUE1
    {SPR_AGB2, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZGEMBLUE2
    {SPR_ABK1, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZBOOK1
    {SPR_ABK2, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZBOOK2
    {SPR_ASK2, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZSKULL2
    {SPR_AFWP, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZFWEAPON
    {SPR_ACWP, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZCWEAPON
    {SPR_AMWP, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTIPUZZMWEAPON
    {SPR_AGER, 32768, 4, NULL, S_ARTIPUZZGEAR_2, 0, 0}, // S_ARTIPUZZGEAR_1
    {SPR_AGER, 32769, 4, NULL, S_ARTIPUZZGEAR_3, 0, 0}, // S_ARTIPUZZGEAR_2
    {SPR_AGER, 32770, 4, NULL, S_ARTIPUZZGEAR_4, 0, 0}, // S_ARTIPUZZGEAR_3
    {SPR_AGER, 32771, 4, NULL, S_ARTIPUZZGEAR_5, 0, 0}, // S_ARTIPUZZGEAR_4
    {SPR_AGER, 32772, 4, NULL, S_ARTIPUZZGEAR_6, 0, 0}, // S_ARTIPUZZGEAR_5
    {SPR_AGER, 32773, 4, NULL, S_ARTIPUZZGEAR_7, 0, 0}, // S_ARTIPUZZGEAR_6
    {SPR_AGER, 32774, 4, NULL, S_ARTIPUZZGEAR_8, 0, 0}, // S_ARTIPUZZGEAR_7
    {SPR_AGER, 32775, 4, NULL, S_ARTIPUZZGEAR_1, 0, 0}, // S_ARTIPUZZGEAR_8
    {SPR_AGR2, 32768, 4, NULL, S_ARTIPUZZGEAR2_2, 0, 0},        // S_ARTIPUZZGEAR2_1
    {SPR_AGR2, 32769, 4, NULL, S_ARTIPUZZGEAR2_3, 0, 0},        // S_ARTIPUZZGEAR2_2
    {SPR_AGR2, 32770, 4, NULL, S_ARTIPUZZGEAR2_4, 0, 0},        // S_ARTIPUZZGEAR2_3
    {SPR_AGR2, 32771, 4, NULL, S_ARTIPUZZGEAR2_5, 0, 0},        // S_ARTIPUZZGEAR2_4
    {SPR_AGR2, 32772, 4, NULL, S_ARTIPUZZGEAR2_6, 0, 0},        // S_ARTIPUZZGEAR2_5
    {SPR_AGR2, 32773, 4, NULL, S_ARTIPUZZGEAR2_7, 0, 0},        // S_ARTIPUZZGEAR2_6
    {SPR_AGR2, 32774, 4, NULL, S_ARTIPUZZGEAR2_8, 0, 0},        // S_ARTIPUZZGEAR2_7
    {SPR_AGR2, 32775, 4, NULL, S_ARTIPUZZGEAR2_1, 0, 0},        // S_ARTIPUZZGEAR2_8
    {SPR_AGR3, 32768, 4, NULL, S_ARTIPUZZGEAR3_2, 0, 0},        // S_ARTIPUZZGEAR3_1
    {SPR_AGR3, 32769, 4, NULL, S_ARTIPUZZGEAR3_3, 0, 0},        // S_ARTIPUZZGEAR3_2
    {SPR_AGR3, 32770, 4, NULL, S_ARTIPUZZGEAR3_4, 0, 0},        // S_ARTIPUZZGEAR3_3
    {SPR_AGR3, 32771, 4, NULL, S_ARTIPUZZGEAR3_5, 0, 0},        // S_ARTIPUZZGEAR3_4
    {SPR_AGR3, 32772, 4, NULL, S_ARTIPUZZGEAR3_6, 0, 0},        // S_ARTIPUZZGEAR3_5
    {SPR_AGR3, 32773, 4, NULL, S_ARTIPUZZGEAR3_7, 0, 0},        // S_ARTIPUZZGEAR3_6
    {SPR_AGR3, 32774, 4, NULL, S_ARTIPUZZGEAR3_8, 0, 0},        // S_ARTIPUZZGEAR3_7
    {SPR_AGR3, 32775, 4, NULL, S_ARTIPUZZGEAR3_1, 0, 0},        // S_ARTIPUZZGEAR3_8
    {SPR_AGR4, 32768, 4, NULL, S_ARTIPUZZGEAR4_2, 0, 0},        // S_ARTIPUZZGEAR4_1
    {SPR_AGR4, 32769, 4, NULL, S_ARTIPUZZGEAR4_3, 0, 0},        // S_ARTIPUZZGEAR4_2
    {SPR_AGR4, 32770, 4, NULL, S_ARTIPUZZGEAR4_4, 0, 0},        // S_ARTIPUZZGEAR4_3
    {SPR_AGR4, 32771, 4, NULL, S_ARTIPUZZGEAR4_5, 0, 0},        // S_ARTIPUZZGEAR4_4
    {SPR_AGR4, 32772, 4, NULL, S_ARTIPUZZGEAR4_6, 0, 0},        // S_ARTIPUZZGEAR4_5
    {SPR_AGR4, 32773, 4, NULL, S_ARTIPUZZGEAR4_7, 0, 0},        // S_ARTIPUZZGEAR4_6
    {SPR_AGR4, 32774, 4, NULL, S_ARTIPUZZGEAR4_8, 0, 0},        // S_ARTIPUZZGEAR4_7
    {SPR_AGR4, 32775, 4, NULL, S_ARTIPUZZGEAR4_1, 0, 0},        // S_ARTIPUZZGEAR4_8
    // [crispy] disable full bright for Torch, use brightmap instead
    {SPR_TRCH, 0, 3, NULL, S_ARTI_TRCH2, 0, 0},     // S_ARTI_TRCH1
    {SPR_TRCH, 1, 3, NULL, S_ARTI_TRCH3, 0, 0},     // S_ARTI_TRCH2
    {SPR_TRCH, 2, 3, NULL, S_ARTI_TRCH1, 0, 0},     // S_ARTI_TRCH3
    {SPR_PSBG, 0, 20, NULL, S_FIREBOMB2, 0, 0}, // S_FIREBOMB1
    {SPR_PSBG, 0, 10, NULL, S_FIREBOMB3, 0, 0}, // S_FIREBOMB2
    {SPR_PSBG, 0, 10, NULL, S_FIREBOMB4, 0, 0}, // S_FIREBOMB3
    {SPR_PSBG, 1, 4, NULL, S_FIREBOMB5, 0, 0},  // S_FIREBOMB4
    {SPR_PSBG, 2, 4, A_Scream, S_FIREBOMB6, 0, 0},      // S_FIREBOMB5
    {SPR_XPL1, 32768, 4, A_Explode, S_FIREBOMB7, 0, 0}, // S_FIREBOMB6
    {SPR_XPL1, 32769, 4, NULL, S_FIREBOMB8, 0, 0},      // S_FIREBOMB7
    {SPR_XPL1, 32770, 4, NULL, S_FIREBOMB9, 0, 0},      // S_FIREBOMB8
    {SPR_XPL1, 32771, 4, NULL, S_FIREBOMB10, 0, 0},     // S_FIREBOMB9
    {SPR_XPL1, 32772, 4, NULL, S_FIREBOMB11, 0, 0},     // S_FIREBOMB10
    {SPR_XPL1, 32773, 4, NULL, S_NULL, 0, 0},   // S_FIREBOMB11
    {SPR_ATLP, 0, 4, NULL, S_ARTI_ATLP2, 0, 0}, // S_ARTI_ATLP1
    {SPR_ATLP, 1, 4, NULL, S_ARTI_ATLP3, 0, 0}, // S_ARTI_ATLP2
    {SPR_ATLP, 2, 4, NULL, S_ARTI_ATLP4, 0, 0}, // S_ARTI_ATLP3
    {SPR_ATLP, 1, 4, NULL, S_ARTI_ATLP1, 0, 0}, // S_ARTI_ATLP4
    {SPR_PSBG, 0, -1, NULL, S_NULL, 0, 0},      // S_ARTI_PSBG1
    {SPR_PSBG, 32768, 18, NULL, S_POISONBAG2, 0, 0},    // S_POISONBAG1
    {SPR_PSBG, 32769, 4, NULL, S_POISONBAG3, 0, 0},     // S_POISONBAG2
    {SPR_PSBG, 2, 3, NULL, S_POISONBAG4, 0, 0}, // S_POISONBAG3
    {SPR_PSBG, 2, 1, A_PoisonBagInit, S_NULL, 0, 0},    // S_POISONBAG4
    {SPR_PSBG, 3, 1, NULL, S_POISONCLOUD2, 0, 0},       // S_POISONCLOUD1
    {SPR_PSBG, 3, 1, A_Scream, S_POISONCLOUD3, 0, 0},   // S_POISONCLOUD2
    {SPR_PSBG, 3, 2, A_PoisonBagDamage, S_POISONCLOUD4, 0, 0},  // S_POISONCLOUD3
    {SPR_PSBG, 4, 2, A_PoisonBagDamage, S_POISONCLOUD5, 0, 0},  // S_POISONCLOUD4
    {SPR_PSBG, 4, 2, A_PoisonBagDamage, S_POISONCLOUD6, 0, 0},  // S_POISONCLOUD5
    {SPR_PSBG, 4, 2, A_PoisonBagDamage, S_POISONCLOUD7, 0, 0},  // S_POISONCLOUD6
    {SPR_PSBG, 5, 2, A_PoisonBagDamage, S_POISONCLOUD8, 0, 0},  // S_POISONCLOUD7
    {SPR_PSBG, 5, 2, A_PoisonBagDamage, S_POISONCLOUD9, 0, 0},  // S_POISONCLOUD8
    {SPR_PSBG, 5, 2, A_PoisonBagDamage, S_POISONCLOUD10, 0, 0}, // S_POISONCLOUD9
    {SPR_PSBG, 6, 2, A_PoisonBagDamage, S_POISONCLOUD11, 0, 0}, // S_POISONCLOUD10
    {SPR_PSBG, 6, 2, A_PoisonBagDamage, S_POISONCLOUD12, 0, 0}, // S_POISONCLOUD11
    {SPR_PSBG, 6, 2, A_PoisonBagDamage, S_POISONCLOUD13, 0, 0}, // S_POISONCLOUD12
    {SPR_PSBG, 7, 2, A_PoisonBagDamage, S_POISONCLOUD14, 0, 0}, // S_POISONCLOUD13
    {SPR_PSBG, 7, 2, A_PoisonBagDamage, S_POISONCLOUD15, 0, 0}, // S_POISONCLOUD14
    {SPR_PSBG, 7, 2, A_PoisonBagDamage, S_POISONCLOUD16, 0, 0}, // S_POISONCLOUD15
    {SPR_PSBG, 8, 2, A_PoisonBagDamage, S_POISONCLOUD17, 0, 0}, // S_POISONCLOUD16
    {SPR_PSBG, 8, 1, A_PoisonBagDamage, S_POISONCLOUD18, 0, 0}, // S_POISONCLOUD17
    {SPR_PSBG, 8, 1, A_PoisonBagCheck, S_POISONCLOUD4, 0, 0},   // S_POISONCLOUD18
    {SPR_PSBG, 7, 7, NULL, S_POISONCLOUD_X2, 0, 0},     // S_POISONCLOUD_X1
    {SPR_PSBG, 6, 7, NULL, S_POISONCLOUD_X3, 0, 0},     // S_POISONCLOUD_X2
    {SPR_PSBG, 5, 6, NULL, S_POISONCLOUD_X4, 0, 0},     // S_POISONCLOUD_X3
    {SPR_PSBG, 3, 6, NULL, S_NULL, 0, 0},       // S_POISONCLOUD_X4
    {SPR_THRW, 0, 4, A_CheckThrowBomb, S_THROWINGBOMB2, 0, 0},  // S_THROWINGBOMB1
    {SPR_THRW, 1, 3, A_CheckThrowBomb, S_THROWINGBOMB3, 0, 0},  // S_THROWINGBOMB2
    {SPR_THRW, 2, 3, A_CheckThrowBomb, S_THROWINGBOMB4, 0, 0},  // S_THROWINGBOMB3
    {SPR_THRW, 3, 3, A_CheckThrowBomb, S_THROWINGBOMB5, 0, 0},  // S_THROWINGBOMB4
    {SPR_THRW, 4, 3, A_CheckThrowBomb, S_THROWINGBOMB6, 0, 0},  // S_THROWINGBOMB5
    {SPR_THRW, 5, 3, A_CheckThrowBomb, S_THROWINGBOMB1, 0, 0},  // S_THROWINGBOMB6
    {SPR_THRW, 6, 6, A_CheckThrowBomb, S_THROWINGBOMB8, 0, 0},  // S_THROWINGBOMB7
    {SPR_THRW, 5, 4, A_CheckThrowBomb, S_THROWINGBOMB9, 0, 0},  // S_THROWINGBOMB8
    {SPR_THRW, 7, 6, A_CheckThrowBomb, S_THROWINGBOMB10, 0, 0}, // S_THROWINGBOMB9
    {SPR_THRW, 5, 4, A_CheckThrowBomb, S_THROWINGBOMB11, 0, 0}, // S_THROWINGBOMB10
    {SPR_THRW, 6, 6, A_CheckThrowBomb, S_THROWINGBOMB12, 0, 0}, // S_THROWINGBOMB11
    {SPR_THRW, 5, 3, A_CheckThrowBomb, S_THROWINGBOMB12, 0, 0}, // S_THROWINGBOMB12
    {SPR_CFCF, 32784, 4, A_NoGravity, S_THROWINGBOMB_X2, 0, 0}, // S_THROWINGBOMB_X1
    {SPR_CFCF, 32785, 3, A_Scream, S_THROWINGBOMB_X3, 0, 0},    // S_THROWINGBOMB_X2
    {SPR_CFCF, 32786, 4, A_Explode, S_THROWINGBOMB_X4, 0, 0},   // S_THROWINGBOMB_X3
    {SPR_CFCF, 32787, 3, NULL, S_THROWINGBOMB_X5, 0, 0},        // S_THROWINGBOMB_X4
    {SPR_CFCF, 32788, 4, NULL, S_THROWINGBOMB_X6, 0, 0},        // S_THROWINGBOMB_X5
    {SPR_CFCF, 32790, 3, NULL, S_THROWINGBOMB_X7, 0, 0},        // S_THROWINGBOMB_X6
    {SPR_CFCF, 32791, 4, NULL, S_THROWINGBOMB_X8, 0, 0},        // S_THROWINGBOMB_X7
    {SPR_CFCF, 32793, 3, NULL, S_NULL, 0, 0},   // S_THROWINGBOMB_X8
    // [crispy] disable full bright for Boots of Speed, use brightmap instead
    {SPR_SPED, 0, 3, NULL, S_ARTI_BOOTS2, 0, 0},    // S_ARTI_BOOTS1
    {SPR_SPED, 1, 3, NULL, S_ARTI_BOOTS3, 0, 0},    // S_ARTI_BOOTS2
    {SPR_SPED, 2, 3, NULL, S_ARTI_BOOTS4, 0, 0},    // S_ARTI_BOOTS3
    {SPR_SPED, 3, 3, NULL, S_ARTI_BOOTS5, 0, 0},    // S_ARTI_BOOTS4
    {SPR_SPED, 4, 3, NULL, S_ARTI_BOOTS6, 0, 0},    // S_ARTI_BOOTS5
    {SPR_SPED, 5, 3, NULL, S_ARTI_BOOTS7, 0, 0},    // S_ARTI_BOOTS6
    {SPR_SPED, 6, 3, NULL, S_ARTI_BOOTS8, 0, 0},    // S_ARTI_BOOTS7
    {SPR_SPED, 7, 3, NULL, S_ARTI_BOOTS1, 0, 0},    // S_ARTI_BOOTS8
    // [crispy] disable full bright for Krater of Might, use brightmap instead
    {SPR_BMAN, 0, -1, NULL, S_NULL, 0, 0},  // S_ARTI_MANA
    // [crispy] disable full bright for Dragonskin Bracers, use brightmap instead
    {SPR_BRAC, 0, 4, NULL, S_ARTI_ARMOR2, 0, 0},    // S_ARTI_ARMOR1
    {SPR_BRAC, 1, 4, NULL, S_ARTI_ARMOR3, 0, 0},    // S_ARTI_ARMOR2
    {SPR_BRAC, 2, 4, NULL, S_ARTI_ARMOR4, 0, 0},    // S_ARTI_ARMOR3
    {SPR_BRAC, 3, 4, NULL, S_ARTI_ARMOR5, 0, 0},    // S_ARTI_ARMOR4
    {SPR_BRAC, 4, 4, NULL, S_ARTI_ARMOR6, 0, 0},    // S_ARTI_ARMOR5
    {SPR_BRAC, 5, 4, NULL, S_ARTI_ARMOR7, 0, 0},    // S_ARTI_ARMOR6
    {SPR_BRAC, 6, 4, NULL, S_ARTI_ARMOR8, 0, 0},    // S_ARTI_ARMOR7
    {SPR_BRAC, 7, 4, NULL, S_ARTI_ARMOR1, 0, 0},    // S_ARTI_ARMOR8
    {SPR_BLST, 32768, 4, NULL, S_ARTI_BLAST2, 0, 0},    // S_ARTI_BLAST1
    {SPR_BLST, 32769, 4, NULL, S_ARTI_BLAST3, 0, 0},    // S_ARTI_BLAST2
    {SPR_BLST, 32770, 4, NULL, S_ARTI_BLAST4, 0, 0},    // S_ARTI_BLAST3
    {SPR_BLST, 32771, 4, NULL, S_ARTI_BLAST5, 0, 0},    // S_ARTI_BLAST4
    {SPR_BLST, 32772, 4, NULL, S_ARTI_BLAST6, 0, 0},    // S_ARTI_BLAST5
    {SPR_BLST, 32773, 4, NULL, S_ARTI_BLAST7, 0, 0},    // S_ARTI_BLAST6
    {SPR_BLST, 32774, 4, NULL, S_ARTI_BLAST8, 0, 0},    // S_ARTI_BLAST7
    {SPR_BLST, 32775, 4, NULL, S_ARTI_BLAST1, 0, 0},    // S_ARTI_BLAST8
    // [crispy] disable full bright for Mystic Ambit Incant, use brightmap instead
    {SPR_HRAD, 0, 4, NULL, S_ARTI_HEALRAD2, 0, 0},  // S_ARTI_HEALRAD1
    {SPR_HRAD, 1, 4, NULL, S_ARTI_HEALRAD3, 0, 0},  // S_ARTI_HEALRAD2
    {SPR_HRAD, 2, 4, NULL, S_ARTI_HEALRAD4, 0, 0},  // S_ARTI_HEALRAD3
    {SPR_HRAD, 3, 4, NULL, S_ARTI_HEALRAD5, 0, 0},  // S_ARTI_HEALRAD4
    {SPR_HRAD, 4, 4, NULL, S_ARTI_HEALRAD6, 0, 0},  // S_ARTI_HEALRAD5
    {SPR_HRAD, 5, 4, NULL, S_ARTI_HEALRAD7, 0, 0},  // S_ARTI_HEALRAD6
    {SPR_HRAD, 6, 4, NULL, S_ARTI_HEALRAD8, 0, 0},  // S_ARTI_HEALRAD7
    {SPR_HRAD, 7, 4, NULL, S_ARTI_HEALRAD9, 0, 0},  // S_ARTI_HEALRAD8
    {SPR_HRAD, 8, 4, NULL, S_ARTI_HEALRAD0, 0, 0},  // S_ARTI_HEALRAD9
    {SPR_HRAD, 9, 4, NULL, S_ARTI_HEALRADA, 0, 0},  // S_ARTI_HEALRAD0
    {SPR_HRAD, 10, 4, NULL, S_ARTI_HEALRADB, 0, 0},  // S_ARTI_HEALRADA
    {SPR_HRAD, 11, 4, NULL, S_ARTI_HEALRADC, 0, 0},  // S_ARTI_HEALRADB
    {SPR_HRAD, 12, 4, NULL, S_ARTI_HEALRADD, 0, 0},  // S_ARTI_HEALRADC
    {SPR_HRAD, 13, 4, NULL, S_ARTI_HEALRADE, 0, 0},  // S_ARTI_HEALRADD
    {SPR_HRAD, 14, 4, NULL, S_ARTI_HEALRADF, 0, 0},  // S_ARTI_HEALRADE
    {SPR_HRAD, 15, 4, NULL, S_ARTI_HEALRAD1, 0, 0},  // S_ARTI_HEALRADF
    {SPR_SPSH, 0, 8, NULL, S_SPLASH2, 0, 0},    // S_SPLASH1
    {SPR_SPSH, 1, 8, NULL, S_SPLASH3, 0, 0},    // S_SPLASH2
    {SPR_SPSH, 2, 8, NULL, S_SPLASH4, 0, 0},    // S_SPLASH3
    {SPR_SPSH, 3, 16, NULL, S_NULL, 0, 0},      // S_SPLASH4
    {SPR_SPSH, 3, 10, NULL, S_NULL, 0, 0},      // S_SPLASHX
    {SPR_SPSH, 4, 5, NULL, S_SPLASHBASE2, 0, 0},        // S_SPLASHBASE1
    {SPR_SPSH, 5, 5, NULL, S_SPLASHBASE3, 0, 0},        // S_SPLASHBASE2
    {SPR_SPSH, 6, 5, NULL, S_SPLASHBASE4, 0, 0},        // S_SPLASHBASE3
    {SPR_SPSH, 7, 5, NULL, S_SPLASHBASE5, 0, 0},        // S_SPLASHBASE4
    {SPR_SPSH, 8, 5, NULL, S_SPLASHBASE6, 0, 0},        // S_SPLASHBASE5
    {SPR_SPSH, 9, 5, NULL, S_SPLASHBASE7, 0, 0},        // S_SPLASHBASE6
    {SPR_SPSH, 10, 5, NULL, S_NULL, 0, 0},      // S_SPLASHBASE7
    {SPR_LVAS, 32768, 5, NULL, S_LAVASPLASH2, 0, 0},    // S_LAVASPLASH1
    {SPR_LVAS, 32769, 5, NULL, S_LAVASPLASH3, 0, 0},    // S_LAVASPLASH2
    {SPR_LVAS, 32770, 5, NULL, S_LAVASPLASH4, 0, 0},    // S_LAVASPLASH3
    {SPR_LVAS, 32771, 5, NULL, S_LAVASPLASH5, 0, 0},    // S_LAVASPLASH4
    {SPR_LVAS, 32772, 5, NULL, S_LAVASPLASH6, 0, 0},    // S_LAVASPLASH5
    {SPR_LVAS, 32773, 5, NULL, S_NULL, 0, 0},   // S_LAVASPLASH6
    {SPR_LVAS, 32774, 5, NULL, S_LAVASMOKE2, 0, 0},     // S_LAVASMOKE1
    {SPR_LVAS, 32775, 5, NULL, S_LAVASMOKE3, 0, 0},     // S_LAVASMOKE2
    {SPR_LVAS, 32776, 5, NULL, S_LAVASMOKE4, 0, 0},     // S_LAVASMOKE3
    {SPR_LVAS, 32777, 5, NULL, S_LAVASMOKE5, 0, 0},     // S_LAVASMOKE4
    {SPR_LVAS, 32778, 5, NULL, S_NULL, 0, 0},   // S_LAVASMOKE5
    {SPR_SLDG, 0, 8, NULL, S_SLUDGECHUNK2, 0, 0},       // S_SLUDGECHUNK1
    {SPR_SLDG, 1, 8, NULL, S_SLUDGECHUNK3, 0, 0},       // S_SLUDGECHUNK2
    {SPR_SLDG, 2, 8, NULL, S_SLUDGECHUNK4, 0, 0},       // S_SLUDGECHUNK3
    {SPR_SLDG, 3, 8, NULL, S_NULL, 0, 0},       // S_SLUDGECHUNK4
    {SPR_SLDG, 3, 6, NULL, S_NULL, 0, 0},       // S_SLUDGECHUNKX
    {SPR_SLDG, 4, 6, NULL, S_SLUDGESPLASH2, 0, 0},      // S_SLUDGESPLASH1
    {SPR_SLDG, 5, 6, NULL, S_SLUDGESPLASH3, 0, 0},      // S_SLUDGESPLASH2
    {SPR_SLDG, 6, 6, NULL, S_SLUDGESPLASH4, 0, 0},      // S_SLUDGESPLASH3
    {SPR_SLDG, 7, 6, NULL, S_NULL, 0, 0},       // S_SLUDGESPLASH4
    {SPR_STTW, 0, -1, NULL, S_NULL, 0, 0},      // S_ZWINGEDSTATUE1
    {SPR_RCK1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZROCK1_1
    {SPR_RCK2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZROCK2_1
    {SPR_RCK3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZROCK3_1
    {SPR_RCK4, 0, -1, NULL, S_NULL, 0, 0},      // S_ZROCK4_1
    {SPR_CDLR, 0, 4, NULL, S_ZCHANDELIER2, 0, 0},       // S_ZCHANDELIER1
    {SPR_CDLR, 1, 4, NULL, S_ZCHANDELIER3, 0, 0},       // S_ZCHANDELIER2
    {SPR_CDLR, 2, 4, NULL, S_ZCHANDELIER1, 0, 0},       // S_ZCHANDELIER3
    {SPR_CDLR, 3, -1, NULL, S_NULL, 0, 0},      // S_ZCHANDELIER_U
    {SPR_TRE1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTREEDEAD1
    {SPR_TRE1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTREE
    {SPR_TRDT, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTREEDESTRUCTIBLE1
    {SPR_TRDT, 1, 5, NULL, S_ZTREEDES_D2, 0, 0},        // S_ZTREEDES_D1
    {SPR_TRDT, 2, 5, A_Scream, S_ZTREEDES_D3, 0, 0},    // S_ZTREEDES_D2
    {SPR_TRDT, 3, 5, NULL, S_ZTREEDES_D4, 0, 0},        // S_ZTREEDES_D3
    {SPR_TRDT, 4, 5, NULL, S_ZTREEDES_D5, 0, 0},        // S_ZTREEDES_D4
    {SPR_TRDT, 5, 5, NULL, S_ZTREEDES_D6, 0, 0},        // S_ZTREEDES_D5
    {SPR_TRDT, 6, -1, NULL, S_NULL, 0, 0},      // S_ZTREEDES_D6
    {SPR_TRDT, 32775, 5, NULL, S_ZTREEDES_X2, 0, 0},    // S_ZTREEDES_X1
    {SPR_TRDT, 32776, 5, NULL, S_ZTREEDES_X3, 0, 0},    // S_ZTREEDES_X2
    {SPR_TRDT, 32777, 5, NULL, S_ZTREEDES_X4, 0, 0},    // S_ZTREEDES_X3
    {SPR_TRDT, 32778, 5, NULL, S_ZTREEDES_X5, 0, 0},    // S_ZTREEDES_X4
    {SPR_TRDT, 32779, 5, NULL, S_ZTREEDES_X6, 0, 0},    // S_ZTREEDES_X5
    {SPR_TRDT, 32780, 5, A_Explode, S_ZTREEDES_X7, 0, 0},       // S_ZTREEDES_X6
    {SPR_TRDT, 32781, 5, NULL, S_ZTREEDES_X8, 0, 0},    // S_ZTREEDES_X7
    {SPR_TRDT, 14, 5, NULL, S_ZTREEDES_X9, 0, 0},       // S_ZTREEDES_X8
    {SPR_TRDT, 15, 5, NULL, S_ZTREEDES_X10, 0, 0},      // S_ZTREEDES_X9
    {SPR_TRDT, 16, -1, NULL, S_NULL, 0, 0},     // S_ZTREEDES_X10
    {SPR_TRE2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTREESWAMP182_1
    {SPR_TRE3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTREESWAMP172_1
    {SPR_STM1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTUMPBURNED1
    {SPR_STM2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTUMPBARE1
    {SPR_STM3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTUMPSWAMP1_1
    {SPR_STM4, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTUMPSWAMP2_1
    {SPR_MSH1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSHROOMLARGE1_1
    {SPR_MSH2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSHROOMLARGE2_1
    {SPR_MSH3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSHROOMLARGE3_1
    {SPR_MSH4, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSHROOMSMALL1_1
    {SPR_MSH5, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSHROOMSMALL2_1
    {SPR_MSH6, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSHROOMSMALL3_1
    {SPR_MSH7, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSHROOMSMALL4_1
    {SPR_MSH8, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSHROOMSMALL5_1
    {SPR_SGMP, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALAGMITEPILLAR1
    {SPR_SGM1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALAGMITELARGE1
    {SPR_SGM2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALAGMITEMEDIUM1
    {SPR_SGM3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALAGMITESMALL1
    {SPR_SLC1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALACTITELARGE1
    {SPR_SLC2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALACTITEMEDIUM1
    {SPR_SLC3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALACTITESMALL1
    {SPR_MSS1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZMOSSCEILING1_1
    {SPR_MSS2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZMOSSCEILING2_1
    {SPR_SWMV, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSWAMPVINE1
    {SPR_CPS1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZCORPSEKABOB1
    {SPR_CPS2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZCORPSESLEEPING1
    {SPR_TMS1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTOMBSTONERIP1
    {SPR_TMS2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTOMBSTONESHANE1
    {SPR_TMS3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTOMBSTONEBIGCROSS1
    {SPR_TMS4, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTOMBSTONEBRIANR1
    {SPR_TMS5, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTOMBSTONECROSSCIRCLE1
    {SPR_TMS6, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTOMBSTONESMALLCROSS1
    {SPR_TMS7, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTOMBSTONEBRIANP1
    {SPR_CPS3, 0, -1, NULL, S_NULL, 0, 0},      // S_CORPSEHANGING_1
    {SPR_STT2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLEGREENTALL_1
    {SPR_STT3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLEBLUETALL_1
    {SPR_STT4, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLEGREENSHORT_1
    {SPR_STT5, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLEBLUESHORT_1
    {SPR_GAR1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLESTRIPETALL_1
    {SPR_GAR2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLEDARKREDTALL_1
    {SPR_GAR3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLEREDTALL_1
    {SPR_GAR4, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLETANTALL_1
    {SPR_GAR5, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLERUSTTALL_1
    {SPR_GAR6, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLEDARKREDSHORT_1
    {SPR_GAR7, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLEREDSHORT_1
    {SPR_GAR8, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLETANSHORT_1
    {SPR_GAR9, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTATUEGARGOYLERUSTSHORT_1
    {SPR_BNR1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZBANNERTATTERED_1
    {SPR_TRE4, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTREELARGE1
    {SPR_TRE5, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTREELARGE2
    {SPR_TRE6, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTREEGNARLED1
    {SPR_TRE7, 0, -1, NULL, S_NULL, 0, 0},      // S_ZTREEGNARLED2
    {SPR_LOGG, 0, -1, NULL, S_NULL, 0, 0},      // S_ZLOG
    {SPR_ICT1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALACTITEICELARGE
    {SPR_ICT2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALACTITEICEMEDIUM
    {SPR_ICT3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALACTITEICESMALL
    {SPR_ICT4, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALACTITEICETINY
    {SPR_ICM1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALAGMITEICELARGE
    {SPR_ICM2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALAGMITEICEMEDIUM
    {SPR_ICM3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALAGMITEICESMALL
    {SPR_ICM4, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSTALAGMITEICETINY
    {SPR_RKBL, 0, -1, NULL, S_NULL, 0, 0},      // S_ZROCKBROWN1
    {SPR_RKBS, 0, -1, NULL, S_NULL, 0, 0},      // S_ZROCKBROWN2
    {SPR_RKBK, 0, -1, NULL, S_NULL, 0, 0},      // S_ZROCKBLACK
    {SPR_RBL1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZRUBBLE1
    {SPR_RBL2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZRUBBLE2
    {SPR_RBL3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZRUBBLE3
    {SPR_VASE, 0, -1, NULL, S_NULL, 0, 0},      // S_ZVASEPILLAR
    {SPR_POT1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZPOTTERY1
    {SPR_POT2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZPOTTERY2
    {SPR_POT3, 0, -1, NULL, S_NULL, 0, 0},      // S_ZPOTTERY3
    {SPR_POT1, 0, 0, A_PotteryExplode, S_NULL, 0, 0},   // S_ZPOTTERY_EXPLODE
    {SPR_PBIT, 0, -1, NULL, S_NULL, 0, 0},      // S_POTTERYBIT_1
    {SPR_PBIT, 1, -1, NULL, S_NULL, 0, 0},      // S_POTTERYBIT_2
    {SPR_PBIT, 2, -1, NULL, S_NULL, 0, 0},      // S_POTTERYBIT_3
    {SPR_PBIT, 3, -1, NULL, S_NULL, 0, 0},      // S_POTTERYBIT_4
    {SPR_PBIT, 4, -1, NULL, S_NULL, 0, 0},      // S_POTTERYBIT_5
    {SPR_PBIT, 5, 0, A_PotteryChooseBit, S_NULL, 0, 0}, // S_POTTERYBIT_EX0
    {SPR_PBIT, 5, 140, NULL, S_POTTERYBIT_EX1_2, 0, 0}, // S_POTTERYBIT_EX1
    {SPR_PBIT, 5, 1, A_PotteryCheck, S_NULL, 0, 0},     // S_POTTERYBIT_EX1_2
    {SPR_PBIT, 6, 140, NULL, S_POTTERYBIT_EX2_2, 0, 0}, // S_POTTERYBIT_EX2
    {SPR_PBIT, 6, 1, A_PotteryCheck, S_NULL, 0, 0},     // S_POTTERYBIT_EX2_2
    {SPR_PBIT, 7, 140, NULL, S_POTTERYBIT_EX3_2, 0, 0}, // S_POTTERYBIT_EX3
    {SPR_PBIT, 7, 1, A_PotteryCheck, S_NULL, 0, 0},     // S_POTTERYBIT_EX3_2
    {SPR_PBIT, 8, 140, NULL, S_POTTERYBIT_EX4_2, 0, 0}, // S_POTTERYBIT_EX4
    {SPR_PBIT, 8, 1, A_PotteryCheck, S_NULL, 0, 0},     // S_POTTERYBIT_EX4_2
    {SPR_PBIT, 9, 140, NULL, S_POTTERYBIT_EX5_2, 0, 0}, // S_POTTERYBIT_EX5
    {SPR_PBIT, 9, 1, A_PotteryCheck, S_NULL, 0, 0},     // S_POTTERYBIT_EX5_2
    {SPR_CPS4, 0, -1, NULL, S_NULL, 0, 0},      // S_ZCORPSELYNCHED1
    {SPR_CPS5, 0, 140, A_CorpseBloodDrip, S_ZCORPSELYNCHED2, 0, 0},     // S_ZCORPSELYNCHED2
    {SPR_CPS6, 0, -1, NULL, S_NULL, 0, 0},      // S_ZCORPSESITTING
    {SPR_CPS6, 0, 1, A_CorpseExplode, S_NULL, 0, 0},    // S_ZCORPSESITTING_X
    {SPR_CPB1, 0, -1, NULL, S_NULL, 0, 0},      // S_CORPSEBIT_1
    {SPR_CPB2, 0, -1, NULL, S_NULL, 0, 0},      // S_CORPSEBIT_2
    {SPR_CPB3, 0, -1, NULL, S_NULL, 0, 0},      // S_CORPSEBIT_3
    {SPR_CPB4, 0, -1, NULL, S_NULL, 0, 0},      // S_CORPSEBIT_4
    {SPR_BDRP, 0, -1, NULL, S_NULL, 0, 0},      // S_CORPSEBLOODDRIP
    {SPR_BDSH, 0, 3, NULL, S_CORPSEBLOODDRIP_X2, 0, 0}, // S_CORPSEBLOODDRIP_X1
    {SPR_BDSH, 1, 3, NULL, S_CORPSEBLOODDRIP_X3, 0, 0}, // S_CORPSEBLOODDRIP_X2
    {SPR_BDSH, 2, 2, NULL, S_CORPSEBLOODDRIP_X4, 0, 0}, // S_CORPSEBLOODDRIP_X3
    {SPR_BDSH, 3, 2, NULL, S_NULL, 0, 0},       // S_CORPSEBLOODDRIP_X4
    {SPR_BDPL, 0, -1, NULL, S_NULL, 0, 0},      // S_BLOODPOOL
    {SPR_CNDL, 32768, 4, NULL, S_ZCANDLE2, 0, 0},       // S_ZCANDLE1
    {SPR_CNDL, 32769, 4, NULL, S_ZCANDLE3, 0, 0},       // S_ZCANDLE2
    {SPR_CNDL, 32770, 4, NULL, S_ZCANDLE1, 0, 0},       // S_ZCANDLE3
    {SPR_MAN1, 0, 20, A_LeafSpawn, S_ZLEAFSPAWNER, 0, 0},       // S_ZLEAFSPAWNER
    {SPR_LEF1, 0, 4, NULL, S_LEAF1_2, 0, 0},    // S_LEAF1_1
    {SPR_LEF1, 1, 4, NULL, S_LEAF1_3, 0, 0},    // S_LEAF1_2
    {SPR_LEF1, 2, 4, NULL, S_LEAF1_4, 0, 0},    // S_LEAF1_3
    {SPR_LEF1, 3, 4, A_LeafThrust, S_LEAF1_5, 0, 0},    // S_LEAF1_4
    {SPR_LEF1, 4, 4, NULL, S_LEAF1_6, 0, 0},    // S_LEAF1_5
    {SPR_LEF1, 5, 4, NULL, S_LEAF1_7, 0, 0},    // S_LEAF1_6
    {SPR_LEF1, 6, 4, NULL, S_LEAF1_8, 0, 0},    // S_LEAF1_7
    {SPR_LEF1, 7, 4, A_LeafThrust, S_LEAF1_9, 0, 0},    // S_LEAF1_8
    {SPR_LEF1, 8, 4, NULL, S_LEAF1_10, 0, 0},   // S_LEAF1_9
    {SPR_LEF1, 0, 4, NULL, S_LEAF1_11, 0, 0},   // S_LEAF1_10
    {SPR_LEF1, 1, 4, NULL, S_LEAF1_12, 0, 0},   // S_LEAF1_11
    {SPR_LEF1, 2, 4, A_LeafThrust, S_LEAF1_13, 0, 0},   // S_LEAF1_12
    {SPR_LEF1, 3, 4, NULL, S_LEAF1_14, 0, 0},   // S_LEAF1_13
    {SPR_LEF1, 4, 4, NULL, S_LEAF1_15, 0, 0},   // S_LEAF1_14
    {SPR_LEF1, 5, 4, NULL, S_LEAF1_16, 0, 0},   // S_LEAF1_15
    {SPR_LEF1, 6, 4, A_LeafThrust, S_LEAF1_17, 0, 0},   // S_LEAF1_16
    {SPR_LEF1, 7, 4, NULL, S_LEAF1_18, 0, 0},   // S_LEAF1_17
    {SPR_LEF1, 8, 4, NULL, S_NULL, 0, 0},       // S_LEAF1_18
    {SPR_LEF3, 3, 10, A_LeafCheck, S_LEAF_X1, 0, 0},    // S_LEAF_X1
    {SPR_LEF2, 0, 4, NULL, S_LEAF2_2, 0, 0},    // S_LEAF2_1
    {SPR_LEF2, 1, 4, NULL, S_LEAF2_3, 0, 0},    // S_LEAF2_2
    {SPR_LEF2, 2, 4, NULL, S_LEAF2_4, 0, 0},    // S_LEAF2_3
    {SPR_LEF2, 3, 4, A_LeafThrust, S_LEAF2_5, 0, 0},    // S_LEAF2_4
    {SPR_LEF2, 4, 4, NULL, S_LEAF2_6, 0, 0},    // S_LEAF2_5
    {SPR_LEF2, 5, 4, NULL, S_LEAF2_7, 0, 0},    // S_LEAF2_6
    {SPR_LEF2, 6, 4, NULL, S_LEAF2_8, 0, 0},    // S_LEAF2_7
    {SPR_LEF2, 7, 4, A_LeafThrust, S_LEAF2_9, 0, 0},    // S_LEAF2_8
    {SPR_LEF2, 8, 4, NULL, S_LEAF2_10, 0, 0},   // S_LEAF2_9
    {SPR_LEF2, 0, 4, NULL, S_LEAF2_11, 0, 0},   // S_LEAF2_10
    {SPR_LEF2, 1, 4, NULL, S_LEAF2_12, 0, 0},   // S_LEAF2_11
    {SPR_LEF2, 2, 4, A_LeafThrust, S_LEAF2_13, 0, 0},   // S_LEAF2_12
    {SPR_LEF2, 3, 4, NULL, S_LEAF2_14, 0, 0},   // S_LEAF2_13
    {SPR_LEF2, 4, 4, NULL, S_LEAF2_15, 0, 0},   // S_LEAF2_14
    {SPR_LEF2, 5, 4, NULL, S_LEAF2_16, 0, 0},   // S_LEAF2_15
    {SPR_LEF2, 6, 4, A_LeafThrust, S_LEAF2_17, 0, 0},   // S_LEAF2_16
    {SPR_LEF2, 7, 4, NULL, S_LEAF2_18, 0, 0},   // S_LEAF2_17
    {SPR_LEF2, 8, 4, NULL, S_NULL, 0, 0},       // S_LEAF2_18
    // [crispy] disable full bright for Twined Torch, use brightmap instead
    {SPR_TWTR, 0, 4, NULL, S_ZTWINEDTORCH_2, 0, 0}, // S_ZTWINEDTORCH_1
    {SPR_TWTR, 1, 4, NULL, S_ZTWINEDTORCH_3, 0, 0}, // S_ZTWINEDTORCH_2
    {SPR_TWTR, 2, 4, NULL, S_ZTWINEDTORCH_4, 0, 0}, // S_ZTWINEDTORCH_3
    {SPR_TWTR, 3, 4, NULL, S_ZTWINEDTORCH_5, 0, 0}, // S_ZTWINEDTORCH_4
    {SPR_TWTR, 4, 4, NULL, S_ZTWINEDTORCH_6, 0, 0}, // S_ZTWINEDTORCH_5
    {SPR_TWTR, 5, 4, NULL, S_ZTWINEDTORCH_7, 0, 0}, // S_ZTWINEDTORCH_6
    {SPR_TWTR, 6, 4, NULL, S_ZTWINEDTORCH_8, 0, 0}, // S_ZTWINEDTORCH_7
    {SPR_TWTR, 7, 4, NULL, S_ZTWINEDTORCH_1, 0, 0}, // S_ZTWINEDTORCH_8
    {SPR_TWTR, 8, -1, NULL, S_NULL, 0, 0},      // S_ZTWINEDTORCH_UNLIT
    {SPR_TLGL, 0, 2, NULL, S_BRIDGE2, 0, 0},    // S_BRIDGE1
    {SPR_TLGL, 0, 2, A_BridgeInit, S_BRIDGE3, 0, 0},    // S_BRIDGE2
    {SPR_TLGL, 0, -1, NULL, S_NULL, 0, 0},      // S_BRIDGE3
    {SPR_TLGL, 0, 2, NULL, S_FREE_BRIDGE2, 0, 0},       // S_FREE_BRIDGE1
    {SPR_TLGL, 0, 300, NULL, S_NULL, 0, 0},     // S_FREE_BRIDGE2
    {SPR_TLGL, 0, 2, NULL, S_BBALL2, 0, 0},     // S_BBALL1
    {SPR_TLGL, 0, 5, A_BridgeOrbit, S_BBALL2, 0, 0},    // S_BBALL2
    {SPR_WLTR, 32768, 5, NULL, S_ZWALLTORCH2, 0, 0},    // S_ZWALLTORCH1
    {SPR_WLTR, 32769, 5, NULL, S_ZWALLTORCH3, 0, 0},    // S_ZWALLTORCH2
    {SPR_WLTR, 32770, 5, NULL, S_ZWALLTORCH4, 0, 0},    // S_ZWALLTORCH3
    {SPR_WLTR, 32771, 5, NULL, S_ZWALLTORCH5, 0, 0},    // S_ZWALLTORCH4
    {SPR_WLTR, 32772, 5, NULL, S_ZWALLTORCH6, 0, 0},    // S_ZWALLTORCH5
    {SPR_WLTR, 32773, 5, NULL, S_ZWALLTORCH7, 0, 0},    // S_ZWALLTORCH6
    {SPR_WLTR, 32774, 5, NULL, S_ZWALLTORCH8, 0, 0},    // S_ZWALLTORCH7
    {SPR_WLTR, 32775, 5, NULL, S_ZWALLTORCH1, 0, 0},    // S_ZWALLTORCH8
    {SPR_WLTR, 8, -1, NULL, S_NULL, 0, 0},      // S_ZWALLTORCH_U
    {SPR_BARL, 0, -1, NULL, S_NULL, 0, 0},      // S_ZBARREL1
    {SPR_SHB1, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSHRUB1
    {SPR_SHB1, 0, 1, A_TreeDeath, S_ZSHRUB1, 0, 0},     // S_ZSHRUB1_DIE
    {SPR_SHB1, 32769, 7, NULL, S_ZSHRUB1_X2, 0, 0},     // S_ZSHRUB1_X1
    {SPR_SHB1, 32770, 6, A_Scream, S_ZSHRUB1_X3, 0, 0}, // S_ZSHRUB1_X2
    {SPR_SHB1, 32771, 5, NULL, S_NULL, 0, 0},   // S_ZSHRUB1_X3
    {SPR_SHB2, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSHRUB2
    {SPR_SHB2, 0, 1, A_TreeDeath, S_ZSHRUB2, 0, 0},     // S_ZSHRUB2_DIE
    {SPR_SHB2, 32769, 7, NULL, S_ZSHRUB2_X2, 0, 0},     // S_ZSHRUB2_X1
    {SPR_SHB2, 32770, 6, A_Scream, S_ZSHRUB2_X3, 0, 0}, // S_ZSHRUB2_X2
    {SPR_SHB2, 32771, 5, A_Explode, S_ZSHRUB2_X4, 0, 0},        // S_ZSHRUB2_X3
    {SPR_SHB2, 32772, 5, NULL, S_NULL, 0, 0},   // S_ZSHRUB2_X4
    {SPR_BCKT, 0, -1, NULL, S_NULL, 0, 0},      // S_ZBUCKET1
    {SPR_SHRM, 0, 5, A_PoisonShroom, S_ZPOISONSHROOM_P2, 0, 0}, // S_ZPOISONSHROOM1
    {SPR_SHRM, 0, 6, NULL, S_ZPOISONSHROOM_P2, 0, 0},   // S_ZPOISONSHROOM_P1
    {SPR_SHRM, 1, 8, A_Pain, S_ZPOISONSHROOM1, 0, 0},   // S_ZPOISONSHROOM_P2
    {SPR_SHRM, 2, 5, NULL, S_ZPOISONSHROOM_X2, 0, 0},   // S_ZPOISONSHROOM_X1
    {SPR_SHRM, 3, 5, NULL, S_ZPOISONSHROOM_X3, 0, 0},   // S_ZPOISONSHROOM_X2
    {SPR_SHRM, 4, 5, A_PoisonBagInit, S_ZPOISONSHROOM_X4, 0, 0},        // S_ZPOISONSHROOM_X3
    {SPR_SHRM, 5, -1, NULL, S_NULL, 0, 0},      // S_ZPOISONSHROOM_X4
    // [crispy] disable full bright for Fire Bull, use brightmap instead
    {SPR_FBUL, 0, 4, NULL, S_ZFIREBULL2, 0, 0},     // S_ZFIREBULL1
    {SPR_FBUL, 1, 4, NULL, S_ZFIREBULL3, 0, 0},     // S_ZFIREBULL2
    {SPR_FBUL, 2, 4, NULL, S_ZFIREBULL4, 0, 0},     // S_ZFIREBULL3
    {SPR_FBUL, 3, 4, NULL, S_ZFIREBULL5, 0, 0},     // S_ZFIREBULL4
    {SPR_FBUL, 4, 4, NULL, S_ZFIREBULL6, 0, 0},     // S_ZFIREBULL5
    {SPR_FBUL, 5, 4, NULL, S_ZFIREBULL7, 0, 0},     // S_ZFIREBULL6
    {SPR_FBUL, 6, 4, NULL, S_ZFIREBULL1, 0, 0},     // S_ZFIREBULL7
    {SPR_FBUL, 9, 4, NULL, S_ZFIREBULL_DEATH2, 0, 0},       // S_ZFIREBULL_DEATH
    {SPR_FBUL, 8, 4, NULL, S_ZFIREBULL_U, 0, 0},    // S_ZFIREBULL_DEATH2
    {SPR_FBUL, 7, -1, NULL, S_NULL, 0, 0},      // S_ZFIREBULL_U
    {SPR_FBUL, 8, 4, NULL, S_ZFIREBULL_BIRTH2, 0, 0},       // S_ZFIREBULL_BIRTH
    {SPR_FBUL, 9, 4, NULL, S_ZFIREBULL1, 0, 0},     // S_ZFIREBULL_BIRTH2
    // [crispy] disable full bright for Skull with Flame, use brightmap instead
    {SPR_FSKL, 0, 4, NULL, S_ZFIRETHING2, 0, 0},    // S_ZFIRETHING1
    {SPR_FSKL, 1, 3, NULL, S_ZFIRETHING3, 0, 0},    // S_ZFIRETHING2
    {SPR_FSKL, 2, 4, NULL, S_ZFIRETHING4, 0, 0},    // S_ZFIRETHING3
    {SPR_FSKL, 3, 3, NULL, S_ZFIRETHING5, 0, 0},    // S_ZFIRETHING4
    {SPR_FSKL, 4, 4, NULL, S_ZFIRETHING6, 0, 0},    // S_ZFIRETHING5
    {SPR_FSKL, 5, 3, NULL, S_ZFIRETHING7, 0, 0},    // S_ZFIRETHING6
    {SPR_FSKL, 6, 4, NULL, S_ZFIRETHING8, 0, 0},    // S_ZFIRETHING7
    {SPR_FSKL, 7, 3, NULL, S_ZFIRETHING9, 0, 0},    // S_ZFIRETHING8
    {SPR_FSKL, 8, 4, NULL, S_ZFIRETHING1, 0, 0},    // S_ZFIRETHING9
    // [crispy] disable full bright for Brazier, use brightmap instead
    {SPR_BRTR, 0, 4, NULL, S_ZBRASSTORCH2, 0, 0},   // S_ZBRASSTORCH1
    {SPR_BRTR, 1, 4, NULL, S_ZBRASSTORCH3, 0, 0},   // S_ZBRASSTORCH2
    {SPR_BRTR, 2, 4, NULL, S_ZBRASSTORCH4, 0, 0},   // S_ZBRASSTORCH3
    {SPR_BRTR, 3, 4, NULL, S_ZBRASSTORCH5, 0, 0},   // S_ZBRASSTORCH4
    {SPR_BRTR, 4, 4, NULL, S_ZBRASSTORCH6, 0, 0},   // S_ZBRASSTORCH5
    {SPR_BRTR, 5, 4, NULL, S_ZBRASSTORCH7, 0, 0},   // S_ZBRASSTORCH6
    {SPR_BRTR, 6, 4, NULL, S_ZBRASSTORCH8, 0, 0},   // S_ZBRASSTORCH7
    {SPR_BRTR, 7, 4, NULL, S_ZBRASSTORCH9, 0, 0},   // S_ZBRASSTORCH8
    {SPR_BRTR, 8, 4, NULL, S_ZBRASSTORCH10, 0, 0},  // S_ZBRASSTORCH9
    {SPR_BRTR, 9, 4, NULL, S_ZBRASSTORCH11, 0, 0},  // S_ZBRASSTORCH10
    {SPR_BRTR, 10, 4, NULL, S_ZBRASSTORCH12, 0, 0},  // S_ZBRASSTORCH11
    {SPR_BRTR, 11, 4, NULL, S_ZBRASSTORCH13, 0, 0},  // S_ZBRASSTORCH12
    {SPR_BRTR, 12, 4, NULL, S_ZBRASSTORCH1, 0, 0},   // S_ZBRASSTORCH13
    {SPR_SUIT, 0, -1, NULL, S_NULL, 0, 0},      // S_ZSUITOFARMOR
    {SPR_SUIT, 0, 1, A_SoAExplode, S_NULL, 0, 0},       // S_ZSUITOFARMOR_X1
    {SPR_SUIT, 1, -1, NULL, S_NULL, 0, 0},      // S_ZARMORCHUNK1
    {SPR_SUIT, 2, -1, NULL, S_NULL, 0, 0},      // S_ZARMORCHUNK2
    {SPR_SUIT, 3, -1, NULL, S_NULL, 0, 0},      // S_ZARMORCHUNK3
    {SPR_SUIT, 4, -1, NULL, S_NULL, 0, 0},      // S_ZARMORCHUNK4
    {SPR_SUIT, 5, -1, NULL, S_NULL, 0, 0},      // S_ZARMORCHUNK5
    {SPR_SUIT, 6, -1, NULL, S_NULL, 0, 0},      // S_ZARMORCHUNK6
    {SPR_SUIT, 7, -1, NULL, S_NULL, 0, 0},      // S_ZARMORCHUNK7
    {SPR_SUIT, 8, -1, NULL, S_NULL, 0, 0},      // S_ZARMORCHUNK8
    {SPR_SUIT, 9, -1, NULL, S_NULL, 0, 0},      // S_ZARMORCHUNK9
    {SPR_SUIT, 10, -1, NULL, S_NULL, 0, 0},     // S_ZARMORCHUNK10
    {SPR_BBLL, 5, -1, NULL, S_NULL, 0, 0},      // S_ZBELL
    {SPR_BBLL, 0, 4, A_BellReset1, S_ZBELL_X2, 0, 0},   // S_ZBELL_X1
    {SPR_BBLL, 1, 4, NULL, S_ZBELL_X3, 0, 0},   // S_ZBELL_X2
    {SPR_BBLL, 2, 4, NULL, S_ZBELL_X4, 0, 0},   // S_ZBELL_X3
    {SPR_BBLL, 3, 5, A_Scream, S_ZBELL_X5, 0, 0},       // S_ZBELL_X4
    {SPR_BBLL, 2, 4, NULL, S_ZBELL_X6, 0, 0},   // S_ZBELL_X5
    {SPR_BBLL, 1, 4, NULL, S_ZBELL_X7, 0, 0},   // S_ZBELL_X6
    {SPR_BBLL, 0, 3, NULL, S_ZBELL_X8, 0, 0},   // S_ZBELL_X7
    {SPR_BBLL, 4, 4, NULL, S_ZBELL_X9, 0, 0},   // S_ZBELL_X8
    {SPR_BBLL, 5, 5, NULL, S_ZBELL_X10, 0, 0},  // S_ZBELL_X9
    {SPR_BBLL, 6, 6, A_Scream, S_ZBELL_X11, 0, 0},      // S_ZBELL_X10
    {SPR_BBLL, 5, 5, NULL, S_ZBELL_X12, 0, 0},  // S_ZBELL_X11
    {SPR_BBLL, 4, 4, NULL, S_ZBELL_X13, 0, 0},  // S_ZBELL_X12
    {SPR_BBLL, 0, 4, NULL, S_ZBELL_X14, 0, 0},  // S_ZBELL_X13
    {SPR_BBLL, 1, 5, NULL, S_ZBELL_X15, 0, 0},  // S_ZBELL_X14
    {SPR_BBLL, 2, 5, NULL, S_ZBELL_X16, 0, 0},  // S_ZBELL_X15
    {SPR_BBLL, 3, 6, A_Scream, S_ZBELL_X17, 0, 0},      // S_ZBELL_X16
    {SPR_BBLL, 2, 5, NULL, S_ZBELL_X18, 0, 0},  // S_ZBELL_X17
    {SPR_BBLL, 1, 5, NULL, S_ZBELL_X19, 0, 0},  // S_ZBELL_X18
    {SPR_BBLL, 0, 4, NULL, S_ZBELL_X20, 0, 0},  // S_ZBELL_X19
    {SPR_BBLL, 4, 5, NULL, S_ZBELL_X21, 0, 0},  // S_ZBELL_X20
    {SPR_BBLL, 5, 5, NULL, S_ZBELL_X22, 0, 0},  // S_ZBELL_X21
    {SPR_BBLL, 6, 7, A_Scream, S_ZBELL_X23, 0, 0},      // S_ZBELL_X22
    {SPR_BBLL, 5, 5, NULL, S_ZBELL_X24, 0, 0},  // S_ZBELL_X23
    {SPR_BBLL, 4, 5, NULL, S_ZBELL_X25, 0, 0},  // S_ZBELL_X24
    {SPR_BBLL, 0, 5, NULL, S_ZBELL_X26, 0, 0},  // S_ZBELL_X25
    {SPR_BBLL, 1, 6, NULL, S_ZBELL_X27, 0, 0},  // S_ZBELL_X26
    {SPR_BBLL, 2, 6, NULL, S_ZBELL_X28, 0, 0},  // S_ZBELL_X27
    {SPR_BBLL, 3, 7, A_Scream, S_ZBELL_X29, 0, 0},      // S_ZBELL_X28
    {SPR_BBLL, 2, 6, NULL, S_ZBELL_X30, 0, 0},  // S_ZBELL_X29
    {SPR_BBLL, 1, 6, NULL, S_ZBELL_X31, 0, 0},  // S_ZBELL_X30
    {SPR_BBLL, 0, 5, NULL, S_ZBELL_X32, 0, 0},  // S_ZBELL_X31
    {SPR_BBLL, 4, 6, NULL, S_ZBELL_X33, 0, 0},  // S_ZBELL_X32
    {SPR_BBLL, 5, 6, NULL, S_ZBELL_X34, 0, 0},  // S_ZBELL_X33
    {SPR_BBLL, 6, 7, A_Scream, S_ZBELL_X35, 0, 0},      // S_ZBELL_X34
    {SPR_BBLL, 5, 6, NULL, S_ZBELL_X36, 0, 0},  // S_ZBELL_X35
    {SPR_BBLL, 4, 6, NULL, S_ZBELL_X37, 0, 0},  // S_ZBELL_X36
    {SPR_BBLL, 0, 6, NULL, S_ZBELL_X38, 0, 0},  // S_ZBELL_X37
    {SPR_BBLL, 1, 6, NULL, S_ZBELL_X39, 0, 0},  // S_ZBELL_X38
    {SPR_BBLL, 2, 6, NULL, S_ZBELL_X40, 0, 0},  // S_ZBELL_X39
    {SPR_BBLL, 1, 7, NULL, S_ZBELL_X41, 0, 0},  // S_ZBELL_X40
    {SPR_BBLL, 0, 8, NULL, S_ZBELL_X42, 0, 0},  // S_ZBELL_X41
    {SPR_BBLL, 4, 12, NULL, S_ZBELL_X43, 0, 0}, // S_ZBELL_X42
    {SPR_BBLL, 0, 10, NULL, S_ZBELL_X44, 0, 0}, // S_ZBELL_X43
    {SPR_BBLL, 1, 12, NULL, S_ZBELL_X45, 0, 0}, // S_ZBELL_X44
    {SPR_BBLL, 0, 12, NULL, S_ZBELL_X46, 0, 0}, // S_ZBELL_X45
    {SPR_BBLL, 4, 14, NULL, S_ZBELL_X47, 0, 0}, // S_ZBELL_X46
    {SPR_BBLL, 0, 1, A_BellReset2, S_ZBELL, 0, 0},      // S_ZBELL_X47
    {SPR_CAND, 32768, 5, NULL, S_ZBLUE_CANDLE2, 0, 0},  // S_ZBLUE_CANDLE1
    {SPR_CAND, 32769, 5, NULL, S_ZBLUE_CANDLE3, 0, 0},  // S_ZBLUE_CANDLE2
    {SPR_CAND, 32770, 5, NULL, S_ZBLUE_CANDLE4, 0, 0},  // S_ZBLUE_CANDLE3
    {SPR_CAND, 32771, 5, NULL, S_ZBLUE_CANDLE5, 0, 0},  // S_ZBLUE_CANDLE4
    {SPR_CAND, 32772, 5, NULL, S_ZBLUE_CANDLE1, 0, 0},  // S_ZBLUE_CANDLE5
    {SPR_IRON, 0, -1, NULL, S_NULL, 0, 0},      // S_ZIRON_MAIDEN
    {SPR_XMAS, 0, -1, NULL, S_NULL, 0, 0},      // S_ZXMAS_TREE
    {SPR_XMAS, 0, 4, A_TreeDeath, S_ZXMAS_TREE, 0, 0},  // S_ZXMAS_TREE_DIE
    {SPR_XMAS, 32769, 6, NULL, S_ZXMAS_TREE_X2, 0, 0},  // S_ZXMAS_TREE_X1
    {SPR_XMAS, 32770, 6, A_Scream, S_ZXMAS_TREE_X3, 0, 0},      // S_ZXMAS_TREE_X2
    {SPR_XMAS, 32771, 5, NULL, S_ZXMAS_TREE_X4, 0, 0},  // S_ZXMAS_TREE_X3
    {SPR_XMAS, 32772, 5, A_Explode, S_ZXMAS_TREE_X5, 0, 0},     // S_ZXMAS_TREE_X4
    {SPR_XMAS, 32773, 5, NULL, S_ZXMAS_TREE_X6, 0, 0},  // S_ZXMAS_TREE_X5
    {SPR_XMAS, 32774, 4, NULL, S_ZXMAS_TREE_X7, 0, 0},  // S_ZXMAS_TREE_X6
    {SPR_XMAS, 7, 5, NULL, S_ZXMAS_TREE_X8, 0, 0},      // S_ZXMAS_TREE_X7
    {SPR_XMAS, 8, 4, A_NoBlocking, S_ZXMAS_TREE_X9, 0, 0},      // S_ZXMAS_TREE_X8
    {SPR_XMAS, 9, 4, NULL, S_ZXMAS_TREE_X10, 0, 0},     // S_ZXMAS_TREE_X9
    {SPR_XMAS, 10, -1, NULL, S_NULL, 0, 0},     // S_ZXMAS_TREE_X10
    {SPR_CDRN, 32769, 4, NULL, S_ZCAULDRON2, 0, 0},     // S_ZCAULDRON1
    {SPR_CDRN, 32770, 4, NULL, S_ZCAULDRON3, 0, 0},     // S_ZCAULDRON2
    {SPR_CDRN, 32771, 4, NULL, S_ZCAULDRON4, 0, 0},     // S_ZCAULDRON3
    {SPR_CDRN, 32772, 4, NULL, S_ZCAULDRON5, 0, 0},     // S_ZCAULDRON4
    {SPR_CDRN, 32773, 4, NULL, S_ZCAULDRON6, 0, 0},     // S_ZCAULDRON5
    {SPR_CDRN, 32774, 4, NULL, S_ZCAULDRON7, 0, 0},     // S_ZCAULDRON6
    {SPR_CDRN, 32775, 4, NULL, S_ZCAULDRON1, 0, 0},     // S_ZCAULDRON7
    {SPR_CDRN, 0, -1, NULL, S_NULL, 0, 0},      // S_ZCAULDRON_U
    {SPR_CHNS, 0, -1, NULL, S_NULL, 0, 0},      // S_ZCHAINBIT32
    {SPR_CHNS, 1, -1, NULL, S_NULL, 0, 0},      // S_ZCHAINBIT64
    {SPR_CHNS, 2, -1, NULL, S_NULL, 0, 0},      // S_ZCHAINEND_HEART
    {SPR_CHNS, 3, -1, NULL, S_NULL, 0, 0},      // S_ZCHAINEND_HOOK1
    {SPR_CHNS, 4, -1, NULL, S_NULL, 0, 0},      // S_ZCHAINEND_HOOK2
    {SPR_CHNS, 5, -1, NULL, S_NULL, 0, 0},      // S_ZCHAINEND_SPIKE
    {SPR_CHNS, 6, -1, NULL, S_NULL, 0, 0},      // S_ZCHAINEND_SKULL
    {SPR_TST1, 0, -1, NULL, S_NULL, 0, 0},      // S_TABLE_SHIT1
    {SPR_TST2, 0, -1, NULL, S_NULL, 0, 0},      // S_TABLE_SHIT2
    {SPR_TST3, 0, -1, NULL, S_NULL, 0, 0},      // S_TABLE_SHIT3
    {SPR_TST4, 0, -1, NULL, S_NULL, 0, 0},      // S_TABLE_SHIT4
    {SPR_TST5, 0, -1, NULL, S_NULL, 0, 0},      // S_TABLE_SHIT5
    {SPR_TST6, 0, -1, NULL, S_NULL, 0, 0},      // S_TABLE_SHIT6
    {SPR_TST7, 0, -1, NULL, S_NULL, 0, 0},      // S_TABLE_SHIT7
    {SPR_TST8, 0, -1, NULL, S_NULL, 0, 0},      // S_TABLE_SHIT8
    {SPR_TST9, 0, -1, NULL, S_NULL, 0, 0},      // S_TABLE_SHIT9
    {SPR_TST0, 0, -1, NULL, S_NULL, 0, 0},      // S_TABLE_SHIT10
    {SPR_TELE, 32768, 6, NULL, S_TFOG2, 0, 0},  // S_TFOG1
    {SPR_TELE, 32769, 6, NULL, S_TFOG3, 0, 0},  // S_TFOG2
    {SPR_TELE, 32770, 6, NULL, S_TFOG4, 0, 0},  // S_TFOG3
    {SPR_TELE, 32771, 6, NULL, S_TFOG5, 0, 0},  // S_TFOG4
    {SPR_TELE, 32772, 6, NULL, S_TFOG6, 0, 0},  // S_TFOG5
    {SPR_TELE, 32773, 6, NULL, S_TFOG7, 0, 0},  // S_TFOG6
    {SPR_TELE, 32774, 6, NULL, S_TFOG8, 0, 0},  // S_TFOG7
    {SPR_TELE, 32775, 6, NULL, S_TFOG9, 0, 0},  // S_TFOG8
    {SPR_TELE, 32774, 6, NULL, S_TFOG10, 0, 0}, // S_TFOG9
    {SPR_TELE, 32773, 6, NULL, S_TFOG11, 0, 0}, // S_TFOG10
    {SPR_TELE, 32772, 6, NULL, S_TFOG12, 0, 0}, // S_TFOG11
    {SPR_TELE, 32771, 6, NULL, S_TFOG13, 0, 0}, // S_TFOG12
    {SPR_TELE, 32770, 6, NULL, S_NULL, 0, 0},   // S_TFOG13
    {SPR_TSMK, 0, 4, NULL, S_TELESMOKE2, 0, 0}, // S_TELESMOKE1
    {SPR_TSMK, 1, 3, NULL, S_TELESMOKE3, 0, 0}, // S_TELESMOKE2
    {SPR_TSMK, 2, 4, NULL, S_TELESMOKE4, 0, 0}, // S_TELESMOKE3
    {SPR_TSMK, 3, 3, NULL, S_TELESMOKE5, 0, 0}, // S_TELESMOKE4
    {SPR_TSMK, 4, 4, NULL, S_TELESMOKE6, 0, 0}, // S_TELESMOKE5
    {SPR_TSMK, 5, 3, NULL, S_TELESMOKE7, 0, 0}, // S_TELESMOKE6
    {SPR_TSMK, 6, 4, NULL, S_TELESMOKE8, 0, 0}, // S_TELESMOKE7
    {SPR_TSMK, 7, 3, NULL, S_TELESMOKE9, 0, 0}, // S_TELESMOKE8
    {SPR_TSMK, 8, 4, NULL, S_TELESMOKE10, 0, 0},        // S_TELESMOKE9
    {SPR_TSMK, 9, 3, NULL, S_TELESMOKE11, 0, 0},        // S_TELESMOKE10
    {SPR_TSMK, 10, 4, NULL, S_TELESMOKE12, 0, 0},       // S_TELESMOKE11
    {SPR_TSMK, 11, 3, NULL, S_TELESMOKE13, 0, 0},       // S_TELESMOKE12
    {SPR_TSMK, 12, 4, NULL, S_TELESMOKE14, 0, 0},       // S_TELESMOKE13
    {SPR_TSMK, 13, 3, NULL, S_TELESMOKE15, 0, 0},       // S_TELESMOKE14
    {SPR_TSMK, 14, 4, NULL, S_TELESMOKE16, 0, 0},       // S_TELESMOKE15
    {SPR_TSMK, 15, 3, NULL, S_TELESMOKE17, 0, 0},       // S_TELESMOKE16
    {SPR_TSMK, 16, 4, NULL, S_TELESMOKE18, 0, 0},       // S_TELESMOKE17
    {SPR_TSMK, 17, 3, NULL, S_TELESMOKE19, 0, 0},       // S_TELESMOKE18
    {SPR_TSMK, 18, 4, NULL, S_TELESMOKE20, 0, 0},       // S_TELESMOKE19
    {SPR_TSMK, 19, 3, NULL, S_TELESMOKE21, 0, 0},       // S_TELESMOKE20
    {SPR_TSMK, 20, 4, NULL, S_TELESMOKE22, 0, 0},       // S_TELESMOKE21
    {SPR_TSMK, 21, 3, NULL, S_TELESMOKE23, 0, 0},       // S_TELESMOKE22
    {SPR_TSMK, 22, 4, NULL, S_TELESMOKE24, 0, 0},       // S_TELESMOKE23
    {SPR_TSMK, 23, 3, NULL, S_TELESMOKE25, 0, 0},       // S_TELESMOKE24
    {SPR_TSMK, 24, 4, NULL, S_TELESMOKE26, 0, 0},       // S_TELESMOKE25
    {SPR_TSMK, 25, 3, NULL, S_TELESMOKE1, 0, 0},        // S_TELESMOKE26
    {SPR_FPCH, 0, 0, A_Light0, S_NULL, 0, 0},   // S_LIGHTDONE
    {SPR_FPCH, 0, 1, A_WeaponReady, S_PUNCHREADY, 0, 0},        // S_PUNCHREADY
    {SPR_FPCH, 0, 1, A_Lower, S_PUNCHDOWN, 0, 0},       // S_PUNCHDOWN
    {SPR_FPCH, 0, 1, A_Raise, S_PUNCHUP, 0, 0}, // S_PUNCHUP
    {SPR_FPCH, 1, 5, NULL, S_PUNCHATK1_2, 5, 40},       // S_PUNCHATK1_1
    {SPR_FPCH, 2, 4, NULL, S_PUNCHATK1_3, 5, 40},       // S_PUNCHATK1_2
    {SPR_FPCH, 3, 4, A_FPunchAttack, S_PUNCHATK1_4, 5, 40},     // S_PUNCHATK1_3
    {SPR_FPCH, 2, 4, NULL, S_PUNCHATK1_5, 5, 40},       // S_PUNCHATK1_4
    {SPR_FPCH, 1, 5, A_ReFire, S_PUNCHREADY, 5, 40},    // S_PUNCHATK1_5
    {SPR_FPCH, 3, 4, NULL, S_PUNCHATK2_2, 5, 40},       // S_PUNCHATK2_1
    {SPR_FPCH, 4, 4, NULL, S_PUNCHATK2_3, 5, 40},       // S_PUNCHATK2_2
    {SPR_FPCH, 4, 1, NULL, S_PUNCHATK2_4, 15, 50},      // S_PUNCHATK2_3
    {SPR_FPCH, 4, 1, NULL, S_PUNCHATK2_5, 25, 60},      // S_PUNCHATK2_4
    {SPR_FPCH, 4, 1, NULL, S_PUNCHATK2_6, 35, 70},      // S_PUNCHATK2_5
    {SPR_FPCH, 4, 1, NULL, S_PUNCHATK2_7, 45, 80},      // S_PUNCHATK2_6
    {SPR_FPCH, 4, 1, NULL, S_PUNCHATK2_8, 55, 90},      // S_PUNCHATK2_7
    {SPR_FPCH, 4, 1, NULL, S_PUNCHATK2_9, 65, 100},     // S_PUNCHATK2_8
    {SPR_FPCH, 4, 10, NULL, S_PUNCHREADY, 0, 150},      // S_PUNCHATK2_9
    {SPR_FHFX, 18, 4, NULL, S_PUNCHPUFF2, 0, 0},        // S_PUNCHPUFF1
    {SPR_FHFX, 19, 4, NULL, S_PUNCHPUFF3, 0, 0},        // S_PUNCHPUFF2
    {SPR_FHFX, 20, 4, NULL, S_PUNCHPUFF4, 0, 0},        // S_PUNCHPUFF3
    {SPR_FHFX, 21, 4, NULL, S_PUNCHPUFF5, 0, 0},        // S_PUNCHPUFF4
    {SPR_FHFX, 22, 4, NULL, S_NULL, 0, 0},      // S_PUNCHPUFF5
    {SPR_WFAX, 0, -1, NULL, S_NULL, 0, 0},      // S_AXE
    {SPR_FAXE, 0, 1, A_WeaponReady, S_FAXEREADY, 0, 0}, // S_FAXEREADY
    {SPR_FAXE, 0, 1, A_Lower, S_FAXEDOWN, 0, 0},        // S_FAXEDOWN
    {SPR_FAXE, 0, 1, A_Raise, S_FAXEUP, 0, 0},  // S_FAXEUP
    {SPR_FAXE, 1, 4, NULL, S_FAXEATK_2, 15, 32},        // S_FAXEATK_1
    {SPR_FAXE, 2, 3, NULL, S_FAXEATK_3, 15, 32},        // S_FAXEATK_2
    {SPR_FAXE, 3, 2, NULL, S_FAXEATK_4, 15, 32},        // S_FAXEATK_3
    {SPR_FAXE, 3, 1, A_FAxeAttack, S_FAXEATK_5, -5, 70},        // S_FAXEATK_4
    {SPR_FAXE, 3, 2, NULL, S_FAXEATK_6, -25, 90},       // S_FAXEATK_5
    {SPR_FAXE, 4, 1, NULL, S_FAXEATK_7, 15, 32},        // S_FAXEATK_6
    {SPR_FAXE, 4, 2, NULL, S_FAXEATK_8, 10, 54},        // S_FAXEATK_7
    {SPR_FAXE, 4, 7, NULL, S_FAXEATK_9, 10, 150},       // S_FAXEATK_8
    {SPR_FAXE, 0, 1, A_ReFire, S_FAXEATK_10, 0, 60},    // S_FAXEATK_9
    {SPR_FAXE, 0, 1, NULL, S_FAXEATK_11, 0, 52},        // S_FAXEATK_10
    {SPR_FAXE, 0, 1, NULL, S_FAXEATK_12, 0, 44},        // S_FAXEATK_11
    {SPR_FAXE, 0, 1, NULL, S_FAXEATK_13, 0, 36},        // S_FAXEATK_12
    {SPR_FAXE, 0, 1, NULL, S_FAXEREADY, 0, 0},  // S_FAXEATK_13
    {SPR_FAXE, 11, 1, A_WeaponReady, S_FAXEREADY_G1, 0, 0},     // S_FAXEREADY_G
    {SPR_FAXE, 11, 1, A_WeaponReady, S_FAXEREADY_G2, 0, 0},     // S_FAXEREADY_G1
    {SPR_FAXE, 11, 1, A_WeaponReady, S_FAXEREADY_G3, 0, 0},     // S_FAXEREADY_G2
    {SPR_FAXE, 12, 1, A_WeaponReady, S_FAXEREADY_G4, 0, 0},     // S_FAXEREADY_G3
    {SPR_FAXE, 12, 1, A_WeaponReady, S_FAXEREADY_G5, 0, 0},     // S_FAXEREADY_G4
    {SPR_FAXE, 12, 1, A_WeaponReady, S_FAXEREADY_G, 0, 0},      // S_FAXEREADY_G5
    {SPR_FAXE, 11, 1, A_Lower, S_FAXEDOWN_G, 0, 0},     // S_FAXEDOWN_G
    {SPR_FAXE, 11, 1, A_Raise, S_FAXEUP_G, 0, 0},       // S_FAXEUP_G
    {SPR_FAXE, 13, 4, NULL, S_FAXEATK_G2, 15, 32},      // S_FAXEATK_G1
    {SPR_FAXE, 14, 3, NULL, S_FAXEATK_G3, 15, 32},      // S_FAXEATK_G2
    {SPR_FAXE, 15, 2, NULL, S_FAXEATK_G4, 15, 32},      // S_FAXEATK_G3
    {SPR_FAXE, 15, 1, A_FAxeAttack, S_FAXEATK_G5, -5, 70},      // S_FAXEATK_G4
    {SPR_FAXE, 15, 2, NULL, S_FAXEATK_G6, -25, 90},     // S_FAXEATK_G5
    {SPR_FAXE, 16, 1, NULL, S_FAXEATK_G7, 15, 32},      // S_FAXEATK_G6
    {SPR_FAXE, 16, 2, NULL, S_FAXEATK_G8, 10, 54},      // S_FAXEATK_G7
    {SPR_FAXE, 16, 7, NULL, S_FAXEATK_G9, 10, 150},     // S_FAXEATK_G8
    {SPR_FAXE, 0, 1, A_ReFire, S_FAXEATK_G10, 0, 60},   // S_FAXEATK_G9
    {SPR_FAXE, 0, 1, NULL, S_FAXEATK_G11, 0, 52},       // S_FAXEATK_G10
    {SPR_FAXE, 0, 1, NULL, S_FAXEATK_G12, 0, 44},       // S_FAXEATK_G11
    {SPR_FAXE, 0, 1, NULL, S_FAXEATK_G13, 0, 36},       // S_FAXEATK_G12
    {SPR_FAXE, 0, 1, NULL, S_FAXEREADY_G, 0, 0},        // S_FAXEATK_G13
    {SPR_FAXE, 32785, 4, NULL, S_AXEPUFF_GLOW2, 0, 0},  // S_AXEPUFF_GLOW1
    {SPR_FAXE, 32786, 4, NULL, S_AXEPUFF_GLOW3, 0, 0},  // S_AXEPUFF_GLOW2
    {SPR_FAXE, 32787, 4, NULL, S_AXEPUFF_GLOW4, 0, 0},  // S_AXEPUFF_GLOW3
    {SPR_FAXE, 32788, 4, NULL, S_AXEPUFF_GLOW5, 0, 0},  // S_AXEPUFF_GLOW4
    {SPR_FAXE, 32789, 4, NULL, S_AXEPUFF_GLOW6, 0, 0},  // S_AXEPUFF_GLOW5
    {SPR_FAXE, 32790, 4, NULL, S_AXEPUFF_GLOW7, 0, 0},  // S_AXEPUFF_GLOW6
    {SPR_FAXE, 32791, 4, NULL, S_NULL, 0, 0},   // S_AXEPUFF_GLOW7
    {SPR_FAXE, 5, 3, NULL, S_AXEBLOOD2, 0, 0},  // S_AXEBLOOD1
    {SPR_FAXE, 6, 3, NULL, S_AXEBLOOD3, 0, 0},  // S_AXEBLOOD2
    {SPR_FAXE, 7, 3, NULL, S_AXEBLOOD4, 0, 0},  // S_AXEBLOOD3
    {SPR_FAXE, 8, 3, NULL, S_AXEBLOOD5, 0, 0},  // S_AXEBLOOD4
    {SPR_FAXE, 9, 3, NULL, S_AXEBLOOD6, 0, 0},  // S_AXEBLOOD5
    {SPR_FAXE, 10, 3, NULL, S_NULL, 0, 0},      // S_AXEBLOOD6
    {SPR_WFHM, 0, -1, NULL, S_NULL, 0, 0},      // S_HAMM
    {SPR_FHMR, 0, 1, A_WeaponReady, S_FHAMMERREADY, 0, 0},      // S_FHAMMERREADY
    {SPR_FHMR, 0, 1, A_Lower, S_FHAMMERDOWN, 0, 0},     // S_FHAMMERDOWN
    {SPR_FHMR, 0, 1, A_Raise, S_FHAMMERUP, 0, 0},       // S_FHAMMERUP
    {SPR_FHMR, 1, 6, NULL, S_FHAMMERATK_2, 5, 0},       // S_FHAMMERATK_1
    {SPR_FHMR, 2, 3, A_FHammerAttack, S_FHAMMERATK_3, 5, 0},    // S_FHAMMERATK_2
    {SPR_FHMR, 3, 3, NULL, S_FHAMMERATK_4, 5, 0},       // S_FHAMMERATK_3
    {SPR_FHMR, 4, 2, NULL, S_FHAMMERATK_5, 5, 0},       // S_FHAMMERATK_4
    {SPR_FHMR, 4, 10, A_FHammerThrow, S_FHAMMERATK_6, 5, 150},  // S_FHAMMERATK_5
    {SPR_FHMR, 0, 1, NULL, S_FHAMMERATK_7, 0, 60},      // S_FHAMMERATK_6
    {SPR_FHMR, 0, 1, NULL, S_FHAMMERATK_8, 0, 55},      // S_FHAMMERATK_7
    {SPR_FHMR, 0, 1, NULL, S_FHAMMERATK_9, 0, 50},      // S_FHAMMERATK_8
    {SPR_FHMR, 0, 1, NULL, S_FHAMMERATK_10, 0, 45},     // S_FHAMMERATK_9
    {SPR_FHMR, 0, 1, NULL, S_FHAMMERATK_11, 0, 40},     // S_FHAMMERATK_10
    {SPR_FHMR, 0, 1, NULL, S_FHAMMERATK_12, 0, 35},     // S_FHAMMERATK_11
    {SPR_FHMR, 0, 1, NULL, S_FHAMMERREADY, 0, 0},       // S_FHAMMERATK_12
    {SPR_FHFX, 32768, 2, NULL, S_HAMMER_MISSILE_2, 0, 0},       // S_HAMMER_MISSILE_1
    {SPR_FHFX, 32769, 2, A_ContMobjSound, S_HAMMER_MISSILE_3, 0, 0},    // S_HAMMER_MISSILE_2
    {SPR_FHFX, 32770, 2, NULL, S_HAMMER_MISSILE_4, 0, 0},       // S_HAMMER_MISSILE_3
    {SPR_FHFX, 32771, 2, NULL, S_HAMMER_MISSILE_5, 0, 0},       // S_HAMMER_MISSILE_4
    {SPR_FHFX, 32772, 2, NULL, S_HAMMER_MISSILE_6, 0, 0},       // S_HAMMER_MISSILE_5
    {SPR_FHFX, 32773, 2, NULL, S_HAMMER_MISSILE_7, 0, 0},       // S_HAMMER_MISSILE_6
    {SPR_FHFX, 32774, 2, NULL, S_HAMMER_MISSILE_8, 0, 0},       // S_HAMMER_MISSILE_7
    {SPR_FHFX, 32775, 2, NULL, S_HAMMER_MISSILE_1, 0, 0},       // S_HAMMER_MISSILE_8
    {SPR_FHFX, 32776, 3, NULL, S_HAMMER_MISSILE_X2, 0, 0},      // S_HAMMER_MISSILE_X1
    {SPR_FHFX, 32777, 3, NULL, S_HAMMER_MISSILE_X3, 0, 0},      // S_HAMMER_MISSILE_X2
    {SPR_FHFX, 32778, 3, A_Explode, S_HAMMER_MISSILE_X4, 0, 0}, // S_HAMMER_MISSILE_X3
    {SPR_FHFX, 32779, 3, NULL, S_HAMMER_MISSILE_X5, 0, 0},      // S_HAMMER_MISSILE_X4
    {SPR_FHFX, 32780, 3, NULL, S_HAMMER_MISSILE_X6, 0, 0},      // S_HAMMER_MISSILE_X5
    {SPR_FHFX, 13, 3, NULL, S_HAMMER_MISSILE_X7, 0, 0}, // S_HAMMER_MISSILE_X6
    {SPR_FHFX, 32782, 3, NULL, S_HAMMER_MISSILE_X8, 0, 0},      // S_HAMMER_MISSILE_X7
    {SPR_FHFX, 32783, 3, NULL, S_HAMMER_MISSILE_X9, 0, 0},      // S_HAMMER_MISSILE_X8
    {SPR_FHFX, 32784, 3, NULL, S_HAMMER_MISSILE_X10, 0, 0},     // S_HAMMER_MISSILE_X9
    {SPR_FHFX, 32785, 3, NULL, S_NULL, 0, 0},   // S_HAMMER_MISSILE_X10
    {SPR_FHFX, 18, 4, NULL, S_HAMMERPUFF2, 0, 0},       // S_HAMMERPUFF1
    {SPR_FHFX, 19, 4, NULL, S_HAMMERPUFF3, 0, 0},       // S_HAMMERPUFF2
    {SPR_FHFX, 20, 4, NULL, S_HAMMERPUFF4, 0, 0},       // S_HAMMERPUFF3
    {SPR_FHFX, 21, 4, NULL, S_HAMMERPUFF5, 0, 0},       // S_HAMMERPUFF4
    {SPR_FHFX, 22, 4, NULL, S_NULL, 0, 0},      // S_HAMMERPUFF5
    // [crispy] disable full bright for Quietus, use brightmap instead
    {SPR_FSRD, 0, 1, A_WeaponReady, S_FSWORDREADY1, 0, 0},  // S_FSWORDREADY
    {SPR_FSRD, 0, 1, A_WeaponReady, S_FSWORDREADY2, 0, 0},  // S_FSWORDREADY1
    {SPR_FSRD, 0, 1, A_WeaponReady, S_FSWORDREADY3, 0, 0},  // S_FSWORDREADY2
    {SPR_FSRD, 0, 1, A_WeaponReady, S_FSWORDREADY4, 0, 0},  // S_FSWORDREADY3
    {SPR_FSRD, 1, 1, A_WeaponReady, S_FSWORDREADY5, 0, 0},  // S_FSWORDREADY4
    {SPR_FSRD, 1, 1, A_WeaponReady, S_FSWORDREADY6, 0, 0},  // S_FSWORDREADY5
    {SPR_FSRD, 1, 1, A_WeaponReady, S_FSWORDREADY7, 0, 0},  // S_FSWORDREADY6
    {SPR_FSRD, 1, 1, A_WeaponReady, S_FSWORDREADY8, 0, 0},  // S_FSWORDREADY7
    {SPR_FSRD, 2, 1, A_WeaponReady, S_FSWORDREADY9, 0, 0},  // S_FSWORDREADY8
    {SPR_FSRD, 2, 1, A_WeaponReady, S_FSWORDREADY10, 0, 0}, // S_FSWORDREADY9
    {SPR_FSRD, 2, 1, A_WeaponReady, S_FSWORDREADY11, 0, 0}, // S_FSWORDREADY10
    {SPR_FSRD, 2, 1, A_WeaponReady, S_FSWORDREADY, 0, 0},   // S_FSWORDREADY11
    {SPR_FSRD, 0, 1, A_Lower, S_FSWORDDOWN, 0, 0},  // S_FSWORDDOWN
    {SPR_FSRD, 0, 1, A_Raise, S_FSWORDUP, 0, 0},    // S_FSWORDUP
    {SPR_FSRD, 3, 3, NULL, S_FSWORDATK_2, 5, 36},   // S_FSWORDATK_1
    {SPR_FSRD, 4, 3, NULL, S_FSWORDATK_3, 5, 36},   // S_FSWORDATK_2
    {SPR_FSRD, 5, 2, NULL, S_FSWORDATK_4, 5, 36},   // S_FSWORDATK_3
    {SPR_FSRD, 6, 3, A_FSwordAttack, S_FSWORDATK_5, 5, 36}, // S_FSWORDATK_4
    {SPR_FSRD, 7, 2, NULL, S_FSWORDATK_6, 5, 36},   // S_FSWORDATK_5
    {SPR_FSRD, 8, 2, NULL, S_FSWORDATK_7, 5, 36},   // S_FSWORDATK_6
    {SPR_FSRD, 8, 10, NULL, S_FSWORDATK_8, 5, 150}, // S_FSWORDATK_7
    {SPR_FSRD, 0, 1, NULL, S_FSWORDATK_9, 5, 60},   // S_FSWORDATK_8
    {SPR_FSRD, 1, 1, NULL, S_FSWORDATK_10, 5, 55},  // S_FSWORDATK_9
    {SPR_FSRD, 2, 1, NULL, S_FSWORDATK_11, 5, 50},  // S_FSWORDATK_10
    {SPR_FSRD, 0, 1, NULL, S_FSWORDATK_12, 5, 45},  // S_FSWORDATK_11
    {SPR_FSRD, 1, 1, NULL, S_FSWORDREADY, 5, 40},   // S_FSWORDATK_12
    {SPR_FSFX, 32768, 3, NULL, S_FSWORD_MISSILE2, 0, 0},        // S_FSWORD_MISSILE1
    {SPR_FSFX, 32769, 3, NULL, S_FSWORD_MISSILE3, 0, 0},        // S_FSWORD_MISSILE2
    {SPR_FSFX, 32770, 3, NULL, S_FSWORD_MISSILE1, 0, 0},        // S_FSWORD_MISSILE3
    {SPR_FSFX, 32771, 4, NULL, S_FSWORD_MISSILE_X2, 0, 0},      // S_FSWORD_MISSILE_X1
    {SPR_FSFX, 32772, 3, A_FSwordFlames, S_FSWORD_MISSILE_X3, 0, 0},    // S_FSWORD_MISSILE_X2
    {SPR_FSFX, 32773, 4, A_Explode, S_FSWORD_MISSILE_X4, 0, 0}, // S_FSWORD_MISSILE_X3
    {SPR_FSFX, 32774, 3, NULL, S_FSWORD_MISSILE_X5, 0, 0},      // S_FSWORD_MISSILE_X4
    {SPR_FSFX, 32775, 4, NULL, S_FSWORD_MISSILE_X6, 0, 0},      // S_FSWORD_MISSILE_X5
    {SPR_FSFX, 32776, 3, NULL, S_FSWORD_MISSILE_X7, 0, 0},      // S_FSWORD_MISSILE_X6
    {SPR_FSFX, 32777, 4, NULL, S_FSWORD_MISSILE_X8, 0, 0},      // S_FSWORD_MISSILE_X7
    {SPR_FSFX, 32778, 3, NULL, S_FSWORD_MISSILE_X9, 0, 0},      // S_FSWORD_MISSILE_X8
    {SPR_FSFX, 32779, 3, NULL, S_FSWORD_MISSILE_X10, 0, 0},     // S_FSWORD_MISSILE_X9
    {SPR_FSFX, 32780, 3, NULL, S_NULL, 0, 0},   // S_FSWORD_MISSILE_X10
    {SPR_FSFX, 32781, 3, NULL, S_FSWORD_FLAME2, 0, 0},  // S_FSWORD_FLAME1
    {SPR_FSFX, 32782, 3, NULL, S_FSWORD_FLAME3, 0, 0},  // S_FSWORD_FLAME2
    {SPR_FSFX, 32783, 3, NULL, S_FSWORD_FLAME4, 0, 0},  // S_FSWORD_FLAME3
    {SPR_FSFX, 32784, 3, NULL, S_FSWORD_FLAME5, 0, 0},  // S_FSWORD_FLAME4
    {SPR_FSFX, 32785, 3, NULL, S_FSWORD_FLAME6, 0, 0},  // S_FSWORD_FLAME5
    {SPR_FSFX, 32786, 3, NULL, S_FSWORD_FLAME7, 0, 0},  // S_FSWORD_FLAME6
    {SPR_FSFX, 32787, 3, NULL, S_FSWORD_FLAME8, 0, 0},  // S_FSWORD_FLAME7
    {SPR_FSFX, 32788, 3, NULL, S_FSWORD_FLAME9, 0, 0},  // S_FSWORD_FLAME8
    {SPR_FSFX, 32789, 3, NULL, S_FSWORD_FLAME10, 0, 0}, // S_FSWORD_FLAME9
    {SPR_FSFX, 32790, 3, NULL, S_NULL, 0, 0},   // S_FSWORD_FLAME10
    {SPR_CMCE, 0, 1, A_WeaponReady, S_CMACEREADY, 0, 0},        // S_CMACEREADY
    {SPR_CMCE, 0, 1, A_Lower, S_CMACEDOWN, 0, 0},       // S_CMACEDOWN
    {SPR_CMCE, 0, 1, A_Raise, S_CMACEUP, 0, 0}, // S_CMACEUP
    {SPR_CMCE, 1, 2, NULL, S_CMACEATK_2, 60, 20},       // S_CMACEATK_1
    {SPR_CMCE, 1, 1, NULL, S_CMACEATK_3, 30, 33},       // S_CMACEATK_2
    {SPR_CMCE, 1, 2, NULL, S_CMACEATK_4, 8, 45},        // S_CMACEATK_3
    {SPR_CMCE, 2, 1, NULL, S_CMACEATK_5, 8, 45},        // S_CMACEATK_4
    {SPR_CMCE, 3, 1, NULL, S_CMACEATK_6, 8, 45},        // S_CMACEATK_5
    {SPR_CMCE, 4, 1, NULL, S_CMACEATK_7, 8, 45},        // S_CMACEATK_6
    {SPR_CMCE, 4, 1, A_CMaceAttack, S_CMACEATK_8, -11, 58},     // S_CMACEATK_7
    {SPR_CMCE, 5, 1, NULL, S_CMACEATK_9, 8, 45},        // S_CMACEATK_8
    {SPR_CMCE, 5, 2, NULL, S_CMACEATK_10, -8, 74},      // S_CMACEATK_9
    {SPR_CMCE, 5, 1, NULL, S_CMACEATK_11, -20, 96},     // S_CMACEATK_10
    {SPR_CMCE, 5, 8, NULL, S_CMACEATK_12, -33, 160},    // S_CMACEATK_11
    {SPR_CMCE, 0, 2, A_ReFire, S_CMACEATK_13, 8, 75},   // S_CMACEATK_12
    {SPR_CMCE, 0, 1, NULL, S_CMACEATK_14, 8, 65},       // S_CMACEATK_13
    {SPR_CMCE, 0, 2, NULL, S_CMACEATK_15, 8, 60},       // S_CMACEATK_14
    {SPR_CMCE, 0, 1, NULL, S_CMACEATK_16, 8, 55},       // S_CMACEATK_15
    {SPR_CMCE, 0, 2, NULL, S_CMACEATK_17, 8, 50},       // S_CMACEATK_16
    {SPR_CMCE, 0, 1, NULL, S_CMACEREADY, 8, 45},        // S_CMACEATK_17
    {SPR_WCSS, 0, -1, NULL, S_NULL, 0, 0},      // S_CSTAFF
    {SPR_CSSF, 2, 4, NULL, S_CSTAFFREADY1, 0, 0},       // S_CSTAFFREADY
    {SPR_CSSF, 1, 3, A_CStaffInitBlink, S_CSTAFFREADY2, 0, 0},  // S_CSTAFFREADY1
    {SPR_CSSF, 0, 1, A_WeaponReady, S_CSTAFFREADY3, 0, 0},      // S_CSTAFFREADY2
    {SPR_CSSF, 0, 1, A_WeaponReady, S_CSTAFFREADY4, 0, 0},      // S_CSTAFFREADY3
    {SPR_CSSF, 0, 1, A_WeaponReady, S_CSTAFFREADY5, 0, 0},      // S_CSTAFFREADY4
    {SPR_CSSF, 0, 1, A_WeaponReady, S_CSTAFFREADY6, 0, 0},      // S_CSTAFFREADY5
    {SPR_CSSF, 0, 1, A_WeaponReady, S_CSTAFFREADY7, 0, 0},      // S_CSTAFFREADY6
    {SPR_CSSF, 0, 1, A_WeaponReady, S_CSTAFFREADY8, 0, 0},      // S_CSTAFFREADY7
    {SPR_CSSF, 0, 1, A_WeaponReady, S_CSTAFFREADY9, 0, 0},      // S_CSTAFFREADY8
    {SPR_CSSF, 0, 1, A_CStaffCheckBlink, S_CSTAFFREADY2, 0, 0}, // S_CSTAFFREADY9
    {SPR_CSSF, 1, 1, A_WeaponReady, S_CSTAFFBLINK2, 0, 0},      // S_CSTAFFBLINK1
    {SPR_CSSF, 1, 1, A_WeaponReady, S_CSTAFFBLINK3, 0, 0},      // S_CSTAFFBLINK2
    {SPR_CSSF, 1, 1, A_WeaponReady, S_CSTAFFBLINK4, 0, 0},      // S_CSTAFFBLINK3
    {SPR_CSSF, 2, 1, A_WeaponReady, S_CSTAFFBLINK5, 0, 0},      // S_CSTAFFBLINK4
    {SPR_CSSF, 2, 1, A_WeaponReady, S_CSTAFFBLINK6, 0, 0},      // S_CSTAFFBLINK5
    {SPR_CSSF, 2, 1, A_WeaponReady, S_CSTAFFBLINK7, 0, 0},      // S_CSTAFFBLINK6
    {SPR_CSSF, 2, 1, A_WeaponReady, S_CSTAFFBLINK8, 0, 0},      // S_CSTAFFBLINK7
    {SPR_CSSF, 2, 1, A_WeaponReady, S_CSTAFFBLINK9, 0, 0},      // S_CSTAFFBLINK8
    {SPR_CSSF, 1, 1, A_WeaponReady, S_CSTAFFBLINK10, 0, 0},     // S_CSTAFFBLINK9
    {SPR_CSSF, 1, 1, A_WeaponReady, S_CSTAFFBLINK11, 0, 0},     // S_CSTAFFBLINK10
    {SPR_CSSF, 1, 1, A_WeaponReady, S_CSTAFFREADY2, 0, 0},      // S_CSTAFFBLINK11
    {SPR_CSSF, 1, 3, NULL, S_CSTAFFDOWN2, 0, 0},        // S_CSTAFFDOWN
    {SPR_CSSF, 2, 4, NULL, S_CSTAFFDOWN3, 0, 0},        // S_CSTAFFDOWN2
    {SPR_CSSF, 2, 1, A_Lower, S_CSTAFFDOWN3, 0, 0},     // S_CSTAFFDOWN3
    {SPR_CSSF, 2, 1, A_Raise, S_CSTAFFUP, 0, 0},        // S_CSTAFFUP
    {SPR_CSSF, 0, 1, A_CStaffCheck, S_CSTAFFATK_2, 0, 45},      // S_CSTAFFATK_1
    {SPR_CSSF, 9, 1, A_CStaffAttack, S_CSTAFFATK_3, 0, 50},     // S_CSTAFFATK_2
    {SPR_CSSF, 9, 2, NULL, S_CSTAFFATK_4, 0, 50},       // S_CSTAFFATK_3
    {SPR_CSSF, 9, 2, NULL, S_CSTAFFATK_5, 0, 45},       // S_CSTAFFATK_4
    {SPR_CSSF, 0, 2, NULL, S_CSTAFFATK_6, 0, 40},       // S_CSTAFFATK_5
    {SPR_CSSF, 0, 2, NULL, S_CSTAFFREADY2, 0, 36},      // S_CSTAFFATK_6
    {SPR_CSSF, 10, 10, NULL, S_CSTAFFREADY2, 0, 36},    // S_CSTAFFATK2_1
    {SPR_CSSF, 32771, 1, A_CStaffMissileSlither, S_CSTAFF_MISSILE2, 0, 0},      // S_CSTAFF_MISSILE1
    {SPR_CSSF, 32771, 1, A_CStaffMissileSlither, S_CSTAFF_MISSILE3, 0, 0},      // S_CSTAFF_MISSILE2
    {SPR_CSSF, 32772, 1, A_CStaffMissileSlither, S_CSTAFF_MISSILE4, 0, 0},      // S_CSTAFF_MISSILE3
    {SPR_CSSF, 32772, 1, A_CStaffMissileSlither, S_CSTAFF_MISSILE1, 0, 0},      // S_CSTAFF_MISSILE4
    {SPR_CSSF, 32773, 4, NULL, S_CSTAFF_MISSILE_X2, 0, 0},      // S_CSTAFF_MISSILE_X1
    {SPR_CSSF, 32774, 4, NULL, S_CSTAFF_MISSILE_X3, 0, 0},      // S_CSTAFF_MISSILE_X2
    {SPR_CSSF, 32775, 3, NULL, S_CSTAFF_MISSILE_X4, 0, 0},      // S_CSTAFF_MISSILE_X3
    {SPR_CSSF, 32776, 3, NULL, S_NULL, 0, 0},   // S_CSTAFF_MISSILE_X4
    {SPR_FHFX, 18, 4, NULL, S_CSTAFFPUFF2, 0, 0},       // S_CSTAFFPUFF1
    {SPR_FHFX, 19, 4, NULL, S_CSTAFFPUFF3, 0, 0},       // S_CSTAFFPUFF2
    {SPR_FHFX, 20, 4, NULL, S_CSTAFFPUFF4, 0, 0},       // S_CSTAFFPUFF3
    {SPR_FHFX, 21, 4, NULL, S_CSTAFFPUFF5, 0, 0},       // S_CSTAFFPUFF4
    {SPR_FHFX, 22, 4, NULL, S_NULL, 0, 0},      // S_CSTAFFPUFF5
    {SPR_WCFM, 32768, 4, NULL, S_CFLAME2, 0, 0},        // S_CFLAME1
    {SPR_WCFM, 32769, 4, NULL, S_CFLAME3, 0, 0},        // S_CFLAME2
    {SPR_WCFM, 32770, 4, NULL, S_CFLAME4, 0, 0},        // S_CFLAME3
    {SPR_WCFM, 32771, 4, NULL, S_CFLAME5, 0, 0},        // S_CFLAME4
    {SPR_WCFM, 32772, 4, NULL, S_CFLAME6, 0, 0},        // S_CFLAME5
    {SPR_WCFM, 32773, 4, NULL, S_CFLAME7, 0, 0},        // S_CFLAME6
    {SPR_WCFM, 32774, 4, NULL, S_CFLAME8, 0, 0},        // S_CFLAME7
    {SPR_WCFM, 32775, 4, NULL, S_CFLAME1, 0, 0},        // S_CFLAME8
    {SPR_CFLM, 0, 1, A_WeaponReady, S_CFLAMEREADY2, 0, 0},      // S_CFLAMEREADY1
    {SPR_CFLM, 0, 1, A_WeaponReady, S_CFLAMEREADY3, 0, 0},      // S_CFLAMEREADY2
    {SPR_CFLM, 0, 1, A_WeaponReady, S_CFLAMEREADY4, 0, 0},      // S_CFLAMEREADY3
    {SPR_CFLM, 0, 1, A_WeaponReady, S_CFLAMEREADY5, 0, 0},      // S_CFLAMEREADY4
    {SPR_CFLM, 1, 1, A_WeaponReady, S_CFLAMEREADY6, 0, 0},      // S_CFLAMEREADY5
    {SPR_CFLM, 1, 1, A_WeaponReady, S_CFLAMEREADY7, 0, 0},      // S_CFLAMEREADY6
    {SPR_CFLM, 1, 1, A_WeaponReady, S_CFLAMEREADY8, 0, 0},      // S_CFLAMEREADY7
    {SPR_CFLM, 1, 1, A_WeaponReady, S_CFLAMEREADY9, 0, 0},      // S_CFLAMEREADY8
    {SPR_CFLM, 2, 1, A_WeaponReady, S_CFLAMEREADY10, 0, 0},     // S_CFLAMEREADY9
    {SPR_CFLM, 2, 1, A_WeaponReady, S_CFLAMEREADY11, 0, 0},     // S_CFLAMEREADY10
    {SPR_CFLM, 2, 1, A_WeaponReady, S_CFLAMEREADY12, 0, 0},     // S_CFLAMEREADY11
    {SPR_CFLM, 2, 1, A_WeaponReady, S_CFLAMEREADY1, 0, 0},      // S_CFLAMEREADY12
    {SPR_CFLM, 0, 1, A_Lower, S_CFLAMEDOWN, 0, 0},      // S_CFLAMEDOWN
    {SPR_CFLM, 0, 1, A_Raise, S_CFLAMEUP, 0, 0},        // S_CFLAMEUP
    {SPR_CFLM, 0, 2, NULL, S_CFLAMEATK_2, 0, 40},       // S_CFLAMEATK_1
    {SPR_CFLM, 3, 2, NULL, S_CFLAMEATK_3, 0, 50},       // S_CFLAMEATK_2
    {SPR_CFLM, 3, 2, NULL, S_CFLAMEATK_4, 0, 36},       // S_CFLAMEATK_3
    {SPR_CFLM, 32772, 4, NULL, S_CFLAMEATK_5, 0, 0},    // S_CFLAMEATK_4
    {SPR_CFLM, 32773, 4, A_CFlameAttack, S_CFLAMEATK_6, 0, 0},  // S_CFLAMEATK_5
    {SPR_CFLM, 32772, 4, NULL, S_CFLAMEATK_7, 0, 0},    // S_CFLAMEATK_6
    {SPR_CFLM, 6, 2, NULL, S_CFLAMEATK_8, 0, 40},       // S_CFLAMEATK_7
    {SPR_CFLM, 6, 2, NULL, S_CFLAMEREADY1, 0, 0},       // S_CFLAMEATK_8
    {SPR_CFFX, 32781, 5, NULL, S_CFLAMEFLOOR2, 0, 0},   // S_CFLAMEFLOOR1
    {SPR_CFFX, 32782, 4, NULL, S_CFLAMEFLOOR3, 0, 0},   // S_CFLAMEFLOOR2
    {SPR_CFFX, 32783, 3, NULL, S_NULL, 0, 0},   // S_CFLAMEFLOOR3
    {SPR_CFFX, 32768, 3, NULL, S_FLAMEPUFF2, 0, 0},     // S_FLAMEPUFF1
    {SPR_CFFX, 32769, 3, NULL, S_FLAMEPUFF3, 0, 0},     // S_FLAMEPUFF2
    {SPR_CFFX, 32770, 3, NULL, S_FLAMEPUFF4, 0, 0},     // S_FLAMEPUFF3
    {SPR_CFFX, 32771, 4, NULL, S_FLAMEPUFF5, 0, 0},     // S_FLAMEPUFF4
    {SPR_CFFX, 32772, 3, NULL, S_FLAMEPUFF6, 0, 0},     // S_FLAMEPUFF5
    {SPR_CFFX, 32773, 4, NULL, S_FLAMEPUFF7, 0, 0},     // S_FLAMEPUFF6
    {SPR_CFFX, 32774, 3, NULL, S_FLAMEPUFF8, 0, 0},     // S_FLAMEPUFF7
    {SPR_CFFX, 32775, 4, NULL, S_FLAMEPUFF9, 0, 0},     // S_FLAMEPUFF8
    {SPR_CFFX, 32776, 3, NULL, S_FLAMEPUFF10, 0, 0},    // S_FLAMEPUFF9
    {SPR_CFFX, 32777, 4, NULL, S_FLAMEPUFF11, 0, 0},    // S_FLAMEPUFF10
    {SPR_CFFX, 32778, 3, NULL, S_FLAMEPUFF12, 0, 0},    // S_FLAMEPUFF11
    {SPR_CFFX, 32779, 4, NULL, S_FLAMEPUFF13, 0, 0},    // S_FLAMEPUFF12
    {SPR_CFFX, 32780, 3, NULL, S_NULL, 0, 0},   // S_FLAMEPUFF13
    {SPR_CFFX, 32768, 3, NULL, S_FLAMEPUFF2_2, 0, 0},   // S_FLAMEPUFF2_1
    {SPR_CFFX, 32769, 3, NULL, S_FLAMEPUFF2_3, 0, 0},   // S_FLAMEPUFF2_2
    {SPR_CFFX, 32770, 3, NULL, S_FLAMEPUFF2_4, 0, 0},   // S_FLAMEPUFF2_3
    {SPR_CFFX, 32771, 4, NULL, S_FLAMEPUFF2_5, 0, 0},   // S_FLAMEPUFF2_4
    {SPR_CFFX, 32772, 3, NULL, S_FLAMEPUFF2_6, 0, 0},   // S_FLAMEPUFF2_5
    {SPR_CFFX, 32773, 4, NULL, S_FLAMEPUFF2_7, 0, 0},   // S_FLAMEPUFF2_6
    {SPR_CFFX, 32774, 3, NULL, S_FLAMEPUFF2_8, 0, 0},   // S_FLAMEPUFF2_7
    {SPR_CFFX, 32775, 4, NULL, S_FLAMEPUFF2_9, 0, 0},   // S_FLAMEPUFF2_8
    {SPR_CFFX, 32776, 3, NULL, S_FLAMEPUFF2_10, 0, 0},  // S_FLAMEPUFF2_9
    {SPR_CFFX, 32770, 3, NULL, S_FLAMEPUFF2_11, 0, 0},  // S_FLAMEPUFF2_10
    {SPR_CFFX, 32771, 4, NULL, S_FLAMEPUFF2_12, 0, 0},  // S_FLAMEPUFF2_11
    {SPR_CFFX, 32772, 3, NULL, S_FLAMEPUFF2_13, 0, 0},  // S_FLAMEPUFF2_12
    {SPR_CFFX, 32773, 4, NULL, S_FLAMEPUFF2_14, 0, 0},  // S_FLAMEPUFF2_13
    {SPR_CFFX, 32774, 3, NULL, S_FLAMEPUFF2_15, 0, 0},  // S_FLAMEPUFF2_14
    {SPR_CFFX, 32775, 4, NULL, S_FLAMEPUFF2_16, 0, 0},  // S_FLAMEPUFF2_15
    {SPR_CFFX, 32776, 3, NULL, S_FLAMEPUFF2_17, 0, 0},  // S_FLAMEPUFF2_16
    {SPR_CFFX, 32777, 4, NULL, S_FLAMEPUFF2_18, 0, 0},  // S_FLAMEPUFF2_17
    {SPR_CFFX, 32778, 3, NULL, S_FLAMEPUFF2_19, 0, 0},  // S_FLAMEPUFF2_18
    {SPR_CFFX, 32779, 4, NULL, S_FLAMEPUFF2_20, 0, 0},  // S_FLAMEPUFF2_19
    {SPR_CFFX, 32780, 3, NULL, S_NULL, 0, 0},   // S_FLAMEPUFF2_20
    {SPR_CFCF, 32768, 4, NULL, S_CIRCLE_FLAME2, 0, 0},  // S_CIRCLE_FLAME1
    {SPR_CFCF, 32769, 2, A_CFlameRotate, S_CIRCLE_FLAME3, 0, 0},        // S_CIRCLE_FLAME2
    {SPR_CFCF, 32770, 2, NULL, S_CIRCLE_FLAME4, 0, 0},  // S_CIRCLE_FLAME3
    {SPR_CFCF, 32771, 1, NULL, S_CIRCLE_FLAME5, 0, 0},  // S_CIRCLE_FLAME4
    {SPR_CFCF, 32772, 2, NULL, S_CIRCLE_FLAME6, 0, 0},  // S_CIRCLE_FLAME5
    {SPR_CFCF, 32773, 2, A_CFlameRotate, S_CIRCLE_FLAME7, 0, 0},        // S_CIRCLE_FLAME6
    {SPR_CFCF, 32774, 1, NULL, S_CIRCLE_FLAME8, 0, 0},  // S_CIRCLE_FLAME7
    {SPR_CFCF, 32775, 2, NULL, S_CIRCLE_FLAME9, 0, 0},  // S_CIRCLE_FLAME8
    {SPR_CFCF, 32776, 2, NULL, S_CIRCLE_FLAME10, 0, 0}, // S_CIRCLE_FLAME9
    {SPR_CFCF, 32777, 1, A_CFlameRotate, S_CIRCLE_FLAME11, 0, 0},       // S_CIRCLE_FLAME10
    {SPR_CFCF, 32778, 2, NULL, S_CIRCLE_FLAME12, 0, 0}, // S_CIRCLE_FLAME11
    {SPR_CFCF, 32779, 3, NULL, S_CIRCLE_FLAME13, 0, 0}, // S_CIRCLE_FLAME12
    {SPR_CFCF, 32780, 3, NULL, S_CIRCLE_FLAME14, 0, 0}, // S_CIRCLE_FLAME13
    {SPR_CFCF, 32781, 2, A_CFlameRotate, S_CIRCLE_FLAME15, 0, 0},       // S_CIRCLE_FLAME14
    {SPR_CFCF, 32782, 3, NULL, S_CIRCLE_FLAME16, 0, 0}, // S_CIRCLE_FLAME15
    {SPR_CFCF, 32783, 2, NULL, S_NULL, 0, 0},   // S_CIRCLE_FLAME16
    {SPR_CFCF, 32784, 3, NULL, S_CIRCLE_FLAME_X2, 0, 0},        // S_CIRCLE_FLAME_X1
    {SPR_CFCF, 32785, 3, NULL, S_CIRCLE_FLAME_X3, 0, 0},        // S_CIRCLE_FLAME_X2
    {SPR_CFCF, 32786, 3, A_Explode, S_CIRCLE_FLAME_X4, 0, 0},   // S_CIRCLE_FLAME_X3
    {SPR_CFCF, 32787, 3, NULL, S_CIRCLE_FLAME_X5, 0, 0},        // S_CIRCLE_FLAME_X4
    {SPR_CFCF, 32788, 3, NULL, S_CIRCLE_FLAME_X6, 0, 0},        // S_CIRCLE_FLAME_X5
    {SPR_CFCF, 32789, 3, NULL, S_CIRCLE_FLAME_X7, 0, 0},        // S_CIRCLE_FLAME_X6
    {SPR_CFCF, 32790, 3, NULL, S_CIRCLE_FLAME_X8, 0, 0},        // S_CIRCLE_FLAME_X7
    {SPR_CFCF, 32791, 3, NULL, S_CIRCLE_FLAME_X9, 0, 0},        // S_CIRCLE_FLAME_X8
    {SPR_CFCF, 32792, 3, NULL, S_CIRCLE_FLAME_X10, 0, 0},       // S_CIRCLE_FLAME_X9
    {SPR_CFCF, 32793, 3, NULL, S_NULL, 0, 0},   // S_CIRCLE_FLAME_X10
    {SPR_CFFX, 32768, 4, NULL, S_CFLAME_MISSILE2, 0, 0},        // S_CFLAME_MISSILE1
    {SPR_CFFX, 0, 1, A_CFlamePuff, S_FLAMEPUFF1, 0, 0}, // S_CFLAME_MISSILE2
    {SPR_CFFX, 32768, 1, A_CFlameMissile, S_FLAMEPUFF1, 0, 0},  // S_CFLAME_MISSILE_X
    {SPR_CHLY, 0, 1, A_WeaponReady, S_CHOLYREADY, 0, 0},        // S_CHOLYREADY
    {SPR_CHLY, 0, 1, A_Lower, S_CHOLYDOWN, 0, 0},       // S_CHOLYDOWN
    {SPR_CHLY, 0, 1, A_Raise, S_CHOLYUP, 0, 0}, // S_CHOLYUP
    {SPR_CHLY, 32768, 1, NULL, S_CHOLYATK_2, 0, 40},    // S_CHOLYATK_1
    {SPR_CHLY, 32769, 1, NULL, S_CHOLYATK_3, 0, 40},    // S_CHOLYATK_2
    {SPR_CHLY, 32770, 2, NULL, S_CHOLYATK_4, 0, 43},    // S_CHOLYATK_3
    {SPR_CHLY, 32771, 2, NULL, S_CHOLYATK_5, 0, 43},    // S_CHOLYATK_4
    {SPR_CHLY, 32772, 2, NULL, S_CHOLYATK_6, 0, 45},    // S_CHOLYATK_5
    {SPR_CHLY, 32773, 6, A_CHolyAttack, S_CHOLYATK_7, 0, 48},   // S_CHOLYATK_6
    {SPR_CHLY, 32774, 2, A_CHolyPalette, S_CHOLYATK_8, 0, 40},  // S_CHOLYATK_7
    {SPR_CHLY, 32774, 2, A_CHolyPalette, S_CHOLYATK_9, 0, 40},  // S_CHOLYATK_8
    {SPR_CHLY, 32774, 2, A_CHolyPalette, S_CHOLYREADY, 0, 36},  // S_CHOLYATK_9
    {SPR_SPIR, 0, 2, A_CHolySeek, S_HOLY_FX2, 0, 0},    // S_HOLY_FX1
    {SPR_SPIR, 0, 2, A_CHolySeek, S_HOLY_FX3, 0, 0},    // S_HOLY_FX2
    {SPR_SPIR, 1, 2, A_CHolySeek, S_HOLY_FX4, 0, 0},    // S_HOLY_FX3
    {SPR_SPIR, 1, 2, A_CHolyCheckScream, S_HOLY_FX1, 0, 0},     // S_HOLY_FX4
    {SPR_SPIR, 3, 4, NULL, S_HOLY_FX_X2, 0, 0}, // S_HOLY_FX_X1
    {SPR_SPIR, 4, 4, A_Scream, S_HOLY_FX_X3, 0, 0},     // S_HOLY_FX_X2
    {SPR_SPIR, 5, 4, NULL, S_HOLY_FX_X4, 0, 0}, // S_HOLY_FX_X3
    {SPR_SPIR, 6, 4, NULL, S_HOLY_FX_X5, 0, 0}, // S_HOLY_FX_X4
    {SPR_SPIR, 7, 4, NULL, S_HOLY_FX_X6, 0, 0}, // S_HOLY_FX_X5
    {SPR_SPIR, 8, 4, NULL, S_NULL, 0, 0},       // S_HOLY_FX_X6
    {SPR_SPIR, 2, 1, A_CHolyTail, S_HOLY_TAIL1, 0, 0},  // S_HOLY_TAIL1
    {SPR_SPIR, 3, -1, NULL, S_NULL, 0, 0},      // S_HOLY_TAIL2
    {SPR_SPIR, 10, 3, NULL, S_HOLY_PUFF2, 0, 0},        // S_HOLY_PUFF1
    {SPR_SPIR, 11, 3, NULL, S_HOLY_PUFF3, 0, 0},        // S_HOLY_PUFF2
    {SPR_SPIR, 12, 3, NULL, S_HOLY_PUFF4, 0, 0},        // S_HOLY_PUFF3
    {SPR_SPIR, 13, 3, NULL, S_HOLY_PUFF5, 0, 0},        // S_HOLY_PUFF4
    {SPR_SPIR, 14, 3, NULL, S_NULL, 0, 0},      // S_HOLY_PUFF5
    {SPR_SPIR, 32783, 3, A_CHolySpawnPuff, S_HOLY_MISSILE2, 0, 0},      // S_HOLY_MISSILE1
    {SPR_SPIR, 32783, 3, A_CHolySpawnPuff, S_HOLY_MISSILE3, 0, 0},      // S_HOLY_MISSILE2
    {SPR_SPIR, 32783, 3, A_CHolySpawnPuff, S_HOLY_MISSILE4, 0, 0},      // S_HOLY_MISSILE3
    {SPR_SPIR, 32783, 3, A_CHolySpawnPuff, S_HOLY_MISSILE_X, 0, 0},     // S_HOLY_MISSILE4
    {SPR_SPIR, 32783, 1, A_CHolyAttack2, S_NULL, 0, 0}, // S_HOLY_MISSILE_X
    {SPR_SPIR, 16, 3, NULL, S_HOLY_MISSILE_P2, 0, 0},   // S_HOLY_MISSILE_P1
    {SPR_SPIR, 17, 3, NULL, S_HOLY_MISSILE_P3, 0, 0},   // S_HOLY_MISSILE_P2
    {SPR_SPIR, 18, 3, NULL, S_HOLY_MISSILE_P4, 0, 0},   // S_HOLY_MISSILE_P3
    {SPR_SPIR, 19, 3, NULL, S_HOLY_MISSILE_P5, 0, 0},   // S_HOLY_MISSILE_P4
    {SPR_SPIR, 20, 3, NULL, S_NULL, 0, 0},      // S_HOLY_MISSILE_P5
    {SPR_MWND, 0, 1, A_WeaponReady, S_MWANDREADY, 0, 0},        // S_MWANDREADY
    {SPR_MWND, 0, 1, A_Lower, S_MWANDDOWN, 0, 0},       // S_MWANDDOWN
    {SPR_MWND, 0, 1, A_Raise, S_MWANDUP, 0, 0}, // S_MWANDUP
    {SPR_MWND, 0, 6, NULL, S_MWANDATK_2, 0, 0}, // S_MWANDATK_1
    {SPR_MWND, 32769, 6, A_MWandAttack, S_MWANDATK_3, 0, 48},   // S_MWANDATK_2
    {SPR_MWND, 0, 3, NULL, S_MWANDATK_4, 0, 40},        // S_MWANDATK_3
    {SPR_MWND, 0, 3, A_ReFire, S_MWANDREADY, 0, 36},    // S_MWANDATK_4
    {SPR_MWND, 32772, 4, NULL, S_MWANDPUFF2, 0, 0},     // S_MWANDPUFF1
    {SPR_MWND, 32773, 3, NULL, S_MWANDPUFF3, 0, 0},     // S_MWANDPUFF2
    {SPR_MWND, 32774, 4, NULL, S_MWANDPUFF4, 0, 0},     // S_MWANDPUFF3
    {SPR_MWND, 32775, 3, NULL, S_MWANDPUFF5, 0, 0},     // S_MWANDPUFF4
    {SPR_MWND, 32776, 4, NULL, S_NULL, 0, 0},   // S_MWANDPUFF5
    {SPR_MWND, 2, 4, NULL, S_MWANDSMOKE2, 0, 0},        // S_MWANDSMOKE1
    {SPR_MWND, 3, 4, NULL, S_MWANDSMOKE3, 0, 0},        // S_MWANDSMOKE2
    {SPR_MWND, 2, 4, NULL, S_MWANDSMOKE4, 0, 0},        // S_MWANDSMOKE3
    {SPR_MWND, 3, 4, NULL, S_NULL, 0, 0},       // S_MWANDSMOKE4
    {SPR_MWND, 32770, 4, NULL, S_MWAND_MISSILE2, 0, 0}, // S_MWAND_MISSILE1
    {SPR_MWND, 32771, 4, NULL, S_MWAND_MISSILE1, 0, 0}, // S_MWAND_MISSILE2
    {SPR_WMLG, 32768, 4, NULL, S_MW_LIGHTNING2, 0, 0},  // S_MW_LIGHTNING1
    {SPR_WMLG, 32769, 4, NULL, S_MW_LIGHTNING3, 0, 0},  // S_MW_LIGHTNING2
    {SPR_WMLG, 32770, 4, NULL, S_MW_LIGHTNING4, 0, 0},  // S_MW_LIGHTNING3
    {SPR_WMLG, 32771, 4, NULL, S_MW_LIGHTNING5, 0, 0},  // S_MW_LIGHTNING4
    {SPR_WMLG, 32772, 4, NULL, S_MW_LIGHTNING6, 0, 0},  // S_MW_LIGHTNING5
    {SPR_WMLG, 32773, 4, NULL, S_MW_LIGHTNING7, 0, 0},  // S_MW_LIGHTNING6
    {SPR_WMLG, 32774, 4, NULL, S_MW_LIGHTNING8, 0, 0},  // S_MW_LIGHTNING7
    {SPR_WMLG, 32775, 4, NULL, S_MW_LIGHTNING1, 0, 0},  // S_MW_LIGHTNING8
    // [crispy] disable full bright for Arc of Death, use brightmap instead
    {SPR_MLNG, 0, 1, A_WeaponReady, S_MLIGHTNINGREADY2, 0, 0},      // S_MLIGHTNINGREADY
    {SPR_MLNG, 0, 1, A_WeaponReady, S_MLIGHTNINGREADY3, 0, 0},      // S_MLIGHTNINGREADY2
    {SPR_MLNG, 0, 1, A_WeaponReady, S_MLIGHTNINGREADY4, 0, 0},      // S_MLIGHTNINGREADY3
    {SPR_MLNG, 0, 1, A_WeaponReady, S_MLIGHTNINGREADY5, 0, 0},      // S_MLIGHTNINGREADY4
    {SPR_MLNG, 0, 1, A_WeaponReady, S_MLIGHTNINGREADY6, 0, 0},      // S_MLIGHTNINGREADY5
    {SPR_MLNG, 0, 1, A_LightningReady, S_MLIGHTNINGREADY7, 0, 0},   // S_MLIGHTNINGREADY6
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY8, 0, 0},      // S_MLIGHTNINGREADY7
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY9, 0, 0},      // S_MLIGHTNINGREADY8
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY10, 0, 0},     // S_MLIGHTNINGREADY9
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY11, 0, 0},     // S_MLIGHTNINGREADY10
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY12, 0, 0},     // S_MLIGHTNINGREADY11
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY13, 0, 0},     // S_MLIGHTNINGREADY12
    {SPR_MLNG, 2, 1, A_WeaponReady, S_MLIGHTNINGREADY14, 0, 0},     // S_MLIGHTNINGREADY13
    {SPR_MLNG, 2, 1, A_WeaponReady, S_MLIGHTNINGREADY15, 0, 0},     // S_MLIGHTNINGREADY14
    {SPR_MLNG, 2, 1, A_WeaponReady, S_MLIGHTNINGREADY16, 0, 0},     // S_MLIGHTNINGREADY15
    {SPR_MLNG, 2, 1, A_WeaponReady, S_MLIGHTNINGREADY17, 0, 0},     // S_MLIGHTNINGREADY16
    {SPR_MLNG, 2, 1, A_WeaponReady, S_MLIGHTNINGREADY18, 0, 0},     // S_MLIGHTNINGREADY17
    {SPR_MLNG, 2, 1, A_LightningReady, S_MLIGHTNINGREADY19, 0, 0},  // S_MLIGHTNINGREADY18
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY20, 0, 0},     // S_MLIGHTNINGREADY19
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY21, 0, 0},     // S_MLIGHTNINGREADY20
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY22, 0, 0},     // S_MLIGHTNINGREADY21
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY23, 0, 0},     // S_MLIGHTNINGREADY22
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY24, 0, 0},     // S_MLIGHTNINGREADY23
    {SPR_MLNG, 1, 1, A_WeaponReady, S_MLIGHTNINGREADY, 0, 0},       // S_MLIGHTNINGREADY24
    {SPR_MLNG, 0, 1, A_Lower, S_MLIGHTNINGDOWN, 0, 0},      // S_MLIGHTNINGDOWN
    {SPR_MLNG, 0, 1, A_Raise, S_MLIGHTNINGUP, 0, 0},        // S_MLIGHTNINGUP
    {SPR_MLNG, 3, 3, NULL, S_MLIGHTNINGATK_2, 0, 0},        // S_MLIGHTNINGATK_1
    {SPR_MLNG, 4, 3, NULL, S_MLIGHTNINGATK_3, 0, 0},        // S_MLIGHTNINGATK_2
    {SPR_MLNG, 5, 4, A_MLightningAttack, S_MLIGHTNINGATK_4, 0, 0},  // S_MLIGHTNINGATK_3
    {SPR_MLNG, 6, 4, NULL, S_MLIGHTNINGATK_5, 0, 0},        // S_MLIGHTNINGATK_4
    {SPR_MLNG, 7, 3, NULL, S_MLIGHTNINGATK_6, 0, 0},        // S_MLIGHTNINGATK_5
    {SPR_MLNG, 8, 3, NULL, S_MLIGHTNINGATK_7, 0, 0},        // S_MLIGHTNINGATK_6
    {SPR_MLNG, 8, 6, NULL, S_MLIGHTNINGATK_8, 0, 199},      // S_MLIGHTNINGATK_7
    {SPR_MLNG, 2, 2, NULL, S_MLIGHTNINGATK_9, 0, 55},       // S_MLIGHTNINGATK_8
    {SPR_MLNG, 1, 2, NULL, S_MLIGHTNINGATK_10, 0, 50},      // S_MLIGHTNINGATK_9
    {SPR_MLNG, 1, 2, NULL, S_MLIGHTNINGATK_11, 0, 45},      // S_MLIGHTNINGATK_10
    {SPR_MLNG, 1, 2, NULL, S_MLIGHTNINGREADY, 0, 40},       // S_MLIGHTNINGATK_11
    {SPR_MLFX, 32768, 2, A_LightningZap, S_LIGHTNING_CEILING2, 0, 0},   // S_LIGHTNING_CEILING1
    {SPR_MLFX, 32769, 2, A_LightningClip, S_LIGHTNING_CEILING3, 0, 0},  // S_LIGHTNING_CEILING2
    {SPR_MLFX, 32770, 2, A_LightningClip, S_LIGHTNING_CEILING4, 0, 0},  // S_LIGHTNING_CEILING3
    {SPR_MLFX, 32771, 2, A_LightningClip, S_LIGHTNING_CEILING1, 0, 0},  // S_LIGHTNING_CEILING4
    {SPR_MLF2, 32768, 2, A_LightningRemove, S_LIGHTNING_C_X2, 0, 0},    // S_LIGHTNING_C_X1
    {SPR_MLF2, 32769, 3, NULL, S_LIGHTNING_C_X3, 0, 0}, // S_LIGHTNING_C_X2
    {SPR_MLF2, 32770, 3, NULL, S_LIGHTNING_C_X4, 0, 0}, // S_LIGHTNING_C_X3
    {SPR_MLF2, 32771, 3, NULL, S_LIGHTNING_C_X5, 0, 0}, // S_LIGHTNING_C_X4
    {SPR_MLF2, 32772, 3, NULL, S_LIGHTNING_C_X6, 0, 0}, // S_LIGHTNING_C_X5
    {SPR_MLF2, 32778, 3, NULL, S_LIGHTNING_C_X7, 0, 0}, // S_LIGHTNING_C_X6
    {SPR_MLF2, 32779, 3, NULL, S_LIGHTNING_C_X8, 0, 0}, // S_LIGHTNING_C_X7
    {SPR_MLF2, 32780, 3, NULL, S_LIGHTNING_C_X9, 0, 0}, // S_LIGHTNING_C_X8
    {SPR_ACLO, 4, 35, NULL, S_LIGHTNING_C_X10, 0, 0},   // S_LIGHTNING_C_X9
    {SPR_MLF2, 32781, 3, NULL, S_LIGHTNING_C_X11, 0, 0},        // S_LIGHTNING_C_X10
    {SPR_MLF2, 32782, 3, NULL, S_LIGHTNING_C_X12, 0, 0},        // S_LIGHTNING_C_X11
    {SPR_MLF2, 32783, 4, NULL, S_LIGHTNING_C_X13, 0, 0},        // S_LIGHTNING_C_X12
    {SPR_MLF2, 32784, 3, NULL, S_LIGHTNING_C_X14, 0, 0},        // S_LIGHTNING_C_X13
    {SPR_MLF2, 32783, 3, NULL, S_LIGHTNING_C_X15, 0, 0},        // S_LIGHTNING_C_X14
    {SPR_MLF2, 32784, 4, NULL, S_LIGHTNING_C_X16, 0, 0},        // S_LIGHTNING_C_X15
    {SPR_MLF2, 32783, 3, NULL, S_LIGHTNING_C_X17, 0, 0},        // S_LIGHTNING_C_X16
    {SPR_MLF2, 32782, 3, NULL, S_LIGHTNING_C_X18, 0, 0},        // S_LIGHTNING_C_X17
    {SPR_MLF2, 32783, 3, NULL, S_LIGHTNING_C_X19, 0, 0},        // S_LIGHTNING_C_X18
    {SPR_MLF2, 32783, 1, A_HideThing, S_FREETARGMOBJ, 0, 0},    // S_LIGHTNING_C_X19
    {SPR_MLFX, 32772, 2, A_LightningZap, S_LIGHTNING_FLOOR2, 0, 0},     // S_LIGHTNING_FLOOR1
    {SPR_MLFX, 32773, 2, A_LightningClip, S_LIGHTNING_FLOOR3, 0, 0},    // S_LIGHTNING_FLOOR2
    {SPR_MLFX, 32774, 2, A_LightningClip, S_LIGHTNING_FLOOR4, 0, 0},    // S_LIGHTNING_FLOOR3
    {SPR_MLFX, 32775, 2, A_LightningClip, S_LIGHTNING_FLOOR1, 0, 0},    // S_LIGHTNING_FLOOR4
    {SPR_MLF2, 32773, 2, A_LightningRemove, S_LIGHTNING_F_X2, 0, 0},    // S_LIGHTNING_F_X1
    {SPR_MLF2, 32774, 3, NULL, S_LIGHTNING_F_X3, 0, 0}, // S_LIGHTNING_F_X2
    {SPR_MLF2, 32775, 3, NULL, S_LIGHTNING_F_X4, 0, 0}, // S_LIGHTNING_F_X3
    {SPR_MLF2, 32776, 3, NULL, S_LIGHTNING_F_X5, 0, 0}, // S_LIGHTNING_F_X4
    {SPR_MLF2, 32777, 3, NULL, S_LIGHTNING_F_X6, 0, 0}, // S_LIGHTNING_F_X5
    {SPR_MLF2, 32778, 3, NULL, S_LIGHTNING_F_X7, 0, 0}, // S_LIGHTNING_F_X6
    {SPR_MLF2, 32779, 3, NULL, S_LIGHTNING_F_X8, 0, 0}, // S_LIGHTNING_F_X7
    {SPR_MLF2, 32780, 3, NULL, S_LIGHTNING_F_X9, 0, 0}, // S_LIGHTNING_F_X8
    {SPR_ACLO, 4, 20, NULL, S_LIGHTNING_F_X10, 0, 0},   // S_LIGHTNING_F_X9
    {SPR_MLF2, 32781, 3, NULL, S_LIGHTNING_F_X11, 0, 0},        // S_LIGHTNING_F_X10
    {SPR_MLF2, 32782, 3, NULL, S_LIGHTNING_F_X12, 0, 0},        // S_LIGHTNING_F_X11
    {SPR_MLF2, 32783, 4, NULL, S_LIGHTNING_F_X13, 0, 0},        // S_LIGHTNING_F_X12
    {SPR_MLF2, 32784, 3, NULL, S_LIGHTNING_F_X14, 0, 0},        // S_LIGHTNING_F_X13
    {SPR_MLF2, 32783, 3, NULL, S_LIGHTNING_F_X15, 0, 0},        // S_LIGHTNING_F_X14
    {SPR_MLF2, 32784, 4, A_LastZap, S_LIGHTNING_F_X16, 0, 0},   // S_LIGHTNING_F_X15
    {SPR_MLF2, 32783, 3, NULL, S_LIGHTNING_F_X17, 0, 0},        // S_LIGHTNING_F_X16
    {SPR_MLF2, 32782, 3, NULL, S_LIGHTNING_F_X18, 0, 0},        // S_LIGHTNING_F_X17
    {SPR_MLF2, 32783, 3, NULL, S_LIGHTNING_F_X19, 0, 0},        // S_LIGHTNING_F_X18
    {SPR_MLF2, 32783, 1, A_HideThing, S_FREETARGMOBJ, 0, 0},    // S_LIGHTNING_F_X19
    {SPR_MLFX, 32776, 2, A_ZapMimic, S_LIGHTNING_ZAP2, 0, 0},   // S_LIGHTNING_ZAP1
    {SPR_MLFX, 32777, 2, A_ZapMimic, S_LIGHTNING_ZAP3, 0, 0},   // S_LIGHTNING_ZAP2
    {SPR_MLFX, 32778, 2, A_ZapMimic, S_LIGHTNING_ZAP4, 0, 0},   // S_LIGHTNING_ZAP3
    {SPR_MLFX, 32779, 2, A_ZapMimic, S_LIGHTNING_ZAP5, 0, 0},   // S_LIGHTNING_ZAP4
    {SPR_MLFX, 32780, 2, A_ZapMimic, S_LIGHTNING_ZAP1, 0, 0},   // S_LIGHTNING_ZAP5
    {SPR_MLFX, 32781, 2, NULL, S_LIGHTNING_ZAP_X2, 0, 0},       // S_LIGHTNING_ZAP_X1
    {SPR_MLFX, 32782, 2, NULL, S_LIGHTNING_ZAP_X3, 0, 0},       // S_LIGHTNING_ZAP_X2
    {SPR_MLFX, 32783, 2, NULL, S_LIGHTNING_ZAP_X4, 0, 0},       // S_LIGHTNING_ZAP_X3
    {SPR_MLFX, 32784, 2, NULL, S_LIGHTNING_ZAP_X5, 0, 0},       // S_LIGHTNING_ZAP_X4
    {SPR_MLFX, 32785, 2, NULL, S_LIGHTNING_ZAP_X6, 0, 0},       // S_LIGHTNING_ZAP_X5
    {SPR_MLFX, 32786, 2, NULL, S_LIGHTNING_ZAP_X7, 0, 0},       // S_LIGHTNING_ZAP_X6
    {SPR_MLFX, 32787, 2, NULL, S_LIGHTNING_ZAP_X8, 0, 0},       // S_LIGHTNING_ZAP_X7
    {SPR_MLFX, 32788, 2, NULL, S_NULL, 0, 0},   // S_LIGHTNING_ZAP_X8
    {SPR_MSTF, 0, 1, A_WeaponReady, S_MSTAFFREADY2, 0, 0},      // S_MSTAFFREADY
    {SPR_MSTF, 0, 1, A_WeaponReady, S_MSTAFFREADY3, 0, 0},      // S_MSTAFFREADY2
    {SPR_MSTF, 0, 1, A_WeaponReady, S_MSTAFFREADY4, 0, 0},      // S_MSTAFFREADY3
    {SPR_MSTF, 0, 1, A_WeaponReady, S_MSTAFFREADY5, 0, 0},      // S_MSTAFFREADY4
    {SPR_MSTF, 0, 1, A_WeaponReady, S_MSTAFFREADY6, 0, 0},      // S_MSTAFFREADY5
    {SPR_MSTF, 0, 1, A_WeaponReady, S_MSTAFFREADY7, 0, 0},      // S_MSTAFFREADY6
    {SPR_MSTF, 1, 1, A_WeaponReady, S_MSTAFFREADY8, 0, 0},      // S_MSTAFFREADY7
    {SPR_MSTF, 1, 1, A_WeaponReady, S_MSTAFFREADY9, 0, 0},      // S_MSTAFFREADY8
    {SPR_MSTF, 1, 1, A_WeaponReady, S_MSTAFFREADY10, 0, 0},     // S_MSTAFFREADY9
    {SPR_MSTF, 1, 1, A_WeaponReady, S_MSTAFFREADY11, 0, 0},     // S_MSTAFFREADY10
    {SPR_MSTF, 1, 1, A_WeaponReady, S_MSTAFFREADY12, 0, 0},     // S_MSTAFFREADY11
    {SPR_MSTF, 1, 1, A_WeaponReady, S_MSTAFFREADY13, 0, 0},     // S_MSTAFFREADY12
    {SPR_MSTF, 2, 1, A_WeaponReady, S_MSTAFFREADY14, 0, 0},     // S_MSTAFFREADY13
    {SPR_MSTF, 2, 1, A_WeaponReady, S_MSTAFFREADY15, 0, 0},     // S_MSTAFFREADY14
    {SPR_MSTF, 2, 1, A_WeaponReady, S_MSTAFFREADY16, 0, 0},     // S_MSTAFFREADY15
    {SPR_MSTF, 2, 1, A_WeaponReady, S_MSTAFFREADY17, 0, 0},     // S_MSTAFFREADY16
    {SPR_MSTF, 2, 1, A_WeaponReady, S_MSTAFFREADY18, 0, 0},     // S_MSTAFFREADY17
    {SPR_MSTF, 2, 1, A_WeaponReady, S_MSTAFFREADY19, 0, 0},     // S_MSTAFFREADY18
    {SPR_MSTF, 3, 1, A_WeaponReady, S_MSTAFFREADY20, 0, 0},     // S_MSTAFFREADY19
    {SPR_MSTF, 3, 1, A_WeaponReady, S_MSTAFFREADY21, 0, 0},     // S_MSTAFFREADY20
    {SPR_MSTF, 3, 1, A_WeaponReady, S_MSTAFFREADY22, 0, 0},     // S_MSTAFFREADY21
    {SPR_MSTF, 3, 1, A_WeaponReady, S_MSTAFFREADY23, 0, 0},     // S_MSTAFFREADY22
    {SPR_MSTF, 3, 1, A_WeaponReady, S_MSTAFFREADY24, 0, 0},     // S_MSTAFFREADY23
    {SPR_MSTF, 3, 1, A_WeaponReady, S_MSTAFFREADY25, 0, 0},     // S_MSTAFFREADY24
    {SPR_MSTF, 4, 1, A_WeaponReady, S_MSTAFFREADY26, 0, 0},     // S_MSTAFFREADY25
    {SPR_MSTF, 4, 1, A_WeaponReady, S_MSTAFFREADY27, 0, 0},     // S_MSTAFFREADY26
    {SPR_MSTF, 4, 1, A_WeaponReady, S_MSTAFFREADY28, 0, 0},     // S_MSTAFFREADY27
    {SPR_MSTF, 4, 1, A_WeaponReady, S_MSTAFFREADY29, 0, 0},     // S_MSTAFFREADY28
    {SPR_MSTF, 4, 1, A_WeaponReady, S_MSTAFFREADY30, 0, 0},     // S_MSTAFFREADY29
    {SPR_MSTF, 4, 1, A_WeaponReady, S_MSTAFFREADY31, 0, 0},     // S_MSTAFFREADY30
    {SPR_MSTF, 5, 1, A_WeaponReady, S_MSTAFFREADY32, 0, 0},     // S_MSTAFFREADY31
    {SPR_MSTF, 5, 1, A_WeaponReady, S_MSTAFFREADY33, 0, 0},     // S_MSTAFFREADY32
    {SPR_MSTF, 5, 1, A_WeaponReady, S_MSTAFFREADY34, 0, 0},     // S_MSTAFFREADY33
    {SPR_MSTF, 5, 1, A_WeaponReady, S_MSTAFFREADY35, 0, 0},     // S_MSTAFFREADY34
    {SPR_MSTF, 5, 1, A_WeaponReady, S_MSTAFFREADY, 0, 0},       // S_MSTAFFREADY35
    {SPR_MSTF, 0, 1, A_Lower, S_MSTAFFDOWN, 0, 0},      // S_MSTAFFDOWN
    {SPR_MSTF, 0, 1, A_Raise, S_MSTAFFUP, 0, 0},        // S_MSTAFFUP
    {SPR_MSTF, 6, 4, NULL, S_MSTAFFATK_2, 0, 40},       // S_MSTAFFATK_1
    {SPR_MSTF, 32775, 4, A_MStaffAttack, S_MSTAFFATK_3, 0, 48}, // S_MSTAFFATK_2
    {SPR_MSTF, 32775, 2, A_MStaffPalette, S_MSTAFFATK_4, 0, 48},        // S_MSTAFFATK_3
    {SPR_MSTF, 8, 2, A_MStaffPalette, S_MSTAFFATK_5, 0, 48},    // S_MSTAFFATK_4
    {SPR_MSTF, 8, 2, A_MStaffPalette, S_MSTAFFATK_6, 0, 48},    // S_MSTAFFATK_5
    {SPR_MSTF, 8, 1, NULL, S_MSTAFFATK_7, 0, 40},       // S_MSTAFFATK_6
    {SPR_MSTF, 9, 5, NULL, S_MSTAFFREADY, 0, 36},       // S_MSTAFFATK_7
    {SPR_MSP1, 32768, 3, A_MStaffWeave, S_MSTAFF_FX1_2, 0, 0},  // S_MSTAFF_FX1_1
    {SPR_MSP1, 32769, 3, A_MStaffWeave, S_MSTAFF_FX1_3, 0, 0},  // S_MSTAFF_FX1_2
    {SPR_MSP1, 32770, 3, A_MStaffWeave, S_MSTAFF_FX1_4, 0, 0},  // S_MSTAFF_FX1_3
    {SPR_MSP1, 32771, 3, A_MStaffWeave, S_MSTAFF_FX1_5, 0, 0},  // S_MSTAFF_FX1_4
    {SPR_MSP1, 32772, 3, A_MStaffWeave, S_MSTAFF_FX1_6, 0, 0},  // S_MSTAFF_FX1_5
    {SPR_MSP1, 32773, 3, A_MStaffWeave, S_MSTAFF_FX1_1, 0, 0},  // S_MSTAFF_FX1_6
    {SPR_MSP1, 32774, 4, NULL, S_MSTAFF_FX_X2, 0, 0},   // S_MSTAFF_FX_X1
    {SPR_MSP1, 32775, 5, A_Explode, S_MSTAFF_FX_X3, 0, 0},      // S_MSTAFF_FX_X2
    {SPR_MSP1, 32776, 4, NULL, S_MSTAFF_FX_X4, 0, 0},   // S_MSTAFF_FX_X3
    {SPR_MSP1, 32777, 5, NULL, S_MSTAFF_FX_X5, 0, 0},   // S_MSTAFF_FX_X4
    {SPR_MSP1, 32778, 4, NULL, S_MSTAFF_FX_X6, 0, 0},   // S_MSTAFF_FX_X5
    {SPR_MSP1, 32779, 5, NULL, S_MSTAFF_FX_X7, 0, 0},   // S_MSTAFF_FX_X6
    {SPR_MSP1, 32780, 4, NULL, S_MSTAFF_FX_X8, 0, 0},   // S_MSTAFF_FX_X7
    {SPR_MSP1, 32781, 5, NULL, S_MSTAFF_FX_X9, 0, 0},   // S_MSTAFF_FX_X8
    {SPR_MSP1, 32782, 4, NULL, S_MSTAFF_FX_X10, 0, 0},  // S_MSTAFF_FX_X9
    {SPR_MSP1, 32783, 4, NULL, S_NULL, 0, 0},   // S_MSTAFF_FX_X10
    {SPR_MSP2, 32768, 2, A_MStaffTrack, S_MSTAFF_FX2_2, 0, 0},  // S_MSTAFF_FX2_1
    {SPR_MSP2, 32769, 2, A_MStaffTrack, S_MSTAFF_FX2_3, 0, 0},  // S_MSTAFF_FX2_2
    {SPR_MSP2, 32770, 2, A_MStaffTrack, S_MSTAFF_FX2_4, 0, 0},  // S_MSTAFF_FX2_3
    {SPR_MSP2, 32771, 2, A_MStaffTrack, S_MSTAFF_FX2_1, 0, 0},  // S_MSTAFF_FX2_4
    {SPR_MSP2, 32772, 4, NULL, S_MSTAFF_FX2_X2, 0, 0},  // S_MSTAFF_FX2_X1
    {SPR_MSP2, 32773, 5, A_Explode, S_MSTAFF_FX2_X3, 0, 0},     // S_MSTAFF_FX2_X2
    {SPR_MSP2, 32774, 5, NULL, S_MSTAFF_FX2_X4, 0, 0},  // S_MSTAFF_FX2_X3
    {SPR_MSP2, 32775, 5, NULL, S_MSTAFF_FX2_X5, 0, 0},  // S_MSTAFF_FX2_X4
    {SPR_MSP2, 32776, 4, NULL, S_NULL, 0, 0},   // S_MSTAFF_FX2_X5
    {SPR_WFR1, 32768, -1, NULL, S_NULL, 0, 0},  // S_FSWORD1
    {SPR_WFR2, 32768, -1, NULL, S_NULL, 0, 0},  // S_FSWORD2
    {SPR_WFR3, 32768, -1, NULL, S_NULL, 0, 0},  // S_FSWORD3
    {SPR_WCH1, 32768, -1, NULL, S_NULL, 0, 0},  // S_CHOLY1
    {SPR_WCH2, 32768, -1, NULL, S_NULL, 0, 0},  // S_CHOLY2
    {SPR_WCH3, 32768, -1, NULL, S_NULL, 0, 0},  // S_CHOLY3
    {SPR_WMS1, 32768, -1, NULL, S_NULL, 0, 0},  // S_MSTAFF1
    {SPR_WMS2, 32768, -1, NULL, S_NULL, 0, 0},  // S_MSTAFF2
    {SPR_WMS3, 32768, -1, NULL, S_NULL, 0, 0},  // S_MSTAFF3
    {SPR_WPIG, 0, 1, A_WeaponReady, S_SNOUTREADY, 0, 0},        // S_SNOUTREADY
    {SPR_WPIG, 0, 1, A_Lower, S_SNOUTDOWN, 0, 0},       // S_SNOUTDOWN
    {SPR_WPIG, 0, 1, A_Raise, S_SNOUTUP, 0, 0}, // S_SNOUTUP
    {SPR_WPIG, 0, 4, A_SnoutAttack, S_SNOUTATK2, 0, 0}, // S_SNOUTATK1
    {SPR_WPIG, 1, 8, A_SnoutAttack, S_SNOUTREADY, 0, 0},        // S_SNOUTATK2
    {SPR_WMCS, 32768, 8, NULL, S_COS2, 0, 0},   // S_COS1
    {SPR_WMCS, 32769, 8, NULL, S_COS3, 0, 0},   // S_COS2
    {SPR_WMCS, 32770, 8, NULL, S_COS1, 0, 0},   // S_COS3
    {SPR_CONE, 0, 1, A_WeaponReady, S_CONEREADY, 0, 0}, // S_CONEREADY
    {SPR_CONE, 0, 1, A_Lower, S_CONEDOWN, 0, 0},        // S_CONEDOWN
    {SPR_CONE, 0, 1, A_Raise, S_CONEUP, 0, 0},  // S_CONEUP
    {SPR_CONE, 1, 3, NULL, S_CONEATK1_2, 0, 0}, // S_CONEATK1_1
    {SPR_CONE, 2, 4, NULL, S_CONEATK1_3, 0, 0}, // S_CONEATK1_2
    {SPR_CONE, 3, 3, NULL, S_CONEATK1_4, 0, 0}, // S_CONEATK1_3
    {SPR_CONE, 4, 5, NULL, S_CONEATK1_5, 0, 0}, // S_CONEATK1_4
    {SPR_CONE, 5, 3, A_FireConePL1, S_CONEATK1_6, 0, 0},        // S_CONEATK1_5
    {SPR_CONE, 6, 3, NULL, S_CONEATK1_7, 0, 0}, // S_CONEATK1_6
    {SPR_CONE, 0, 9, NULL, S_CONEATK1_8, 0, 0}, // S_CONEATK1_7
    {SPR_CONE, 0, 10, A_ReFire, S_CONEREADY, 0, 0},     // S_CONEATK1_8
    {SPR_SHRD, 32768, 2, NULL, S_SHARDFX1_2, 0, 0},     // S_SHARDFX1_1
    {SPR_SHRD, 32768, 3, A_ShedShard, S_SHARDFX1_3, 0, 0},      // S_SHARDFX1_2
    {SPR_SHRD, 32769, 3, NULL, S_SHARDFX1_4, 0, 0},     // S_SHARDFX1_3
    {SPR_SHRD, 32770, 3, NULL, S_SHARDFX1_1, 0, 0},     // S_SHARDFX1_4
    {SPR_SHEX, 32768, 5, NULL, S_SHARDFXE1_2, 0, 0},    // S_SHARDFXE1_1
    {SPR_SHEX, 32769, 5, NULL, S_SHARDFXE1_3, 0, 0},    // S_SHARDFXE1_2
    {SPR_SHEX, 32770, 5, NULL, S_SHARDFXE1_4, 0, 0},    // S_SHARDFXE1_3
    {SPR_SHEX, 32771, 5, NULL, S_SHARDFXE1_5, 0, 0},    // S_SHARDFXE1_4
    {SPR_SHEX, 32772, 5, NULL, S_NULL, 0, 0},   // S_SHARDFXE1_5
    {SPR_BLOD, 2, 8, NULL, S_BLOOD2, 0, 0},     // S_BLOOD1
    {SPR_BLOD, 1, 8, NULL, S_BLOOD3, 0, 0},     // S_BLOOD2
    {SPR_BLOD, 0, 8, NULL, S_NULL, 0, 0},       // S_BLOOD3
    {SPR_BLOD, 2, 8, NULL, S_BLOODSPLATTER2, 0, 0},     // S_BLOODSPLATTER1
    {SPR_BLOD, 1, 8, NULL, S_BLOODSPLATTER3, 0, 0},     // S_BLOODSPLATTER2
    {SPR_BLOD, 0, 8, NULL, S_NULL, 0, 0},       // S_BLOODSPLATTER3
    {SPR_BLOD, 0, 6, NULL, S_NULL, 0, 0},       // S_BLOODSPLATTERX
    {SPR_GIBS, 0, -1, NULL, S_NULL, 0, 0},      // S_GIBS1
    {SPR_PLAY, 0, -1, NULL, S_NULL, 0, 0},      // S_FPLAY
    {SPR_PLAY, 0, 4, NULL, S_FPLAY_RUN2, 0, 0}, // S_FPLAY_RUN1
    {SPR_PLAY, 1, 4, NULL, S_FPLAY_RUN3, 0, 0}, // S_FPLAY_RUN2
    {SPR_PLAY, 2, 4, NULL, S_FPLAY_RUN4, 0, 0}, // S_FPLAY_RUN3
    {SPR_PLAY, 3, 4, NULL, S_FPLAY_RUN1, 0, 0}, // S_FPLAY_RUN4
    {SPR_PLAY, 4, 8, NULL, S_FPLAY_ATK2, 0, 0}, // S_FPLAY_ATK1
    {SPR_PLAY, 5, 8, NULL, S_FPLAY, 0, 0},      // S_FPLAY_ATK2
    {SPR_PLAY, 6, 4, NULL, S_FPLAY_PAIN2, 0, 0},        // S_FPLAY_PAIN
    {SPR_PLAY, 6, 4, A_Pain, S_FPLAY, 0, 0},    // S_FPLAY_PAIN2
    {SPR_PLAY, 7, 6, NULL, S_FPLAY_DIE2, 0, 0}, // S_FPLAY_DIE1
    {SPR_PLAY, 8, 6, A_Scream, S_FPLAY_DIE3, 0, 0},     // S_FPLAY_DIE2
    {SPR_PLAY, 9, 6, NULL, S_FPLAY_DIE4, 0, 0}, // S_FPLAY_DIE3
    {SPR_PLAY, 10, 6, NULL, S_FPLAY_DIE5, 0, 0},        // S_FPLAY_DIE4
    {SPR_PLAY, 11, 6, A_NoBlocking, S_FPLAY_DIE6, 0, 0},        // S_FPLAY_DIE5
    {SPR_PLAY, 12, 6, NULL, S_FPLAY_DIE7, 0, 0},        // S_FPLAY_DIE6
    {SPR_PLAY, 13, -1, A_AddPlayerCorpse, S_NULL, 0, 0},        // S_FPLAY_DIE7
    {SPR_PLAY, 14, 5, A_Scream, S_FPLAY_XDIE2, 0, 0},   // S_FPLAY_XDIE1
    {SPR_PLAY, 15, 5, A_SkullPop, S_FPLAY_XDIE3, 0, 0}, // S_FPLAY_XDIE2
    {SPR_PLAY, 17, 5, A_NoBlocking, S_FPLAY_XDIE4, 0, 0},       // S_FPLAY_XDIE3
    {SPR_PLAY, 18, 5, NULL, S_FPLAY_XDIE5, 0, 0},       // S_FPLAY_XDIE4
    {SPR_PLAY, 19, 5, NULL, S_FPLAY_XDIE6, 0, 0},       // S_FPLAY_XDIE5
    {SPR_PLAY, 20, 5, NULL, S_FPLAY_XDIE7, 0, 0},       // S_FPLAY_XDIE6
    {SPR_PLAY, 21, 5, NULL, S_FPLAY_XDIE8, 0, 0},       // S_FPLAY_XDIE7
    {SPR_PLAY, 22, -1, A_AddPlayerCorpse, S_NULL, 0, 0},        // S_FPLAY_XDIE8
    {SPR_PLAY, 23, 5, A_FreezeDeath, S_FPLAY_ICE2, 0, 0},       // S_FPLAY_ICE
    {SPR_PLAY, 23, 1, A_FreezeDeathChunks, S_FPLAY_ICE2, 0, 0}, // S_FPLAY_ICE2
    {SPR_FDTH, 32768, 5, NULL, S_PLAY_F_FDTH2, 0, 0},   // S_PLAY_F_FDTH1
    {SPR_FDTH, 32769, 4, NULL, S_PLAY_FDTH3, 0, 0},     // S_PLAY_F_FDTH2
    {SPR_FDTH, 32770, 5, NULL, S_PLAY_C_FDTH2, 0, 0},   // S_PLAY_C_FDTH1
    {SPR_FDTH, 32771, 4, NULL, S_PLAY_FDTH3, 0, 0},     // S_PLAY_C_FDTH2
    {SPR_FDTH, 32772, 5, NULL, S_PLAY_M_FDTH2, 0, 0},   // S_PLAY_M_FDTH1
    {SPR_FDTH, 32773, 4, NULL, S_PLAY_FDTH3, 0, 0},     // S_PLAY_M_FDTH2
    {SPR_FDTH, 32774, 5, NULL, S_PLAY_FDTH4, 0, 0},     // S_PLAY_FDTH3
    {SPR_FDTH, 32775, 4, A_Scream, S_PLAY_FDTH5, 0, 0}, // S_PLAY_FDTH4
    {SPR_FDTH, 32776, 5, NULL, S_PLAY_FDTH6, 0, 0},     // S_PLAY_FDTH5
    {SPR_FDTH, 32777, 4, NULL, S_PLAY_FDTH7, 0, 0},     // S_PLAY_FDTH6
    {SPR_FDTH, 32778, 5, NULL, S_PLAY_FDTH8, 0, 0},     // S_PLAY_FDTH7
    {SPR_FDTH, 32779, 4, NULL, S_PLAY_FDTH9, 0, 0},     // S_PLAY_FDTH8
    {SPR_FDTH, 32780, 5, NULL, S_PLAY_FDTH10, 0, 0},    // S_PLAY_FDTH9
    {SPR_FDTH, 32781, 4, NULL, S_PLAY_FDTH11, 0, 0},    // S_PLAY_FDTH10
    {SPR_FDTH, 32782, 5, NULL, S_PLAY_FDTH12, 0, 0},    // S_PLAY_FDTH11
    {SPR_FDTH, 32783, 4, NULL, S_PLAY_FDTH13, 0, 0},    // S_PLAY_FDTH12
    {SPR_FDTH, 32784, 5, NULL, S_PLAY_FDTH14, 0, 0},    // S_PLAY_FDTH13
    {SPR_FDTH, 32785, 4, NULL, S_PLAY_FDTH15, 0, 0},    // S_PLAY_FDTH14
    {SPR_FDTH, 32786, 5, A_NoBlocking, S_PLAY_FDTH16, 0, 0},    // S_PLAY_FDTH15
    {SPR_FDTH, 32787, 4, NULL, S_PLAY_FDTH17, 0, 0},    // S_PLAY_FDTH16
    {SPR_FDTH, 32788, 5, NULL, S_PLAY_FDTH18, 0, 0},    // S_PLAY_FDTH17
    {SPR_FDTH, 32789, 4, NULL, S_PLAY_FDTH19, 0, 0},    // S_PLAY_FDTH18
    {SPR_ACLO, 4, 35, A_CheckBurnGone, S_PLAY_FDTH19, 0, 0},    // S_PLAY_FDTH19
    {SPR_ACLO, 4, 8, NULL, S_NULL, 0, 0},       // S_PLAY_FDTH20
    {SPR_BSKL, 0, 5, A_CheckSkullFloor, S_BLOODYSKULL2, 0, 0},  // S_BLOODYSKULL1
    {SPR_BSKL, 1, 5, A_CheckSkullFloor, S_BLOODYSKULL3, 0, 0},  // S_BLOODYSKULL2
    {SPR_BSKL, 2, 5, A_CheckSkullFloor, S_BLOODYSKULL4, 0, 0},  // S_BLOODYSKULL3
    {SPR_BSKL, 3, 5, A_CheckSkullFloor, S_BLOODYSKULL5, 0, 0},  // S_BLOODYSKULL4
    {SPR_BSKL, 5, 5, A_CheckSkullFloor, S_BLOODYSKULL6, 0, 0},  // S_BLOODYSKULL5
    {SPR_BSKL, 6, 5, A_CheckSkullFloor, S_BLOODYSKULL7, 0, 0},  // S_BLOODYSKULL6
    {SPR_BSKL, 7, 5, A_CheckSkullFloor, S_BLOODYSKULL1, 0, 0},  // S_BLOODYSKULL7
    {SPR_BSKL, 8, 16, A_CheckSkullDone, S_BLOODYSKULLX1, 0, 0}, // S_BLOODYSKULLX1
    {SPR_BSKL, 8, 1050, NULL, S_NULL, 0, 0},    // S_BLOODYSKULLX2
    {SPR_PLAY, 0, 5, NULL, S_PLAYER_SPEED2, 0, 0},      // S_PLAYER_SPEED1
    {SPR_PLAY, 0, 3, A_SpeedFade, S_NULL, 0, 0},        // S_PLAYER_SPEED2
    {SPR_ICEC, 0, 10, NULL, S_ICECHUNK2, 0, 0}, // S_ICECHUNK1
    {SPR_ICEC, 1, 10, A_IceSetTics, S_ICECHUNK3, 0, 0}, // S_ICECHUNK2
    {SPR_ICEC, 2, 10, A_IceSetTics, S_ICECHUNK4, 0, 0}, // S_ICECHUNK3
    {SPR_ICEC, 3, 10, A_IceSetTics, S_NULL, 0, 0},      // S_ICECHUNK4
    {SPR_ICEC, 0, 10, A_IceCheckHeadDone, S_ICECHUNK_HEAD, 0, 0},       // S_ICECHUNK_HEAD
    {SPR_ICEC, 0, 1050, NULL, S_NULL, 0, 0},    // S_ICECHUNK_HEAD2
    {SPR_CLER, 0, -1, NULL, S_NULL, 0, 0},      // S_CPLAY
    {SPR_CLER, 0, 4, NULL, S_CPLAY_RUN2, 0, 0}, // S_CPLAY_RUN1
    {SPR_CLER, 1, 4, NULL, S_CPLAY_RUN3, 0, 0}, // S_CPLAY_RUN2
    {SPR_CLER, 2, 4, NULL, S_CPLAY_RUN4, 0, 0}, // S_CPLAY_RUN3
    {SPR_CLER, 3, 4, NULL, S_CPLAY_RUN1, 0, 0}, // S_CPLAY_RUN4
    {SPR_CLER, 4, 6, NULL, S_CPLAY_ATK2, 0, 0}, // S_CPLAY_ATK1
    {SPR_CLER, 5, 6, NULL, S_CPLAY_ATK3, 0, 0}, // S_CPLAY_ATK2
    {SPR_CLER, 6, 6, NULL, S_CPLAY, 0, 0},      // S_CPLAY_ATK3
    {SPR_CLER, 7, 4, NULL, S_CPLAY_PAIN2, 0, 0},        // S_CPLAY_PAIN
    {SPR_CLER, 7, 4, A_Pain, S_CPLAY, 0, 0},    // S_CPLAY_PAIN2
    {SPR_CLER, 8, 6, NULL, S_CPLAY_DIE2, 0, 0}, // S_CPLAY_DIE1
    {SPR_CLER, 10, 6, A_Scream, S_CPLAY_DIE3, 0, 0},    // S_CPLAY_DIE2
    {SPR_CLER, 11, 6, NULL, S_CPLAY_DIE4, 0, 0},        // S_CPLAY_DIE3
    {SPR_CLER, 11, 6, NULL, S_CPLAY_DIE5, 0, 0},        // S_CPLAY_DIE4
    {SPR_CLER, 12, 6, A_NoBlocking, S_CPLAY_DIE6, 0, 0},        // S_CPLAY_DIE5
    {SPR_CLER, 13, 6, NULL, S_CPLAY_DIE7, 0, 0},        // S_CPLAY_DIE6
    {SPR_CLER, 14, 6, NULL, S_CPLAY_DIE8, 0, 0},        // S_CPLAY_DIE7
    {SPR_CLER, 15, 6, NULL, S_CPLAY_DIE9, 0, 0},        // S_CPLAY_DIE8
    {SPR_CLER, 16, -1, A_AddPlayerCorpse, S_NULL, 0, 0},        // S_CPLAY_DIE9
    {SPR_CLER, 17, 5, A_Scream, S_CPLAY_XDIE2, 0, 0},   // S_CPLAY_XDIE1
    {SPR_CLER, 18, 5, NULL, S_CPLAY_XDIE3, 0, 0},       // S_CPLAY_XDIE2
    {SPR_CLER, 19, 5, A_NoBlocking, S_CPLAY_XDIE4, 0, 0},       // S_CPLAY_XDIE3
    {SPR_CLER, 20, 5, NULL, S_CPLAY_XDIE5, 0, 0},       // S_CPLAY_XDIE4
    {SPR_CLER, 21, 5, NULL, S_CPLAY_XDIE6, 0, 0},       // S_CPLAY_XDIE5
    {SPR_CLER, 22, 5, NULL, S_CPLAY_XDIE7, 0, 0},       // S_CPLAY_XDIE6
    {SPR_CLER, 23, 5, NULL, S_CPLAY_XDIE8, 0, 0},       // S_CPLAY_XDIE7
    {SPR_CLER, 24, 5, NULL, S_CPLAY_XDIE9, 0, 0},       // S_CPLAY_XDIE8
    {SPR_CLER, 25, 5, NULL, S_CPLAY_XDIE10, 0, 0},      // S_CPLAY_XDIE9
    {SPR_CLER, 26, -1, A_AddPlayerCorpse, S_NULL, 0, 0},        // S_CPLAY_XDIE10
    {SPR_CLER, 27, 5, A_FreezeDeath, S_CPLAY_ICE2, 0, 0},       // S_CPLAY_ICE
    {SPR_CLER, 27, 1, A_FreezeDeathChunks, S_CPLAY_ICE2, 0, 0}, // S_CPLAY_ICE2
    {SPR_MAGE, 0, -1, NULL, S_NULL, 0, 0},      // S_MPLAY
    {SPR_MAGE, 0, 4, NULL, S_MPLAY_RUN2, 0, 0}, // S_MPLAY_RUN1
    {SPR_MAGE, 1, 4, NULL, S_MPLAY_RUN3, 0, 0}, // S_MPLAY_RUN2
    {SPR_MAGE, 2, 4, NULL, S_MPLAY_RUN4, 0, 0}, // S_MPLAY_RUN3
    {SPR_MAGE, 3, 4, NULL, S_MPLAY_RUN1, 0, 0}, // S_MPLAY_RUN4
    {SPR_MAGE, 4, 8, NULL, S_MPLAY_ATK2, 0, 0}, // S_MPLAY_ATK1
    {SPR_MAGE, 32773, 8, NULL, S_MPLAY, 0, 0},  // S_MPLAY_ATK2
    {SPR_MAGE, 6, 4, NULL, S_MPLAY_PAIN2, 0, 0},        // S_MPLAY_PAIN
    {SPR_MAGE, 6, 4, A_Pain, S_MPLAY, 0, 0},    // S_MPLAY_PAIN2
    {SPR_MAGE, 7, 6, NULL, S_MPLAY_DIE2, 0, 0}, // S_MPLAY_DIE1
    {SPR_MAGE, 8, 6, A_Scream, S_MPLAY_DIE3, 0, 0},     // S_MPLAY_DIE2
    {SPR_MAGE, 9, 6, NULL, S_MPLAY_DIE4, 0, 0}, // S_MPLAY_DIE3
    {SPR_MAGE, 10, 6, NULL, S_MPLAY_DIE5, 0, 0},        // S_MPLAY_DIE4
    {SPR_MAGE, 11, 6, A_NoBlocking, S_MPLAY_DIE6, 0, 0},        // S_MPLAY_DIE5
    {SPR_MAGE, 12, 6, NULL, S_MPLAY_DIE7, 0, 0},        // S_MPLAY_DIE6
    {SPR_MAGE, 13, -1, A_AddPlayerCorpse, S_NULL, 0, 0},        // S_MPLAY_DIE7
    {SPR_MAGE, 14, 5, A_Scream, S_MPLAY_XDIE2, 0, 0},   // S_MPLAY_XDIE1
    {SPR_MAGE, 15, 5, NULL, S_MPLAY_XDIE3, 0, 0},       // S_MPLAY_XDIE2
    {SPR_MAGE, 17, 5, A_NoBlocking, S_MPLAY_XDIE4, 0, 0},       // S_MPLAY_XDIE3
    {SPR_MAGE, 18, 5, NULL, S_MPLAY_XDIE5, 0, 0},       // S_MPLAY_XDIE4
    {SPR_MAGE, 19, 5, NULL, S_MPLAY_XDIE6, 0, 0},       // S_MPLAY_XDIE5
    {SPR_MAGE, 20, 5, NULL, S_MPLAY_XDIE7, 0, 0},       // S_MPLAY_XDIE6
    {SPR_MAGE, 21, 5, NULL, S_MPLAY_XDIE8, 0, 0},       // S_MPLAY_XDIE7
    {SPR_MAGE, 22, 5, NULL, S_MPLAY_XDIE9, 0, 0},       // S_MPLAY_XDIE8
    {SPR_MAGE, 23, -1, A_AddPlayerCorpse, S_NULL, 0, 0},        // S_MPLAY_XDIE9
    {SPR_MAGE, 24, 5, A_FreezeDeath, S_MPLAY_ICE2, 0, 0},       // S_MPLAY_ICE
    {SPR_MAGE, 24, 1, A_FreezeDeathChunks, S_MPLAY_ICE2, 0, 0}, // S_MPLAY_ICE2
    {SPR_PIGY, 0, -1, NULL, S_NULL, 0, 0},      // S_PIGPLAY
    {SPR_PIGY, 0, 3, NULL, S_PIGPLAY_RUN2, 0, 0},       // S_PIGPLAY_RUN1
    {SPR_PIGY, 1, 3, NULL, S_PIGPLAY_RUN3, 0, 0},       // S_PIGPLAY_RUN2
    {SPR_PIGY, 2, 3, NULL, S_PIGPLAY_RUN4, 0, 0},       // S_PIGPLAY_RUN3
    {SPR_PIGY, 3, 3, NULL, S_PIGPLAY_RUN1, 0, 0},       // S_PIGPLAY_RUN4
    {SPR_PIGY, 0, 12, NULL, S_PIGPLAY, 0, 0},   // S_PIGPLAY_ATK1
    {SPR_PIGY, 3, 4, A_PigPain, S_PIGPLAY, 0, 0},       // S_PIGPLAY_PAIN
    {SPR_PIGY, 1, 10, A_PigLook, S_PIG_LOOK1, 0, 0},    // S_PIG_LOOK1
    {SPR_PIGY, 0, 3, A_PigChase, S_PIG_WALK2, 0, 0},    // S_PIG_WALK1
    {SPR_PIGY, 1, 3, A_PigChase, S_PIG_WALK3, 0, 0},    // S_PIG_WALK2
    {SPR_PIGY, 2, 3, A_PigChase, S_PIG_WALK4, 0, 0},    // S_PIG_WALK3
    {SPR_PIGY, 3, 3, A_PigChase, S_PIG_WALK1, 0, 0},    // S_PIG_WALK4
    {SPR_PIGY, 3, 4, A_PigPain, S_PIG_WALK1, 0, 0},     // S_PIG_PAIN
    {SPR_PIGY, 0, 5, A_FaceTarget, S_PIG_ATK2, 0, 0},   // S_PIG_ATK1
    {SPR_PIGY, 0, 10, A_PigAttack, S_PIG_WALK1, 0, 0},  // S_PIG_ATK2
    {SPR_PIGY, 4, 4, A_Scream, S_PIG_DIE2, 0, 0},       // S_PIG_DIE1
    {SPR_PIGY, 5, 3, A_NoBlocking, S_PIG_DIE3, 0, 0},   // S_PIG_DIE2
    {SPR_PIGY, 6, 4, A_QueueCorpse, S_PIG_DIE4, 0, 0},  // S_PIG_DIE3
    {SPR_PIGY, 7, 3, NULL, S_PIG_DIE5, 0, 0},   // S_PIG_DIE4
    {SPR_PIGY, 8, 4, NULL, S_PIG_DIE6, 0, 0},   // S_PIG_DIE5
    {SPR_PIGY, 9, 4, NULL, S_PIG_DIE7, 0, 0},   // S_PIG_DIE6
    {SPR_PIGY, 10, 4, NULL, S_PIG_DIE8, 0, 0},  // S_PIG_DIE7
    {SPR_PIGY, 11, -1, NULL, S_NULL, 0, 0},     // S_PIG_DIE8
    {SPR_PIGY, 12, 5, A_FreezeDeath, S_PIG_ICE2, 0, 0}, // S_PIG_ICE
    {SPR_PIGY, 12, 1, A_FreezeDeathChunks, S_PIG_ICE2, 0, 0},   // S_PIG_ICE2
    {SPR_CENT, 0, 10, A_Look, S_CENTAUR_LOOK2, 0, 0},   // S_CENTAUR_LOOK1
    {SPR_CENT, 1, 10, A_Look, S_CENTAUR_LOOK1, 0, 0},   // S_CENTAUR_LOOK2
    {SPR_CENT, 0, 4, A_Chase, S_CENTAUR_WALK2, 0, 0},   // S_CENTAUR_WALK1
    {SPR_CENT, 1, 4, A_Chase, S_CENTAUR_WALK3, 0, 0},   // S_CENTAUR_WALK2
    {SPR_CENT, 2, 4, A_Chase, S_CENTAUR_WALK4, 0, 0},   // S_CENTAUR_WALK3
    {SPR_CENT, 3, 4, A_Chase, S_CENTAUR_WALK1, 0, 0},   // S_CENTAUR_WALK4
    {SPR_CENT, 7, 5, A_FaceTarget, S_CENTAUR_ATK2, 0, 0},       // S_CENTAUR_ATK1
    {SPR_CENT, 8, 4, A_FaceTarget, S_CENTAUR_ATK3, 0, 0},       // S_CENTAUR_ATK2
    {SPR_CENT, 9, 7, A_CentaurAttack, S_CENTAUR_WALK1, 0, 0},   // S_CENTAUR_ATK3
    {SPR_CENT, 4, 10, A_FaceTarget, S_CENTAUR_MISSILE2, 0, 0},  // S_CENTAUR_MISSILE1
    {SPR_CENT, 32773, 8, A_CentaurAttack2, S_CENTAUR_MISSILE3, 0, 0},   // S_CENTAUR_MISSILE2
    {SPR_CENT, 4, 10, A_FaceTarget, S_CENTAUR_MISSILE4, 0, 0},  // S_CENTAUR_MISSILE3
    {SPR_CENT, 32773, 8, A_CentaurAttack2, S_CENTAUR_WALK1, 0, 0},      // S_CENTAUR_MISSILE4
    {SPR_CENT, 6, 6, A_Pain, S_CENTAUR_PAIN2, 0, 0},    // S_CENTAUR_PAIN1
    {SPR_CENT, 6, 6, A_SetReflective, S_CENTAUR_PAIN3, 0, 0},   // S_CENTAUR_PAIN2
    {SPR_CENT, 4, 15, A_CentaurDefend, S_CENTAUR_PAIN4, 0, 0},  // S_CENTAUR_PAIN3
    {SPR_CENT, 4, 15, A_CentaurDefend, S_CENTAUR_PAIN5, 0, 0},  // S_CENTAUR_PAIN4
    {SPR_CENT, 4, 15, A_CentaurDefend, S_CENTAUR_PAIN6, 0, 0},  // S_CENTAUR_PAIN5
    {SPR_CENT, 4, 1, A_UnSetReflective, S_CENTAUR_WALK1, 0, 0}, // S_CENTAUR_PAIN6
    {SPR_CENT, 10, 4, NULL, S_CENTAUR_DEATH2, 0, 0},    // S_CENTAUR_DEATH1
    {SPR_CENT, 11, 4, A_Scream, S_CENTAUR_DEATH3, 0, 0},        // S_CENTAUR_DEATH2
    {SPR_CENT, 12, 4, NULL, S_CENTAUR_DEATH4, 0, 0},    // S_CENTAUR_DEATH3
    {SPR_CENT, 13, 4, NULL, S_CENTAUR_DEATH5, 0, 0},    // S_CENTAUR_DEATH4
    {SPR_CENT, 14, 4, A_NoBlocking, S_CENTAUR_DEATH6, 0, 0},    // S_CENTAUR_DEATH5
    {SPR_CENT, 15, 4, NULL, S_CENTAUR_DEATH7, 0, 0},    // S_CENTAUR_DEATH6
    {SPR_CENT, 16, 4, NULL, S_CENTAUR_DEATH8, 0, 0},    // S_CENTAUR_DEATH7
    {SPR_CENT, 17, 4, A_QueueCorpse, S_CENTAUR_DEATH9, 0, 0},   // S_CENTAUR_DEATH8
    {SPR_CENT, 18, 4, NULL, S_CENTAUR_DEATH0, 0, 0},    // S_CENTAUR_DEATH9
    {SPR_CENT, 19, -1, NULL, S_NULL, 0, 0},     // S_CENTAUR_DEATH0
    {SPR_CTXD, 0, 4, NULL, S_CENTAUR_DEATH_X2, 0, 0},   // S_CENTAUR_DEATH_X1
    {SPR_CTXD, 1, 4, A_NoBlocking, S_CENTAUR_DEATH_X3, 0, 0},   // S_CENTAUR_DEATH_X2
    {SPR_CTXD, 2, 4, A_CentaurDropStuff, S_CENTAUR_DEATH_X4, 0, 0},     // S_CENTAUR_DEATH_X3
    {SPR_CTXD, 3, 3, A_Scream, S_CENTAUR_DEATH_X5, 0, 0},       // S_CENTAUR_DEATH_X4
    {SPR_CTXD, 4, 4, A_QueueCorpse, S_CENTAUR_DEATH_X6, 0, 0},  // S_CENTAUR_DEATH_X5
    {SPR_CTXD, 5, 3, NULL, S_CENTAUR_DEATH_X7, 0, 0},   // S_CENTAUR_DEATH_X6
    {SPR_CTXD, 6, 4, NULL, S_CENTAUR_DEATH_X8, 0, 0},   // S_CENTAUR_DEATH_X7
    {SPR_CTXD, 7, 3, NULL, S_CENTAUR_DEATH_X9, 0, 0},   // S_CENTAUR_DEATH_X8
    {SPR_CTXD, 8, 4, NULL, S_CENTAUR_DEATH_X10, 0, 0},  // S_CENTAUR_DEATH_X9
    {SPR_CTXD, 9, 3, NULL, S_CENTAUR_DEATH_X11, 0, 0},  // S_CENTAUR_DEATH_X10
    {SPR_CTXD, 10, -1, NULL, S_NULL, 0, 0},     // S_CENTAUR_DEATH_X11
    {SPR_CENT, 20, 5, A_FreezeDeath, S_CENTAUR_ICE2, 0, 0},     // S_CENTAUR_ICE
    {SPR_CENT, 20, 1, A_FreezeDeathChunks, S_CENTAUR_ICE2, 0, 0},       // S_CENTAUR_ICE2
    {SPR_CTFX, 32768, -1, NULL, S_NULL, 0, 0},  // S_CENTAUR_FX1
    {SPR_CTFX, 32769, 4, NULL, S_CENTAUR_FX_X2, 0, 0},  // S_CENTAUR_FX_X1
    {SPR_CTFX, 32770, 3, NULL, S_CENTAUR_FX_X3, 0, 0},  // S_CENTAUR_FX_X2
    {SPR_CTFX, 32771, 4, NULL, S_CENTAUR_FX_X4, 0, 0},  // S_CENTAUR_FX_X3
    {SPR_CTFX, 32772, 3, NULL, S_CENTAUR_FX_X5, 0, 0},  // S_CENTAUR_FX_X4
    {SPR_CTFX, 32773, 2, NULL, S_NULL, 0, 0},   // S_CENTAUR_FX_X5
    {SPR_CTDP, 0, 3, A_CheckFloor, S_CENTAUR_SHIELD2, 0, 0},    // S_CENTAUR_SHIELD1
    {SPR_CTDP, 1, 3, A_CheckFloor, S_CENTAUR_SHIELD3, 0, 0},    // S_CENTAUR_SHIELD2
    {SPR_CTDP, 2, 3, A_CheckFloor, S_CENTAUR_SHIELD4, 0, 0},    // S_CENTAUR_SHIELD3
    {SPR_CTDP, 3, 3, A_CheckFloor, S_CENTAUR_SHIELD5, 0, 0},    // S_CENTAUR_SHIELD4
    {SPR_CTDP, 4, 3, A_CheckFloor, S_CENTAUR_SHIELD6, 0, 0},    // S_CENTAUR_SHIELD5
    {SPR_CTDP, 5, 3, A_CheckFloor, S_CENTAUR_SHIELD3, 0, 0},    // S_CENTAUR_SHIELD6
    {SPR_CTDP, 6, 4, NULL, S_CENTAUR_SHIELD_X2, 0, 0},  // S_CENTAUR_SHIELD_X1
    {SPR_CTDP, 7, 4, A_QueueCorpse, S_CENTAUR_SHIELD_X3, 0, 0}, // S_CENTAUR_SHIELD_X2
    {SPR_CTDP, 8, 4, NULL, S_CENTAUR_SHIELD_X4, 0, 0},  // S_CENTAUR_SHIELD_X3
    {SPR_CTDP, 9, -1, NULL, S_NULL, 0, 0},      // S_CENTAUR_SHIELD_X4
    {SPR_CTDP, 10, 3, A_CheckFloor, S_CENTAUR_SWORD2, 0, 0},    // S_CENTAUR_SWORD1
    {SPR_CTDP, 11, 3, A_CheckFloor, S_CENTAUR_SWORD3, 0, 0},    // S_CENTAUR_SWORD2
    {SPR_CTDP, 12, 3, A_CheckFloor, S_CENTAUR_SWORD4, 0, 0},    // S_CENTAUR_SWORD3
    {SPR_CTDP, 13, 3, A_CheckFloor, S_CENTAUR_SWORD5, 0, 0},    // S_CENTAUR_SWORD4
    {SPR_CTDP, 14, 3, A_CheckFloor, S_CENTAUR_SWORD6, 0, 0},    // S_CENTAUR_SWORD5
    {SPR_CTDP, 15, 3, A_CheckFloor, S_CENTAUR_SWORD7, 0, 0},    // S_CENTAUR_SWORD6
    {SPR_CTDP, 16, 3, A_CheckFloor, S_CENTAUR_SWORD3, 0, 0},    // S_CENTAUR_SWORD7
    {SPR_CTDP, 17, 4, NULL, S_CENTAUR_SWORD_X2, 0, 0},  // S_CENTAUR_SWORD_X1
    {SPR_CTDP, 18, 4, A_QueueCorpse, S_CENTAUR_SWORD_X3, 0, 0}, // S_CENTAUR_SWORD_X2
    {SPR_CTDP, 19, -1, NULL, S_NULL, 0, 0},     // S_CENTAUR_SWORD_X3
    {SPR_DEMN, 0, 10, A_Look, S_DEMN_LOOK2, 0, 0},      // S_DEMN_LOOK1
    {SPR_DEMN, 0, 10, A_Look, S_DEMN_LOOK1, 0, 0},      // S_DEMN_LOOK2
    {SPR_DEMN, 0, 4, A_Chase, S_DEMN_CHASE2, 0, 0},     // S_DEMN_CHASE1
    {SPR_DEMN, 1, 4, A_Chase, S_DEMN_CHASE3, 0, 0},     // S_DEMN_CHASE2
    {SPR_DEMN, 2, 4, A_Chase, S_DEMN_CHASE4, 0, 0},     // S_DEMN_CHASE3
    {SPR_DEMN, 3, 4, A_Chase, S_DEMN_CHASE1, 0, 0},     // S_DEMN_CHASE4
    {SPR_DEMN, 4, 6, A_FaceTarget, S_DEMN_ATK1_2, 0, 0},        // S_DEMN_ATK1_1
    {SPR_DEMN, 5, 8, A_FaceTarget, S_DEMN_ATK1_3, 0, 0},        // S_DEMN_ATK1_2
    {SPR_DEMN, 6, 6, A_DemonAttack1, S_DEMN_CHASE1, 0, 0},      // S_DEMN_ATK1_3
    {SPR_DEMN, 4, 5, A_FaceTarget, S_DEMN_ATK2_2, 0, 0},        // S_DEMN_ATK2_1
    {SPR_DEMN, 5, 6, A_FaceTarget, S_DEMN_ATK2_3, 0, 0},        // S_DEMN_ATK2_2
    {SPR_DEMN, 6, 5, A_DemonAttack2, S_DEMN_CHASE1, 0, 0},      // S_DEMN_ATK2_3
    {SPR_DEMN, 4, 4, NULL, S_DEMN_PAIN2, 0, 0}, // S_DEMN_PAIN1
    {SPR_DEMN, 4, 4, A_Pain, S_DEMN_CHASE1, 0, 0},      // S_DEMN_PAIN2
    {SPR_DEMN, 7, 6, NULL, S_DEMN_DEATH2, 0, 0},        // S_DEMN_DEATH1
    {SPR_DEMN, 8, 6, NULL, S_DEMN_DEATH3, 0, 0},        // S_DEMN_DEATH2
    {SPR_DEMN, 9, 6, A_Scream, S_DEMN_DEATH4, 0, 0},    // S_DEMN_DEATH3
    {SPR_DEMN, 10, 6, A_NoBlocking, S_DEMN_DEATH5, 0, 0},       // S_DEMN_DEATH4
    {SPR_DEMN, 11, 6, A_QueueCorpse, S_DEMN_DEATH6, 0, 0},      // S_DEMN_DEATH5
    {SPR_DEMN, 12, 6, NULL, S_DEMN_DEATH7, 0, 0},       // S_DEMN_DEATH6
    {SPR_DEMN, 13, 6, NULL, S_DEMN_DEATH8, 0, 0},       // S_DEMN_DEATH7
    {SPR_DEMN, 14, 6, NULL, S_DEMN_DEATH9, 0, 0},       // S_DEMN_DEATH8
    {SPR_DEMN, 15, -1, NULL, S_NULL, 0, 0},     // S_DEMN_DEATH9
    {SPR_DEMN, 7, 6, NULL, S_DEMN_XDEATH2, 0, 0},       // S_DEMN_XDEATH1
    {SPR_DEMN, 8, 6, A_DemonDeath, S_DEMN_XDEATH3, 0, 0},       // S_DEMN_XDEATH2
    {SPR_DEMN, 9, 6, A_Scream, S_DEMN_XDEATH4, 0, 0},   // S_DEMN_XDEATH3
    {SPR_DEMN, 10, 6, A_NoBlocking, S_DEMN_XDEATH5, 0, 0},      // S_DEMN_XDEATH4
    {SPR_DEMN, 11, 6, A_QueueCorpse, S_DEMN_XDEATH6, 0, 0},     // S_DEMN_XDEATH5
    {SPR_DEMN, 12, 6, NULL, S_DEMN_XDEATH7, 0, 0},      // S_DEMN_XDEATH6
    {SPR_DEMN, 13, 6, NULL, S_DEMN_XDEATH8, 0, 0},      // S_DEMN_XDEATH7
    {SPR_DEMN, 14, 6, NULL, S_DEMN_XDEATH9, 0, 0},      // S_DEMN_XDEATH8
    {SPR_DEMN, 15, -1, NULL, S_NULL, 0, 0},     // S_DEMN_XDEATH9
    {SPR_DEMN, 16, 5, A_FreezeDeath, S_DEMON_ICE2, 0, 0},       // S_DEMON_ICE
    {SPR_DEMN, 16, 1, A_FreezeDeathChunks, S_DEMON_ICE2, 0, 0}, // S_DEMON_ICE2
    {SPR_DEMA, 0, 4, NULL, S_DEMONCHUNK1_2, 0, 0},      // S_DEMONCHUNK1_1
    {SPR_DEMA, 0, 10, A_QueueCorpse, S_DEMONCHUNK1_3, 0, 0},    // S_DEMONCHUNK1_2
    {SPR_DEMA, 0, 20, NULL, S_DEMONCHUNK1_3, 0, 0},     // S_DEMONCHUNK1_3
    {SPR_DEMA, 0, -1, NULL, S_NULL, 0, 0},      // S_DEMONCHUNK1_4
    {SPR_DEMB, 0, 4, NULL, S_DEMONCHUNK2_2, 0, 0},      // S_DEMONCHUNK2_1
    {SPR_DEMB, 0, 10, A_QueueCorpse, S_DEMONCHUNK2_3, 0, 0},    // S_DEMONCHUNK2_2
    {SPR_DEMB, 0, 20, NULL, S_DEMONCHUNK2_3, 0, 0},     // S_DEMONCHUNK2_3
    {SPR_DEMB, 0, -1, NULL, S_NULL, 0, 0},      // S_DEMONCHUNK2_4
    {SPR_DEMC, 0, 4, NULL, S_DEMONCHUNK3_2, 0, 0},      // S_DEMONCHUNK3_1
    {SPR_DEMC, 0, 10, A_QueueCorpse, S_DEMONCHUNK3_3, 0, 0},    // S_DEMONCHUNK3_2
    {SPR_DEMC, 0, 20, NULL, S_DEMONCHUNK3_3, 0, 0},     // S_DEMONCHUNK3_3
    {SPR_DEMC, 0, -1, NULL, S_NULL, 0, 0},      // S_DEMONCHUNK3_4
    {SPR_DEMD, 0, 4, NULL, S_DEMONCHUNK4_2, 0, 0},      // S_DEMONCHUNK4_1
    {SPR_DEMD, 0, 10, A_QueueCorpse, S_DEMONCHUNK4_3, 0, 0},    // S_DEMONCHUNK4_2
    {SPR_DEMD, 0, 20, NULL, S_DEMONCHUNK4_3, 0, 0},     // S_DEMONCHUNK4_3
    {SPR_DEMD, 0, -1, NULL, S_NULL, 0, 0},      // S_DEMONCHUNK4_4
    {SPR_DEME, 0, 4, NULL, S_DEMONCHUNK5_2, 0, 0},      // S_DEMONCHUNK5_1
    {SPR_DEME, 0, 10, A_QueueCorpse, S_DEMONCHUNK5_3, 0, 0},    // S_DEMONCHUNK5_2
    {SPR_DEME, 0, 20, NULL, S_DEMONCHUNK5_3, 0, 0},     // S_DEMONCHUNK5_3
    {SPR_DEME, 0, -1, NULL, S_NULL, 0, 0},      // S_DEMONCHUNK5_4
    {SPR_DMFX, 32768, 4, NULL, S_DEMONFX_MOVE2, 0, 0},  // S_DEMONFX_MOVE1
    {SPR_DMFX, 32769, 4, NULL, S_DEMONFX_MOVE3, 0, 0},  // S_DEMONFX_MOVE2
    {SPR_DMFX, 32770, 4, NULL, S_DEMONFX_MOVE1, 0, 0},  // S_DEMONFX_MOVE3
    {SPR_DMFX, 32771, 4, NULL, S_DEMONFX_BOOM2, 0, 0},  // S_DEMONFX_BOOM1
    {SPR_DMFX, 32772, 4, NULL, S_DEMONFX_BOOM3, 0, 0},  // S_DEMONFX_BOOM2
    {SPR_DMFX, 32773, 3, NULL, S_DEMONFX_BOOM4, 0, 0},  // S_DEMONFX_BOOM3
    {SPR_DMFX, 32774, 3, NULL, S_DEMONFX_BOOM5, 0, 0},  // S_DEMONFX_BOOM4
    {SPR_DMFX, 32775, 3, NULL, S_NULL, 0, 0},   // S_DEMONFX_BOOM5
    {SPR_DEM2, 0, 10, A_Look, S_DEMN2_LOOK2, 0, 0},     // S_DEMN2_LOOK1
    {SPR_DEM2, 0, 10, A_Look, S_DEMN2_LOOK1, 0, 0},     // S_DEMN2_LOOK2
    {SPR_DEM2, 0, 4, A_Chase, S_DEMN2_CHASE2, 0, 0},    // S_DEMN2_CHASE1
    {SPR_DEM2, 1, 4, A_Chase, S_DEMN2_CHASE3, 0, 0},    // S_DEMN2_CHASE2
    {SPR_DEM2, 2, 4, A_Chase, S_DEMN2_CHASE4, 0, 0},    // S_DEMN2_CHASE3
    {SPR_DEM2, 3, 4, A_Chase, S_DEMN2_CHASE1, 0, 0},    // S_DEMN2_CHASE4
    {SPR_DEM2, 4, 6, A_FaceTarget, S_DEMN2_ATK1_2, 0, 0},       // S_DEMN2_ATK1_1
    {SPR_DEM2, 5, 8, A_FaceTarget, S_DEMN2_ATK1_3, 0, 0},       // S_DEMN2_ATK1_2
    {SPR_DEM2, 6, 6, A_DemonAttack1, S_DEMN2_CHASE1, 0, 0},     // S_DEMN2_ATK1_3
    {SPR_DEM2, 4, 5, A_FaceTarget, S_DEMN2_ATK2_2, 0, 0},       // S_DEMN2_ATK2_1
    {SPR_DEM2, 5, 6, A_FaceTarget, S_DEMN2_ATK2_3, 0, 0},       // S_DEMN2_ATK2_2
    {SPR_DEM2, 6, 5, A_DemonAttack2, S_DEMN2_CHASE1, 0, 0},     // S_DEMN2_ATK2_3
    {SPR_DEM2, 4, 4, NULL, S_DEMN2_PAIN2, 0, 0},        // S_DEMN2_PAIN1
    {SPR_DEM2, 4, 4, A_Pain, S_DEMN2_CHASE1, 0, 0},     // S_DEMN2_PAIN2
    {SPR_DEM2, 7, 6, NULL, S_DEMN2_DEATH2, 0, 0},       // S_DEMN2_DEATH1
    {SPR_DEM2, 8, 6, NULL, S_DEMN2_DEATH3, 0, 0},       // S_DEMN2_DEATH2
    {SPR_DEM2, 9, 6, A_Scream, S_DEMN2_DEATH4, 0, 0},   // S_DEMN2_DEATH3
    {SPR_DEM2, 10, 6, A_NoBlocking, S_DEMN2_DEATH5, 0, 0},      // S_DEMN2_DEATH4
    {SPR_DEM2, 11, 6, A_QueueCorpse, S_DEMN2_DEATH6, 0, 0},     // S_DEMN2_DEATH5
    {SPR_DEM2, 12, 6, NULL, S_DEMN2_DEATH7, 0, 0},      // S_DEMN2_DEATH6
    {SPR_DEM2, 13, 6, NULL, S_DEMN2_DEATH8, 0, 0},      // S_DEMN2_DEATH7
    {SPR_DEM2, 14, 6, NULL, S_DEMN2_DEATH9, 0, 0},      // S_DEMN2_DEATH8
    {SPR_DEM2, 15, -1, NULL, S_NULL, 0, 0},     // S_DEMN2_DEATH9
    {SPR_DEM2, 7, 6, NULL, S_DEMN2_XDEATH2, 0, 0},      // S_DEMN2_XDEATH1
    {SPR_DEM2, 8, 6, A_Demon2Death, S_DEMN2_XDEATH3, 0, 0},     // S_DEMN2_XDEATH2
    {SPR_DEM2, 9, 6, A_Scream, S_DEMN2_XDEATH4, 0, 0},  // S_DEMN2_XDEATH3
    {SPR_DEM2, 10, 6, A_NoBlocking, S_DEMN2_XDEATH5, 0, 0},     // S_DEMN2_XDEATH4
    {SPR_DEM2, 11, 6, A_QueueCorpse, S_DEMN2_XDEATH6, 0, 0},    // S_DEMN2_XDEATH5
    {SPR_DEM2, 12, 6, NULL, S_DEMN2_XDEATH7, 0, 0},     // S_DEMN2_XDEATH6
    {SPR_DEM2, 13, 6, NULL, S_DEMN2_XDEATH8, 0, 0},     // S_DEMN2_XDEATH7
    {SPR_DEM2, 14, 6, NULL, S_DEMN2_XDEATH9, 0, 0},     // S_DEMN2_XDEATH8
    {SPR_DEM2, 15, -1, NULL, S_NULL, 0, 0},     // S_DEMN2_XDEATH9
    {SPR_DMBA, 0, 4, NULL, S_DEMON2CHUNK1_2, 0, 0},     // S_DEMON2CHUNK1_1
    {SPR_DMBA, 0, 10, A_QueueCorpse, S_DEMON2CHUNK1_3, 0, 0},   // S_DEMON2CHUNK1_2
    {SPR_DMBA, 0, 20, NULL, S_DEMON2CHUNK1_3, 0, 0},    // S_DEMON2CHUNK1_3
    {SPR_DMBA, 0, -1, NULL, S_NULL, 0, 0},      // S_DEMON2CHUNK1_4
    {SPR_DMBB, 0, 4, NULL, S_DEMON2CHUNK2_2, 0, 0},     // S_DEMON2CHUNK2_1
    {SPR_DMBB, 0, 10, A_QueueCorpse, S_DEMON2CHUNK2_3, 0, 0},   // S_DEMON2CHUNK2_2
    {SPR_DMBB, 0, 20, NULL, S_DEMON2CHUNK2_3, 0, 0},    // S_DEMON2CHUNK2_3
    {SPR_DMBB, 0, -1, NULL, S_NULL, 0, 0},      // S_DEMON2CHUNK2_4
    {SPR_DMBC, 0, 4, NULL, S_DEMON2CHUNK3_2, 0, 0},     // S_DEMON2CHUNK3_1
    {SPR_DMBC, 0, 10, A_QueueCorpse, S_DEMON2CHUNK3_3, 0, 0},   // S_DEMON2CHUNK3_2
    {SPR_DMBC, 0, 20, NULL, S_DEMON2CHUNK3_3, 0, 0},    // S_DEMON2CHUNK3_3
    {SPR_DMBC, 0, -1, NULL, S_NULL, 0, 0},      // S_DEMON2CHUNK3_4
    {SPR_DMBD, 0, 4, NULL, S_DEMON2CHUNK4_2, 0, 0},     // S_DEMON2CHUNK4_1
    {SPR_DMBD, 0, 10, A_QueueCorpse, S_DEMON2CHUNK4_3, 0, 0},   // S_DEMON2CHUNK4_2
    {SPR_DMBD, 0, 20, NULL, S_DEMON2CHUNK4_3, 0, 0},    // S_DEMON2CHUNK4_3
    {SPR_DMBD, 0, -1, NULL, S_NULL, 0, 0},      // S_DEMON2CHUNK4_4
    {SPR_DMBE, 0, 4, NULL, S_DEMON2CHUNK5_2, 0, 0},     // S_DEMON2CHUNK5_1
    {SPR_DMBE, 0, 10, NULL, S_DEMON2CHUNK5_3, 0, 0},    // S_DEMON2CHUNK5_2
    {SPR_DMBE, 0, 20, NULL, S_DEMON2CHUNK5_3, 0, 0},    // S_DEMON2CHUNK5_3
    {SPR_DMBE, 0, -1, NULL, S_NULL, 0, 0},      // S_DEMON2CHUNK5_4
    {SPR_D2FX, 32768, 4, NULL, S_DEMON2FX_MOVE2, 0, 0}, // S_DEMON2FX_MOVE1
    {SPR_D2FX, 32769, 4, NULL, S_DEMON2FX_MOVE3, 0, 0}, // S_DEMON2FX_MOVE2
    {SPR_D2FX, 32770, 4, NULL, S_DEMON2FX_MOVE4, 0, 0}, // S_DEMON2FX_MOVE3
    {SPR_D2FX, 32771, 4, NULL, S_DEMON2FX_MOVE5, 0, 0}, // S_DEMON2FX_MOVE4
    {SPR_D2FX, 32772, 4, NULL, S_DEMON2FX_MOVE6, 0, 0}, // S_DEMON2FX_MOVE5
    {SPR_D2FX, 32773, 4, NULL, S_DEMON2FX_MOVE1, 0, 0}, // S_DEMON2FX_MOVE6
    {SPR_D2FX, 32774, 4, NULL, S_DEMON2FX_BOOM2, 0, 0}, // S_DEMON2FX_BOOM1
    {SPR_D2FX, 32775, 4, NULL, S_DEMON2FX_BOOM3, 0, 0}, // S_DEMON2FX_BOOM2
    {SPR_D2FX, 32776, 4, NULL, S_DEMON2FX_BOOM4, 0, 0}, // S_DEMON2FX_BOOM3
    {SPR_D2FX, 32777, 4, NULL, S_DEMON2FX_BOOM5, 0, 0}, // S_DEMON2FX_BOOM4
    {SPR_D2FX, 32778, 3, NULL, S_DEMON2FX_BOOM6, 0, 0}, // S_DEMON2FX_BOOM5
    {SPR_D2FX, 32779, 3, NULL, S_NULL, 0, 0},   // S_DEMON2FX_BOOM6
    {SPR_WRTH, 0, 2, A_WraithRaiseInit, S_WRAITH_RAISE2, 0, 0}, // S_WRAITH_RAISE1
    {SPR_WRTH, 0, 2, A_WraithRaise, S_WRAITH_RAISE3, 0, 0},     // S_WRAITH_RAISE2
    {SPR_WRTH, 0, 2, A_FaceTarget, S_WRAITH_RAISE4, 0, 0},      // S_WRAITH_RAISE3
    {SPR_WRTH, 1, 2, A_WraithRaise, S_WRAITH_RAISE5, 0, 0},     // S_WRAITH_RAISE4
    {SPR_WRTH, 1, 2, A_WraithRaise, S_WRAITH_RAISE2, 0, 0},     // S_WRAITH_RAISE5
    {SPR_WRTH, 0, 10, NULL, S_WRAITH_INIT2, 0, 0},      // S_WRAITH_INIT1
    {SPR_WRTH, 1, 5, A_WraithInit, S_WRAITH_LOOK1, 0, 0},       // S_WRAITH_INIT2
    {SPR_WRTH, 0, 15, A_WraithLook, S_WRAITH_LOOK2, 0, 0},      // S_WRAITH_LOOK1
    {SPR_WRTH, 1, 15, A_WraithLook, S_WRAITH_LOOK1, 0, 0},      // S_WRAITH_LOOK2
    {SPR_WRTH, 0, 4, A_WraithChase, S_WRAITH_CHASE2, 0, 0},     // S_WRAITH_CHASE1
    {SPR_WRTH, 1, 4, A_WraithChase, S_WRAITH_CHASE3, 0, 0},     // S_WRAITH_CHASE2
    {SPR_WRTH, 2, 4, A_WraithChase, S_WRAITH_CHASE4, 0, 0},     // S_WRAITH_CHASE3
    {SPR_WRTH, 3, 4, A_WraithChase, S_WRAITH_CHASE1, 0, 0},     // S_WRAITH_CHASE4
    {SPR_WRTH, 4, 6, A_FaceTarget, S_WRAITH_ATK1_2, 0, 0},      // S_WRAITH_ATK1_1
    {SPR_WRTH, 5, 6, A_WraithFX3, S_WRAITH_ATK1_3, 0, 0},       // S_WRAITH_ATK1_2
    {SPR_WRTH, 6, 6, A_WraithMelee, S_WRAITH_CHASE1, 0, 0},     // S_WRAITH_ATK1_3
    {SPR_WRTH, 4, 6, A_FaceTarget, S_WRAITH_ATK2_2, 0, 0},      // S_WRAITH_ATK2_1
    {SPR_WRTH, 5, 6, NULL, S_WRAITH_ATK2_3, 0, 0},      // S_WRAITH_ATK2_2
    {SPR_WRTH, 6, 6, A_WraithMissile, S_WRAITH_CHASE1, 0, 0},   // S_WRAITH_ATK2_3
    {SPR_WRTH, 0, 2, NULL, S_WRAITH_PAIN2, 0, 0},       // S_WRAITH_PAIN1
    {SPR_WRTH, 7, 6, A_Pain, S_WRAITH_CHASE1, 0, 0},    // S_WRAITH_PAIN2
    {SPR_WRTH, 8, 4, NULL, S_WRAITH_DEATH1_2, 0, 0},    // S_WRAITH_DEATH1_1
    {SPR_WRTH, 9, 4, A_Scream, S_WRAITH_DEATH1_3, 0, 0},        // S_WRAITH_DEATH1_2
    {SPR_WRTH, 10, 4, NULL, S_WRAITH_DEATH1_4, 0, 0},   // S_WRAITH_DEATH1_3
    {SPR_WRTH, 11, 4, NULL, S_WRAITH_DEATH1_5, 0, 0},   // S_WRAITH_DEATH1_4
    {SPR_WRTH, 12, 4, A_NoBlocking, S_WRAITH_DEATH1_6, 0, 0},   // S_WRAITH_DEATH1_5
    {SPR_WRTH, 13, 4, A_QueueCorpse, S_WRAITH_DEATH1_7, 0, 0},  // S_WRAITH_DEATH1_6
    {SPR_WRTH, 14, 4, NULL, S_WRAITH_DEATH1_8, 0, 0},   // S_WRAITH_DEATH1_7
    {SPR_WRTH, 15, 5, NULL, S_WRAITH_DEATH1_9, 0, 0},   // S_WRAITH_DEATH1_8
    {SPR_WRTH, 16, 5, NULL, S_WRAITH_DEATH1_0, 0, 0},   // S_WRAITH_DEATH1_9
    {SPR_WRTH, 17, -1, NULL, S_NULL, 0, 0},     // S_WRAITH_DEATH1_0
    {SPR_WRT2, 0, 5, NULL, S_WRAITH_DEATH2_2, 0, 0},    // S_WRAITH_DEATH2_1
    {SPR_WRT2, 1, 5, A_Scream, S_WRAITH_DEATH2_3, 0, 0},        // S_WRAITH_DEATH2_2
    {SPR_WRT2, 2, 5, NULL, S_WRAITH_DEATH2_4, 0, 0},    // S_WRAITH_DEATH2_3
    {SPR_WRT2, 3, 5, NULL, S_WRAITH_DEATH2_5, 0, 0},    // S_WRAITH_DEATH2_4
    {SPR_WRT2, 4, 5, A_NoBlocking, S_WRAITH_DEATH2_6, 0, 0},    // S_WRAITH_DEATH2_5
    {SPR_WRT2, 5, 5, A_QueueCorpse, S_WRAITH_DEATH2_7, 0, 0},   // S_WRAITH_DEATH2_6
    {SPR_WRT2, 6, 5, NULL, S_WRAITH_DEATH2_8, 0, 0},    // S_WRAITH_DEATH2_7
    {SPR_WRT2, 7, -1, NULL, S_NULL, 0, 0},      // S_WRAITH_DEATH2_8
    {SPR_WRT2, 8, 5, A_FreezeDeath, S_WRAITH_ICE2, 0, 0},       // S_WRAITH_ICE
    {SPR_WRT2, 8, 1, A_FreezeDeathChunks, S_WRAITH_ICE2, 0, 0}, // S_WRAITH_ICE2
    {SPR_WRBL, 32768, 3, NULL, S_WRTHFX_MOVE2, 0, 0},   // S_WRTHFX_MOVE1
    {SPR_WRBL, 32769, 3, A_WraithFX2, S_WRTHFX_MOVE3, 0, 0},    // S_WRTHFX_MOVE2
    {SPR_WRBL, 32770, 3, NULL, S_WRTHFX_MOVE1, 0, 0},   // S_WRTHFX_MOVE3
    {SPR_WRBL, 32771, 4, NULL, S_WRTHFX_BOOM2, 0, 0},   // S_WRTHFX_BOOM1
    {SPR_WRBL, 32772, 4, A_WraithFX2, S_WRTHFX_BOOM3, 0, 0},    // S_WRTHFX_BOOM2
    {SPR_WRBL, 32773, 4, NULL, S_WRTHFX_BOOM4, 0, 0},   // S_WRTHFX_BOOM3
    {SPR_WRBL, 32774, 3, A_WraithFX2, S_WRTHFX_BOOM5, 0, 0},    // S_WRTHFX_BOOM4
    {SPR_WRBL, 32775, 3, A_WraithFX2, S_WRTHFX_BOOM6, 0, 0},    // S_WRTHFX_BOOM5
    {SPR_WRBL, 32776, 3, NULL, S_NULL, 0, 0},   // S_WRTHFX_BOOM6
    {SPR_WRBL, 32777, 4, NULL, S_WRTHFX_SIZZLE2, 0, 0}, // S_WRTHFX_SIZZLE1
    {SPR_WRBL, 32778, 4, NULL, S_WRTHFX_SIZZLE3, 0, 0}, // S_WRTHFX_SIZZLE2
    {SPR_WRBL, 32779, 4, NULL, S_WRTHFX_SIZZLE4, 0, 0}, // S_WRTHFX_SIZZLE3
    {SPR_WRBL, 32780, 4, NULL, S_WRTHFX_SIZZLE5, 0, 0}, // S_WRTHFX_SIZZLE4
    {SPR_WRBL, 32781, 4, NULL, S_WRTHFX_SIZZLE6, 0, 0}, // S_WRTHFX_SIZZLE5
    {SPR_WRBL, 32782, 4, NULL, S_WRTHFX_SIZZLE7, 0, 0}, // S_WRTHFX_SIZZLE6
    {SPR_WRBL, 32783, 4, NULL, S_NULL, 0, 0},   // S_WRTHFX_SIZZLE7
    {SPR_WRBL, 32784, 4, NULL, S_WRTHFX_DROP2, 0, 0},   // S_WRTHFX_DROP1
    {SPR_WRBL, 32785, 4, NULL, S_WRTHFX_DROP3, 0, 0},   // S_WRTHFX_DROP2
    {SPR_WRBL, 32786, 4, NULL, S_WRTHFX_DROP1, 0, 0},   // S_WRTHFX_DROP3
    {SPR_WRBL, 32786, 4, NULL, S_NULL, 0, 0},   // S_WRTHFX_DEAD1
    {SPR_WRBL, 19, 4, NULL, S_WRTHFX_ADROP2, 0, 0},     // S_WRTHFX_ADROP1
    {SPR_WRBL, 20, 4, NULL, S_WRTHFX_ADROP3, 0, 0},     // S_WRTHFX_ADROP2
    {SPR_WRBL, 21, 4, NULL, S_WRTHFX_ADROP4, 0, 0},     // S_WRTHFX_ADROP3
    {SPR_WRBL, 22, 4, NULL, S_WRTHFX_ADROP1, 0, 0},     // S_WRTHFX_ADROP4
    {SPR_WRBL, 22, 10, NULL, S_NULL, 0, 0},     // S_WRTHFX_ADEAD1
    {SPR_WRBL, 23, 7, NULL, S_WRTHFX_BDROP2, 0, 0},     // S_WRTHFX_BDROP1
    {SPR_WRBL, 24, 7, NULL, S_WRTHFX_BDROP3, 0, 0},     // S_WRTHFX_BDROP2
    {SPR_WRBL, 25, 7, NULL, S_WRTHFX_BDROP1, 0, 0},     // S_WRTHFX_BDROP3
    {SPR_WRBL, 25, 35, NULL, S_NULL, 0, 0},     // S_WRTHFX_BDEAD1
    {SPR_MNTR, 0, 15, NULL, S_MNTR_SPAWN2, 0, 0},       // S_MNTR_SPAWN1
    {SPR_MNTR, 0, 15, A_MinotaurFade1, S_MNTR_SPAWN3, 0, 0},    // S_MNTR_SPAWN2
    {SPR_MNTR, 0, 3, A_MinotaurFade2, S_MNTR_LOOK1, 0, 0},      // S_MNTR_SPAWN3
    {SPR_MNTR, 0, 10, A_MinotaurLook, S_MNTR_LOOK2, 0, 0},      // S_MNTR_LOOK1
    {SPR_MNTR, 1, 10, A_MinotaurLook, S_MNTR_LOOK1, 0, 0},      // S_MNTR_LOOK2
    {SPR_MNTR, 0, 5, A_MinotaurChase, S_MNTR_WALK2, 0, 0},      // S_MNTR_WALK1
    {SPR_MNTR, 1, 5, A_MinotaurChase, S_MNTR_WALK3, 0, 0},      // S_MNTR_WALK2
    {SPR_MNTR, 2, 5, A_MinotaurChase, S_MNTR_WALK4, 0, 0},      // S_MNTR_WALK3
    {SPR_MNTR, 3, 5, A_MinotaurChase, S_MNTR_WALK1, 0, 0},      // S_MNTR_WALK4
    {SPR_MNTR, 0, 5, A_MinotaurRoam, S_MNTR_ROAM2, 0, 0},       // S_MNTR_ROAM1
    {SPR_MNTR, 1, 5, A_MinotaurRoam, S_MNTR_ROAM3, 0, 0},       // S_MNTR_ROAM2
    {SPR_MNTR, 2, 5, A_MinotaurRoam, S_MNTR_ROAM4, 0, 0},       // S_MNTR_ROAM3
    {SPR_MNTR, 3, 5, A_MinotaurRoam, S_MNTR_ROAM1, 0, 0},       // S_MNTR_ROAM4
    {SPR_MNTR, 6, 10, A_FaceTarget, S_MNTR_ATK1_2, 0, 0},       // S_MNTR_ATK1_1
    {SPR_MNTR, 7, 7, A_FaceTarget, S_MNTR_ATK1_3, 0, 0},        // S_MNTR_ATK1_2
    {SPR_MNTR, 8, 12, A_MinotaurAtk1, S_MNTR_WALK1, 0, 0},      // S_MNTR_ATK1_3
    {SPR_MNTR, 6, 10, A_MinotaurDecide, S_MNTR_ATK2_2, 0, 0},   // S_MNTR_ATK2_1
    {SPR_MNTR, 9, 4, A_FaceTarget, S_MNTR_ATK2_3, 0, 0},        // S_MNTR_ATK2_2
    {SPR_MNTR, 10, 9, A_MinotaurAtk2, S_MNTR_WALK1, 0, 0},      // S_MNTR_ATK2_3
    {SPR_MNTR, 6, 10, A_FaceTarget, S_MNTR_ATK3_2, 0, 0},       // S_MNTR_ATK3_1
    {SPR_MNTR, 7, 7, A_FaceTarget, S_MNTR_ATK3_3, 0, 0},        // S_MNTR_ATK3_2
    {SPR_MNTR, 8, 12, A_MinotaurAtk3, S_MNTR_WALK1, 0, 0},      // S_MNTR_ATK3_3
    {SPR_MNTR, 8, 12, NULL, S_MNTR_ATK3_1, 0, 0},       // S_MNTR_ATK3_4
    {SPR_MNTR, 5, 2, A_MinotaurCharge, S_MNTR_ATK4_1, 0, 0},    // S_MNTR_ATK4_1
    {SPR_MNTR, 4, 3, NULL, S_MNTR_PAIN2, 0, 0}, // S_MNTR_PAIN1
    {SPR_MNTR, 4, 6, A_Pain, S_MNTR_WALK1, 0, 0},       // S_MNTR_PAIN2
    {SPR_MNTR, 4, 6, NULL, S_MNTR_DIE2, 0, 0},  // S_MNTR_DIE1
    {SPR_MNTR, 4, 2, A_Scream, S_MNTR_DIE3, 0, 0},      // S_MNTR_DIE2
    {SPR_MNTR, 4, 5, A_SmokePuffExit, S_MNTR_DIE4, 0, 0},       // S_MNTR_DIE3
    {SPR_MNTR, 4, 5, NULL, S_MNTR_DIE5, 0, 0},  // S_MNTR_DIE4
    {SPR_MNTR, 4, 5, A_NoBlocking, S_MNTR_DIE6, 0, 0},  // S_MNTR_DIE5
    {SPR_MNTR, 4, 5, NULL, S_MNTR_DIE7, 0, 0},  // S_MNTR_DIE6
    {SPR_MNTR, 4, 5, A_MinotaurFade1, S_MNTR_DIE8, 0, 0},       // S_MNTR_DIE7
    {SPR_MNTR, 4, 5, A_MinotaurFade0, S_MNTR_DIE9, 0, 0},       // S_MNTR_DIE8
    {SPR_MNTR, 4, 10, NULL, S_NULL, 0, 0},      // S_MNTR_DIE9
    {SPR_FX12, 32768, 6, NULL, S_MNTRFX1_2, 0, 0},      // S_MNTRFX1_1
    {SPR_FX12, 32769, 6, NULL, S_MNTRFX1_1, 0, 0},      // S_MNTRFX1_2
    {SPR_FX12, 32770, 5, NULL, S_MNTRFXI1_2, 0, 0},     // S_MNTRFXI1_1
    {SPR_FX12, 32771, 5, NULL, S_MNTRFXI1_3, 0, 0},     // S_MNTRFXI1_2
    {SPR_FX12, 32772, 5, NULL, S_MNTRFXI1_4, 0, 0},     // S_MNTRFXI1_3
    {SPR_FX12, 32773, 5, NULL, S_MNTRFXI1_5, 0, 0},     // S_MNTRFXI1_4
    {SPR_FX12, 32774, 5, NULL, S_MNTRFXI1_6, 0, 0},     // S_MNTRFXI1_5
    {SPR_FX12, 32775, 5, NULL, S_NULL, 0, 0},   // S_MNTRFXI1_6
    {SPR_FX13, 0, 2, A_MntrFloorFire, S_MNTRFX2_1, 0, 0},       // S_MNTRFX2_1
    {SPR_FX13, 32776, 4, A_Explode, S_MNTRFXI2_2, 0, 0},        // S_MNTRFXI2_1
    {SPR_FX13, 32777, 4, NULL, S_MNTRFXI2_3, 0, 0},     // S_MNTRFXI2_2
    {SPR_FX13, 32778, 4, NULL, S_MNTRFXI2_4, 0, 0},     // S_MNTRFXI2_3
    {SPR_FX13, 32779, 4, NULL, S_MNTRFXI2_5, 0, 0},     // S_MNTRFXI2_4
    {SPR_FX13, 32780, 4, NULL, S_NULL, 0, 0},   // S_MNTRFXI2_5
    {SPR_FX13, 32771, 4, NULL, S_MNTRFX3_2, 0, 0},      // S_MNTRFX3_1
    {SPR_FX13, 32770, 4, NULL, S_MNTRFX3_3, 0, 0},      // S_MNTRFX3_2
    {SPR_FX13, 32769, 5, NULL, S_MNTRFX3_4, 0, 0},      // S_MNTRFX3_3
    {SPR_FX13, 32770, 5, NULL, S_MNTRFX3_5, 0, 0},      // S_MNTRFX3_4
    {SPR_FX13, 32771, 5, NULL, S_MNTRFX3_6, 0, 0},      // S_MNTRFX3_5
    {SPR_FX13, 32772, 5, NULL, S_MNTRFX3_7, 0, 0},      // S_MNTRFX3_6
    {SPR_FX13, 32773, 4, NULL, S_MNTRFX3_8, 0, 0},      // S_MNTRFX3_7
    {SPR_FX13, 32774, 4, NULL, S_MNTRFX3_9, 0, 0},      // S_MNTRFX3_8
    {SPR_FX13, 32775, 4, NULL, S_NULL, 0, 0},   // S_MNTRFX3_9
    {SPR_MNSM, 0, 3, NULL, S_MINOSMOKE2, 0, 0}, // S_MINOSMOKE1
    {SPR_MNSM, 1, 3, NULL, S_MINOSMOKE3, 0, 0}, // S_MINOSMOKE2
    {SPR_MNSM, 2, 3, NULL, S_MINOSMOKE4, 0, 0}, // S_MINOSMOKE3
    {SPR_MNSM, 3, 3, NULL, S_MINOSMOKE5, 0, 0}, // S_MINOSMOKE4
    {SPR_MNSM, 4, 3, NULL, S_MINOSMOKE6, 0, 0}, // S_MINOSMOKE5
    {SPR_MNSM, 5, 3, NULL, S_MINOSMOKE7, 0, 0}, // S_MINOSMOKE6
    {SPR_MNSM, 6, 3, NULL, S_MINOSMOKE8, 0, 0}, // S_MINOSMOKE7
    {SPR_MNSM, 7, 3, NULL, S_MINOSMOKE9, 0, 0}, // S_MINOSMOKE8
    {SPR_MNSM, 8, 3, NULL, S_MINOSMOKE0, 0, 0}, // S_MINOSMOKE9
    {SPR_MNSM, 9, 3, NULL, S_MINOSMOKEA, 0, 0}, // S_MINOSMOKE0
    {SPR_MNSM, 10, 3, NULL, S_MINOSMOKEB, 0, 0},        // S_MINOSMOKEA
    {SPR_MNSM, 11, 3, NULL, S_MINOSMOKEC, 0, 0},        // S_MINOSMOKEB
    {SPR_MNSM, 12, 3, NULL, S_MINOSMOKED, 0, 0},        // S_MINOSMOKEC
    {SPR_MNSM, 13, 3, NULL, S_MINOSMOKEE, 0, 0},        // S_MINOSMOKED
    {SPR_MNSM, 14, 3, NULL, S_MINOSMOKEF, 0, 0},        // S_MINOSMOKEE
    {SPR_MNSM, 15, 3, NULL, S_MINOSMOKEG, 0, 0},        // S_MINOSMOKEF
    {SPR_MNSM, 16, 3, NULL, S_NULL, 0, 0},      // S_MINOSMOKEG
    {SPR_MNSM, 0, 3, NULL, S_MINOSMOKEX2, 0, 0},        // S_MINOSMOKEX1
    {SPR_MNSM, 1, 3, NULL, S_MINOSMOKEX3, 0, 0},        // S_MINOSMOKEX2
    {SPR_MNSM, 2, 3, NULL, S_MINOSMOKEX4, 0, 0},        // S_MINOSMOKEX3
    {SPR_MNSM, 3, 3, NULL, S_MINOSMOKEX5, 0, 0},        // S_MINOSMOKEX4
    {SPR_MNSM, 4, 3, NULL, S_MINOSMOKEX6, 0, 0},        // S_MINOSMOKEX5
    {SPR_MNSM, 5, 3, NULL, S_MINOSMOKEX7, 0, 0},        // S_MINOSMOKEX6
    {SPR_MNSM, 6, 3, NULL, S_MINOSMOKEX8, 0, 0},        // S_MINOSMOKEX7
    {SPR_MNSM, 7, 3, NULL, S_MINOSMOKEX9, 0, 0},        // S_MINOSMOKEX8
    {SPR_MNSM, 8, 3, NULL, S_MINOSMOKEX0, 0, 0},        // S_MINOSMOKEX9
    {SPR_MNSM, 9, 3, NULL, S_MINOSMOKEXA, 0, 0},        // S_MINOSMOKEX0
    {SPR_MNSM, 8, 3, NULL, S_MINOSMOKEXB, 0, 0},        // S_MINOSMOKEXA
    {SPR_MNSM, 7, 3, NULL, S_MINOSMOKEXC, 0, 0},        // S_MINOSMOKEXB
    {SPR_MNSM, 6, 3, NULL, S_MINOSMOKEXD, 0, 0},        // S_MINOSMOKEXC
    {SPR_MNSM, 5, 3, NULL, S_MINOSMOKEXE, 0, 0},        // S_MINOSMOKEXD
    {SPR_MNSM, 4, 3, NULL, S_MINOSMOKEXF, 0, 0},        // S_MINOSMOKEXE
    {SPR_MNSM, 3, 3, NULL, S_MINOSMOKEXG, 0, 0},        // S_MINOSMOKEXF
    {SPR_MNSM, 2, 3, NULL, S_MINOSMOKEXH, 0, 0},        // S_MINOSMOKEXG
    {SPR_MNSM, 1, 3, NULL, S_MINOSMOKEXI, 0, 0},        // S_MINOSMOKEXH
    {SPR_MNSM, 0, 3, NULL, S_NULL, 0, 0},       // S_MINOSMOKEXI
    {SPR_SSPT, 7, 10, A_Look, S_SERPENT_LOOK1, 0, 0},   // S_SERPENT_LOOK1
    {SPR_SSPT, 7, 1, A_SerpentChase, S_SERPENT_SWIM2, 0, 0},    // S_SERPENT_SWIM1
    {SPR_SSPT, 7, 1, A_SerpentChase, S_SERPENT_SWIM3, 0, 0},    // S_SERPENT_SWIM2
    {SPR_SSPT, 7, 2, A_SerpentHumpDecide, S_SERPENT_SWIM1, 0, 0},       // S_SERPENT_SWIM3
    {SPR_SSPT, 7, 3, A_SerpentUnHide, S_SERPENT_HUMP2, 0, 0},   // S_SERPENT_HUMP1
    {SPR_SSPT, 4, 3, A_SerpentRaiseHump, S_SERPENT_HUMP3, 0, 0},        // S_SERPENT_HUMP2
    {SPR_SSPT, 5, 3, A_SerpentRaiseHump, S_SERPENT_HUMP4, 0, 0},        // S_SERPENT_HUMP3
    {SPR_SSPT, 6, 3, A_SerpentRaiseHump, S_SERPENT_HUMP5, 0, 0},        // S_SERPENT_HUMP4
    {SPR_SSPT, 4, 3, A_SerpentRaiseHump, S_SERPENT_HUMP6, 0, 0},        // S_SERPENT_HUMP5
    {SPR_SSPT, 5, 3, A_SerpentRaiseHump, S_SERPENT_HUMP7, 0, 0},        // S_SERPENT_HUMP6
    {SPR_SSPT, 6, 3, NULL, S_SERPENT_HUMP8, 0, 0},      // S_SERPENT_HUMP7
    {SPR_SSPT, 4, 3, NULL, S_SERPENT_HUMP9, 0, 0},      // S_SERPENT_HUMP8
    {SPR_SSPT, 5, 3, NULL, S_SERPENT_HUMP10, 0, 0},     // S_SERPENT_HUMP9
    {SPR_SSPT, 6, 3, A_SerpentLowerHump, S_SERPENT_HUMP11, 0, 0},       // S_SERPENT_HUMP10
    {SPR_SSPT, 4, 3, A_SerpentLowerHump, S_SERPENT_HUMP12, 0, 0},       // S_SERPENT_HUMP11
    {SPR_SSPT, 5, 3, A_SerpentLowerHump, S_SERPENT_HUMP13, 0, 0},       // S_SERPENT_HUMP12
    {SPR_SSPT, 6, 3, A_SerpentLowerHump, S_SERPENT_HUMP14, 0, 0},       // S_SERPENT_HUMP13
    {SPR_SSPT, 4, 3, A_SerpentLowerHump, S_SERPENT_HUMP15, 0, 0},       // S_SERPENT_HUMP14
    {SPR_SSPT, 5, 3, A_SerpentHide, S_SERPENT_SWIM1, 0, 0},     // S_SERPENT_HUMP15
    {SPR_SSPT, 0, 1, A_UnHideThing, S_SERPENT_SURFACE2, 0, 0},  // S_SERPENT_SURFACE1
    {SPR_SSPT, 0, 1, A_SerpentBirthScream, S_SERPENT_SURFACE3, 0, 0},   // S_SERPENT_SURFACE2
    {SPR_SSPT, 1, 3, A_SetShootable, S_SERPENT_SURFACE4, 0, 0}, // S_SERPENT_SURFACE3
    {SPR_SSPT, 2, 3, NULL, S_SERPENT_SURFACE5, 0, 0},   // S_SERPENT_SURFACE4
    {SPR_SSPT, 3, 4, A_SerpentCheckForAttack, S_SERPENT_DIVE1, 0, 0},   // S_SERPENT_SURFACE5
    {SPR_SSDV, 0, 4, NULL, S_SERPENT_DIVE2, 0, 0},      // S_SERPENT_DIVE1
    {SPR_SSDV, 1, 4, NULL, S_SERPENT_DIVE3, 0, 0},      // S_SERPENT_DIVE2
    {SPR_SSDV, 2, 4, NULL, S_SERPENT_DIVE4, 0, 0},      // S_SERPENT_DIVE3
    {SPR_SSDV, 3, 4, A_UnSetShootable, S_SERPENT_DIVE5, 0, 0},  // S_SERPENT_DIVE4
    {SPR_SSDV, 4, 3, A_SerpentDiveSound, S_SERPENT_DIVE6, 0, 0},        // S_SERPENT_DIVE5
    {SPR_SSDV, 5, 3, NULL, S_SERPENT_DIVE7, 0, 0},      // S_SERPENT_DIVE6
    {SPR_SSDV, 6, 4, NULL, S_SERPENT_DIVE8, 0, 0},      // S_SERPENT_DIVE7
    {SPR_SSDV, 7, 4, NULL, S_SERPENT_DIVE9, 0, 0},      // S_SERPENT_DIVE8
    {SPR_SSDV, 8, 3, NULL, S_SERPENT_DIVE10, 0, 0},     // S_SERPENT_DIVE9
    {SPR_SSDV, 9, 3, A_SerpentHide, S_SERPENT_SWIM1, 0, 0},     // S_SERPENT_DIVE10
    {SPR_SSPT, 8, 5, A_SerpentWalk, S_SERPENT_WALK2, 0, 0},     // S_SERPENT_WALK1
    {SPR_SSPT, 9, 5, A_SerpentWalk, S_SERPENT_WALK3, 0, 0},     // S_SERPENT_WALK2
    {SPR_SSPT, 8, 5, A_SerpentWalk, S_SERPENT_WALK4, 0, 0},     // S_SERPENT_WALK3
    {SPR_SSPT, 9, 5, A_SerpentCheckForAttack, S_SERPENT_DIVE1, 0, 0},   // S_SERPENT_WALK4
    {SPR_SSPT, 11, 5, NULL, S_SERPENT_PAIN2, 0, 0},     // S_SERPENT_PAIN1
    {SPR_SSPT, 11, 5, A_Pain, S_SERPENT_DIVE1, 0, 0},   // S_SERPENT_PAIN2
    {SPR_SSPT, 10, 6, A_FaceTarget, S_SERPENT_ATK2, 0, 0},      // S_SERPENT_ATK1
    {SPR_SSPT, 11, 5, A_SerpentChooseAttack, S_SERPENT_MELEE1, 0, 0},   // S_SERPENT_ATK2
    {SPR_SSPT, 13, 5, A_SerpentMeleeAttack, S_SERPENT_DIVE1, 0, 0},     // S_SERPENT_MELEE1
    {SPR_SSPT, 13, 5, A_SerpentMissileAttack, S_SERPENT_DIVE1, 0, 0},   // S_SERPENT_MISSILE1
    {SPR_SSPT, 14, 4, NULL, S_SERPENT_DIE2, 0, 0},      // S_SERPENT_DIE1
    {SPR_SSPT, 15, 4, A_Scream, S_SERPENT_DIE3, 0, 0},  // S_SERPENT_DIE2
    {SPR_SSPT, 16, 4, A_NoBlocking, S_SERPENT_DIE4, 0, 0},      // S_SERPENT_DIE3
    {SPR_SSPT, 17, 4, NULL, S_SERPENT_DIE5, 0, 0},      // S_SERPENT_DIE4
    {SPR_SSPT, 18, 4, NULL, S_SERPENT_DIE6, 0, 0},      // S_SERPENT_DIE5
    {SPR_SSPT, 19, 4, NULL, S_SERPENT_DIE7, 0, 0},      // S_SERPENT_DIE6
    {SPR_SSPT, 20, 4, NULL, S_SERPENT_DIE8, 0, 0},      // S_SERPENT_DIE7
    {SPR_SSPT, 21, 4, NULL, S_SERPENT_DIE9, 0, 0},      // S_SERPENT_DIE8
    {SPR_SSPT, 22, 4, NULL, S_SERPENT_DIE10, 0, 0},     // S_SERPENT_DIE9
    {SPR_SSPT, 23, 4, NULL, S_SERPENT_DIE11, 0, 0},     // S_SERPENT_DIE10
    {SPR_SSPT, 24, 4, NULL, S_SERPENT_DIE12, 0, 0},     // S_SERPENT_DIE11
    {SPR_SSPT, 25, 4, NULL, S_NULL, 0, 0},      // S_SERPENT_DIE12
    {SPR_SSXD, 0, 4, NULL, S_SERPENT_XDIE2, 0, 0},      // S_SERPENT_XDIE1
    {SPR_SSXD, 1, 4, A_SerpentHeadPop, S_SERPENT_XDIE3, 0, 0},  // S_SERPENT_XDIE2
    {SPR_SSXD, 2, 4, A_NoBlocking, S_SERPENT_XDIE4, 0, 0},      // S_SERPENT_XDIE3
    {SPR_SSXD, 3, 4, NULL, S_SERPENT_XDIE5, 0, 0},      // S_SERPENT_XDIE4
    {SPR_SSXD, 4, 4, NULL, S_SERPENT_XDIE6, 0, 0},      // S_SERPENT_XDIE5
    {SPR_SSXD, 5, 3, NULL, S_SERPENT_XDIE7, 0, 0},      // S_SERPENT_XDIE6
    {SPR_SSXD, 6, 3, NULL, S_SERPENT_XDIE8, 0, 0},      // S_SERPENT_XDIE7
    {SPR_SSXD, 7, 3, A_SerpentSpawnGibs, S_NULL, 0, 0}, // S_SERPENT_XDIE8
    {SPR_SSPT, 26, 5, A_FreezeDeath, S_SERPENT_ICE2, 0, 0},     // S_SERPENT_ICE
    {SPR_SSPT, 26, 1, A_FreezeDeathChunks, S_SERPENT_ICE2, 0, 0},       // S_SERPENT_ICE2
    {SPR_SSFX, 32768, 3, A_ContMobjSound, S_SERPENT_FX2, 0, 0}, // S_SERPENT_FX1
    {SPR_SSFX, 32769, 3, NULL, S_SERPENT_FX3, 0, 0},    // S_SERPENT_FX2
    {SPR_SSFX, 32768, 3, NULL, S_SERPENT_FX4, 0, 0},    // S_SERPENT_FX3
    {SPR_SSFX, 32769, 3, NULL, S_SERPENT_FX1, 0, 0},    // S_SERPENT_FX4
    {SPR_SSFX, 32770, 4, NULL, S_SERPENT_FX_X2, 0, 0},  // S_SERPENT_FX_X1
    {SPR_SSFX, 32771, 4, NULL, S_SERPENT_FX_X3, 0, 0},  // S_SERPENT_FX_X2
    {SPR_SSFX, 32772, 4, NULL, S_SERPENT_FX_X4, 0, 0},  // S_SERPENT_FX_X3
    {SPR_SSFX, 32773, 4, NULL, S_SERPENT_FX_X5, 0, 0},  // S_SERPENT_FX_X4
    {SPR_SSFX, 32774, 4, NULL, S_SERPENT_FX_X6, 0, 0},  // S_SERPENT_FX_X5
    {SPR_SSFX, 32775, 4, NULL, S_NULL, 0, 0},   // S_SERPENT_FX_X6
    {SPR_SSXD, 8, 4, A_SerpentHeadCheck, S_SERPENT_HEAD2, 0, 0},        // S_SERPENT_HEAD1
    {SPR_SSXD, 9, 4, A_SerpentHeadCheck, S_SERPENT_HEAD3, 0, 0},        // S_SERPENT_HEAD2
    {SPR_SSXD, 10, 4, A_SerpentHeadCheck, S_SERPENT_HEAD4, 0, 0},       // S_SERPENT_HEAD3
    {SPR_SSXD, 11, 4, A_SerpentHeadCheck, S_SERPENT_HEAD5, 0, 0},       // S_SERPENT_HEAD4
    {SPR_SSXD, 12, 4, A_SerpentHeadCheck, S_SERPENT_HEAD6, 0, 0},       // S_SERPENT_HEAD5
    {SPR_SSXD, 13, 4, A_SerpentHeadCheck, S_SERPENT_HEAD7, 0, 0},       // S_SERPENT_HEAD6
    {SPR_SSXD, 14, 4, A_SerpentHeadCheck, S_SERPENT_HEAD8, 0, 0},       // S_SERPENT_HEAD7
    {SPR_SSXD, 15, 4, A_SerpentHeadCheck, S_SERPENT_HEAD1, 0, 0},       // S_SERPENT_HEAD8
    {SPR_SSXD, 18, -1, NULL, S_SERPENT_HEAD_X1, 0, 0},  // S_SERPENT_HEAD_X1
    {SPR_SSXD, 16, 6, NULL, S_SERPENT_GIB1_2, 0, 0},    // S_SERPENT_GIB1_1
    {SPR_SSXD, 16, 6, A_FloatGib, S_SERPENT_GIB1_3, 0, 0},      // S_SERPENT_GIB1_2
    {SPR_SSXD, 16, 8, A_FloatGib, S_SERPENT_GIB1_4, 0, 0},      // S_SERPENT_GIB1_3
    {SPR_SSXD, 16, 8, A_FloatGib, S_SERPENT_GIB1_5, 0, 0},      // S_SERPENT_GIB1_4
    {SPR_SSXD, 16, 12, A_FloatGib, S_SERPENT_GIB1_6, 0, 0},     // S_SERPENT_GIB1_5
    {SPR_SSXD, 16, 12, A_FloatGib, S_SERPENT_GIB1_7, 0, 0},     // S_SERPENT_GIB1_6
    {SPR_SSXD, 16, 232, A_DelayGib, S_SERPENT_GIB1_8, 0, 0},    // S_SERPENT_GIB1_7
    {SPR_SSXD, 16, 12, A_SinkGib, S_SERPENT_GIB1_9, 0, 0},      // S_SERPENT_GIB1_8
    {SPR_SSXD, 16, 12, A_SinkGib, S_SERPENT_GIB1_10, 0, 0},     // S_SERPENT_GIB1_9
    {SPR_SSXD, 16, 8, A_SinkGib, S_SERPENT_GIB1_11, 0, 0},      // S_SERPENT_GIB1_10
    {SPR_SSXD, 16, 8, A_SinkGib, S_SERPENT_GIB1_12, 0, 0},      // S_SERPENT_GIB1_11
    {SPR_SSXD, 16, 8, A_SinkGib, S_NULL, 0, 0}, // S_SERPENT_GIB1_12
    {SPR_SSXD, 17, 6, NULL, S_SERPENT_GIB2_2, 0, 0},    // S_SERPENT_GIB2_1
    {SPR_SSXD, 17, 6, A_FloatGib, S_SERPENT_GIB2_3, 0, 0},      // S_SERPENT_GIB2_2
    {SPR_SSXD, 17, 8, A_FloatGib, S_SERPENT_GIB2_4, 0, 0},      // S_SERPENT_GIB2_3
    {SPR_SSXD, 17, 8, A_FloatGib, S_SERPENT_GIB2_5, 0, 0},      // S_SERPENT_GIB2_4
    {SPR_SSXD, 17, 12, A_FloatGib, S_SERPENT_GIB2_6, 0, 0},     // S_SERPENT_GIB2_5
    {SPR_SSXD, 17, 12, A_FloatGib, S_SERPENT_GIB2_7, 0, 0},     // S_SERPENT_GIB2_6
    {SPR_SSXD, 17, 232, A_DelayGib, S_SERPENT_GIB2_8, 0, 0},    // S_SERPENT_GIB2_7
    {SPR_SSXD, 17, 12, A_SinkGib, S_SERPENT_GIB2_9, 0, 0},      // S_SERPENT_GIB2_8
    {SPR_SSXD, 17, 12, A_SinkGib, S_SERPENT_GIB2_10, 0, 0},     // S_SERPENT_GIB2_9
    {SPR_SSXD, 17, 8, A_SinkGib, S_SERPENT_GIB2_11, 0, 0},      // S_SERPENT_GIB2_10
    {SPR_SSXD, 17, 8, A_SinkGib, S_SERPENT_GIB2_12, 0, 0},      // S_SERPENT_GIB2_11
    {SPR_SSXD, 17, 8, A_SinkGib, S_NULL, 0, 0}, // S_SERPENT_GIB2_12
    {SPR_SSXD, 19, 6, NULL, S_SERPENT_GIB3_2, 0, 0},    // S_SERPENT_GIB3_1
    {SPR_SSXD, 19, 6, A_FloatGib, S_SERPENT_GIB3_3, 0, 0},      // S_SERPENT_GIB3_2
    {SPR_SSXD, 19, 8, A_FloatGib, S_SERPENT_GIB3_4, 0, 0},      // S_SERPENT_GIB3_3
    {SPR_SSXD, 19, 8, A_FloatGib, S_SERPENT_GIB3_5, 0, 0},      // S_SERPENT_GIB3_4
    {SPR_SSXD, 19, 12, A_FloatGib, S_SERPENT_GIB3_6, 0, 0},     // S_SERPENT_GIB3_5
    {SPR_SSXD, 19, 12, A_FloatGib, S_SERPENT_GIB3_7, 0, 0},     // S_SERPENT_GIB3_6
    {SPR_SSXD, 19, 232, A_DelayGib, S_SERPENT_GIB3_8, 0, 0},    // S_SERPENT_GIB3_7
    {SPR_SSXD, 19, 12, A_SinkGib, S_SERPENT_GIB3_9, 0, 0},      // S_SERPENT_GIB3_8
    {SPR_SSXD, 19, 12, A_SinkGib, S_SERPENT_GIB3_10, 0, 0},     // S_SERPENT_GIB3_9
    {SPR_SSXD, 19, 8, A_SinkGib, S_SERPENT_GIB3_11, 0, 0},      // S_SERPENT_GIB3_10
    {SPR_SSXD, 19, 8, A_SinkGib, S_SERPENT_GIB3_12, 0, 0},      // S_SERPENT_GIB3_11
    {SPR_SSXD, 19, 8, A_SinkGib, S_NULL, 0, 0}, // S_SERPENT_GIB3_12
    {SPR_BISH, 0, 10, A_Look, S_BISHOP_LOOK1, 0, 0},    // S_BISHOP_LOOK1
    {SPR_BISH, 0, 1, A_BishopDecide, S_BISHOP_WALK1, 0, 0},     // S_BISHOP_DECIDE
    {SPR_BISH, 0, 2, A_BishopDoBlur, S_BISHOP_BLUR2, 0, 0},     // S_BISHOP_BLUR1
    {SPR_BISH, 0, 4, A_BishopSpawnBlur, S_BISHOP_BLUR2, 0, 0},  // S_BISHOP_BLUR2
    {SPR_BISH, 0, 2, A_Chase, S_BISHOP_WALK2, 0, 0},    // S_BISHOP_WALK1
    {SPR_BISH, 0, 2, A_BishopChase, S_BISHOP_WALK3, 0, 0},      // S_BISHOP_WALK2
    {SPR_BISH, 0, 2, NULL, S_BISHOP_WALK4, 0, 0},       // S_BISHOP_WALK3
    {SPR_BISH, 1, 2, A_BishopChase, S_BISHOP_WALK5, 0, 0},      // S_BISHOP_WALK4
    {SPR_BISH, 1, 2, A_Chase, S_BISHOP_WALK6, 0, 0},    // S_BISHOP_WALK5
    {SPR_BISH, 1, 2, A_BishopChase, S_BISHOP_DECIDE, 0, 0},     // S_BISHOP_WALK6
    {SPR_BISH, 0, 3, A_FaceTarget, S_BISHOP_ATK2, 0, 0},        // S_BISHOP_ATK1
    {SPR_BISH, 32771, 3, A_FaceTarget, S_BISHOP_ATK3, 0, 0},    // S_BISHOP_ATK2
    {SPR_BISH, 32772, 3, A_FaceTarget, S_BISHOP_ATK4, 0, 0},    // S_BISHOP_ATK3
    {SPR_BISH, 32773, 3, A_BishopAttack, S_BISHOP_ATK5, 0, 0},  // S_BISHOP_ATK4
    {SPR_BISH, 32773, 5, A_BishopAttack2, S_BISHOP_ATK5, 0, 0}, // S_BISHOP_ATK5
    {SPR_BISH, 2, 6, A_Pain, S_BISHOP_PAIN2, 0, 0},     // S_BISHOP_PAIN1
    {SPR_BISH, 2, 6, A_BishopPainBlur, S_BISHOP_PAIN3, 0, 0},   // S_BISHOP_PAIN2
    {SPR_BISH, 2, 6, A_BishopPainBlur, S_BISHOP_PAIN4, 0, 0},   // S_BISHOP_PAIN3
    {SPR_BISH, 2, 6, A_BishopPainBlur, S_BISHOP_PAIN5, 0, 0},   // S_BISHOP_PAIN4
    {SPR_BISH, 2, 0, NULL, S_BISHOP_WALK1, 0, 0},       // S_BISHOP_PAIN5
    {SPR_BISH, 6, 6, NULL, S_BISHOP_DEATH2, 0, 0},      // S_BISHOP_DEATH1
    {SPR_BISH, 32775, 6, A_Scream, S_BISHOP_DEATH3, 0, 0},      // S_BISHOP_DEATH2
    {SPR_BISH, 32776, 5, A_NoBlocking, S_BISHOP_DEATH4, 0, 0},  // S_BISHOP_DEATH3
    {SPR_BISH, 32777, 5, A_Explode, S_BISHOP_DEATH5, 0, 0},     // S_BISHOP_DEATH4
    {SPR_BISH, 32778, 5, NULL, S_BISHOP_DEATH6, 0, 0},  // S_BISHOP_DEATH5
    {SPR_BISH, 32779, 4, NULL, S_BISHOP_DEATH7, 0, 0},  // S_BISHOP_DEATH6
    {SPR_BISH, 32780, 4, NULL, S_BISHOP_DEATH8, 0, 0},  // S_BISHOP_DEATH7
    {SPR_BISH, 13, 4, A_BishopPuff, S_BISHOP_DEATH9, 0, 0},     // S_BISHOP_DEATH8
    {SPR_BISH, 14, 4, A_QueueCorpse, S_BISHOP_DEATH10, 0, 0},   // S_BISHOP_DEATH9
    {SPR_BISH, 15, -1, NULL, S_NULL, 0, 0},     // S_BISHOP_DEATH10
    {SPR_BISH, 23, 5, A_FreezeDeath, S_BISHOP_ICE2, 0, 0},      // S_BISHOP_ICE
    {SPR_BISH, 23, 1, A_FreezeDeathChunks, S_BISHOP_ICE2, 0, 0},        // S_BISHOP_ICE2
    {SPR_BISH, 16, 5, NULL, S_BISHOP_PUFF2, 0, 0},      // S_BISHOP_PUFF1
    {SPR_BISH, 17, 5, NULL, S_BISHOP_PUFF3, 0, 0},      // S_BISHOP_PUFF2
    {SPR_BISH, 18, 5, NULL, S_BISHOP_PUFF4, 0, 0},      // S_BISHOP_PUFF3
    {SPR_BISH, 19, 5, NULL, S_BISHOP_PUFF5, 0, 0},      // S_BISHOP_PUFF4
    {SPR_BISH, 20, 6, NULL, S_BISHOP_PUFF6, 0, 0},      // S_BISHOP_PUFF5
    {SPR_BISH, 21, 6, NULL, S_BISHOP_PUFF7, 0, 0},      // S_BISHOP_PUFF6
    {SPR_BISH, 22, 5, NULL, S_NULL, 0, 0},      // S_BISHOP_PUFF7
    {SPR_BISH, 0, 16, NULL, S_BISHOPBLUR2, 0, 0},       // S_BISHOPBLUR1
    {SPR_BISH, 0, 8, A_SetAltShadow, S_NULL, 0, 0},     // S_BISHOPBLUR2
    {SPR_BISH, 2, 8, NULL, S_NULL, 0, 0},       // S_BISHOPPAINBLUR1
    {SPR_BPFX, 32768, 1, A_BishopMissileWeave, S_BISHFX1_2, 0, 0},      // S_BISHFX1_1
    {SPR_BPFX, 32769, 1, A_BishopMissileWeave, S_BISHFX1_3, 0, 0},      // S_BISHFX1_2
    {SPR_BPFX, 32768, 1, A_BishopMissileWeave, S_BISHFX1_4, 0, 0},      // S_BISHFX1_3
    {SPR_BPFX, 32769, 1, A_BishopMissileWeave, S_BISHFX1_5, 0, 0},      // S_BISHFX1_4
    {SPR_BPFX, 32769, 0, A_BishopMissileSeek, S_BISHFX1_1, 0, 0},       // S_BISHFX1_5
    {SPR_BPFX, 32770, 4, NULL, S_BISHFXI1_2, 0, 0},     // S_BISHFXI1_1
    {SPR_BPFX, 32771, 4, NULL, S_BISHFXI1_3, 0, 0},     // S_BISHFXI1_2
    {SPR_BPFX, 32772, 4, NULL, S_BISHFXI1_4, 0, 0},     // S_BISHFXI1_3
    {SPR_BPFX, 32773, 4, NULL, S_BISHFXI1_5, 0, 0},     // S_BISHFXI1_4
    {SPR_BPFX, 32774, 3, NULL, S_BISHFXI1_6, 0, 0},     // S_BISHFXI1_5
    {SPR_BPFX, 32775, 3, NULL, S_NULL, 0, 0},   // S_BISHFXI1_6
    {SPR_DRAG, 3, 10, A_Look, S_DRAGON_LOOK1, 0, 0},    // S_DRAGON_LOOK1
    {SPR_DRAG, 2, 5, NULL, S_DRAGON_INIT2, 0, 0},       // S_DRAGON_INIT
    {SPR_DRAG, 1, 5, NULL, S_DRAGON_INIT3, 0, 0},       // S_DRAGON_INIT2
    {SPR_DRAG, 0, 5, A_DragonInitFlight, S_DRAGON_WALK1, 0, 0}, // S_DRAGON_INIT3
    {SPR_DRAG, 1, 3, A_DragonFlap, S_DRAGON_WALK2, 0, 0},       // S_DRAGON_WALK1
    {SPR_DRAG, 1, 3, A_DragonFlight, S_DRAGON_WALK3, 0, 0},     // S_DRAGON_WALK2
    {SPR_DRAG, 2, 3, A_DragonFlight, S_DRAGON_WALK4, 0, 0},     // S_DRAGON_WALK3
    {SPR_DRAG, 2, 3, A_DragonFlight, S_DRAGON_WALK5, 0, 0},     // S_DRAGON_WALK4
    {SPR_DRAG, 3, 3, A_DragonFlight, S_DRAGON_WALK6, 0, 0},     // S_DRAGON_WALK5
    {SPR_DRAG, 3, 3, A_DragonFlight, S_DRAGON_WALK7, 0, 0},     // S_DRAGON_WALK6
    {SPR_DRAG, 2, 3, A_DragonFlight, S_DRAGON_WALK8, 0, 0},     // S_DRAGON_WALK7
    {SPR_DRAG, 2, 3, A_DragonFlight, S_DRAGON_WALK9, 0, 0},     // S_DRAGON_WALK8
    {SPR_DRAG, 1, 3, A_DragonFlight, S_DRAGON_WALK10, 0, 0},    // S_DRAGON_WALK9
    {SPR_DRAG, 1, 3, A_DragonFlight, S_DRAGON_WALK11, 0, 0},    // S_DRAGON_WALK10
    {SPR_DRAG, 0, 3, A_DragonFlight, S_DRAGON_WALK12, 0, 0},    // S_DRAGON_WALK11
    {SPR_DRAG, 0, 3, A_DragonFlight, S_DRAGON_WALK1, 0, 0},     // S_DRAGON_WALK12
    {SPR_DRAG, 4, 8, A_DragonAttack, S_DRAGON_WALK1, 0, 0},     // S_DRAGON_ATK1
    {SPR_DRAG, 5, 10, A_DragonPain, S_DRAGON_WALK1, 0, 0},      // S_DRAGON_PAIN1
    {SPR_DRAG, 6, 5, A_Scream, S_DRAGON_DEATH2, 0, 0},  // S_DRAGON_DEATH1
    {SPR_DRAG, 7, 4, A_NoBlocking, S_DRAGON_DEATH3, 0, 0},      // S_DRAGON_DEATH2
    {SPR_DRAG, 8, 4, NULL, S_DRAGON_DEATH4, 0, 0},      // S_DRAGON_DEATH3
    {SPR_DRAG, 9, 4, A_DragonCheckCrash, S_DRAGON_DEATH4, 0, 0},        // S_DRAGON_DEATH4
    {SPR_DRAG, 10, 5, NULL, S_DRAGON_CRASH2, 0, 0},     // S_DRAGON_CRASH1
    {SPR_DRAG, 11, 5, NULL, S_DRAGON_CRASH3, 0, 0},     // S_DRAGON_CRASH2
    {SPR_DRAG, 12, -1, NULL, S_NULL, 0, 0},     // S_DRAGON_CRASH3
    {SPR_DRFX, 32768, 4, NULL, S_DRAGON_FX1_2, 0, 0},   // S_DRAGON_FX1_1
    {SPR_DRFX, 32769, 4, NULL, S_DRAGON_FX1_3, 0, 0},   // S_DRAGON_FX1_2
    {SPR_DRFX, 32770, 4, NULL, S_DRAGON_FX1_4, 0, 0},   // S_DRAGON_FX1_3
    {SPR_DRFX, 32771, 4, NULL, S_DRAGON_FX1_5, 0, 0},   // S_DRAGON_FX1_4
    {SPR_DRFX, 32772, 4, NULL, S_DRAGON_FX1_6, 0, 0},   // S_DRAGON_FX1_5
    {SPR_DRFX, 32773, 4, NULL, S_DRAGON_FX1_1, 0, 0},   // S_DRAGON_FX1_6
    {SPR_DRFX, 32774, 4, NULL, S_DRAGON_FX1_X2, 0, 0},  // S_DRAGON_FX1_X1
    {SPR_DRFX, 32775, 4, NULL, S_DRAGON_FX1_X3, 0, 0},  // S_DRAGON_FX1_X2
    {SPR_DRFX, 32776, 4, NULL, S_DRAGON_FX1_X4, 0, 0},  // S_DRAGON_FX1_X3
    {SPR_DRFX, 32777, 4, A_DragonFX2, S_DRAGON_FX1_X5, 0, 0},   // S_DRAGON_FX1_X4
    {SPR_DRFX, 32778, 3, NULL, S_DRAGON_FX1_X6, 0, 0},  // S_DRAGON_FX1_X5
    {SPR_DRFX, 32779, 3, NULL, S_NULL, 0, 0},   // S_DRAGON_FX1_X6
    {SPR_CFCF, 32784, 1, NULL, S_DRAGON_FX2_2, 0, 0},   // S_DRAGON_FX2_1
    {SPR_CFCF, 32784, 4, A_UnHideThing, S_DRAGON_FX2_3, 0, 0},  // S_DRAGON_FX2_2
    {SPR_CFCF, 32785, 3, A_Scream, S_DRAGON_FX2_4, 0, 0},       // S_DRAGON_FX2_3
    {SPR_CFCF, 32786, 4, NULL, S_DRAGON_FX2_5, 0, 0},   // S_DRAGON_FX2_4
    {SPR_CFCF, 32787, 3, A_Explode, S_DRAGON_FX2_6, 0, 0},      // S_DRAGON_FX2_5
    {SPR_CFCF, 32788, 4, NULL, S_DRAGON_FX2_7, 0, 0},   // S_DRAGON_FX2_6
    {SPR_CFCF, 32789, 3, NULL, S_DRAGON_FX2_8, 0, 0},   // S_DRAGON_FX2_7
    {SPR_CFCF, 32790, 4, NULL, S_DRAGON_FX2_9, 0, 0},   // S_DRAGON_FX2_8
    {SPR_CFCF, 32791, 3, NULL, S_DRAGON_FX2_10, 0, 0},  // S_DRAGON_FX2_9
    {SPR_CFCF, 32792, 4, NULL, S_DRAGON_FX2_11, 0, 0},  // S_DRAGON_FX2_10
    {SPR_CFCF, 32793, 3, NULL, S_NULL, 0, 0},   // S_DRAGON_FX2_11
    {SPR_ARM1, 0, -1, NULL, S_NULL, 0, 0},      // S_ARMOR_1
    {SPR_ARM2, 0, -1, NULL, S_NULL, 0, 0},      // S_ARMOR_2
    {SPR_ARM3, 0, -1, NULL, S_NULL, 0, 0},      // S_ARMOR_3
    {SPR_ARM4, 0, -1, NULL, S_NULL, 0, 0},      // S_ARMOR_4
    {SPR_MAN1, 32768, 4, NULL, S_MANA1_2, 0, 0},        // S_MANA1_1
    {SPR_MAN1, 32769, 4, NULL, S_MANA1_3, 0, 0},        // S_MANA1_2
    {SPR_MAN1, 32770, 4, NULL, S_MANA1_4, 0, 0},        // S_MANA1_3
    {SPR_MAN1, 32771, 4, NULL, S_MANA1_5, 0, 0},        // S_MANA1_4
    {SPR_MAN1, 32772, 4, NULL, S_MANA1_6, 0, 0},        // S_MANA1_5
    {SPR_MAN1, 32773, 4, NULL, S_MANA1_7, 0, 0},        // S_MANA1_6
    {SPR_MAN1, 32774, 4, NULL, S_MANA1_8, 0, 0},        // S_MANA1_7
    {SPR_MAN1, 32775, 4, NULL, S_MANA1_9, 0, 0},        // S_MANA1_8
    {SPR_MAN1, 32776, 4, NULL, S_MANA1_1, 0, 0},        // S_MANA1_9
    {SPR_MAN2, 32768, 4, NULL, S_MANA2_2, 0, 0},        // S_MANA2_1
    {SPR_MAN2, 32769, 4, NULL, S_MANA2_3, 0, 0},        // S_MANA2_2
    {SPR_MAN2, 32770, 4, NULL, S_MANA2_4, 0, 0},        // S_MANA2_3
    {SPR_MAN2, 32771, 4, NULL, S_MANA2_5, 0, 0},        // S_MANA2_4
    {SPR_MAN2, 32772, 4, NULL, S_MANA2_6, 0, 0},        // S_MANA2_5
    {SPR_MAN2, 32773, 4, NULL, S_MANA2_7, 0, 0},        // S_MANA2_6
    {SPR_MAN2, 32774, 4, NULL, S_MANA2_8, 0, 0},        // S_MANA2_7
    {SPR_MAN2, 32775, 4, NULL, S_MANA2_9, 0, 0},        // S_MANA2_8
    {SPR_MAN2, 32776, 4, NULL, S_MANA2_10, 0, 0},       // S_MANA2_9
    {SPR_MAN2, 32777, 4, NULL, S_MANA2_11, 0, 0},       // S_MANA2_10
    {SPR_MAN2, 32778, 4, NULL, S_MANA2_12, 0, 0},       // S_MANA2_11
    {SPR_MAN2, 32779, 4, NULL, S_MANA2_13, 0, 0},       // S_MANA2_12
    {SPR_MAN2, 32780, 4, NULL, S_MANA2_14, 0, 0},       // S_MANA2_13
    {SPR_MAN2, 32781, 4, NULL, S_MANA2_15, 0, 0},       // S_MANA2_14
    {SPR_MAN2, 32782, 4, NULL, S_MANA2_16, 0, 0},       // S_MANA2_15
    {SPR_MAN2, 32783, 4, NULL, S_MANA2_1, 0, 0},        // S_MANA2_16
    {SPR_MAN3, 32768, 4, NULL, S_MANA3_2, 0, 0},        // S_MANA3_1
    {SPR_MAN3, 32769, 4, NULL, S_MANA3_3, 0, 0},        // S_MANA3_2
    {SPR_MAN3, 32770, 4, NULL, S_MANA3_4, 0, 0},        // S_MANA3_3
    {SPR_MAN3, 32771, 4, NULL, S_MANA3_5, 0, 0},        // S_MANA3_4
    {SPR_MAN3, 32772, 4, NULL, S_MANA3_6, 0, 0},        // S_MANA3_5
    {SPR_MAN3, 32773, 4, NULL, S_MANA3_7, 0, 0},        // S_MANA3_6
    {SPR_MAN3, 32774, 4, NULL, S_MANA3_8, 0, 0},        // S_MANA3_7
    {SPR_MAN3, 32775, 4, NULL, S_MANA3_9, 0, 0},        // S_MANA3_8
    {SPR_MAN3, 32776, 4, NULL, S_MANA3_10, 0, 0},       // S_MANA3_9
    {SPR_MAN3, 32777, 4, NULL, S_MANA3_11, 0, 0},       // S_MANA3_10
    {SPR_MAN3, 32778, 4, NULL, S_MANA3_12, 0, 0},       // S_MANA3_11
    {SPR_MAN3, 32779, 4, NULL, S_MANA3_13, 0, 0},       // S_MANA3_12
    {SPR_MAN3, 32780, 4, NULL, S_MANA3_14, 0, 0},       // S_MANA3_13
    {SPR_MAN3, 32781, 4, NULL, S_MANA3_15, 0, 0},       // S_MANA3_14
    {SPR_MAN3, 32782, 4, NULL, S_MANA3_16, 0, 0},       // S_MANA3_15
    {SPR_MAN3, 32783, 4, NULL, S_MANA3_1, 0, 0},        // S_MANA3_16
    {SPR_KEY1, 0, -1, NULL, S_NULL, 0, 0},      // S_KEY1
    {SPR_KEY2, 0, -1, NULL, S_NULL, 0, 0},      // S_KEY2
    {SPR_KEY3, 0, -1, NULL, S_NULL, 0, 0},      // S_KEY3
    {SPR_KEY4, 0, -1, NULL, S_NULL, 0, 0},      // S_KEY4
    {SPR_KEY5, 0, -1, NULL, S_NULL, 0, 0},      // S_KEY5
    {SPR_KEY6, 0, -1, NULL, S_NULL, 0, 0},      // S_KEY6
    {SPR_KEY7, 0, -1, NULL, S_NULL, 0, 0},      // S_KEY7
    {SPR_KEY8, 0, -1, NULL, S_NULL, 0, 0},      // S_KEY8
    {SPR_KEY9, 0, -1, NULL, S_NULL, 0, 0},      // S_KEY9
    {SPR_KEYA, 0, -1, NULL, S_NULL, 0, 0},      // S_KEYA
    {SPR_KEYB, 0, -1, NULL, S_NULL, 0, 0},      // S_KEYB
    {SPR_TLGL, 0, 1, NULL, S_SND_WIND2, 0, 0},  // S_SND_WIND1
    {SPR_TLGL, 0, 200, A_ESound, S_SND_WIND2, 0, 0},    // S_SND_WIND2
    {SPR_TLGL, 0, 85, A_ESound, S_SND_WATERFALL, 0, 0}, // S_SND_WATERFALL
    {SPR_ETTN, 0, 10, A_Look, S_ETTIN_LOOK2, 0, 0},     // S_ETTIN_LOOK1
    {SPR_ETTN, 0, 10, A_Look, S_ETTIN_LOOK1, 0, 0},     // S_ETTIN_LOOK2
    {SPR_ETTN, 0, 5, A_Chase, S_ETTIN_CHASE2, 0, 0},    // S_ETTIN_CHASE1
    {SPR_ETTN, 1, 5, A_Chase, S_ETTIN_CHASE3, 0, 0},    // S_ETTIN_CHASE2
    {SPR_ETTN, 2, 5, A_Chase, S_ETTIN_CHASE4, 0, 0},    // S_ETTIN_CHASE3
    {SPR_ETTN, 3, 5, A_Chase, S_ETTIN_CHASE1, 0, 0},    // S_ETTIN_CHASE4
    {SPR_ETTN, 7, 7, A_Pain, S_ETTIN_CHASE1, 0, 0},     // S_ETTIN_PAIN1
    {SPR_ETTN, 4, 6, A_FaceTarget, S_ETTIN_ATK1_2, 0, 0},       // S_ETTIN_ATK1_1
    {SPR_ETTN, 5, 6, A_FaceTarget, S_ETTIN_ATK1_3, 0, 0},       // S_ETTIN_ATK1_2
    {SPR_ETTN, 6, 8, A_EttinAttack, S_ETTIN_CHASE1, 0, 0},      // S_ETTIN_ATK1_3
    {SPR_ETTN, 8, 4, NULL, S_ETTIN_DEATH1_2, 0, 0},     // S_ETTIN_DEATH1_1
    {SPR_ETTN, 9, 4, NULL, S_ETTIN_DEATH1_3, 0, 0},     // S_ETTIN_DEATH1_2
    {SPR_ETTN, 10, 4, A_Scream, S_ETTIN_DEATH1_4, 0, 0},        // S_ETTIN_DEATH1_3
    {SPR_ETTN, 11, 4, A_NoBlocking, S_ETTIN_DEATH1_5, 0, 0},    // S_ETTIN_DEATH1_4
    {SPR_ETTN, 12, 4, A_QueueCorpse, S_ETTIN_DEATH1_6, 0, 0},   // S_ETTIN_DEATH1_5
    {SPR_ETTN, 13, 4, NULL, S_ETTIN_DEATH1_7, 0, 0},    // S_ETTIN_DEATH1_6
    {SPR_ETTN, 14, 4, NULL, S_ETTIN_DEATH1_8, 0, 0},    // S_ETTIN_DEATH1_7
    {SPR_ETTN, 15, 4, NULL, S_ETTIN_DEATH1_9, 0, 0},    // S_ETTIN_DEATH1_8
    {SPR_ETTN, 16, -1, NULL, S_NULL, 0, 0},     // S_ETTIN_DEATH1_9
    {SPR_ETTB, 0, 4, NULL, S_ETTIN_DEATH2_2, 0, 0},     // S_ETTIN_DEATH2_1
    {SPR_ETTB, 1, 4, A_NoBlocking, S_ETTIN_DEATH2_3, 0, 0},     // S_ETTIN_DEATH2_2
    {SPR_ETTB, 2, 4, A_DropMace, S_ETTIN_DEATH2_4, 0, 0},       // S_ETTIN_DEATH2_3
    {SPR_ETTB, 3, 4, A_Scream, S_ETTIN_DEATH2_5, 0, 0}, // S_ETTIN_DEATH2_4
    {SPR_ETTB, 4, 4, A_QueueCorpse, S_ETTIN_DEATH2_6, 0, 0},    // S_ETTIN_DEATH2_5
    {SPR_ETTB, 5, 4, NULL, S_ETTIN_DEATH2_7, 0, 0},     // S_ETTIN_DEATH2_6
    {SPR_ETTB, 6, 4, NULL, S_ETTIN_DEATH2_8, 0, 0},     // S_ETTIN_DEATH2_7
    {SPR_ETTB, 7, 4, NULL, S_ETTIN_DEATH2_9, 0, 0},     // S_ETTIN_DEATH2_8
    {SPR_ETTB, 8, 4, NULL, S_ETTIN_DEATH2_0, 0, 0},     // S_ETTIN_DEATH2_9
    {SPR_ETTB, 9, 4, NULL, S_ETTIN_DEATH2_A, 0, 0},     // S_ETTIN_DEATH2_0
    {SPR_ETTB, 10, 4, NULL, S_ETTIN_DEATH2_B, 0, 0},    // S_ETTIN_DEATH2_A
    {SPR_ETTB, 11, -1, NULL, S_NULL, 0, 0},     // S_ETTIN_DEATH2_B
    {SPR_ETTN, 17, 5, A_FreezeDeath, S_ETTIN_ICE2, 0, 0},       // S_ETTIN_ICE1
    {SPR_ETTN, 17, 1, A_FreezeDeathChunks, S_ETTIN_ICE2, 0, 0}, // S_ETTIN_ICE2
    {SPR_ETTB, 12, 5, A_CheckFloor, S_ETTIN_MACE2, 0, 0},       // S_ETTIN_MACE1
    {SPR_ETTB, 13, 5, A_CheckFloor, S_ETTIN_MACE3, 0, 0},       // S_ETTIN_MACE2
    {SPR_ETTB, 14, 5, A_CheckFloor, S_ETTIN_MACE4, 0, 0},       // S_ETTIN_MACE3
    {SPR_ETTB, 15, 5, A_CheckFloor, S_ETTIN_MACE1, 0, 0},       // S_ETTIN_MACE4
    {SPR_ETTB, 16, 5, NULL, S_ETTIN_MACE6, 0, 0},       // S_ETTIN_MACE5
    {SPR_ETTB, 17, 5, A_QueueCorpse, S_ETTIN_MACE7, 0, 0},      // S_ETTIN_MACE6
    {SPR_ETTB, 18, -1, NULL, S_NULL, 0, 0},     // S_ETTIN_MACE7
    {SPR_FDMN, 32791, 5, NULL, S_FIRED_LOOK1, 0, 0},    // S_FIRED_SPAWN1
    {SPR_FDMN, 32772, 10, A_Look, S_FIRED_LOOK2, 0, 0}, // S_FIRED_LOOK1
    {SPR_FDMN, 32773, 10, A_Look, S_FIRED_LOOK3, 0, 0}, // S_FIRED_LOOK2
    {SPR_FDMN, 32774, 10, A_Look, S_FIRED_LOOK1, 0, 0}, // S_FIRED_LOOK3
    {SPR_FDMN, 32772, 8, NULL, S_FIRED_LOOK5, 0, 0},    // S_FIRED_LOOK4
    {SPR_FDMN, 32773, 6, NULL, S_FIRED_LOOK6, 0, 0},    // S_FIRED_LOOK5
    {SPR_FDMN, 32774, 5, NULL, S_FIRED_LOOK7, 0, 0},    // S_FIRED_LOOK6
    {SPR_FDMN, 32773, 8, NULL, S_FIRED_LOOK8, 0, 0},    // S_FIRED_LOOK7
    {SPR_FDMN, 32772, 6, NULL, S_FIRED_LOOK9, 0, 0},    // S_FIRED_LOOK8
    {SPR_FDMN, 32773, 7, A_FiredRocks, S_FIRED_LOOK0, 0, 0},    // S_FIRED_LOOK9
    {SPR_FDMN, 32775, 5, NULL, S_FIRED_LOOKA, 0, 0},    // S_FIRED_LOOK0
    {SPR_FDMN, 32776, 5, NULL, S_FIRED_LOOKB, 0, 0},    // S_FIRED_LOOKA
    {SPR_FDMN, 32777, 5, A_UnSetInvulnerable, S_FIRED_WALK1, 0, 0},     // S_FIRED_LOOKB
    {SPR_FDMN, 32768, 5, A_FiredChase, S_FIRED_WALK2, 0, 0},    // S_FIRED_WALK1
    {SPR_FDMN, 32769, 5, A_FiredChase, S_FIRED_WALK3, 0, 0},    // S_FIRED_WALK2
    {SPR_FDMN, 32770, 5, A_FiredChase, S_FIRED_WALK1, 0, 0},    // S_FIRED_WALK3
    {SPR_FDMN, 32771, 6, A_Pain, S_FIRED_WALK1, 0, 0},  // S_FIRED_PAIN1
    {SPR_FDMN, 32778, 3, A_FaceTarget, S_FIRED_ATTACK2, 0, 0},  // S_FIRED_ATTACK1
    {SPR_FDMN, 32778, 5, A_FiredAttack, S_FIRED_ATTACK3, 0, 0}, // S_FIRED_ATTACK2
    {SPR_FDMN, 32778, 5, A_FiredAttack, S_FIRED_ATTACK4, 0, 0}, // S_FIRED_ATTACK3
    {SPR_FDMN, 32778, 5, A_FiredAttack, S_FIRED_WALK1, 0, 0},   // S_FIRED_ATTACK4
    {SPR_FDMN, 32771, 4, A_FaceTarget, S_FIRED_DEATH2, 0, 0},   // S_FIRED_DEATH1
    {SPR_FDMN, 32779, 4, A_Scream, S_FIRED_DEATH3, 0, 0},       // S_FIRED_DEATH2
    {SPR_FDMN, 32779, 4, A_NoBlocking, S_FIRED_DEATH4, 0, 0},   // S_FIRED_DEATH3
    {SPR_FDMN, 32779, 200, NULL, S_NULL, 0, 0}, // S_FIRED_DEATH4
    {SPR_FDMN, 12, 5, A_FaceTarget, S_FIRED_XDEATH2, 0, 0},     // S_FIRED_XDEATH1
    {SPR_FDMN, 13, 5, A_NoBlocking, S_FIRED_XDEATH3, 0, 0},     // S_FIRED_XDEATH2
    {SPR_FDMN, 14, 5, A_FiredSplotch, S_NULL, 0, 0},    // S_FIRED_XDEATH3
    {SPR_FDMN, 17, 5, A_FreezeDeath, S_FIRED_ICE2, 0, 0},       // S_FIRED_ICE1
    {SPR_FDMN, 17, 1, A_FreezeDeathChunks, S_FIRED_ICE2, 0, 0}, // S_FIRED_ICE2
    {SPR_FDMN, 15, 3, NULL, S_FIRED_CORPSE2, 0, 0},     // S_FIRED_CORPSE1
    {SPR_FDMN, 15, 6, A_QueueCorpse, S_FIRED_CORPSE3, 0, 0},    // S_FIRED_CORPSE2
    {SPR_FDMN, 24, -1, NULL, S_NULL, 0, 0},     // S_FIRED_CORPSE3
    {SPR_FDMN, 16, 3, NULL, S_FIRED_CORPSE5, 0, 0},     // S_FIRED_CORPSE4
    {SPR_FDMN, 16, 6, A_QueueCorpse, S_FIRED_CORPSE6, 0, 0},    // S_FIRED_CORPSE5
    {SPR_FDMN, 25, -1, NULL, S_NULL, 0, 0},     // S_FIRED_CORPSE6
    {SPR_FDMN, 18, 4, NULL, S_FIRED_RDROP1, 0, 0},      // S_FIRED_RDROP1
    {SPR_FDMN, 18, 5, A_SmBounce, S_FIRED_RDEAD1_2, 0, 0},      // S_FIRED_RDEAD1_1
    {SPR_FDMN, 18, 200, NULL, S_NULL, 0, 0},    // S_FIRED_RDEAD1_2
    {SPR_FDMN, 19, 4, NULL, S_FIRED_RDROP2, 0, 0},      // S_FIRED_RDROP2
    {SPR_FDMN, 19, 5, A_SmBounce, S_FIRED_RDEAD2_2, 0, 0},      // S_FIRED_RDEAD2_1
    {SPR_FDMN, 19, 200, NULL, S_NULL, 0, 0},    // S_FIRED_RDEAD2_2
    {SPR_FDMN, 20, 4, NULL, S_FIRED_RDROP3, 0, 0},      // S_FIRED_RDROP3
    {SPR_FDMN, 20, 5, A_SmBounce, S_FIRED_RDEAD3_2, 0, 0},      // S_FIRED_RDEAD3_1
    {SPR_FDMN, 20, 200, NULL, S_NULL, 0, 0},    // S_FIRED_RDEAD3_2
    {SPR_FDMN, 21, 4, NULL, S_FIRED_RDROP4, 0, 0},      // S_FIRED_RDROP4
    {SPR_FDMN, 21, 5, A_SmBounce, S_FIRED_RDEAD4_2, 0, 0},      // S_FIRED_RDEAD4_1
    {SPR_FDMN, 21, 200, NULL, S_NULL, 0, 0},    // S_FIRED_RDEAD4_2
    {SPR_FDMN, 22, 4, NULL, S_FIRED_RDROP5, 0, 0},      // S_FIRED_RDROP5
    {SPR_FDMN, 22, 5, A_SmBounce, S_FIRED_RDEAD5_2, 0, 0},      // S_FIRED_RDEAD5_1
    {SPR_FDMN, 22, 200, NULL, S_NULL, 0, 0},    // S_FIRED_RDEAD5_2
    {SPR_FDMB, 32768, 5, NULL, S_FIRED_FX6_1, 0, 0},    // S_FIRED_FX6_1
    {SPR_FDMB, 32769, 5, NULL, S_FIRED_FX6_3, 0, 0},    // S_FIRED_FX6_2
    {SPR_FDMB, 32770, 5, NULL, S_FIRED_FX6_4, 0, 0},    // S_FIRED_FX6_3
    {SPR_FDMB, 32771, 5, NULL, S_FIRED_FX6_5, 0, 0},    // S_FIRED_FX6_4
    {SPR_FDMB, 32772, 5, NULL, S_NULL, 0, 0},   // S_FIRED_FX6_5
    {SPR_ICEY, 0, 10, A_IceGuyLook, S_ICEGUY_LOOK, 0, 0},       // S_ICEGUY_LOOK
    {SPR_ICEY, 0, -1, NULL, S_ICEGUY_LOOK, 0, 0},       // S_ICEGUY_DORMANT
    {SPR_ICEY, 0, 4, A_Chase, S_ICEGUY_WALK2, 0, 0},    // S_ICEGUY_WALK1
    {SPR_ICEY, 1, 4, A_IceGuyChase, S_ICEGUY_WALK3, 0, 0},      // S_ICEGUY_WALK2
    {SPR_ICEY, 2, 4, A_Chase, S_ICEGUY_WALK4, 0, 0},    // S_ICEGUY_WALK3
    {SPR_ICEY, 3, 4, A_Chase, S_ICEGUY_WALK1, 0, 0},    // S_ICEGUY_WALK4
    {SPR_ICEY, 4, 3, A_FaceTarget, S_ICEGUY_ATK2, 0, 0},        // S_ICEGUY_ATK1
    {SPR_ICEY, 5, 3, A_FaceTarget, S_ICEGUY_ATK3, 0, 0},        // S_ICEGUY_ATK2
    {SPR_ICEY, 32774, 8, A_IceGuyAttack, S_ICEGUY_ATK4, 0, 0},  // S_ICEGUY_ATK3
    {SPR_ICEY, 5, 4, A_FaceTarget, S_ICEGUY_WALK1, 0, 0},       // S_ICEGUY_ATK4
    {SPR_ICEY, 0, 1, A_Pain, S_ICEGUY_WALK1, 0, 0},     // S_ICEGUY_PAIN1
    {SPR_ICEY, 0, 1, A_IceGuyDie, S_NULL, 0, 0},        // S_ICEGUY_DEATH
    {SPR_ICPR, 32768, 3, A_IceGuyMissilePuff, S_ICEGUY_FX2, 0, 0},      // S_ICEGUY_FX1
    {SPR_ICPR, 32769, 3, A_IceGuyMissilePuff, S_ICEGUY_FX3, 0, 0},      // S_ICEGUY_FX2
    {SPR_ICPR, 32770, 3, A_IceGuyMissilePuff, S_ICEGUY_FX1, 0, 0},      // S_ICEGUY_FX3
    {SPR_ICPR, 32771, 4, NULL, S_ICEGUY_FX_X2, 0, 0},   // S_ICEGUY_FX_X1
    {SPR_ICPR, 32772, 4, A_IceGuyMissileExplode, S_ICEGUY_FX_X3, 0, 0}, // S_ICEGUY_FX_X2
    {SPR_ICPR, 32773, 4, NULL, S_ICEGUY_FX_X4, 0, 0},   // S_ICEGUY_FX_X3
    {SPR_ICPR, 32774, 4, NULL, S_ICEGUY_FX_X5, 0, 0},   // S_ICEGUY_FX_X4
    {SPR_ICPR, 32775, 3, NULL, S_NULL, 0, 0},   // S_ICEGUY_FX_X5
    {SPR_ICPR, 8, 3, NULL, S_ICEFX_PUFF2, 0, 0},        // S_ICEFX_PUFF1
    {SPR_ICPR, 9, 3, NULL, S_ICEFX_PUFF3, 0, 0},        // S_ICEFX_PUFF2
    {SPR_ICPR, 10, 3, NULL, S_ICEFX_PUFF4, 0, 0},       // S_ICEFX_PUFF3
    {SPR_ICPR, 11, 2, NULL, S_ICEFX_PUFF5, 0, 0},       // S_ICEFX_PUFF4
    {SPR_ICPR, 12, 2, NULL, S_NULL, 0, 0},      // S_ICEFX_PUFF5
    {SPR_ICPR, 32781, 3, NULL, S_ICEGUY_FX2_2, 0, 0},   // S_ICEGUY_FX2_1
    {SPR_ICPR, 32782, 3, NULL, S_ICEGUY_FX2_3, 0, 0},   // S_ICEGUY_FX2_2
    {SPR_ICPR, 32783, 3, NULL, S_ICEGUY_FX2_1, 0, 0},   // S_ICEGUY_FX2_3
    {SPR_ICPR, 32784, 50, NULL, S_NULL, 0, 0},  // S_ICEGUY_BIT1
    {SPR_ICPR, 32785, 50, NULL, S_NULL, 0, 0},  // S_ICEGUY_BIT2
    {SPR_ICWS, 0, 2, NULL, S_ICEGUY_WISP1_2, 0, 0},     // S_ICEGUY_WISP1_1
    {SPR_ICWS, 1, 2, NULL, S_ICEGUY_WISP1_3, 0, 0},     // S_ICEGUY_WISP1_2
    {SPR_ICWS, 2, 2, NULL, S_ICEGUY_WISP1_4, 0, 0},     // S_ICEGUY_WISP1_3
    {SPR_ICWS, 3, 2, NULL, S_ICEGUY_WISP1_5, 0, 0},     // S_ICEGUY_WISP1_4
    {SPR_ICWS, 4, 2, NULL, S_ICEGUY_WISP1_6, 0, 0},     // S_ICEGUY_WISP1_5
    {SPR_ICWS, 5, 2, NULL, S_ICEGUY_WISP1_7, 0, 0},     // S_ICEGUY_WISP1_6
    {SPR_ICWS, 6, 2, NULL, S_ICEGUY_WISP1_8, 0, 0},     // S_ICEGUY_WISP1_7
    {SPR_ICWS, 7, 2, NULL, S_ICEGUY_WISP1_9, 0, 0},     // S_ICEGUY_WISP1_8
    {SPR_ICWS, 8, 2, NULL, S_NULL, 0, 0},       // S_ICEGUY_WISP1_9
    {SPR_ICWS, 9, 2, NULL, S_ICEGUY_WISP2_2, 0, 0},     // S_ICEGUY_WISP2_1
    {SPR_ICWS, 10, 2, NULL, S_ICEGUY_WISP2_3, 0, 0},    // S_ICEGUY_WISP2_2
    {SPR_ICWS, 11, 2, NULL, S_ICEGUY_WISP2_4, 0, 0},    // S_ICEGUY_WISP2_3
    {SPR_ICWS, 12, 2, NULL, S_ICEGUY_WISP2_5, 0, 0},    // S_ICEGUY_WISP2_4
    {SPR_ICWS, 13, 2, NULL, S_ICEGUY_WISP2_6, 0, 0},    // S_ICEGUY_WISP2_5
    {SPR_ICWS, 14, 2, NULL, S_ICEGUY_WISP2_7, 0, 0},    // S_ICEGUY_WISP2_6
    {SPR_ICWS, 15, 2, NULL, S_ICEGUY_WISP2_8, 0, 0},    // S_ICEGUY_WISP2_7
    {SPR_ICWS, 16, 2, NULL, S_ICEGUY_WISP2_9, 0, 0},    // S_ICEGUY_WISP2_8
    {SPR_ICWS, 17, 2, NULL, S_NULL, 0, 0},      // S_ICEGUY_WISP2_9
    {SPR_PLAY, 0, 2, NULL, S_FIGHTER2, 0, 0},   // S_FIGHTER
    {SPR_PLAY, 0, 3, A_ClassBossHealth, S_FIGHTERLOOK, 0, 0},   // S_FIGHTER2
    {SPR_PLAY, 0, 5, A_Look, S_FIGHTERLOOK, 0, 0},      // S_FIGHTERLOOK
    {SPR_PLAY, 0, 4, A_FastChase, S_FIGHTER_RUN2, 0, 0},        // S_FIGHTER_RUN1
    {SPR_PLAY, 1, 4, A_FastChase, S_FIGHTER_RUN3, 0, 0},        // S_FIGHTER_RUN2
    {SPR_PLAY, 2, 4, A_FastChase, S_FIGHTER_RUN4, 0, 0},        // S_FIGHTER_RUN3
    {SPR_PLAY, 3, 4, A_FastChase, S_FIGHTER_RUN1, 0, 0},        // S_FIGHTER_RUN4
    {SPR_PLAY, 4, 8, A_FaceTarget, S_FIGHTER_ATK2, 0, 0},       // S_FIGHTER_ATK1
    {SPR_PLAY, 5, 8, A_FighterAttack, S_FIGHTER_RUN1, 0, 0},    // S_FIGHTER_ATK2
    {SPR_PLAY, 6, 4, NULL, S_FIGHTER_PAIN2, 0, 0},      // S_FIGHTER_PAIN
    {SPR_PLAY, 6, 4, A_Pain, S_FIGHTER_RUN1, 0, 0},     // S_FIGHTER_PAIN2
    {SPR_PLAY, 7, 6, NULL, S_FIGHTER_DIE2, 0, 0},       // S_FIGHTER_DIE1
    {SPR_PLAY, 8, 6, A_Scream, S_FIGHTER_DIE3, 0, 0},   // S_FIGHTER_DIE2
    {SPR_PLAY, 9, 6, NULL, S_FIGHTER_DIE4, 0, 0},       // S_FIGHTER_DIE3
    {SPR_PLAY, 10, 6, NULL, S_FIGHTER_DIE5, 0, 0},      // S_FIGHTER_DIE4
    {SPR_PLAY, 11, 6, A_NoBlocking, S_FIGHTER_DIE6, 0, 0},      // S_FIGHTER_DIE5
    {SPR_PLAY, 12, 6, NULL, S_FIGHTER_DIE7, 0, 0},      // S_FIGHTER_DIE6
    {SPR_PLAY, 13, -1, NULL, S_NULL, 0, 0},     // S_FIGHTER_DIE7
    {SPR_PLAY, 14, 5, A_Scream, S_FIGHTER_XDIE2, 0, 0}, // S_FIGHTER_XDIE1
    {SPR_PLAY, 15, 5, A_SkullPop, S_FIGHTER_XDIE3, 0, 0},       // S_FIGHTER_XDIE2
    {SPR_PLAY, 17, 5, A_NoBlocking, S_FIGHTER_XDIE4, 0, 0},     // S_FIGHTER_XDIE3
    {SPR_PLAY, 18, 5, NULL, S_FIGHTER_XDIE5, 0, 0},     // S_FIGHTER_XDIE4
    {SPR_PLAY, 19, 5, NULL, S_FIGHTER_XDIE6, 0, 0},     // S_FIGHTER_XDIE5
    {SPR_PLAY, 20, 5, NULL, S_FIGHTER_XDIE7, 0, 0},     // S_FIGHTER_XDIE6
    {SPR_PLAY, 21, 5, NULL, S_FIGHTER_XDIE8, 0, 0},     // S_FIGHTER_XDIE7
    {SPR_PLAY, 22, -1, NULL, S_NULL, 0, 0},     // S_FIGHTER_XDIE8
    {SPR_PLAY, 23, 5, A_FreezeDeath, S_FIGHTER_ICE2, 0, 0},     // S_FIGHTER_ICE
    {SPR_PLAY, 23, 1, A_FreezeDeathChunks, S_FIGHTER_ICE2, 0, 0},       // S_FIGHTER_ICE2
    {SPR_CLER, 0, 2, NULL, S_CLERIC2, 0, 0},    // S_CLERIC
    {SPR_CLER, 0, 3, A_ClassBossHealth, S_CLERICLOOK, 0, 0},    // S_CLERIC2
    {SPR_CLER, 0, 5, A_Look, S_CLERICLOOK, 0, 0},       // S_CLERICLOOK
    {SPR_CLER, 0, 4, A_FastChase, S_CLERIC_RUN2, 0, 0}, // S_CLERIC_RUN1
    {SPR_CLER, 1, 4, A_FastChase, S_CLERIC_RUN3, 0, 0}, // S_CLERIC_RUN2
    {SPR_CLER, 2, 4, A_FastChase, S_CLERIC_RUN4, 0, 0}, // S_CLERIC_RUN3
    {SPR_CLER, 3, 4, A_FastChase, S_CLERIC_RUN1, 0, 0}, // S_CLERIC_RUN4
    {SPR_CLER, 4, 8, A_FaceTarget, S_CLERIC_ATK2, 0, 0},        // S_CLERIC_ATK1
    {SPR_CLER, 5, 8, A_FaceTarget, S_CLERIC_ATK3, 0, 0},        // S_CLERIC_ATK2
    {SPR_CLER, 6, 10, A_ClericAttack, S_CLERIC_RUN1, 0, 0},     // S_CLERIC_ATK3
    {SPR_CLER, 7, 4, NULL, S_CLERIC_PAIN2, 0, 0},       // S_CLERIC_PAIN
    {SPR_CLER, 7, 4, A_Pain, S_CLERIC_RUN1, 0, 0},      // S_CLERIC_PAIN2
    {SPR_CLER, 8, 6, NULL, S_CLERIC_DIE2, 0, 0},        // S_CLERIC_DIE1
    {SPR_CLER, 10, 6, A_Scream, S_CLERIC_DIE3, 0, 0},   // S_CLERIC_DIE2
    {SPR_CLER, 11, 6, NULL, S_CLERIC_DIE4, 0, 0},       // S_CLERIC_DIE3
    {SPR_CLER, 11, 6, NULL, S_CLERIC_DIE5, 0, 0},       // S_CLERIC_DIE4
    {SPR_CLER, 12, 6, A_NoBlocking, S_CLERIC_DIE6, 0, 0},       // S_CLERIC_DIE5
    {SPR_CLER, 13, 6, NULL, S_CLERIC_DIE7, 0, 0},       // S_CLERIC_DIE6
    {SPR_CLER, 14, 6, NULL, S_CLERIC_DIE8, 0, 0},       // S_CLERIC_DIE7
    {SPR_CLER, 15, 6, NULL, S_CLERIC_DIE9, 0, 0},       // S_CLERIC_DIE8
    {SPR_CLER, 16, -1, NULL, S_NULL, 0, 0},     // S_CLERIC_DIE9
    {SPR_CLER, 17, 5, A_Scream, S_CLERIC_XDIE2, 0, 0},  // S_CLERIC_XDIE1
    {SPR_CLER, 18, 5, NULL, S_CLERIC_XDIE3, 0, 0},      // S_CLERIC_XDIE2
    {SPR_CLER, 19, 5, A_NoBlocking, S_CLERIC_XDIE4, 0, 0},      // S_CLERIC_XDIE3
    {SPR_CLER, 20, 5, NULL, S_CLERIC_XDIE5, 0, 0},      // S_CLERIC_XDIE4
    {SPR_CLER, 21, 5, NULL, S_CLERIC_XDIE6, 0, 0},      // S_CLERIC_XDIE5
    {SPR_CLER, 22, 5, NULL, S_CLERIC_XDIE7, 0, 0},      // S_CLERIC_XDIE6
    {SPR_CLER, 23, 5, NULL, S_CLERIC_XDIE8, 0, 0},      // S_CLERIC_XDIE7
    {SPR_CLER, 24, 5, NULL, S_CLERIC_XDIE9, 0, 0},      // S_CLERIC_XDIE8
    {SPR_CLER, 25, 5, NULL, S_CLERIC_XDIE10, 0, 0},     // S_CLERIC_XDIE9
    {SPR_CLER, 26, -1, NULL, S_NULL, 0, 0},     // S_CLERIC_XDIE10
    {SPR_CLER, 27, 5, A_FreezeDeath, S_CLERIC_ICE2, 0, 0},      // S_CLERIC_ICE
    {SPR_CLER, 27, 1, A_FreezeDeathChunks, S_CLERIC_ICE2, 0, 0},        // S_CLERIC_ICE2
    {SPR_MAGE, 0, 2, NULL, S_MAGE2, 0, 0},      // S_MAGE
    {SPR_MAGE, 0, 3, A_ClassBossHealth, S_MAGELOOK, 0, 0},      // S_MAGE2
    {SPR_MAGE, 0, 5, A_Look, S_MAGELOOK, 0, 0}, // S_MAGELOOK
    {SPR_MAGE, 0, 4, A_FastChase, S_MAGE_RUN2, 0, 0},   // S_MAGE_RUN1
    {SPR_MAGE, 1, 4, A_FastChase, S_MAGE_RUN3, 0, 0},   // S_MAGE_RUN2
    {SPR_MAGE, 2, 4, A_FastChase, S_MAGE_RUN4, 0, 0},   // S_MAGE_RUN3
    {SPR_MAGE, 3, 4, A_FastChase, S_MAGE_RUN1, 0, 0},   // S_MAGE_RUN4
    {SPR_MAGE, 4, 8, A_FaceTarget, S_MAGE_ATK2, 0, 0},  // S_MAGE_ATK1
    {SPR_MAGE, 32773, 8, A_MageAttack, S_MAGE_RUN1, 0, 0},      // S_MAGE_ATK2
    {SPR_MAGE, 6, 4, NULL, S_MAGE_PAIN2, 0, 0}, // S_MAGE_PAIN
    {SPR_MAGE, 6, 4, A_Pain, S_MAGE_RUN1, 0, 0},        // S_MAGE_PAIN2
    {SPR_MAGE, 7, 6, NULL, S_MAGE_DIE2, 0, 0},  // S_MAGE_DIE1
    {SPR_MAGE, 8, 6, A_Scream, S_MAGE_DIE3, 0, 0},      // S_MAGE_DIE2
    {SPR_MAGE, 9, 6, NULL, S_MAGE_DIE4, 0, 0},  // S_MAGE_DIE3
    {SPR_MAGE, 10, 6, NULL, S_MAGE_DIE5, 0, 0}, // S_MAGE_DIE4
    {SPR_MAGE, 11, 6, A_NoBlocking, S_MAGE_DIE6, 0, 0}, // S_MAGE_DIE5
    {SPR_MAGE, 12, 6, NULL, S_MAGE_DIE7, 0, 0}, // S_MAGE_DIE6
    {SPR_MAGE, 13, -1, NULL, S_NULL, 0, 0},     // S_MAGE_DIE7
    {SPR_MAGE, 14, 5, A_Scream, S_MAGE_XDIE2, 0, 0},    // S_MAGE_XDIE1
    {SPR_MAGE, 15, 5, NULL, S_MAGE_XDIE3, 0, 0},        // S_MAGE_XDIE2
    {SPR_MAGE, 17, 5, A_NoBlocking, S_MAGE_XDIE4, 0, 0},        // S_MAGE_XDIE3
    {SPR_MAGE, 18, 5, NULL, S_MAGE_XDIE5, 0, 0},        // S_MAGE_XDIE4
    {SPR_MAGE, 19, 5, NULL, S_MAGE_XDIE6, 0, 0},        // S_MAGE_XDIE5
    {SPR_MAGE, 20, 5, NULL, S_MAGE_XDIE7, 0, 0},        // S_MAGE_XDIE6
    {SPR_MAGE, 21, 5, NULL, S_MAGE_XDIE8, 0, 0},        // S_MAGE_XDIE7
    {SPR_MAGE, 22, 5, NULL, S_MAGE_XDIE9, 0, 0},        // S_MAGE_XDIE8
    {SPR_MAGE, 23, -1, NULL, S_NULL, 0, 0},     // S_MAGE_XDIE9
    {SPR_MAGE, 24, 5, A_FreezeDeath, S_MAGE_ICE2, 0, 0},        // S_MAGE_ICE
    {SPR_MAGE, 24, 1, A_FreezeDeathChunks, S_MAGE_ICE2, 0, 0},  // S_MAGE_ICE2
    {SPR_SORC, 0, 3, NULL, S_SORC_SPAWN2, 0, 0},        // S_SORC_SPAWN1
    {SPR_SORC, 0, 2, A_SorcSpinBalls, S_SORC_LOOK1, 0, 0},      // S_SORC_SPAWN2
    {SPR_SORC, 0, 10, A_Look, S_SORC_LOOK1, 0, 0},      // S_SORC_LOOK1
    {SPR_SORC, 0, 5, A_Chase, S_SORC_WALK2, 0, 0},      // S_SORC_WALK1
    {SPR_SORC, 1, 5, A_Chase, S_SORC_WALK3, 0, 0},      // S_SORC_WALK2
    {SPR_SORC, 2, 5, A_Chase, S_SORC_WALK4, 0, 0},      // S_SORC_WALK3
    {SPR_SORC, 3, 5, A_Chase, S_SORC_WALK1, 0, 0},      // S_SORC_WALK4
    {SPR_SORC, 6, 8, NULL, S_SORC_PAIN2, 0, 0}, // S_SORC_PAIN1
    {SPR_SORC, 6, 8, A_Pain, S_SORC_WALK1, 0, 0},       // S_SORC_PAIN2
    {SPR_SORC, 32773, 6, A_FaceTarget, S_SORC_ATK2_2, 0, 0},    // S_SORC_ATK2_1
    {SPR_SORC, 32773, 6, A_SpeedBalls, S_SORC_ATK2_3, 0, 0},    // S_SORC_ATK2_2
    {SPR_SORC, 32773, 6, A_FaceTarget, S_SORC_ATK2_3, 0, 0},    // S_SORC_ATK2_3
    {SPR_SORC, 32772, 6, NULL, S_SORC_ATTACK2, 0, 0},   // S_SORC_ATTACK1
    {SPR_SORC, 32772, 6, A_SpawnFizzle, S_SORC_ATTACK3, 0, 0},  // S_SORC_ATTACK2
    {SPR_SORC, 32772, 5, A_FaceTarget, S_SORC_ATTACK2, 0, 0},   // S_SORC_ATTACK3
    {SPR_SORC, 32772, 2, NULL, S_SORC_ATTACK5, 0, 0},   // S_SORC_ATTACK4
    {SPR_SORC, 32772, 2, A_SorcBossAttack, S_SORC_WALK1, 0, 0}, // S_SORC_ATTACK5
    {SPR_SORC, 32775, 5, NULL, S_SORC_DIE2, 0, 0},      // S_SORC_DIE1
    {SPR_SORC, 32776, 5, A_FaceTarget, S_SORC_DIE3, 0, 0},      // S_SORC_DIE2
    {SPR_SORC, 32777, 5, A_Scream, S_SORC_DIE4, 0, 0},  // S_SORC_DIE3
    {SPR_SORC, 32778, 5, NULL, S_SORC_DIE5, 0, 0},      // S_SORC_DIE4
    {SPR_SORC, 32779, 5, NULL, S_SORC_DIE6, 0, 0},      // S_SORC_DIE5
    {SPR_SORC, 32780, 5, NULL, S_SORC_DIE7, 0, 0},      // S_SORC_DIE6
    {SPR_SORC, 32781, 5, NULL, S_SORC_DIE8, 0, 0},      // S_SORC_DIE7
    {SPR_SORC, 32782, 5, NULL, S_SORC_DIE9, 0, 0},      // S_SORC_DIE8
    {SPR_SORC, 32783, 5, NULL, S_SORC_DIE0, 0, 0},      // S_SORC_DIE9
    {SPR_SORC, 32784, 5, NULL, S_SORC_DIEA, 0, 0},      // S_SORC_DIE0
    {SPR_SORC, 32785, 5, NULL, S_SORC_DIEB, 0, 0},      // S_SORC_DIEA
    {SPR_SORC, 32786, 5, NULL, S_SORC_DIEC, 0, 0},      // S_SORC_DIEB
    {SPR_SORC, 32787, 5, NULL, S_SORC_DIED, 0, 0},      // S_SORC_DIEC
    {SPR_SORC, 32788, 5, A_NoBlocking, S_SORC_DIEE, 0, 0},      // S_SORC_DIED
    {SPR_SORC, 32789, 5, NULL, S_SORC_DIEF, 0, 0},      // S_SORC_DIEE
    {SPR_SORC, 32790, 5, NULL, S_SORC_DIEG, 0, 0},      // S_SORC_DIEF
    {SPR_SORC, 32791, 5, NULL, S_SORC_DIEH, 0, 0},      // S_SORC_DIEG
    {SPR_SORC, 32792, 5, NULL, S_SORC_DIEI, 0, 0},      // S_SORC_DIEH
    {SPR_SORC, 32793, -1, NULL, S_NULL, 0, 0},  // S_SORC_DIEI
    {SPR_SBMP, 0, 2, A_SorcBallOrbit, S_SORCBALL1_2, 0, 0},     // S_SORCBALL1_1
    {SPR_SBMP, 1, 2, A_SorcBallOrbit, S_SORCBALL1_3, 0, 0},     // S_SORCBALL1_2
    {SPR_SBMP, 2, 2, A_SorcBallOrbit, S_SORCBALL1_4, 0, 0},     // S_SORCBALL1_3
    {SPR_SBMP, 3, 2, A_SorcBallOrbit, S_SORCBALL1_5, 0, 0},     // S_SORCBALL1_4
    {SPR_SBMP, 4, 2, A_SorcBallOrbit, S_SORCBALL1_6, 0, 0},     // S_SORCBALL1_5
    {SPR_SBMP, 5, 2, A_SorcBallOrbit, S_SORCBALL1_7, 0, 0},     // S_SORCBALL1_6
    {SPR_SBMP, 6, 2, A_SorcBallOrbit, S_SORCBALL1_8, 0, 0},     // S_SORCBALL1_7
    {SPR_SBMP, 7, 2, A_SorcBallOrbit, S_SORCBALL1_9, 0, 0},     // S_SORCBALL1_8
    {SPR_SBMP, 8, 2, A_SorcBallOrbit, S_SORCBALL1_0, 0, 0},     // S_SORCBALL1_9
    {SPR_SBMP, 9, 2, A_SorcBallOrbit, S_SORCBALL1_A, 0, 0},     // S_SORCBALL1_0
    {SPR_SBMP, 10, 2, A_SorcBallOrbit, S_SORCBALL1_B, 0, 0},    // S_SORCBALL1_A
    {SPR_SBMP, 11, 2, A_SorcBallOrbit, S_SORCBALL1_C, 0, 0},    // S_SORCBALL1_B
    {SPR_SBMP, 12, 2, A_SorcBallOrbit, S_SORCBALL1_D, 0, 0},    // S_SORCBALL1_C
    {SPR_SBMP, 13, 2, A_SorcBallOrbit, S_SORCBALL1_E, 0, 0},    // S_SORCBALL1_D
    {SPR_SBMP, 14, 2, A_SorcBallOrbit, S_SORCBALL1_F, 0, 0},    // S_SORCBALL1_E
    {SPR_SBMP, 15, 2, A_SorcBallOrbit, S_SORCBALL1_1, 0, 0},    // S_SORCBALL1_F
    {SPR_SBMP, 0, 5, A_SorcBallPop, S_SORCBALL1_D2, 0, 0},      // S_SORCBALL1_D1
    {SPR_SBMP, 1, 2, A_BounceCheck, S_SORCBALL1_D2, 0, 0},      // S_SORCBALL1_D2
    {SPR_SBS4, 3, 5, A_Explode, S_SORCBALL1_D6, 0, 0},  // S_SORCBALL1_D5
    {SPR_SBS4, 4, 5, NULL, S_SORCBALL1_D7, 0, 0},       // S_SORCBALL1_D6
    {SPR_SBS4, 5, 6, NULL, S_SORCBALL1_D8, 0, 0},       // S_SORCBALL1_D7
    {SPR_SBS4, 6, 6, NULL, S_SORCBALL1_D9, 0, 0},       // S_SORCBALL1_D8
    {SPR_SBS4, 7, 6, NULL, S_NULL, 0, 0},       // S_SORCBALL1_D9
    {SPR_SBMB, 0, 2, A_SorcBallOrbit, S_SORCBALL2_2, 0, 0},     // S_SORCBALL2_1
    {SPR_SBMB, 1, 2, A_SorcBallOrbit, S_SORCBALL2_3, 0, 0},     // S_SORCBALL2_2
    {SPR_SBMB, 2, 2, A_SorcBallOrbit, S_SORCBALL2_4, 0, 0},     // S_SORCBALL2_3
    {SPR_SBMB, 3, 2, A_SorcBallOrbit, S_SORCBALL2_5, 0, 0},     // S_SORCBALL2_4
    {SPR_SBMB, 4, 2, A_SorcBallOrbit, S_SORCBALL2_6, 0, 0},     // S_SORCBALL2_5
    {SPR_SBMB, 5, 2, A_SorcBallOrbit, S_SORCBALL2_7, 0, 0},     // S_SORCBALL2_6
    {SPR_SBMB, 6, 2, A_SorcBallOrbit, S_SORCBALL2_8, 0, 0},     // S_SORCBALL2_7
    {SPR_SBMB, 7, 2, A_SorcBallOrbit, S_SORCBALL2_9, 0, 0},     // S_SORCBALL2_8
    {SPR_SBMB, 8, 2, A_SorcBallOrbit, S_SORCBALL2_0, 0, 0},     // S_SORCBALL2_9
    {SPR_SBMB, 9, 2, A_SorcBallOrbit, S_SORCBALL2_A, 0, 0},     // S_SORCBALL2_0
    {SPR_SBMB, 10, 2, A_SorcBallOrbit, S_SORCBALL2_B, 0, 0},    // S_SORCBALL2_A
    {SPR_SBMB, 11, 2, A_SorcBallOrbit, S_SORCBALL2_C, 0, 0},    // S_SORCBALL2_B
    {SPR_SBMB, 12, 2, A_SorcBallOrbit, S_SORCBALL2_D, 0, 0},    // S_SORCBALL2_C
    {SPR_SBMB, 13, 2, A_SorcBallOrbit, S_SORCBALL2_E, 0, 0},    // S_SORCBALL2_D
    {SPR_SBMB, 14, 2, A_SorcBallOrbit, S_SORCBALL2_F, 0, 0},    // S_SORCBALL2_E
    {SPR_SBMB, 15, 2, A_SorcBallOrbit, S_SORCBALL2_1, 0, 0},    // S_SORCBALL2_F
    {SPR_SBMB, 0, 5, A_SorcBallPop, S_SORCBALL2_D2, 0, 0},      // S_SORCBALL2_D1
    {SPR_SBMB, 1, 2, A_BounceCheck, S_SORCBALL2_D2, 0, 0},      // S_SORCBALL2_D2
    {SPR_SBS3, 3, 5, A_Explode, S_SORCBALL2_D6, 0, 0},  // S_SORCBALL2_D5
    {SPR_SBS3, 4, 5, NULL, S_SORCBALL2_D7, 0, 0},       // S_SORCBALL2_D6
    {SPR_SBS3, 5, 6, NULL, S_SORCBALL2_D8, 0, 0},       // S_SORCBALL2_D7
    {SPR_SBS3, 6, 6, NULL, S_SORCBALL2_D9, 0, 0},       // S_SORCBALL2_D8
    {SPR_SBS3, 7, 6, NULL, S_NULL, 0, 0},       // S_SORCBALL2_D9
    {SPR_SBMG, 0, 2, A_SorcBallOrbit, S_SORCBALL3_2, 0, 0},     // S_SORCBALL3_1
    {SPR_SBMG, 1, 2, A_SorcBallOrbit, S_SORCBALL3_3, 0, 0},     // S_SORCBALL3_2
    {SPR_SBMG, 2, 2, A_SorcBallOrbit, S_SORCBALL3_4, 0, 0},     // S_SORCBALL3_3
    {SPR_SBMG, 3, 2, A_SorcBallOrbit, S_SORCBALL3_5, 0, 0},     // S_SORCBALL3_4
    {SPR_SBMG, 4, 2, A_SorcBallOrbit, S_SORCBALL3_6, 0, 0},     // S_SORCBALL3_5
    {SPR_SBMG, 5, 2, A_SorcBallOrbit, S_SORCBALL3_7, 0, 0},     // S_SORCBALL3_6
    {SPR_SBMG, 6, 2, A_SorcBallOrbit, S_SORCBALL3_8, 0, 0},     // S_SORCBALL3_7
    {SPR_SBMG, 7, 2, A_SorcBallOrbit, S_SORCBALL3_9, 0, 0},     // S_SORCBALL3_8
    {SPR_SBMG, 8, 2, A_SorcBallOrbit, S_SORCBALL3_0, 0, 0},     // S_SORCBALL3_9
    {SPR_SBMG, 9, 2, A_SorcBallOrbit, S_SORCBALL3_A, 0, 0},     // S_SORCBALL3_0
    {SPR_SBMG, 10, 2, A_SorcBallOrbit, S_SORCBALL3_B, 0, 0},    // S_SORCBALL3_A
    {SPR_SBMG, 11, 2, A_SorcBallOrbit, S_SORCBALL3_C, 0, 0},    // S_SORCBALL3_B
    {SPR_SBMG, 12, 2, A_SorcBallOrbit, S_SORCBALL3_D, 0, 0},    // S_SORCBALL3_C
    {SPR_SBMG, 13, 2, A_SorcBallOrbit, S_SORCBALL3_E, 0, 0},    // S_SORCBALL3_D
    {SPR_SBMG, 14, 2, A_SorcBallOrbit, S_SORCBALL3_F, 0, 0},    // S_SORCBALL3_E
    {SPR_SBMG, 15, 2, A_SorcBallOrbit, S_SORCBALL3_1, 0, 0},    // S_SORCBALL3_F
    {SPR_SBMG, 0, 5, A_SorcBallPop, S_SORCBALL3_D2, 0, 0},      // S_SORCBALL3_D1
    {SPR_SBMG, 1, 2, A_BounceCheck, S_SORCBALL3_D2, 0, 0},      // S_SORCBALL3_D2
    {SPR_SBS3, 3, 5, A_Explode, S_SORCBALL3_D6, 0, 0},  // S_SORCBALL3_D5
    {SPR_SBS3, 4, 5, NULL, S_SORCBALL3_D7, 0, 0},       // S_SORCBALL3_D6
    {SPR_SBS3, 5, 6, NULL, S_SORCBALL3_D8, 0, 0},       // S_SORCBALL3_D7
    {SPR_SBS3, 6, 6, NULL, S_SORCBALL3_D9, 0, 0},       // S_SORCBALL3_D8
    {SPR_SBS3, 7, 6, NULL, S_NULL, 0, 0},       // S_SORCBALL3_D9
    {SPR_SBS1, 32768, 2, NULL, S_SORCFX1_2, 0, 0},      // S_SORCFX1_1
    {SPR_SBS1, 32769, 3, A_SorcFX1Seek, S_SORCFX1_3, 0, 0},     // S_SORCFX1_2
    {SPR_SBS1, 32770, 3, A_SorcFX1Seek, S_SORCFX1_4, 0, 0},     // S_SORCFX1_3
    {SPR_SBS1, 32771, 3, A_SorcFX1Seek, S_SORCFX1_1, 0, 0},     // S_SORCFX1_4
    {SPR_FHFX, 32786, 2, A_Explode, S_SORCFX1_D2, 0, 0},        // S_SORCFX1_D1
    {SPR_FHFX, 32786, 6, NULL, S_SORCFX1_D3, 0, 0},     // S_SORCFX1_D2
    {SPR_FHFX, 32786, 6, NULL, S_NULL, 0, 0},   // S_SORCFX1_D3
    {SPR_SBS2, 32768, 3, A_SorcFX2Split, S_SORCFX2_SPLIT1, 0, 0},       // S_SORCFX2_SPLIT1
    {SPR_SBS2, 32768, 2, A_SorcFX2Orbit, S_SORCFX2_ORBIT2, 0, 0},       // S_SORCFX2_ORBIT1
    {SPR_SBS2, 32769, 2, A_SorcFX2Orbit, S_SORCFX2_ORBIT3, 0, 0},       // S_SORCFX2_ORBIT2
    {SPR_SBS2, 32770, 2, A_SorcFX2Orbit, S_SORCFX2_ORBIT4, 0, 0},       // S_SORCFX2_ORBIT3
    {SPR_SBS2, 32771, 2, A_SorcFX2Orbit, S_SORCFX2_ORBIT5, 0, 0},       // S_SORCFX2_ORBIT4
    {SPR_SBS2, 32772, 2, A_SorcFX2Orbit, S_SORCFX2_ORBIT6, 0, 0},       // S_SORCFX2_ORBIT5
    {SPR_SBS2, 32773, 2, A_SorcFX2Orbit, S_SORCFX2_ORBIT7, 0, 0},       // S_SORCFX2_ORBIT6
    {SPR_SBS2, 32774, 2, A_SorcFX2Orbit, S_SORCFX2_ORBIT8, 0, 0},       // S_SORCFX2_ORBIT7
    {SPR_SBS2, 32775, 2, A_SorcFX2Orbit, S_SORCFX2_ORBIT9, 0, 0},       // S_SORCFX2_ORBIT8
    {SPR_SBS2, 32776, 2, A_SorcFX2Orbit, S_SORCFX2_ORBIT0, 0, 0},       // S_SORCFX2_ORBIT9
    {SPR_SBS2, 32777, 2, A_SorcFX2Orbit, S_SORCFX2_ORBITA, 0, 0},       // S_SORCFX2_ORBIT0
    {SPR_SBS2, 32778, 2, A_SorcFX2Orbit, S_SORCFX2_ORBITB, 0, 0},       // S_SORCFX2_ORBITA
    {SPR_SBS2, 32779, 2, A_SorcFX2Orbit, S_SORCFX2_ORBITC, 0, 0},       // S_SORCFX2_ORBITB
    {SPR_SBS2, 32780, 2, A_SorcFX2Orbit, S_SORCFX2_ORBITD, 0, 0},       // S_SORCFX2_ORBITC
    {SPR_SBS2, 32781, 2, A_SorcFX2Orbit, S_SORCFX2_ORBITE, 0, 0},       // S_SORCFX2_ORBITD
    {SPR_SBS2, 32782, 2, A_SorcFX2Orbit, S_SORCFX2_ORBITF, 0, 0},       // S_SORCFX2_ORBITE
    {SPR_SBS2, 32783, 2, A_SorcFX2Orbit, S_SORCFX2_ORBIT1, 0, 0},       // S_SORCFX2_ORBITF
    {SPR_SBS2, 0, 10, NULL, S_NULL, 0, 0},      // S_SORCFX2T1
    {SPR_SBS3, 32768, 2, NULL, S_SORCFX3_2, 0, 0},      // S_SORCFX3_1
    {SPR_SBS3, 32769, 2, NULL, S_SORCFX3_3, 0, 0},      // S_SORCFX3_2
    {SPR_SBS3, 32770, 2, NULL, S_SORCFX3_1, 0, 0},      // S_SORCFX3_3
    {SPR_SBS3, 32768, 4, NULL, S_BISHMORPHA, 0, 0},     // S_BISHMORPH1
    {SPR_BISH, 15, 4, A_SorcererBishopEntry, S_BISHMORPHB, 0, 0},       // S_BISHMORPHA
    {SPR_BISH, 14, 4, NULL, S_BISHMORPHC, 0, 0},        // S_BISHMORPHB
    {SPR_BISH, 13, 4, NULL, S_BISHMORPHD, 0, 0},        // S_BISHMORPHC
    {SPR_BISH, 12, 3, NULL, S_BISHMORPHE, 0, 0},        // S_BISHMORPHD
    {SPR_BISH, 11, 3, NULL, S_BISHMORPHF, 0, 0},        // S_BISHMORPHE
    {SPR_BISH, 10, 3, NULL, S_BISHMORPHG, 0, 0},        // S_BISHMORPHF
    {SPR_BISH, 9, 3, NULL, S_BISHMORPHH, 0, 0}, // S_BISHMORPHG
    {SPR_BISH, 8, 3, NULL, S_BISHMORPHI, 0, 0}, // S_BISHMORPHH
    {SPR_BISH, 7, 3, NULL, S_BISHMORPHJ, 0, 0}, // S_BISHMORPHI
    {SPR_BISH, 6, 3, A_SpawnBishop, S_NULL, 0, 0},      // S_BISHMORPHJ
    {SPR_SBS3, 3, 3, NULL, S_SORCFX3_EXP2, 0, 0},       // S_SORCFX3_EXP1
    {SPR_SBS3, 4, 3, NULL, S_SORCFX3_EXP3, 0, 0},       // S_SORCFX3_EXP2
    {SPR_SBS3, 5, 3, NULL, S_SORCFX3_EXP4, 0, 0},       // S_SORCFX3_EXP3
    {SPR_SBS3, 6, 3, NULL, S_SORCFX3_EXP5, 0, 0},       // S_SORCFX3_EXP4
    {SPR_SBS3, 7, 3, NULL, S_NULL, 0, 0},       // S_SORCFX3_EXP5
    {SPR_SBS4, 32768, 2, A_SorcFX4Check, S_SORCFX4_2, 0, 0},    // S_SORCFX4_1
    {SPR_SBS4, 32769, 2, A_SorcFX4Check, S_SORCFX4_3, 0, 0},    // S_SORCFX4_2
    {SPR_SBS4, 32770, 2, A_SorcFX4Check, S_SORCFX4_1, 0, 0},    // S_SORCFX4_3
    {SPR_SBS4, 32771, 2, NULL, S_SORCFX4_D2, 0, 0},     // S_SORCFX4_D1
    {SPR_SBS4, 32772, 2, A_Explode, S_SORCFX4_D3, 0, 0},        // S_SORCFX4_D2
    {SPR_SBS4, 32773, 2, NULL, S_SORCFX4_D4, 0, 0},     // S_SORCFX4_D3
    {SPR_SBS4, 32774, 2, NULL, S_SORCFX4_D5, 0, 0},     // S_SORCFX4_D4
    {SPR_SBS4, 32775, 2, NULL, S_NULL, 0, 0},   // S_SORCFX4_D5
    {SPR_SBFX, 32768, 4, NULL, S_SORCSPARK2, 0, 0},     // S_SORCSPARK1
    {SPR_SBFX, 32769, 4, NULL, S_SORCSPARK3, 0, 0},     // S_SORCSPARK2
    {SPR_SBFX, 32770, 4, NULL, S_SORCSPARK4, 0, 0},     // S_SORCSPARK3
    {SPR_SBFX, 32771, 4, NULL, S_SORCSPARK5, 0, 0},     // S_SORCSPARK4
    {SPR_SBFX, 32772, 4, NULL, S_SORCSPARK6, 0, 0},     // S_SORCSPARK5
    {SPR_SBFX, 32773, 4, NULL, S_SORCSPARK7, 0, 0},     // S_SORCSPARK6
    {SPR_SBFX, 32774, 4, NULL, S_NULL, 0, 0},   // S_SORCSPARK7
    {SPR_RADE, 0, 4, NULL, S_BLASTEFFECT2, 0, 0},       // S_BLASTEFFECT1
    {SPR_RADE, 1, 4, NULL, S_BLASTEFFECT3, 0, 0},       // S_BLASTEFFECT2
    {SPR_RADE, 2, 4, NULL, S_BLASTEFFECT4, 0, 0},       // S_BLASTEFFECT3
    {SPR_RADE, 3, 4, NULL, S_BLASTEFFECT5, 0, 0},       // S_BLASTEFFECT4
    {SPR_RADE, 4, 4, NULL, S_BLASTEFFECT6, 0, 0},       // S_BLASTEFFECT5
    {SPR_RADE, 5, 4, NULL, S_BLASTEFFECT7, 0, 0},       // S_BLASTEFFECT6
    {SPR_RADE, 6, 4, NULL, S_BLASTEFFECT8, 0, 0},       // S_BLASTEFFECT7
    {SPR_RADE, 7, 4, NULL, S_BLASTEFFECT9, 0, 0},       // S_BLASTEFFECT8
    {SPR_RADE, 8, 4, NULL, S_NULL, 0, 0},       // S_BLASTEFFECT9
    {SPR_WATR, 0, 5, NULL, S_WATERDRIP1, 0, 0}, // S_WATERDRIP1
    {SPR_KORX, 0, 5, A_Look, S_KORAX_LOOK1, 0, 0},      // S_KORAX_LOOK1
    {SPR_KORX, 0, 3, A_KoraxStep2, S_KORAX_CHASE2, 0, 0},       // S_KORAX_CHASE1
    {SPR_KORX, 0, 3, A_KoraxChase, S_KORAX_CHASE3, 0, 0},       // S_KORAX_CHASE2
    {SPR_KORX, 0, 3, A_KoraxChase, S_KORAX_CHASE4, 0, 0},       // S_KORAX_CHASE3
    {SPR_KORX, 0, 3, A_KoraxChase, S_KORAX_CHASE5, 0, 0},       // S_KORAX_CHASE4
    {SPR_KORX, 1, 3, A_KoraxStep, S_KORAX_CHASE6, 0, 0},        // S_KORAX_CHASE5
    {SPR_KORX, 1, 3, A_KoraxChase, S_KORAX_CHASE7, 0, 0},       // S_KORAX_CHASE6
    {SPR_KORX, 1, 3, A_KoraxChase, S_KORAX_CHASE8, 0, 0},       // S_KORAX_CHASE7
    {SPR_KORX, 1, 3, A_KoraxChase, S_KORAX_CHASE9, 0, 0},       // S_KORAX_CHASE8
    {SPR_KORX, 2, 3, A_KoraxStep2, S_KORAX_CHASE0, 0, 0},       // S_KORAX_CHASE9
    {SPR_KORX, 2, 3, A_KoraxChase, S_KORAX_CHASEA, 0, 0},       // S_KORAX_CHASE0
    {SPR_KORX, 2, 3, A_KoraxChase, S_KORAX_CHASEB, 0, 0},       // S_KORAX_CHASEA
    {SPR_KORX, 2, 3, A_KoraxChase, S_KORAX_CHASEC, 0, 0},       // S_KORAX_CHASEB
    {SPR_KORX, 3, 3, A_KoraxStep, S_KORAX_CHASED, 0, 0},        // S_KORAX_CHASEC
    {SPR_KORX, 3, 3, A_KoraxChase, S_KORAX_CHASEE, 0, 0},       // S_KORAX_CHASED
    {SPR_KORX, 3, 3, A_KoraxChase, S_KORAX_CHASEF, 0, 0},       // S_KORAX_CHASEE
    {SPR_KORX, 3, 3, A_KoraxChase, S_KORAX_CHASE1, 0, 0},       // S_KORAX_CHASEF
    {SPR_KORX, 7, 5, A_Pain, S_KORAX_PAIN2, 0, 0},      // S_KORAX_PAIN1
    {SPR_KORX, 7, 5, NULL, S_KORAX_CHASE2, 0, 0},       // S_KORAX_PAIN2
    {SPR_KORX, 32772, 2, A_FaceTarget, S_KORAX_ATTACK2, 0, 0},  // S_KORAX_ATTACK1
    {SPR_KORX, 32772, 5, A_KoraxDecide, S_KORAX_ATTACK2, 0, 0}, // S_KORAX_ATTACK2
    {SPR_KORX, 32772, 4, A_FaceTarget, S_KORAX_MISSILE2, 0, 0}, // S_KORAX_MISSILE1
    {SPR_KORX, 32773, 8, A_KoraxMissile, S_KORAX_MISSILE3, 0, 0},       // S_KORAX_MISSILE2
    {SPR_KORX, 32772, 8, NULL, S_KORAX_CHASE2, 0, 0},   // S_KORAX_MISSILE3
    {SPR_KORX, 32772, 5, A_FaceTarget, S_KORAX_COMMAND2, 0, 0}, // S_KORAX_COMMAND1
    {SPR_KORX, 32790, 10, A_FaceTarget, S_KORAX_COMMAND3, 0, 0},        // S_KORAX_COMMAND2
    {SPR_KORX, 32774, 15, A_KoraxCommand, S_KORAX_COMMAND4, 0, 0},      // S_KORAX_COMMAND3
    {SPR_KORX, 32790, 10, NULL, S_KORAX_COMMAND5, 0, 0},        // S_KORAX_COMMAND4
    {SPR_KORX, 32772, 5, NULL, S_KORAX_CHASE2, 0, 0},   // S_KORAX_COMMAND5
    {SPR_KORX, 8, 5, NULL, S_KORAX_DEATH2, 0, 0},       // S_KORAX_DEATH1
    {SPR_KORX, 9, 5, A_FaceTarget, S_KORAX_DEATH3, 0, 0},       // S_KORAX_DEATH2
    {SPR_KORX, 10, 5, A_Scream, S_KORAX_DEATH4, 0, 0},  // S_KORAX_DEATH3
    {SPR_KORX, 11, 5, NULL, S_KORAX_DEATH5, 0, 0},      // S_KORAX_DEATH4
    {SPR_KORX, 12, 5, NULL, S_KORAX_DEATH6, 0, 0},      // S_KORAX_DEATH5
    {SPR_KORX, 13, 5, NULL, S_KORAX_DEATH7, 0, 0},      // S_KORAX_DEATH6
    {SPR_KORX, 14, 5, NULL, S_KORAX_DEATH8, 0, 0},      // S_KORAX_DEATH7
    {SPR_KORX, 15, 5, NULL, S_KORAX_DEATH9, 0, 0},      // S_KORAX_DEATH8
    {SPR_KORX, 16, 10, NULL, S_KORAX_DEATH0, 0, 0},     // S_KORAX_DEATH9
    {SPR_KORX, 17, 5, A_KoraxBonePop, S_KORAX_DEATHA, 0, 0},    // S_KORAX_DEATH0
    {SPR_KORX, 18, 5, A_NoBlocking, S_KORAX_DEATHB, 0, 0},      // S_KORAX_DEATHA
    {SPR_KORX, 19, 5, NULL, S_KORAX_DEATHC, 0, 0},      // S_KORAX_DEATHB
    {SPR_KORX, 20, 5, NULL, S_KORAX_DEATHD, 0, 0},      // S_KORAX_DEATHC
    {SPR_KORX, 21, -1, NULL, S_NULL, 0, 0},     // S_KORAX_DEATHD
    {SPR_SPIR, 0, 5, A_KSpiritRoam, S_KSPIRIT_ROAM2, 0, 0},     // S_KSPIRIT_ROAM1
    {SPR_SPIR, 1, 5, A_KSpiritRoam, S_KSPIRIT_ROAM1, 0, 0},     // S_KSPIRIT_ROAM2
    {SPR_SPIR, 3, 5, NULL, S_KSPIRIT_DEATH2, 0, 0},     // S_KSPIRIT_DEATH1
    {SPR_SPIR, 4, 5, NULL, S_KSPIRIT_DEATH3, 0, 0},     // S_KSPIRIT_DEATH2
    {SPR_SPIR, 5, 5, NULL, S_KSPIRIT_DEATH4, 0, 0},     // S_KSPIRIT_DEATH3
    {SPR_SPIR, 6, 5, NULL, S_KSPIRIT_DEATH5, 0, 0},     // S_KSPIRIT_DEATH4
    {SPR_SPIR, 7, 5, NULL, S_KSPIRIT_DEATH6, 0, 0},     // S_KSPIRIT_DEATH5
    {SPR_SPIR, 8, 5, NULL, S_NULL, 0, 0},       // S_KSPIRIT_DEATH6
    {SPR_MLFX, 32776, 2, NULL, S_KBOLT2, 0, 0}, // S_KBOLT1
    {SPR_MLFX, 32777, 2, A_KBoltRaise, S_KBOLT3, 0, 0}, // S_KBOLT2
    {SPR_MLFX, 32776, 2, A_KBolt, S_KBOLT4, 0, 0},      // S_KBOLT3
    {SPR_MLFX, 32777, 2, A_KBolt, S_KBOLT5, 0, 0},      // S_KBOLT4
    {SPR_MLFX, 32778, 2, A_KBolt, S_KBOLT6, 0, 0},      // S_KBOLT5
    {SPR_MLFX, 32779, 2, A_KBolt, S_KBOLT7, 0, 0},      // S_KBOLT6
    {SPR_MLFX, 32780, 2, A_KBolt, S_KBOLT3, 0, 0},      // S_KBOLT7
    {SPR_MAN1, 0, 2, NULL, S_SPAWNBATS2, 0, 0}, // S_SPAWNBATS1
    {SPR_MAN1, 0, 2, A_BatSpawnInit, S_SPAWNBATS3, 0, 0},       // S_SPAWNBATS2
    {SPR_MAN1, 0, 2, A_BatSpawn, S_SPAWNBATS3, 0, 0},   // S_SPAWNBATS3
    {SPR_MAN1, 0, -1, NULL, S_NULL, 0, 0},      // S_SPAWNBATS_OFF
    {SPR_ABAT, 0, 2, A_BatMove, S_BAT2, 0, 0},  // S_BAT1
    {SPR_ABAT, 1, 2, A_BatMove, S_BAT3, 0, 0},  // S_BAT2
    {SPR_ABAT, 2, 2, A_BatMove, S_BAT1, 0, 0},  // S_BAT3
    {SPR_ABAT, 0, 2, NULL, S_NULL, 0, 0}        // S_BAT_DEATH
};


mobjinfo_t mobjinfo[NUMMOBJTYPES] = {

    {                           // MT_MAPSPOT
     9001,                      // doomednum
     S_MAPSPOT,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOSECTOR | MF_NOGRAVITY,        // flags
     0                          // flags2
     },

    {                           // MT_MAPSPOTGRAVITY
     9013,                      // doomednum
     S_MAPSPOT,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     MF2_DONTDRAW               // flags2
     },

    {                           // MT_FIREBALL1
     -1,                        // doomednum
     S_FIREBALL1_1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FIREBALL1_X1,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_FIREBALL,              // deathsound
     2 * FRACUNIT,              // speed
     8 * FRACUNIT,              // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     4,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FIREDAMAGE    // flags2
     },

    {                           // MT_ARROW
     -1,                        // doomednum
     S_ARROW_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ARROW_X1,                // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     6 * FRACUNIT,              // speed
     8 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     4,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_DART
     -1,                        // doomednum
     S_DART_1,                  // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DART_X1,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     6 * FRACUNIT,              // speed
     8 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     2,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_POISONDART
     -1,                        // doomednum
     S_POISONDART_1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_POISONDART_X1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     6 * FRACUNIT,              // speed
     8 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     2,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_RIPPERBALL
     -1,                        // doomednum
     S_RIPPERBALL_1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_RIPPERBALL_X1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     6 * FRACUNIT,              // speed
     8 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     2,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY,    // flags
     MF2_NOTELEPORT | MF2_RIP   // flags2
     },

    {                           // MT_PROJECTILE_BLADE
     -1,                        // doomednum
     S_PRJ_BLADE1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_PRJ_BLADE_X1,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     6 * FRACUNIT,              // speed
     6 * FRACUNIT,              // radius
     6 * FRACUNIT,              // height
     100,                       // mass
     3,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_ICESHARD
     -1,                        // doomednum
     S_ICESHARD1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SHARDFXE1_1,             // deathstate
     S_NULL,                    // xdeathstate
     SFX_MAGE_SHARDS_EXPLODE,   // deathsound
     25 * FRACUNIT,             // speed
     13 * FRACUNIT,             // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     1,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_ICEDAMAGE     // flags2
     },

    {                           // MT_FLAME_SMALL_TEMP
     10500,                     // doomednum
     S_FLAME_TSMALL1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FLAME_LARGE_TEMP
     10502,                     // doomednum
     S_FLAME_TLARGE1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FLAME_SMALL
     10501,                     // doomednum
     S_FLAME_SMALL1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     MF2_NOTELEPORT | MF2_DONTDRAW      // flags2
     },

    {                           // MT_FLAME_LARGE
     10503,                     // doomednum
     S_FLAME_LARGE1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     MF2_NOTELEPORT | MF2_DONTDRAW      // flags2
     },

    {                           // MT_HEALINGBOTTLE
     81,                        // doomednum
     S_ITEM_PTN1_1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_HEALTHFLASK
     82,                        // doomednum
     S_ARTI_PTN2_1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_ARTIFLY
     83,                        // doomednum
     S_ARTI_SOAR1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_ARTIINVULNERABILITY
     84,                        // doomednum
     S_ARTI_INVU1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_SUMMONMAULATOR
     86,                        // doomednum
     S_ARTI_SUMMON,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_SUMMON_FX
     -1,                        // doomednum
     S_SUMMON_FX1_1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SUMMON_FX2_1,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     20 * FRACUNIT,             // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_MISSILE | MF_DROPOFF | MF_NOBLOCKMAP,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_THRUSTFLOOR_UP
     10091,                     // doomednum
     S_THRUSTINIT2_1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     128 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_THRUSTFLOOR_DOWN
     10090,                     // doomednum
     S_THRUSTINIT1_1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     128 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP | MF2_DONTDRAW      // flags2
     },

    {                           // MT_TELEPORTOTHER
     10040,                     // doomednum
     S_ARTI_TELOTHER1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_TELOTHER_FX1
     -1,                        // doomednum
     S_TELO_FX1,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_TELO_FX9,                // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     20 * FRACUNIT,             // speed
     16 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     10001,                     // damage
     SFX_NONE,                  // activesound
     MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_NOBLOCKMAP | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_TELOTHER_FX2
     -1,                        // doomednum
     S_TELO_FX2_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_TELO_FX9,                // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     16 * FRACUNIT,             // speed
     16 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     10001,                     // damage
     SFX_NONE,                  // activesound
     MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_NOBLOCKMAP | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_TELOTHER_FX3
     -1,                        // doomednum
     S_TELO_FX3_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_TELO_FX9,                // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     16 * FRACUNIT,             // speed
     16 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     10001,                     // damage
     SFX_NONE,                  // activesound
     MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_NOBLOCKMAP | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_TELOTHER_FX4
     -1,                        // doomednum
     S_TELO_FX4_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_TELO_FX9,                // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     16 * FRACUNIT,             // speed
     16 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     10001,                     // damage
     SFX_NONE,                  // activesound
     MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_NOBLOCKMAP | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_TELOTHER_FX5
     -1,                        // doomednum
     S_TELO_FX5_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_TELO_FX9,                // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     16 * FRACUNIT,             // speed
     16 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     10001,                     // damage
     SFX_NONE,                  // activesound
     MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_NOBLOCKMAP | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_DIRT1
     -1,                        // doomednum
     S_DIRT1_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DIRT1_D,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_DIRT2
     -1,                        // doomednum
     S_DIRT2_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DIRT2_D,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_DIRT3
     -1,                        // doomednum
     S_DIRT3_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DIRT3_D,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_DIRT4
     -1,                        // doomednum
     S_DIRT4_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DIRT4_D,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE,   // flags
     MF2_NOTELEPORT | MF2_LOGRAV        // flags2
     },

    {                           // MT_DIRT5
     -1,                        // doomednum
     S_DIRT5_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DIRT5_D,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE,   // flags
     MF2_NOTELEPORT | MF2_LOGRAV        // flags2
     },

    {                           // MT_DIRT6
     -1,                        // doomednum
     S_DIRT6_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DIRT6_D,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE,   // flags
     MF2_NOTELEPORT | MF2_LOGRAV        // flags2
     },

    {                           // MT_DIRTCLUMP
     -1,                        // doomednum
     S_DIRTCLUMP1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_ROCK1
     -1,                        // doomednum
     S_ROCK1_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ROCK1_D,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_ROCK2
     -1,                        // doomednum
     S_ROCK2_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ROCK2_D,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_ROCK3
     -1,                        // doomednum
     S_ROCK3_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ROCK3_D,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FOGSPAWNER
     10000,                     // doomednum
     S_SPAWNFOG1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOSECTOR,       // flags
     MF2_DONTDRAW | MF2_FLOATBOB        // flags2
     },

    {                           // MT_FOGPATCHS
     10001,                     // doomednum
     S_FOGPATCHS1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FOGPATCHS0,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     FRACUNIT,                  // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_FLOAT | MF_NOGRAVITY | MF_SHADOW | MF_NOCLIP,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FOGPATCHM
     10002,                     // doomednum
     S_FOGPATCHM1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FOGPATCHM0,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     FRACUNIT,                  // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_FLOAT | MF_NOGRAVITY | MF_SHADOW | MF_NOCLIP,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FOGPATCHL
     10003,                     // doomednum
     S_FOGPATCHL1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FOGPATCHL0,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     FRACUNIT,                  // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_FLOAT | MF_NOGRAVITY | MF_SHADOW | MF_NOCLIP,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_QUAKE_FOCUS
     -1,                        // doomednum
     S_QUAKE_ACTIVE1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOSECTOR,       // flags
     MF2_DONTDRAW               // flags2
     },

    {                           // MT_SGSHARD1
     -1,                        // doomednum
     S_SGSHARD1_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SGSHARD1_D,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SGSHARD2
     -1,                        // doomednum
     S_SGSHARD2_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SGSHARD2_D,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SGSHARD3
     -1,                        // doomednum
     S_SGSHARD3_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SGSHARD3_D,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SGSHARD4
     -1,                        // doomednum
     S_SGSHARD4_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SGSHARD4_D,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SGSHARD5
     -1,                        // doomednum
     S_SGSHARD5_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SGSHARD5_D,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SGSHARD6
     -1,                        // doomednum
     S_SGSHARD6_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SGSHARD6_D,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SGSHARD7
     -1,                        // doomednum
     S_SGSHARD7_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SGSHARD7_D,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SGSHARD8
     -1,                        // doomednum
     S_SGSHARD8_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SGSHARD8_D,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SGSHARD9
     -1,                        // doomednum
     S_SGSHARD9_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SGSHARD9_D,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SGSHARD0
     -1,                        // doomednum
     S_SGSHARD0_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SGSHARD0_D,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_ARTIEGG
     30,                        // doomednum
     S_ARTI_EGGC1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_EGGFX
     -1,                        // doomednum
     S_EGGFX1,                  // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     0,                         // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_EGGFXI1_1,               // deathstate
     S_NULL,                    // xdeathstate
     0,                         // deathsound
     18 * FRACUNIT,             // speed
     8 * FRACUNIT,              // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     1,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_ARTISUPERHEAL
     32,                        // doomednum
     S_ARTI_SPHL1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_ZWINGEDSTATUENOSKULL
     9011,                      // doomednum
     S_ZWINGEDSTATUENOSKULL,    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     62 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_ZGEMPEDESTAL
     9012,                      // doomednum
     S_ZGEMPEDESTAL1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     40 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZSKULL
     9002,                      // doomednum
     S_ARTIPUZZSKULL,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZGEMBIG
     9003,                      // doomednum
     S_ARTIPUZZGEMBIG,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZGEMRED
     9004,                      // doomednum
     S_ARTIPUZZGEMRED,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZGEMGREEN1
     9005,                      // doomednum
     S_ARTIPUZZGEMGREEN1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZGEMGREEN2
     9009,                      // doomednum
     S_ARTIPUZZGEMGREEN2,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZGEMBLUE1
     9006,                      // doomednum
     S_ARTIPUZZGEMBLUE1,        // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZGEMBLUE2
     9010,                      // doomednum
     S_ARTIPUZZGEMBLUE2,        // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZBOOK1
     9007,                      // doomednum
     S_ARTIPUZZBOOK1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZBOOK2
     9008,                      // doomednum
     S_ARTIPUZZBOOK2,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZSKULL2
     9014,                      // doomednum
     S_ARTIPUZZSKULL2,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZFWEAPON
     9015,                      // doomednum
     S_ARTIPUZZFWEAPON,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZCWEAPON
     9016,                      // doomednum
     S_ARTIPUZZCWEAPON,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZMWEAPON
     9017,                      // doomednum
     S_ARTIPUZZMWEAPON,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZGEAR
     9018,                      // doomednum
     S_ARTIPUZZGEAR_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZGEAR2
     9019,                      // doomednum
     S_ARTIPUZZGEAR2_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZGEAR3
     9020,                      // doomednum
     S_ARTIPUZZGEAR3_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTIPUZZGEAR4
     9021,                      // doomednum
     S_ARTIPUZZGEAR4_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARTITORCH
     33,                        // doomednum
     S_ARTI_TRCH1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_FIREBOMB
     -1,                        // doomednum
     S_FIREBOMB1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_FLECHETTE_EXPLODE,     // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOGRAVITY | MF_ALTSHADOW,       // flags
     MF2_FIREDAMAGE             // flags2
     },

    {                           // MT_ARTITELEPORT
     36,                        // doomednum
     S_ARTI_ATLP1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_ARTIPOISONBAG
     8000,                      // doomednum
     S_ARTI_PSBG1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_POISONBAG
     -1,                        // doomednum
     S_POISONBAG1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOGRAVITY | MF_NOBLOCKMAP,      // flags
     0                          // flags2
     },

    {                           // MT_POISONCLOUD
     -1,                        // doomednum
     S_POISONCLOUD1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_POISONSHROOM_DEATH,    // deathsound
     0,                         // speed
     1,                         // radius
     1,                         // height
     INT_MAX,                    // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOGRAVITY | MF_NOBLOCKMAP | MF_SHADOW | MF_NOCLIP | MF_DROPOFF, // flags
     MF2_NODMGTHRUST            // flags2
     },

    {                           // MT_THROWINGBOMB
     -1,                        // doomednum
     S_THROWINGBOMB1,           // spawnstate
     48,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_FLECHETTE_BOUNCE,      // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_THROWINGBOMB_X1,         // deathstate
     S_NULL,                    // xdeathstate
     SFX_FLECHETTE_EXPLODE,     // deathsound
     12 * FRACUNIT,             // speed
     8 * FRACUNIT,              // radius
     10 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE,   // flags
     MF2_FLOORBOUNCE | MF2_FIREDAMAGE   // flags2
     },

    {                           // MT_SPEEDBOOTS
     8002,                      // doomednum
     S_ARTI_BOOTS1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_BOOSTMANA
     8003,                      // doomednum
     S_ARTI_MANA,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_BOOSTARMOR
     8041,                      // doomednum
     S_ARTI_ARMOR1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_BLASTRADIUS
     10110,                     // doomednum
     S_ARTI_BLAST1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_HEALRADIUS
     10120,                     // doomednum
     S_ARTI_HEALRAD1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_SPLASH
     -1,                        // doomednum
     S_SPLASH1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SPLASHX,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     2 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF,   // flags
     MF2_NOTELEPORT | MF2_LOGRAV | MF2_CANNOTPUSH       // flags2
     },

    {                           // MT_SPLASHBASE
     -1,                        // doomednum
     S_SPLASHBASE1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_LAVASPLASH
     -1,                        // doomednum
     S_LAVASPLASH1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_LAVASMOKE
     -1,                        // doomednum
     S_LAVASMOKE1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     0                          // flags2
     },

    {                           // MT_SLUDGECHUNK
     -1,                        // doomednum
     S_SLUDGECHUNK1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SLUDGECHUNKX,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     2 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF,   // flags
     MF2_NOTELEPORT | MF2_LOGRAV | MF2_CANNOTPUSH       // flags2
     },

    {                           // MT_SLUDGESPLASH
     -1,                        // doomednum
     S_SLUDGESPLASH1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_MISC0
     5,                         // doomednum
     S_ZWINGEDSTATUE1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     62 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC1
     6,                         // doomednum
     S_ZROCK1_1,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC2
     7,                         // doomednum
     S_ZROCK2_1,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC3
     9,                         // doomednum
     S_ZROCK3_1,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC4
     15,                        // doomednum
     S_ZROCK4_1,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC5
     17,                        // doomednum
     S_ZCHANDELIER1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     60 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPAWNCEILING | MF_NOGRAVITY,    // flags
     0                          // flags2
     },

    {                           // MT_MISC6
     8063,                      // doomednum
     S_ZCHANDELIER_U,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     60 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPAWNCEILING | MF_NOGRAVITY,    // flags
     0                          // flags2
     },

    {                           // MT_MISC7
     24,                        // doomednum
     S_ZTREEDEAD1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     96 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC8
     25,                        // doomednum
     S_ZTREE,                   // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     15 * FRACUNIT,             // radius
     128 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_TREEDESTRUCTIBLE
     8062,                      // doomednum
     S_ZTREEDESTRUCTIBLE1,      // spawnstate
     70,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZTREEDES_D1,             // deathstate
     S_NULL,                    // xdeathstate
     SFX_TREE_BREAK,            // deathsound
     0,                         // speed
     15 * FRACUNIT,             // radius
     180 * FRACUNIT,            // height
     INT_MAX,                    // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_NOBLOOD,      // flags
     0                          // flags2
     },

    {                           // MT_MISC9
     26,                        // doomednum
     S_ZTREESWAMP182_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     150 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC10
     27,                        // doomednum
     S_ZTREESWAMP172_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     120 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC11
     28,                        // doomednum
     S_ZSTUMPBURNED1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     12 * FRACUNIT,             // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC12
     29,                        // doomednum
     S_ZSTUMPBARE1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     12 * FRACUNIT,             // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC13
     37,                        // doomednum
     S_ZSTUMPSWAMP1_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC14
     38,                        // doomednum
     S_ZSTUMPSWAMP2_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC15
     39,                        // doomednum
     S_ZSHROOMLARGE1_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC16
     40,                        // doomednum
     S_ZSHROOMLARGE2_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC17
     41,                        // doomednum
     S_ZSHROOMLARGE3_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC18
     42,                        // doomednum
     S_ZSHROOMSMALL1_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC19
     44,                        // doomednum
     S_ZSHROOMSMALL2_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC20
     45,                        // doomednum
     S_ZSHROOMSMALL3_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC21
     46,                        // doomednum
     S_ZSHROOMSMALL4_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC22
     47,                        // doomednum
     S_ZSHROOMSMALL5_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC23
     48,                        // doomednum
     S_ZSTALAGMITEPILLAR1,      // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     138 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC24
     49,                        // doomednum
     S_ZSTALAGMITELARGE1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     48 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC25
     50,                        // doomednum
     S_ZSTALAGMITEMEDIUM1,      // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     6 * FRACUNIT,              // radius
     40 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC26
     51,                        // doomednum
     S_ZSTALAGMITESMALL1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     36 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC27
     52,                        // doomednum
     S_ZSTALACTITELARGE1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     66 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_MISC28
     56,                        // doomednum
     S_ZSTALACTITEMEDIUM1,      // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     6 * FRACUNIT,              // radius
     50 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_MISC29
     57,                        // doomednum
     S_ZSTALACTITESMALL1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     40 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_MISC30
     58,                        // doomednum
     S_ZMOSSCEILING1_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPAWNCEILING | MF_NOGRAVITY,    // flags
     0                          // flags2
     },

    {                           // MT_MISC31
     59,                        // doomednum
     S_ZMOSSCEILING2_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     24 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPAWNCEILING | MF_NOGRAVITY,    // flags
     0                          // flags2
     },

    {                           // MT_MISC32
     60,                        // doomednum
     S_ZSWAMPVINE1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     52 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC33
     61,                        // doomednum
     S_ZCORPSEKABOB1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     92 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC34
     62,                        // doomednum
     S_ZCORPSESLEEPING1,        // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC35
     63,                        // doomednum
     S_ZTOMBSTONERIP1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     46 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC36
     64,                        // doomednum
     S_ZTOMBSTONESHANE1,        // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     46 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC37
     65,                        // doomednum
     S_ZTOMBSTONEBIGCROSS1,     // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     46 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC38
     66,                        // doomednum
     S_ZTOMBSTONEBRIANR1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     52 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC39
     67,                        // doomednum
     S_ZTOMBSTONECROSSCIRCLE1,  // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     52 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC40
     68,                        // doomednum
     S_ZTOMBSTONESMALLCROSS1,   // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     46 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC41
     69,                        // doomednum
     S_ZTOMBSTONEBRIANP1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     46 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC42
     71,                        // doomednum
     S_CORPSEHANGING_1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     6 * FRACUNIT,              // radius
     75 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_MISC43
     72,                        // doomednum
     S_ZSTATUEGARGOYLEGREENTALL_1,      // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     108 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC44
     73,                        // doomednum
     S_ZSTATUEGARGOYLEBLUETALL_1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     108 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC45
     74,                        // doomednum
     S_ZSTATUEGARGOYLEGREENSHORT_1,     // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     62 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC46
     76,                        // doomednum
     S_ZSTATUEGARGOYLEBLUESHORT_1,      // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     62 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC47
     8044,                      // doomednum
     S_ZSTATUEGARGOYLESTRIPETALL_1,     // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     108 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC48
     8045,                      // doomednum
     S_ZSTATUEGARGOYLEDARKREDTALL_1,    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     108 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC49
     8046,                      // doomednum
     S_ZSTATUEGARGOYLEREDTALL_1,        // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     108 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC50
     8047,                      // doomednum
     S_ZSTATUEGARGOYLETANTALL_1,        // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     108 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC51
     8048,                      // doomednum
     S_ZSTATUEGARGOYLERUSTTALL_1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     108 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC52
     8049,                      // doomednum
     S_ZSTATUEGARGOYLEDARKREDSHORT_1,   // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     62 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC53
     8050,                      // doomednum
     S_ZSTATUEGARGOYLEREDSHORT_1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     62 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC54
     8051,                      // doomednum
     S_ZSTATUEGARGOYLETANSHORT_1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     62 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC55
     8052,                      // doomednum
     S_ZSTATUEGARGOYLERUSTSHORT_1,      // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     14 * FRACUNIT,             // radius
     62 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC56
     77,                        // doomednum
     S_ZBANNERTATTERED_1,       // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     120 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC57
     78,                        // doomednum
     S_ZTREELARGE1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZTREELARGE1,             // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     15 * FRACUNIT,             // radius
     180 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC58
     79,                        // doomednum
     S_ZTREELARGE2,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZTREELARGE2,             // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     15 * FRACUNIT,             // radius
     180 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC59
     80,                        // doomednum
     S_ZTREEGNARLED1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     22 * FRACUNIT,             // radius
     100 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC60
     87,                        // doomednum
     S_ZTREEGNARLED2,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     22 * FRACUNIT,             // radius
     100 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC61
     88,                        // doomednum
     S_ZLOG,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     25 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC62
     89,                        // doomednum
     S_ZSTALACTITEICELARGE,     // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     66 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_MISC63
     90,                        // doomednum
     S_ZSTALACTITEICEMEDIUM,    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     50 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_MISC64
     91,                        // doomednum
     S_ZSTALACTITEICESMALL,     // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     32 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_MISC65
     92,                        // doomednum
     S_ZSTALACTITEICETINY,      // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_MISC66
     93,                        // doomednum
     S_ZSTALAGMITEICELARGE,     // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     66 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC67
     94,                        // doomednum
     S_ZSTALAGMITEICEMEDIUM,    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     50 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC68
     95,                        // doomednum
     S_ZSTALAGMITEICESMALL,     // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     32 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC69
     96,                        // doomednum
     S_ZSTALAGMITEICETINY,      // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC70
     97,                        // doomednum
     S_ZROCKBROWN1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     17 * FRACUNIT,             // radius
     72 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC71
     98,                        // doomednum
     S_ZROCKBROWN2,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     15 * FRACUNIT,             // radius
     50 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC72
     99,                        // doomednum
     S_ZROCKBLACK,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     40 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_MISC73
     100,                       // doomednum
     S_ZRUBBLE1,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC74
     101,                       // doomednum
     S_ZRUBBLE2,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC75
     102,                       // doomednum
     S_ZRUBBLE3,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_MISC76
     103,                       // doomednum
     S_ZVASEPILLAR,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     12 * FRACUNIT,             // radius
     54 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_POTTERY1
     104,                       // doomednum
     S_ZPOTTERY1,               // spawnstate
     15,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZPOTTERY_EXPLODE,        // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     32 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_NOBLOOD | MF_DROPOFF, // flags
     MF2_SLIDE | MF2_PUSHABLE | MF2_TELESTOMP | MF2_PASSMOBJ    // flags2
     },

    {                           // MT_POTTERY2
     105,                       // doomednum
     S_ZPOTTERY2,               // spawnstate
     15,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZPOTTERY_EXPLODE,        // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     25 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_NOBLOOD | MF_DROPOFF, // flags
     MF2_SLIDE | MF2_PUSHABLE | MF2_TELESTOMP | MF2_PASSMOBJ    // flags2
     },

    {                           // MT_POTTERY3
     106,                       // doomednum
     S_ZPOTTERY3,               // spawnstate
     15,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZPOTTERY_EXPLODE,        // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     15 * FRACUNIT,             // radius
     25 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_NOBLOOD | MF_DROPOFF, // flags
     MF2_SLIDE | MF2_PUSHABLE | MF2_TELESTOMP | MF2_PASSMOBJ    // flags2
     },

    {                           // MT_POTTERYBIT1
     -1,                        // doomednum
     S_POTTERYBIT_1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_POTTERYBIT_EX0,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_MISSILE,                // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_MISC77
     108,                       // doomednum
     S_ZCORPSELYNCHED1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     11 * FRACUNIT,             // radius
     95 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ZLYNCHED_NOHEART
     109,                       // doomednum
     S_ZCORPSELYNCHED2,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     100 * FRACUNIT,            // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_MISC78
     110,                       // doomednum
     S_ZCORPSESITTING,          // spawnstate
     30,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZCORPSESITTING_X,        // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     15 * FRACUNIT,             // radius
     35 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_NOBLOOD,      // flags
     0                          // flags2
     },

    {                           // MT_CORPSEBIT
     -1,                        // doomednum
     S_CORPSEBIT_1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     MF2_TELESTOMP              // flags2
     },

    {                           // MT_CORPSEBLOODDRIP
     -1,                        // doomednum
     S_CORPSEBLOODDRIP,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_CORPSEBLOODDRIP_X1,      // deathstate
     S_NULL,                    // xdeathstate
     SFX_DRIP,                  // deathsound
     0,                         // speed
     FRACUNIT,                  // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_MISSILE,                // flags
     MF2_LOGRAV                 // flags2
     },

    {                           // MT_BLOODPOOL
     111,                       // doomednum
     S_BLOODPOOL,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_MISC79
     119,                       // doomednum
     S_ZCANDLE1,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY,      // flags
     0                          // flags2
     },

    {                           // MT_MISC80
     113,                       // doomednum
     S_ZLEAFSPAWNER,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOSECTOR,       // flags
     MF2_DONTDRAW               // flags2
     },

    {                           // MT_LEAF1
     -1,                        // doomednum
     S_LEAF1_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_LEAF_X1,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     2 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE,        // flags
     MF2_NOTELEPORT | MF2_LOGRAV        // flags2
     },

    {                           // MT_LEAF2
     -1,                        // doomednum
     S_LEAF2_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_LEAF_X1,                 // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     2 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE,        // flags
     MF2_NOTELEPORT | MF2_LOGRAV        // flags2
     },

    {                           // MT_ZTWINEDTORCH
     116,                       // doomednum
     S_ZTWINEDTORCH_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_ZTWINEDTORCH_UNLIT
     117,                       // doomednum
     S_ZTWINEDTORCH_UNLIT,      // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     10 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_BRIDGE
     118,                       // doomednum
     S_BRIDGE1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     32 * FRACUNIT,             // radius
     2 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_NOGRAVITY,   // flags
     MF2_DONTDRAW               // flags2
     },

    {                           // MT_BRIDGEBALL
     -1,                        // doomednum
     S_BBALL1,                  // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY,      // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_ZWALLTORCH
     54,                        // doomednum
     S_ZWALLTORCH1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY,      // flags
     0                          // flags2
     },

    {                           // MT_ZWALLTORCH_UNLIT
     55,                        // doomednum
     S_ZWALLTORCH_U,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY,      // flags
     0                          // flags2
     },

    {                           // MT_ZBARREL
     8100,                      // doomednum
     S_ZBARREL1,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     15 * FRACUNIT,             // radius
     32 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_ZSHRUB1
     8101,                      // doomednum
     S_ZSHRUB1,                 // spawnstate
     20,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_ZSHRUB1_X1,              // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZSHRUB1_DIE,             // deathstate
     S_NULL,                    // xdeathstate
     SFX_TREE_EXPLODE,          // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     24 * FRACUNIT,             // height
     INT_MAX,                    // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_NOBLOOD,      // flags
     0                          // flags2
     },

    {                           // MT_ZSHRUB2
     8102,                      // doomednum
     S_ZSHRUB2,                 // spawnstate
     10,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_ZSHRUB2_X1,              // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZSHRUB2_DIE,             // deathstate
     S_NULL,                    // xdeathstate
     SFX_TREE_EXPLODE,          // deathsound
     0,                         // speed
     16 * FRACUNIT,             // radius
     40 * FRACUNIT,             // height
     INT_MAX,                    // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_NOBLOOD,      // flags
     0                          // flags2
     },

    {                           // MT_ZBUCKET
     8103,                      // doomednum
     S_ZBUCKET1,                // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     72 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SPAWNCEILING | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ZPOISONSHROOM
     8104,                      // doomednum
     S_ZPOISONSHROOM1,          // spawnstate
     30,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_ZPOISONSHROOM_P1,        // painstate
     255,                       // painchance
     SFX_POISONSHROOM_PAIN,     // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZPOISONSHROOM_X1,        // deathstate
     S_NULL,                    // xdeathstate
     SFX_POISONSHROOM_DEATH,    // deathsound
     0,                         // speed
     6 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     INT_MAX,                    // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SHOOTABLE | MF_SOLID | MF_NOBLOOD,      // flags
     0                          // flags2
     },

    {                           // MT_ZFIREBULL
     8042,                      // doomednum
     S_ZFIREBULL1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     80 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_ZFIREBULL_UNLIT
     8043,                      // doomednum
     S_ZFIREBULL_U,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     80 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_FIRETHING
     8060,                      // doomednum
     S_ZFIRETHING1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     10 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_BRASSTORCH
     8061,                      // doomednum
     S_ZBRASSTORCH1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     6 * FRACUNIT,              // radius
     35 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_ZSUITOFARMOR
     8064,                      // doomednum
     S_ZSUITOFARMOR,            // spawnstate
     60,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZSUITOFARMOR_X1,         // deathstate
     S_NULL,                    // xdeathstate
     SFX_SUITOFARMOR_BREAK,     // deathsound
     0,                         // speed
     16 * FRACUNIT,             // radius
     72 * FRACUNIT,             // height
     INT_MAX,                    // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_NOBLOOD,      // flags
     0                          // flags2
     },

    {                           // MT_ZARMORCHUNK
     -1,                        // doomednum
     S_ZARMORCHUNK1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     0,                         // flags
     0                          // flags2
     },

    {                           // MT_ZBELL
     8065,                      // doomednum
     S_ZBELL,                   // spawnstate
     5,                         // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZBELL_X1,                // deathstate
     S_NULL,                    // xdeathstate
     SFX_BELLRING,              // deathsound
     0,                         // speed
     56 * FRACUNIT,             // radius
     120 * FRACUNIT,            // height
     INT_MAX,                    // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_NOBLOOD | MF_NOGRAVITY | MF_SPAWNCEILING,     // flags
     0                          // flags2
     },

    {                           // MT_ZBLUE_CANDLE
     8066,                      // doomednum
     S_ZBLUE_CANDLE1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_ZIRON_MAIDEN
     8067,                      // doomednum
     S_ZIRON_MAIDEN,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     12 * FRACUNIT,             // radius
     60 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_ZXMAS_TREE
     8068,                      // doomednum
     S_ZXMAS_TREE,              // spawnstate
     20,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_ZXMAS_TREE_X1,           // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ZXMAS_TREE_DIE,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_TREE_EXPLODE,          // deathsound
     0,                         // speed
     11 * FRACUNIT,             // radius
     130 * FRACUNIT,            // height
     INT_MAX,                    // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_NOBLOOD,      // flags
     0                          // flags2
     },

    {                           // MT_ZCAULDRON
     8069,                      // doomednum
     S_ZCAULDRON1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     12 * FRACUNIT,             // radius
     26 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_ZCAULDRON_UNLIT
     8070,                      // doomednum
     S_ZCAULDRON_U,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     12 * FRACUNIT,             // radius
     26 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID,                  // flags
     0                          // flags2
     },

    {                           // MT_ZCHAINBIT32
     8071,                      // doomednum
     S_ZCHAINBIT32,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     32 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SPAWNCEILING,    // flags
     0                          // flags2
     },

    {                           // MT_ZCHAINBIT64
     8072,                      // doomednum
     S_ZCHAINBIT64,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     64 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SPAWNCEILING,    // flags
     0                          // flags2
     },

    {                           // MT_ZCHAINEND_HEART
     8073,                      // doomednum
     S_ZCHAINEND_HEART,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     32 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SPAWNCEILING,    // flags
     0                          // flags2
     },

    {                           // MT_ZCHAINEND_HOOK1
     8074,                      // doomednum
     S_ZCHAINEND_HOOK1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     32 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SPAWNCEILING,    // flags
     0                          // flags2
     },

    {                           // MT_ZCHAINEND_HOOK2
     8075,                      // doomednum
     S_ZCHAINEND_HOOK2,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     32 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SPAWNCEILING,    // flags
     0                          // flags2
     },

    {                           // MT_ZCHAINEND_SPIKE
     8076,                      // doomednum
     S_ZCHAINEND_SPIKE,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     32 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SPAWNCEILING,    // flags
     0                          // flags2
     },

    {                           // MT_ZCHAINEND_SKULL
     8077,                      // doomednum
     S_ZCHAINEND_SKULL,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     32 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SPAWNCEILING,    // flags
     0                          // flags2
     },

    {                           // MT_TABLE_SHIT1
     8500,                      // doomednum
     S_TABLE_SHIT1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_TABLE_SHIT2
     8501,                      // doomednum
     S_TABLE_SHIT2,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_TABLE_SHIT3
     8502,                      // doomednum
     S_TABLE_SHIT3,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_TABLE_SHIT4
     8503,                      // doomednum
     S_TABLE_SHIT4,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_TABLE_SHIT5
     8504,                      // doomednum
     S_TABLE_SHIT5,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_TABLE_SHIT6
     8505,                      // doomednum
     S_TABLE_SHIT6,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_TABLE_SHIT7
     8506,                      // doomednum
     S_TABLE_SHIT7,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_TABLE_SHIT8
     8507,                      // doomednum
     S_TABLE_SHIT8,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_TABLE_SHIT9
     8508,                      // doomednum
     S_TABLE_SHIT9,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_TABLE_SHIT10
     8509,                      // doomednum
     S_TABLE_SHIT10,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_TFOG
     -1,                        // doomednum
     S_TFOG1,                   // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_TRANSLUCENT,      // flags
     0                          // flags2
     },

    {                           // MT_MISC81
     140,                       // doomednum
     S_TELESMOKE1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_TRANSLUCENT,      // flags
     0                          // flags2
     },

    {                           // MT_TELEPORTMAN
     14,                        // doomednum
     S_NULL,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOSECTOR | MF_TRANSLUCENT,       // flags
     0                          // flags2
     },

    {                           // MT_PUNCHPUFF
     -1,                        // doomednum
     S_PUNCHPUFF1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_FIGHTER_PUNCH_HITTHING,        // seesound
     8,                         // reactiontime
     SFX_FIGHTER_PUNCH_HITWALL, // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     0                          // flags2
     },

    {                           // MT_FW_AXE
     8010,                      // doomednum
     S_AXE,                     // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_AXEPUFF
     -1,                        // doomednum
     S_HAMMERPUFF1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_FIGHTER_AXE_HITTHING,  // seesound
     8,                         // reactiontime
     SFX_FIGHTER_HAMMER_HITWALL,        // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     0                          // flags2
     },

    {                           // MT_AXEPUFF_GLOW
     -1,                        // doomednum
     S_AXEPUFF_GLOW1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_FIGHTER_AXE_HITTHING,  // seesound
     8,                         // reactiontime
     SFX_FIGHTER_HAMMER_HITWALL,        // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_TRANSLUCENT,      // flags
     0                          // flags2
     },

    {                           // MT_AXEBLOOD
     -1,                        // doomednum
     S_AXEBLOOD1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_AXEBLOOD6,               // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     2 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_DROPOFF, // flags
     MF2_NOTELEPORT | MF2_CANNOTPUSH    // flags2
     },

    {                           // MT_FW_HAMMER
     123,                       // doomednum
     S_HAMM,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_HAMMER_MISSILE
     -1,                        // doomednum
     S_HAMMER_MISSILE_1,        // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_HAMMER_MISSILE_X1,       // deathstate
     S_NULL,                    // xdeathstate
     SFX_FIGHTER_HAMMER_EXPLODE,        // deathsound
     25 * FRACUNIT,             // speed
     14 * FRACUNIT,             // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     10,                        // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS | MF2_FIREDAMAGE  // flags2
     },

    {                           // MT_HAMMERPUFF
     -1,                        // doomednum
     S_HAMMERPUFF1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_FIGHTER_HAMMER_HITTHING,       // seesound
     8,                         // reactiontime
     SFX_FIGHTER_HAMMER_HITWALL,        // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     0                          // flags2
     },

    {                           // MT_FSWORD_MISSILE
     -1,                        // doomednum
     S_FSWORD_MISSILE1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FSWORD_MISSILE_X1,       // deathstate
     S_NULL,                    // xdeathstate
     SFX_FIGHTER_SWORD_EXPLODE, // deathsound
     30 * FRACUNIT,             // speed
     16 * FRACUNIT,             // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     8,                         // damage
     SFX_NONE,                  // activesound
     MF_MISSILE | MF_NOBLOCKMAP | MF_NOGRAVITY | MF_DROPOFF | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS   // flags2
     },

    {                           // MT_FSWORD_FLAME
     -1,                        // doomednum
     S_FSWORD_FLAME1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     0                          // flags2
     },

    {                           // MT_CW_SERPSTAFF
     10,                        // doomednum
     S_CSTAFF,                  // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_CSTAFF_MISSILE
     -1,                        // doomednum
     S_CSTAFF_MISSILE1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_CSTAFF_MISSILE_X1,       // deathstate
     S_NULL,                    // xdeathstate
     SFX_CLERIC_CSTAFF_EXPLODE, // deathsound
     22 * FRACUNIT,             // speed
     12 * FRACUNIT,             // radius
     10 * FRACUNIT,             // height
     100,                       // mass
     5,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS   // flags2
     },

    {                           // MT_CSTAFFPUFF
     -1,                        // doomednum
     S_CSTAFFPUFF1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_CLERIC_CSTAFF_HITTHING,        // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     0                          // flags2
     },

    {                           // MT_CW_FLAME
     8009,                      // doomednum
     S_CFLAME1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_CFLAMEFLOOR
     -1,                        // doomednum
     S_CFLAMEFLOOR1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_TRANSLUCENT,      // flags
     0                          // flags2
     },

    {                           // MT_FLAMEPUFF
     -1,                        // doomednum
     S_FLAMEPUFF1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_CLERIC_FLAME_EXPLODE,  // seesound
     8,                         // reactiontime
     SFX_CLERIC_FLAME_EXPLODE,  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     FRACUNIT,                  // radius
     FRACUNIT,                  // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_TRANSLUCENT,      // flags
     0                          // flags2
     },

    {                           // MT_FLAMEPUFF2
     -1,                        // doomednum
     S_FLAMEPUFF2_1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_CLERIC_FLAME_EXPLODE,  // seesound
     8,                         // reactiontime
     SFX_CLERIC_FLAME_EXPLODE,  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     FRACUNIT,                  // radius
     FRACUNIT,                  // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_TRANSLUCENT,      // flags
     0                          // flags2
     },

    {                           // MT_CIRCLEFLAME
     -1,                        // doomednum
     S_CIRCLE_FLAME1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_CIRCLE_FLAME_X1,         // deathstate
     S_NULL,                    // xdeathstate
     SFX_CLERIC_FLAME_CIRCLE,   // deathsound
     0,                         // speed
     6 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     2,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FIREDAMAGE    // flags2
     },

    {                           // MT_CFLAME_MISSILE
     -1,                        // doomednum
     S_CFLAME_MISSILE1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_CFLAME_MISSILE_X,        // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     200 * FRACUNIT,            // speed
     14 * FRACUNIT,             // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     8,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS | MF2_DONTDRAW | MF2_FIREDAMAGE   // flags2
     },

    {                           // MT_HOLY_FX
     -1,                        // doomednum
     S_HOLY_FX1,                // spawnstate
     105,                       // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_HOLY_FX_X1,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_SPIRIT_DIE,            // deathsound
     12 * FRACUNIT,             // speed
     10 * FRACUNIT,             // radius
     6 * FRACUNIT,              // height
     100,                       // mass
     3,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_ALTSHADOW | MF_MISSILE,     // flags
     MF2_NOTELEPORT | MF2_SEEKERMISSILE | MF2_RIP | MF2_IMPACT | MF2_PCROSS     // flags2
     },

    {                           // MT_HOLY_TAIL
     -1,                        // doomednum
     S_HOLY_TAIL1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     FRACUNIT,                  // radius
     FRACUNIT,                  // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_NOCLIP | MF_ALTSHADOW,      // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_HOLY_PUFF
     -1,                        // doomednum
     S_HOLY_PUFF1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     0                          // flags2
     },

    {                           // MT_HOLY_MISSILE
     -1,                        // doomednum
     S_HOLY_MISSILE1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_HOLY_MISSILE_X,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     30 * FRACUNIT,             // speed
     15 * FRACUNIT,             // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     4,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_MISSILE | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_HOLY_MISSILE_PUFF
     -1,                        // doomednum
     S_HOLY_MISSILE_P1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_ALTSHADOW,  // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_MWANDPUFF
     -1,                        // doomednum
     S_MWANDPUFF1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_TRANSLUCENT,      // flags
     MF2_NOTELEPORT | MF2_CANNOTPUSH | MF2_NODMGTHRUST  // flags2
     },

    {                           // MT_MWANDSMOKE
     -1,                        // doomednum
     S_MWANDSMOKE1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     MF2_NOTELEPORT | MF2_CANNOTPUSH | MF2_NODMGTHRUST  // flags2
     },

    {                           // MT_MWAND_MISSILE
     -1,                        // doomednum
     S_MWAND_MISSILE1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_MWANDPUFF1,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     184 * FRACUNIT,            // speed
     12 * FRACUNIT,             // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     2,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_RIP | MF2_IMPACT | MF2_PCROSS | MF2_NODMGTHRUST | MF2_CANNOTPUSH      // flags2
     },

    {                           // MT_MW_LIGHTNING
     8040,                      // doomednum
     S_MW_LIGHTNING1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_LIGHTNING_CEILING
     -1,                        // doomednum
     S_LIGHTNING_CEILING1,      // spawnstate
     144,                       // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_LIGHTNING_C_X1,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     25 * FRACUNIT,             // speed
     16 * FRACUNIT,             // radius
     40 * FRACUNIT,             // height
     100,                       // mass
     8,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_MISSILE | MF_DROPOFF | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS   // flags2
     },

    {                           // MT_LIGHTNING_FLOOR
     -1,                        // doomednum
     S_LIGHTNING_FLOOR1,        // spawnstate
     144,                       // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_LIGHTNING_F_X1,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     25 * FRACUNIT,             // speed
     16 * FRACUNIT,             // radius
     40 * FRACUNIT,             // height
     100,                       // mass
     8,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_MISSILE | MF_DROPOFF | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS   // flags2
     },

    {                           // MT_LIGHTNING_ZAP
     -1,                        // doomednum
     S_LIGHTNING_ZAP1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_LIGHTNING_ZAP_X8,        // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     15 * FRACUNIT,             // radius
     35 * FRACUNIT,             // height
     100,                       // mass
     2,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_MISSILE | MF_DROPOFF | MF_TRANSLUCENT,    // flags
     0                          // flags2
     },

    {                           // MT_MSTAFF_FX
     -1,                        // doomednum
     S_MSTAFF_FX1_1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_MSTAFF_FX_X1,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_MAGE_STAFF_EXPLODE,    // deathsound
     20 * FRACUNIT,             // speed
     16 * FRACUNIT,             // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     6,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FIREDAMAGE | MF2_RIP | MF2_IMPACT | MF2_PCROSS        // flags2
     },

    {                           // MT_MSTAFF_FX2
     -1,                        // doomednum
     S_MSTAFF_FX2_1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_MSTAFF_FX2_X1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_MAGE_STAFF_EXPLODE,    // deathsound
     17 * FRACUNIT,             // speed
     20 * FRACUNIT,             // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     4,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FIREDAMAGE | MF2_IMPACT | MF2_PCROSS | MF2_SEEKERMISSILE      // flags2
     },

    {                           // MT_FW_SWORD1
     12,                        // doomednum
     S_FSWORD1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_FW_SWORD2
     13,                        // doomednum
     S_FSWORD2,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_FW_SWORD3
     16,                        // doomednum
     S_FSWORD3,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_CW_HOLY1
     18,                        // doomednum
     S_CHOLY1,                  // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_CW_HOLY2
     19,                        // doomednum
     S_CHOLY2,                  // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_CW_HOLY3
     20,                        // doomednum
     S_CHOLY3,                  // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_MW_STAFF1
     21,                        // doomednum
     S_MSTAFF1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_MW_STAFF2
     22,                        // doomednum
     S_MSTAFF2,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_MW_STAFF3
     23,                        // doomednum
     S_MSTAFF3,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_SNOUTPUFF
     -1,                        // doomednum
     S_PUNCHPUFF1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     0                          // flags2
     },

    {                           // MT_MW_CONE
     53,                        // doomednum
     S_COS1,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_SHARDFX1
     -1,                        // doomednum
     S_SHARDFX1_1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SHARDFXE1_1,             // deathstate
     S_NULL,                    // xdeathstate
     SFX_MAGE_SHARDS_EXPLODE,   // deathsound
     25 * FRACUNIT,             // speed
     13 * FRACUNIT,             // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     1,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS | MF2_ICEDAMAGE   // flags2
     },

    {                           // MT_BLOOD
     -1,                        // doomednum
     S_BLOOD1,                  // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     0                          // flags2
     },

    {                           // MT_BLOODSPLATTER
     -1,                        // doomednum
     S_BLOODSPLATTER1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_BLOODSPLATTERX,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     2 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF,   // flags
     MF2_NOTELEPORT | MF2_CANNOTPUSH    // flags2
     },

    {                           // MT_GIBS
     -1,                        // doomednum
     S_GIBS1,                   // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_CORPSE,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_PLAYER_FIGHTER
     -1,                        // doomednum
     S_FPLAY,                   // spawnstate
     100,                       // spawnhealth
     S_FPLAY_RUN1,              // seestate
     SFX_NONE,                  // seesound
     0,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_FPLAY_PAIN,              // painstate
     255,                       // painchance
     SFX_PLAYER_FIGHTER_PAIN,   // painsound
     S_NULL,                    // meleestate
     S_FPLAY_ATK1,              // missilestate
     S_NULL,                    // crashstate
     S_FPLAY_DIE1,              // deathstate
     S_FPLAY_XDIE1,             // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     16 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_DROPOFF | MF_PICKUP | MF_NOTDMATCH,   // flags
     MF2_WINDTHRUST | MF2_FLOORCLIP | MF2_SLIDE | MF2_PASSMOBJ | MF2_TELESTOMP | MF2_PUSHWALL   // flags2
     },

    {                           // MT_BLOODYSKULL
     -1,                        // doomednum
     S_BLOODYSKULL1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     4 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF,        // flags
     MF2_LOGRAV | MF2_CANNOTPUSH        // flags2
     },

    {                           // MT_PLAYER_SPEED
     -1,                        // doomednum
     S_PLAYER_SPEED1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_ALTSHADOW,       // flags
     0                          // flags2
     },

    {                           // MT_ICECHUNK
     -1,                        // doomednum
     S_ICECHUNK1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF,        // flags
     MF2_LOGRAV | MF2_CANNOTPUSH | MF2_FLOORCLIP        // flags2
     },

    {                           // MT_PLAYER_CLERIC
     -1,                        // doomednum
     S_CPLAY,                   // spawnstate
     100,                       // spawnhealth
     S_CPLAY_RUN1,              // seestate
     SFX_NONE,                  // seesound
     0,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_CPLAY_PAIN,              // painstate
     255,                       // painchance
     SFX_PLAYER_CLERIC_PAIN,    // painsound
     S_NULL,                    // meleestate
     S_CPLAY_ATK1,              // missilestate
     S_NULL,                    // crashstate
     S_CPLAY_DIE1,              // deathstate
     S_CPLAY_XDIE1,             // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     16 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_DROPOFF | MF_PICKUP | MF_NOTDMATCH,   // flags
     MF2_WINDTHRUST | MF2_FLOORCLIP | MF2_SLIDE | MF2_PASSMOBJ | MF2_TELESTOMP | MF2_PUSHWALL   // flags2
     },

    {                           // MT_PLAYER_MAGE
     -1,                        // doomednum
     S_MPLAY,                   // spawnstate
     100,                       // spawnhealth
     S_MPLAY_RUN1,              // seestate
     SFX_NONE,                  // seesound
     0,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_MPLAY_PAIN,              // painstate
     255,                       // painchance
     SFX_PLAYER_MAGE_PAIN,      // painsound
     S_NULL,                    // meleestate
     S_MPLAY_ATK1,              // missilestate
     S_NULL,                    // crashstate
     S_MPLAY_DIE1,              // deathstate
     S_MPLAY_XDIE1,             // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     16 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_DROPOFF | MF_PICKUP | MF_NOTDMATCH,   // flags
     MF2_WINDTHRUST | MF2_FLOORCLIP | MF2_SLIDE | MF2_PASSMOBJ | MF2_TELESTOMP | MF2_PUSHWALL   // flags2
     },

    {                           // MT_PIGPLAYER
     -1,                        // doomednum
     S_PIGPLAY,                 // spawnstate
     100,                       // spawnhealth
     S_PIGPLAY_RUN1,            // seestate
     SFX_NONE,                  // seesound
     0,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_PIGPLAY_PAIN,            // painstate
     255,                       // painchance
     SFX_PIG_PAIN,              // painsound
     S_NULL,                    // meleestate
     S_PIGPLAY_ATK1,            // missilestate
     S_NULL,                    // crashstate
     S_PIG_DIE1,                // deathstate
     S_NULL,                    // xdeathstate
     SFX_PIG_DEATH,             // deathsound
     0,                         // speed
     16 * FRACUNIT,             // radius
     24 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_DROPOFF | MF_NOTDMATCH,       // flags
     MF2_WINDTHRUST | MF2_SLIDE | MF2_PASSMOBJ | MF2_FLOORCLIP | MF2_TELESTOMP | MF2_PUSHWALL   // flags2
     },

    {                           // MT_PIG
     -1,                        // doomednum
     S_PIG_LOOK1,               // spawnstate
     25,                        // spawnhealth
     S_PIG_WALK1,               // seestate
     SFX_PIG_ACTIVE1,           // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_PIG_PAIN,                // painstate
     128,                       // painchance
     SFX_PIG_PAIN,              // painsound
     S_PIG_ATK1,                // meleestate
     0,                         // missilestate
     S_NULL,                    // crashstate
     S_PIG_DIE1,                // deathstate
     S_NULL,                    // xdeathstate
     SFX_PIG_DEATH,             // deathsound
     10,                        // speed
     12 * FRACUNIT,             // radius
     22 * FRACUNIT,             // height
     60,                        // mass
     0,                         // damage
     SFX_PIG_ACTIVE1,           // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL,    // flags
     MF2_WINDTHRUST | MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_PUSHWALL | MF2_TELESTOMP       // flags2
     },

    {                           // MT_CENTAUR
     107,                       // doomednum
     S_CENTAUR_LOOK1,           // spawnstate
     200,                       // spawnhealth
     S_CENTAUR_WALK1,           // seestate
     SFX_CENTAUR_SIGHT,         // seesound
     8,                         // reactiontime
     SFX_CENTAUR_ATTACK,        // attacksound
     S_CENTAUR_PAIN1,           // painstate
     135,                       // painchance
     SFX_CENTAUR_PAIN,          // painsound
     S_CENTAUR_ATK1,            // meleestate
     0,                         // missilestate
     S_NULL,                    // crashstate
     S_CENTAUR_DEATH1,          // deathstate
     S_CENTAUR_DEATH_X1,        // xdeathstate
     SFX_CENTAUR_DEATH,         // deathsound
     13,                        // speed
     20 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     120,                       // mass
     0,                         // damage
     SFX_CENTAUR_ACTIVE,        // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL,    // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_PUSHWALL | MF2_MCROSS | MF2_TELESTOMP   // flags2
     },

    {                           // MT_CENTAURLEADER
     115,                       // doomednum
     S_CENTAUR_LOOK1,           // spawnstate
     250,                       // spawnhealth
     S_CENTAUR_WALK1,           // seestate
     SFX_CENTAUR_SIGHT,         // seesound
     8,                         // reactiontime
     SFX_CENTAUR_ATTACK,        // attacksound
     S_CENTAUR_PAIN1,           // painstate
     96,                        // painchance
     SFX_CENTAUR_PAIN,          // painsound
     S_CENTAUR_ATK1,            // meleestate
     S_CENTAUR_MISSILE1,        // missilestate
     S_NULL,                    // crashstate
     S_CENTAUR_DEATH1,          // deathstate
     S_CENTAUR_DEATH_X1,        // xdeathstate
     SFX_CENTAUR_DEATH,         // deathsound
     10,                        // speed
     20 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     120,                       // mass
     0,                         // damage
     SFX_CENTAUR_ACTIVE,        // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL,    // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_PUSHWALL | MF2_MCROSS | MF2_TELESTOMP   // flags2
     },

    {                           // MT_CENTAUR_FX
     -1,                        // doomednum
     S_CENTAUR_FX1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_CENTAUR_FX_X1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_CENTAUR_MISSILE_EXPLODE,       // deathsound
     20 * FRACUNIT,             // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     4,                         // damage
     SFX_NONE,                  // activesound
     MF_MISSILE | MF_NOBLOCKMAP | MF_NOGRAVITY | MF_DROPOFF | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS   // flags2
     },

    {                           // MT_CENTAUR_SHIELD
     -1,                        // doomednum
     S_CENTAUR_SHIELD1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_CENTAUR_SHIELD_X1,       // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_DROPOFF | MF_CORPSE,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_CENTAUR_SWORD
     -1,                        // doomednum
     S_CENTAUR_SWORD1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_CENTAUR_SWORD_X1,        // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_DROPOFF | MF_CORPSE,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_DEMON
     31,                        // doomednum
     S_DEMN_LOOK1,              // spawnstate
     250,                       // spawnhealth
     S_DEMN_CHASE1,             // seestate
     SFX_DEMON_SIGHT,           // seesound
     8,                         // reactiontime
     SFX_DEMON_ATTACK,          // attacksound
     S_DEMN_PAIN1,              // painstate
     50,                        // painchance
     SFX_DEMON_PAIN,            // painsound
     S_DEMN_ATK1_1,             // meleestate
     S_DEMN_ATK2_1,             // missilestate
     S_NULL,                    // crashstate
     S_DEMN_DEATH1,             // deathstate
     S_DEMN_XDEATH1,            // xdeathstate
     SFX_DEMON_DEATH,           // deathsound
     13,                        // speed
     32 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     220,                       // mass
     0,                         // damage
     SFX_DEMON_ACTIVE,          // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL,    // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_MCROSS | MF2_TELESTOMP  // flags2
     },

    {                           // MT_DEMONCHUNK1
     -1,                        // doomednum
     S_DEMONCHUNK1_1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMONCHUNK1_4,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_CORPSE,       // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_DEMONCHUNK2
     -1,                        // doomednum
     S_DEMONCHUNK2_1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMONCHUNK2_4,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_CORPSE,       // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_DEMONCHUNK3
     -1,                        // doomednum
     S_DEMONCHUNK3_1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMONCHUNK3_4,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_CORPSE,       // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_DEMONCHUNK4
     -1,                        // doomednum
     S_DEMONCHUNK4_1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMONCHUNK4_4,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_CORPSE,       // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_DEMONCHUNK5
     -1,                        // doomednum
     S_DEMONCHUNK5_1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMONCHUNK5_4,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_CORPSE,       // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_DEMONFX1
     -1,                        // doomednum
     S_DEMONFX_MOVE1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMONFX_BOOM1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_DEMON_MISSILE_EXPLODE, // deathsound
     15 * FRACUNIT,             // speed
     10 * FRACUNIT,             // radius
     6 * FRACUNIT,              // height
     100,                       // mass
     5,                         // damage
     SFX_NONE,                  // activesound
     MF_MISSILE | MF_NOBLOCKMAP | MF_NOGRAVITY | MF_DROPOFF | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS | MF2_FIREDAMAGE  // flags2
     },

    {                           // MT_DEMON2
     8080,                      // doomednum
     S_DEMN2_LOOK1,             // spawnstate
     250,                       // spawnhealth
     S_DEMN2_CHASE1,            // seestate
     SFX_DEMON_SIGHT,           // seesound
     8,                         // reactiontime
     SFX_DEMON_ATTACK,          // attacksound
     S_DEMN2_PAIN1,             // painstate
     50,                        // painchance
     SFX_DEMON_PAIN,            // painsound
     S_DEMN2_ATK1_1,            // meleestate
     S_DEMN2_ATK2_1,            // missilestate
     S_NULL,                    // crashstate
     S_DEMN2_DEATH1,            // deathstate
     S_DEMN2_XDEATH1,           // xdeathstate
     SFX_DEMON_DEATH,           // deathsound
     13,                        // speed
     32 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     220,                       // mass
     0,                         // damage
     SFX_DEMON_ACTIVE,          // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL,    // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_MCROSS | MF2_TELESTOMP  // flags2
     },

    {                           // MT_DEMON2CHUNK1
     -1,                        // doomednum
     S_DEMON2CHUNK1_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMON2CHUNK1_4,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_CORPSE,       // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_DEMON2CHUNK2
     -1,                        // doomednum
     S_DEMON2CHUNK2_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMON2CHUNK2_4,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_CORPSE,       // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_DEMON2CHUNK3
     -1,                        // doomednum
     S_DEMON2CHUNK3_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMON2CHUNK3_4,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_CORPSE,       // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_DEMON2CHUNK4
     -1,                        // doomednum
     S_DEMON2CHUNK4_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMON2CHUNK4_4,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_CORPSE,       // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_DEMON2CHUNK5
     -1,                        // doomednum
     S_DEMON2CHUNK5_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMON2CHUNK5_4,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_CORPSE,       // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_DEMON2FX1
     -1,                        // doomednum
     S_DEMON2FX_MOVE1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DEMON2FX_BOOM1,          // deathstate
     S_NULL,                    // xdeathstate
     SFX_DEMON_MISSILE_EXPLODE, // deathsound
     15 * FRACUNIT,             // speed
     10 * FRACUNIT,             // radius
     6 * FRACUNIT,              // height
     100,                       // mass
     5,                         // damage
     SFX_NONE,                  // activesound
     MF_MISSILE | MF_NOBLOCKMAP | MF_NOGRAVITY | MF_DROPOFF | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS | MF2_FIREDAMAGE  // flags2
     },

    {                           // MT_WRAITHB
     10011,                     // doomednum
     S_WRAITH_LOOK1,            // spawnstate
     150,                       // spawnhealth
     S_WRAITH_RAISE1,           // seestate
     SFX_WRAITH_SIGHT,          // seesound
     8,                         // reactiontime
     SFX_WRAITH_ATTACK,         // attacksound
     S_WRAITH_PAIN1,            // painstate
     25,                        // painchance
     SFX_WRAITH_PAIN,           // painsound
     S_WRAITH_ATK1_1,           // meleestate
     S_WRAITH_ATK2_1,           // missilestate
     S_NULL,                    // crashstate
     S_WRAITH_DEATH1_1,         // deathstate
     S_WRAITH_DEATH2_1,         // xdeathstate
     SFX_WRAITH_DEATH,          // deathsound
     11,                        // speed
     20 * FRACUNIT,             // radius
     68 * FRACUNIT,             // height
     75,                        // mass
     10,                        // damage
     SFX_WRAITH_ACTIVE,         // activesound
     MF_DROPOFF | MF_NOGRAVITY | MF_FLOAT | MF_COUNTKILL,       // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_PUSHWALL | MF2_TELESTOMP | MF2_DONTDRAW // flags2
     },

    {                           // MT_WRAITH
     34,                        // doomednum
     S_WRAITH_INIT1,            // spawnstate
     150,                       // spawnhealth
     S_WRAITH_CHASE1,           // seestate
     SFX_WRAITH_SIGHT,          // seesound
     8,                         // reactiontime
     SFX_WRAITH_ATTACK,         // attacksound
     S_WRAITH_PAIN1,            // painstate
     25,                        // painchance
     SFX_WRAITH_PAIN,           // painsound
     S_WRAITH_ATK1_1,           // meleestate
     S_WRAITH_ATK2_1,           // missilestate
     S_NULL,                    // crashstate
     S_WRAITH_DEATH1_1,         // deathstate
     S_WRAITH_DEATH2_1,         // xdeathstate
     SFX_WRAITH_DEATH,          // deathsound
     11,                        // speed
     20 * FRACUNIT,             // radius
     55 * FRACUNIT,             // height
     75,                        // mass
     10,                        // damage
     SFX_WRAITH_ACTIVE,         // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_DROPOFF | MF_NOGRAVITY | MF_FLOAT,     // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_PUSHWALL | MF2_TELESTOMP        // flags2
     },

    {                           // MT_WRAITHFX1
     -1,                        // doomednum
     S_WRTHFX_MOVE1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_WRTHFX_BOOM1,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_WRAITH_MISSILE_EXPLODE,        // deathsound
     14 * FRACUNIT,             // speed
     10 * FRACUNIT,             // radius
     6 * FRACUNIT,              // height
     5,                         // mass
     5,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS | MF2_FLOORCLIP | MF2_FIREDAMAGE  // flags2
     },

    {                           // MT_WRAITHFX2
     -1,                        // doomednum
     S_WRTHFX_SIZZLE1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     2 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_TRANSLUCENT,        // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_WRAITHFX3
     -1,                        // doomednum
     S_WRTHFX_DROP1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_WRTHFX_DEAD1,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_DRIP,                  // deathsound
     0,                         // speed
     2 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,   // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_WRAITHFX4
     -1,                        // doomednum
     S_WRTHFX_ADROP1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_WRTHFX_ADEAD1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_DRIP,                  // deathsound
     0,                         // speed
     2 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_WRAITHFX5
     -1,                        // doomednum
     S_WRTHFX_BDROP1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_WRTHFX_BDEAD1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_DRIP,                  // deathsound
     0,                         // speed
     2 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     5,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_MINOTAUR
     9,                         // doomednum
     S_MNTR_SPAWN1,             // spawnstate
     2500,                      // spawnhealth
     S_MNTR_WALK1,              // seestate
     SFX_MAULATOR_SIGHT,        // seesound
     8,                         // reactiontime
     SFX_MAULATOR_HAMMER_SWING, // attacksound
     S_MNTR_PAIN1,              // painstate
     25,                        // painchance
     SFX_MAULATOR_PAIN,         // painsound
     S_MNTR_ATK1_1,             // meleestate
     S_MNTR_ATK2_1,             // missilestate
     S_NULL,                    // crashstate
     S_MNTR_DIE1,               // deathstate
     S_NULL,                    // xdeathstate
     SFX_MAULATOR_DEATH,        // deathsound
     16,                        // speed
     28 * FRACUNIT,             // radius
     100 * FRACUNIT,            // height
     800,                       // mass
     7,                         // damage
     SFX_MAULATOR_ACTIVE,       // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_SHADOW,        // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_PUSHWALL | MF2_TELESTOMP        // flags2
     },

    {                           // MT_MNTRFX1
     -1,                        // doomednum
     S_MNTRFX1_1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_MNTRFXI1_1,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     20 * FRACUNIT,             // speed
     10 * FRACUNIT,             // radius
     6 * FRACUNIT,              // height
     100,                       // mass
     3,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FIREDAMAGE    // flags2
     },

    {                           // MT_MNTRFX2
     -1,                        // doomednum
     S_MNTRFX2_1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_MNTRFXI2_1,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     14 * FRACUNIT,             // speed
     5 * FRACUNIT,              // radius
     12 * FRACUNIT,             // height
     100,                       // mass
     4,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FIREDAMAGE    // flags2
     },

    {                           // MT_MNTRFX3
     -1,                        // doomednum
     S_MNTRFX3_1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     0,                         // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_MNTRFXI2_1,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     4,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FIREDAMAGE    // flags2
     },

    {                           // MT_MNTRSMOKE
     -1,                        // doomednum
     S_MINOSMOKE1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_MNTRSMOKEEXIT
     -1,                        // doomednum
     S_MINOSMOKEX1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_SERPENT
     121,                       // doomednum
     S_SERPENT_LOOK1,           // spawnstate
     90,                        // spawnhealth
     S_SERPENT_SWIM1,           // seestate
     SFX_SERPENT_SIGHT,         // seesound
     8,                         // reactiontime
     SFX_SERPENT_ATTACK,        // attacksound
     S_SERPENT_PAIN1,           // painstate
     96,                        // painchance
     SFX_SERPENT_PAIN,          // painsound
     S_SERPENT_SURFACE1,        // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SERPENT_DIE1,            // deathstate
     S_SERPENT_XDIE1,           // xdeathstate
     SFX_SERPENT_DEATH,         // deathsound
     12,                        // speed
     32 * FRACUNIT,             // radius
     70 * FRACUNIT,             // height
     INT_MAX,                    // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_COUNTKILL | MF_NOBLOOD,      // flags
     MF2_PASSMOBJ | MF2_DONTDRAW | MF2_CANTLEAVEFLOORPIC | MF2_NONSHOOTABLE | MF2_MCROSS        // flags2
     },

    {                           // MT_SERPENTLEADER
     120,                       // doomednum
     S_SERPENT_LOOK1,           // spawnstate
     90,                        // spawnhealth
     S_SERPENT_SWIM1,           // seestate
     SFX_SERPENT_SIGHT,         // seesound
     8,                         // reactiontime
     SFX_SERPENT_ATTACK,        // attacksound
     S_SERPENT_PAIN1,           // painstate
     96,                        // painchance
     SFX_SERPENT_PAIN,          // painsound
     S_SERPENT_SURFACE1,        // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SERPENT_DIE1,            // deathstate
     S_SERPENT_XDIE1,           // xdeathstate
     SFX_SERPENT_DEATH,         // deathsound
     12,                        // speed
     32 * FRACUNIT,             // radius
     70 * FRACUNIT,             // height
     200,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_COUNTKILL | MF_NOBLOOD,      // flags
     MF2_PASSMOBJ | MF2_DONTDRAW | MF2_CANTLEAVEFLOORPIC | MF2_NONSHOOTABLE | MF2_MCROSS        // flags2
     },

    {                           // MT_SERPENTFX
     -1,                        // doomednum
     S_SERPENT_FX1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     0,                         // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SERPENT_FX_X1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_SERPENTFX_HIT,         // deathsound
     15 * FRACUNIT,             // speed
     8 * FRACUNIT,              // radius
     10 * FRACUNIT,             // height
     100,                       // mass
     4,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_SERPENT_HEAD
     -1,                        // doomednum
     S_SERPENT_HEAD1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     10 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP,             // flags
     MF2_LOGRAV                 // flags2
     },

    {                           // MT_SERPENT_GIB1
     -1,                        // doomednum
     S_SERPENT_GIB1_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     3 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY,      // flags
     0                          // flags2
     },

    {                           // MT_SERPENT_GIB2
     -1,                        // doomednum
     S_SERPENT_GIB2_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     3 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY,      // flags
     0                          // flags2
     },

    {                           // MT_SERPENT_GIB3
     -1,                        // doomednum
     S_SERPENT_GIB3_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     3 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY,      // flags
     0                          // flags2
     },

    {                           // MT_BISHOP
     114,                       // doomednum
     S_BISHOP_LOOK1,            // spawnstate
     130,                       // spawnhealth
     S_BISHOP_WALK1,            // seestate
     SFX_BISHOP_SIGHT,          // seesound
     8,                         // reactiontime
     SFX_BISHOP_ATTACK,         // attacksound
     S_BISHOP_PAIN1,            // painstate
     110,                       // painchance
     SFX_BISHOP_PAIN,           // painsound
     0,                         // meleestate
     S_BISHOP_ATK1,             // missilestate
     S_NULL,                    // crashstate
     S_BISHOP_DEATH1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_BISHOP_DEATH,          // deathsound
     10,                        // speed
     22 * FRACUNIT,             // radius
     65 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_BISHOP_ACTIVE,         // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_FLOAT | MF_NOGRAVITY | MF_NOBLOOD,     // flags
     MF2_PASSMOBJ | MF2_PUSHWALL | MF2_TELESTOMP        // flags2
     },

    {                           // MT_BISHOP_PUFF
     -1,                        // doomednum
     S_BISHOP_PUFF1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SHADOW | MF_NOBLOCKMAP | MF_NOGRAVITY,  // flags
     0                          // flags2
     },

    {                           // MT_BISHOPBLUR
     -1,                        // doomednum
     S_BISHOPBLUR1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     0                          // flags2
     },

    {                           // MT_BISHOPPAINBLUR
     -1,                        // doomednum
     S_BISHOPPAINBLUR1,         // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW,  // flags
     0                          // flags2
     },

    {                           // MT_BISH_FX
     -1,                        // doomednum
     S_BISHFX1_1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_BISHFXI1_1,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_BISHOP_MISSILE_EXPLODE,        // deathsound
     10 * FRACUNIT,             // speed
     10 * FRACUNIT,             // radius
     6 * FRACUNIT,              // height
     100,                       // mass
     1,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_SEEKERMISSILE // flags2
     },

    {                           // MT_DRAGON
     254,                       // doomednum
     S_DRAGON_LOOK1,            // spawnstate
     640,                       // spawnhealth
     S_DRAGON_INIT,             // seestate
     SFX_DRAGON_SIGHT,          // seesound
     8,                         // reactiontime
     SFX_DRAGON_ATTACK,         // attacksound
     S_DRAGON_PAIN1,            // painstate
     128,                       // painchance
     SFX_DRAGON_PAIN,           // painsound
     S_NULL,                    // meleestate
     S_DRAGON_ATK1,             // missilestate
     S_NULL,                    // crashstate
     S_DRAGON_DEATH1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_DRAGON_DEATH,          // deathsound
     10 * FRACUNIT,             // speed
     20 * FRACUNIT,             // radius
     65 * FRACUNIT,             // height
     INT_MAX,                    // mass
     0,                         // damage
     SFX_DRAGON_ACTIVE,         // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_FLOAT | MF_NOGRAVITY | MF_NOBLOOD,     // flags
     MF2_PASSMOBJ | MF2_BOSS    // flags2
     },

    {                           // MT_DRAGON_FX
     -1,                        // doomednum
     S_DRAGON_FX1_1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_DRAGON_FX1_X1,           // deathstate
     S_NULL,                    // xdeathstate
     SFX_DRAGON_FIREBALL_EXPLODE,       // deathsound
     24 * FRACUNIT,             // speed
     12 * FRACUNIT,             // radius
     10 * FRACUNIT,             // height
     100,                       // mass
     6,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_FIREDAMAGE    // flags2
     },

    {                           // MT_DRAGON_FX2
     -1,                        // doomednum
     S_DRAGON_FX2_1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_DRAGON_FIREBALL_EXPLODE,       // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_TRANSLUCENT,             // flags
     MF2_NOTELEPORT | MF2_FIREDAMAGE | MF2_DONTDRAW     // flags2
     },

    {                           // MT_ARMOR_1
     8005,                      // doomednum
     S_ARMOR_1,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARMOR_2
     8006,                      // doomednum
     S_ARMOR_2,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARMOR_3
     8007,                      // doomednum
     S_ARMOR_3,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_ARMOR_4
     8008,                      // doomednum
     S_ARMOR_4,                 // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL | MF_NOGRAVITY, // flags
     0                          // flags2
     },

    {                           // MT_MANA1
     122,                       // doomednum
     S_MANA1_1,                 // spawnstate
     10,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_MANA2
     124,                       // doomednum
     S_MANA2_1,                 // spawnstate
     10,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_MANA3
     8004,                      // doomednum
     S_MANA3_1,                 // spawnstate
     20,                        // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     8 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     MF2_FLOATBOB               // flags2
     },

    {                           // MT_KEY1
     8030,                      // doomednum
     S_KEY1,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_KEY2
     8031,                      // doomednum
     S_KEY2,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_KEY3
     8032,                      // doomednum
     S_KEY3,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_KEY4
     8033,                      // doomednum
     S_KEY4,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_KEY5
     8034,                      // doomednum
     S_KEY5,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_KEY6
     8035,                      // doomednum
     S_KEY6,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_KEY7
     8036,                      // doomednum
     S_KEY7,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_KEY8
     8037,                      // doomednum
     S_KEY8,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_KEY9
     8038,                      // doomednum
     S_KEY9,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_KEYA
     8039,                      // doomednum
     S_KEYA,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_KEYB
     8200,                      // doomednum
     S_KEYB,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     8 * FRACUNIT,              // radius
     20 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SPECIAL,                // flags
     0                          // flags2
     },

    {                           // MT_SOUNDWIND
     1410,                      // doomednum
     S_SND_WIND1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOSECTOR,       // flags
     0                          // flags2
     },

    {                           // MT_SOUNDWATERFALL
     41,                        // doomednum
     S_SND_WATERFALL,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOSECTOR,       // flags
     0                          // flags2
     },

    {                           // MT_ETTIN
     10030,                     // doomednum
     S_ETTIN_LOOK1,             // spawnstate
     175,                       // spawnhealth
     S_ETTIN_CHASE1,            // seestate
     SFX_ETTIN_SIGHT,           // seesound
     8,                         // reactiontime
     SFX_ETTIN_ATTACK,          // attacksound
     S_ETTIN_PAIN1,             // painstate
     60,                        // painchance
     SFX_ETTIN_PAIN,            // painsound
     S_ETTIN_ATK1_1,            // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ETTIN_DEATH1_1,          // deathstate
     S_ETTIN_DEATH2_1,          // xdeathstate
     SFX_ETTIN_DEATH,           // deathsound
     13,                        // speed
     25 * FRACUNIT,             // radius
     68 * FRACUNIT,             // height
     175,                       // mass
     3,                         // damage
     SFX_ETTIN_ACTIVE,          // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL,    // flags
     MF2_FLOORCLIP | MF2_PUSHWALL | MF2_MCROSS | MF2_TELESTOMP  // flags2
     },

    {                           // MT_ETTIN_MACE
     -1,                        // doomednum
     S_ETTIN_MACE1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ETTIN_MACE5,             // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_DROPOFF | MF_CORPSE,    // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_FIREDEMON
     10060,                     // doomednum
     S_FIRED_SPAWN1,            // spawnstate
     80,                        // spawnhealth
     S_FIRED_LOOK4,             // seestate
     SFX_FIRED_SPAWN,           // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_FIRED_PAIN1,             // painstate
     1,                         // painchance
     SFX_FIRED_PAIN,            // painsound
     S_NULL,                    // meleestate
     S_FIRED_ATTACK1,           // missilestate
     S_FIRED_XDEATH1,           // crashstate
     S_FIRED_DEATH1,            // deathstate
     S_FIRED_XDEATH1,           // xdeathstate
     SFX_FIRED_DEATH,           // deathsound
     13,                        // speed
     20 * FRACUNIT,             // radius
     68 * FRACUNIT,             // height
     75,                        // mass
     1,                         // damage
     SFX_FIRED_ACTIVE,          // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_DROPOFF | MF_NOGRAVITY | MF_FLOAT,     // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_PUSHWALL | MF2_INVULNERABLE | MF2_MCROSS | MF2_TELESTOMP        // flags2
     },

    {                           // MT_FIREDEMON_SPLOTCH1
     -1,                        // doomednum
     S_FIRED_CORPSE1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_DROPOFF | MF_CORPSE,    // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_FIREDEMON_SPLOTCH2
     -1,                        // doomednum
     S_FIRED_CORPSE4,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_DROPOFF | MF_CORPSE,    // flags
     MF2_NOTELEPORT | MF2_FLOORCLIP     // flags2
     },

    {                           // MT_FIREDEMON_FX1
     -1,                        // doomednum
     S_FIRED_RDROP1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FIRED_RDEAD1_1,          // deathstate
     S_FIRED_RDEAD1_2,          // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     16,                        // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FIREDEMON_FX2
     -1,                        // doomednum
     S_FIRED_RDROP2,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FIRED_RDEAD2_1,          // deathstate
     S_FIRED_RDEAD2_2,          // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     16,                        // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FIREDEMON_FX3
     -1,                        // doomednum
     S_FIRED_RDROP3,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FIRED_RDEAD3_1,          // deathstate
     S_FIRED_RDEAD3_2,          // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     16,                        // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FIREDEMON_FX4
     -1,                        // doomednum
     S_FIRED_RDROP4,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FIRED_RDEAD4_1,          // deathstate
     S_FIRED_RDEAD4_2,          // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     16,                        // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FIREDEMON_FX5
     -1,                        // doomednum
     S_FIRED_RDROP5,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FIRED_RDEAD5_1,          // deathstate
     S_FIRED_RDEAD5_2,          // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     3 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     16,                        // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FIREDEMON_FX6
     -1,                        // doomednum
     S_FIRED_FX6_1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_FIRED_FX6_2,             // deathstate
     S_NULL,                    // xdeathstate
     SFX_FIRED_MISSILE_HIT,     // deathsound
     10 * FRACUNIT,             // speed
     10 * FRACUNIT,             // radius
     6 * FRACUNIT,              // height
     15,                        // mass
     1,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_IMPACT | MF2_PCROSS | MF2_FLOORCLIP | MF2_FIREDAMAGE  // flags2
     },

    {                           // MT_ICEGUY
     8020,                      // doomednum
     S_ICEGUY_LOOK,             // spawnstate
     120,                       // spawnhealth
     S_ICEGUY_WALK1,            // seestate
     SFX_ICEGUY_SIGHT,          // seesound
     8,                         // reactiontime
     SFX_ICEGUY_ATTACK,         // attacksound
     S_ICEGUY_PAIN1,            // painstate
     144,                       // painchance
     SFX_NONE,                  // painsound
     0,                         // meleestate
     S_ICEGUY_ATK1,             // missilestate
     S_NULL,                    // crashstate
     S_ICEGUY_DEATH,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     14,                        // speed
     22 * FRACUNIT,             // radius
     75 * FRACUNIT,             // height
     150,                       // mass
     0,                         // damage
     SFX_ICEGUY_ACTIVE,         // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_NOBLOOD,       // flags
     MF2_PASSMOBJ | MF2_PUSHWALL | MF2_ICEDAMAGE | MF2_MCROSS | MF2_TELESTOMP   // flags2
     },

    {                           // MT_ICEGUY_FX
     -1,                        // doomednum
     S_ICEGUY_FX1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_ICEGUY_FX_X1,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_ICEGUY_FX_EXPLODE,     // deathsound
     14 * FRACUNIT,             // speed
     8 * FRACUNIT,              // radius
     10 * FRACUNIT,             // height
     100,                       // mass
     1,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_DROPOFF | MF_NOGRAVITY | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT | MF2_ICEDAMAGE     // flags2
     },

    {                           // MT_ICEFX_PUFF
     -1,                        // doomednum
     S_ICEFX_PUFF1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     FRACUNIT,                  // radius
     FRACUNIT,                  // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_SHADOW | MF_DROPOFF,     // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_ICEGUY_FX2
     -1,                        // doomednum
     S_ICEGUY_FX2_1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     10 * FRACUNIT,             // speed
     4 * FRACUNIT,              // radius
     4 * FRACUNIT,              // height
     100,                       // mass
     1,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_MISSILE | MF_TRANSLUCENT,   // flags
     MF2_NOTELEPORT | MF2_LOGRAV | MF2_ICEDAMAGE        // flags2
     },

    {                           // MT_ICEGUY_BIT
     -1,                        // doomednum
     S_ICEGUY_BIT1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     FRACUNIT,                  // radius
     FRACUNIT,                  // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF,        // flags
     MF2_NOTELEPORT | MF2_LOGRAV        // flags2
     },

    {                           // MT_ICEGUY_WISP1
     -1,                        // doomednum
     S_ICEGUY_WISP1_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_ALTSHADOW | MF_MISSILE,     // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_ICEGUY_WISP2
     -1,                        // doomednum
     S_ICEGUY_WISP2_1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_ALTSHADOW | MF_MISSILE,     // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_FIGHTER_BOSS
     10100,                     // doomednum
     S_FIGHTER,                 // spawnstate
     800,                       // spawnhealth
     S_FIGHTER_RUN1,            // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_FIGHTER_PAIN,            // painstate
     50,                        // painchance
     SFX_PLAYER_FIGHTER_PAIN,   // painsound
     S_FIGHTER_ATK1,            // meleestate
     S_FIGHTER_ATK1,            // missilestate
     S_NULL,                    // crashstate
     S_FIGHTER_DIE1,            // deathstate
     S_FIGHTER_XDIE1,           // xdeathstate
     SFX_PLAYER_FIGHTER_CRAZY_DEATH,    // deathsound
     25,                        // speed
     16 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL,    // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_TELESTOMP | MF2_PUSHWALL | MF2_MCROSS   // flags2
     },

    {                           // MT_CLERIC_BOSS
     10101,                     // doomednum
     S_CLERIC,                  // spawnstate
     800,                       // spawnhealth
     S_CLERIC_RUN1,             // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_CLERIC_PAIN,             // painstate
     50,                        // painchance
     SFX_PLAYER_CLERIC_PAIN,    // painsound
     S_CLERIC_ATK1,             // meleestate
     S_CLERIC_ATK1,             // missilestate
     S_NULL,                    // crashstate
     S_CLERIC_DIE1,             // deathstate
     S_CLERIC_XDIE1,            // xdeathstate
     SFX_PLAYER_CLERIC_CRAZY_DEATH,     // deathsound
     25,                        // speed
     16 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL,    // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_TELESTOMP | MF2_PUSHWALL | MF2_MCROSS   // flags2
     },

    {                           // MT_MAGE_BOSS
     10102,                     // doomednum
     S_MAGE,                    // spawnstate
     800,                       // spawnhealth
     S_MAGE_RUN1,               // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_MAGE_PAIN,               // painstate
     50,                        // painchance
     SFX_PLAYER_MAGE_PAIN,      // painsound
     S_MAGE_ATK1,               // meleestate
     S_MAGE_ATK1,               // missilestate
     S_NULL,                    // crashstate
     S_MAGE_DIE1,               // deathstate
     S_MAGE_XDIE1,              // xdeathstate
     SFX_PLAYER_MAGE_CRAZY_DEATH,       // deathsound
     25,                        // speed
     16 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL,    // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_TELESTOMP | MF2_PUSHWALL | MF2_MCROSS   // flags2
     },

    {                           // MT_SORCBOSS
     10080,                     // doomednum
     S_SORC_SPAWN1,             // spawnstate
     5000,                      // spawnhealth
     S_SORC_WALK1,              // seestate
     SFX_SORCERER_SIGHT,        // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_SORC_PAIN1,              // painstate
     10,                        // painchance
     SFX_SORCERER_PAIN,         // painsound
     S_NULL,                    // meleestate
     S_SORC_ATK2_1,             // missilestate
     S_NULL,                    // crashstate
     S_SORC_DIE1,               // deathstate
     S_NULL,                    // xdeathstate
     SFX_SORCERER_DEATHSCREAM,  // deathsound
     16,                        // speed
     40 * FRACUNIT,             // radius
     110 * FRACUNIT,            // height
     500,                       // mass
     9,                         // damage
     SFX_SORCERER_ACTIVE,       // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_NOBLOOD,       // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_PUSHWALL | MF2_BOSS | MF2_MCROSS        // flags2
     },

    {                           // MT_SORCBALL1
     -1,                        // doomednum
     S_SORCBALL1_1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_SORCERER_BALLBOUNCE,   // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_SORCBALL1_D1,            // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SORCBALL1_D5,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     10 * FRACUNIT,             // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_MISSILE, // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SORCBALL2
     -1,                        // doomednum
     S_SORCBALL2_1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_SORCERER_BALLBOUNCE,   // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_SORCBALL2_D1,            // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SORCBALL2_D5,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     10 * FRACUNIT,             // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_MISSILE, // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SORCBALL3
     -1,                        // doomednum
     S_SORCBALL3_1,             // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_SORCERER_BALLBOUNCE,   // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_SORCBALL3_D1,            // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SORCBALL3_D5,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     10 * FRACUNIT,             // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_MISSILE, // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SORCFX1
     -1,                        // doomednum
     S_SORCFX1_1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_SORCERER_BALLBOUNCE,   // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SORCFX1_D1,              // deathstate
     S_SORCFX1_D1,              // xdeathstate
     SFX_NONE,                  // deathsound
     7 * FRACUNIT,              // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_TRANSLUCENT,        // flags
     MF2_NOTELEPORT | MF2_FLOORBOUNCE   // flags2
     },

    {                           // MT_SORCFX2
     -1,                        // doomednum
     S_SORCFX2_SPLIT1,          // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SORCFX2T1,               // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     15 * FRACUNIT,             // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY,      // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_SORCFX2_T1
     -1,                        // doomednum
     S_SORCFX2T1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_ALTSHADOW,       // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_SORCFX3
     -1,                        // doomednum
     S_SORCFX3_1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_SORCERER_BISHOPSPAWN,  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_BISHMORPH1,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     15 * FRACUNIT,             // speed
     22 * FRACUNIT,             // radius
     65 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_TRANSLUCENT,        // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_SORCFX3_EXPLOSION
     -1,                        // doomednum
     S_SORCFX3_EXP1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_ALTSHADOW,       // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_SORCFX4
     -1,                        // doomednum
     S_SORCFX4_1,               // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_SORCFX4_D1,              // deathstate
     S_NULL,                    // xdeathstate
     SFX_SORCERER_BALLEXPLODE,  // deathsound
     12 * FRACUNIT,             // speed
     10 * FRACUNIT,             // radius
     10 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_MISSILE | MF_NOGRAVITY | MF_TRANSLUCENT, // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_SORCSPARK1
     -1,                        // doomednum
     S_SORCSPARK1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     5 * FRACUNIT,              // radius
     5 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_TRANSLUCENT,        // flags
     MF2_NOTELEPORT | MF2_LOGRAV        // flags2
     },

    {                           // MT_BLASTEFFECT
     -1,                        // doomednum
     S_BLASTEFFECT1,            // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_NOCLIP | MF_ALTSHADOW,   // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_WATER_DRIP
     -1,                        // doomednum
     S_WATERDRIP1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_DRIP,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     1,                         // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_MISSILE,                // flags
     MF2_LOGRAV | MF2_NOTELEPORT        // flags2
     },

    {                           // MT_KORAX
     10200,                     // doomednum
     S_KORAX_LOOK1,             // spawnstate
     5000,                      // spawnhealth
     S_KORAX_CHASE2,            // seestate
     SFX_KORAX_SIGHT,           // seesound
     8,                         // reactiontime
     SFX_KORAX_ATTACK,          // attacksound
     S_KORAX_PAIN1,             // painstate
     20,                        // painchance
     SFX_KORAX_PAIN,            // painsound
     S_NULL,                    // meleestate
     S_KORAX_ATTACK1,           // missilestate
     S_NULL,                    // crashstate
     S_KORAX_DEATH1,            // deathstate
     S_NULL,                    // xdeathstate
     SFX_KORAX_DEATH,           // deathsound
     10,                        // speed
     65 * FRACUNIT,             // radius
     115 * FRACUNIT,            // height
     2000,                      // mass
     15,                        // damage
     SFX_KORAX_ACTIVE,          // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL,    // flags
     MF2_FLOORCLIP | MF2_PUSHWALL | MF2_MCROSS | MF2_TELESTOMP | MF2_BOSS       // flags2
     },

    {                           // MT_KORAX_SPIRIT1
     -1,                        // doomednum
     S_KSPIRIT_ROAM1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     8 * FRACUNIT,              // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_ALTSHADOW | MF_MISSILE | MF_NOCLIP, // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_KORAX_SPIRIT2
     -1,                        // doomednum
     S_KSPIRIT_ROAM1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     8 * FRACUNIT,              // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_ALTSHADOW | MF_MISSILE | MF_NOCLIP, // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_KORAX_SPIRIT3
     -1,                        // doomednum
     S_KSPIRIT_ROAM1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     8 * FRACUNIT,              // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_ALTSHADOW | MF_MISSILE | MF_NOCLIP, // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_KORAX_SPIRIT4
     -1,                        // doomednum
     S_KSPIRIT_ROAM1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     8 * FRACUNIT,              // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_ALTSHADOW | MF_MISSILE | MF_NOCLIP, // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_KORAX_SPIRIT5
     -1,                        // doomednum
     S_KSPIRIT_ROAM1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     8 * FRACUNIT,              // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_ALTSHADOW | MF_MISSILE | MF_NOCLIP, // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_KORAX_SPIRIT6
     -1,                        // doomednum
     S_KSPIRIT_ROAM1,           // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     8 * FRACUNIT,              // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_DROPOFF | MF_NOGRAVITY | MF_ALTSHADOW | MF_MISSILE | MF_NOCLIP, // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_DEMON_MASH
     -1,                        // doomednum
     S_DEMN_LOOK1,              // spawnstate
     250,                       // spawnhealth
     S_DEMN_CHASE1,             // seestate
     SFX_DEMON_SIGHT,           // seesound
     8,                         // reactiontime
     SFX_DEMON_ATTACK,          // attacksound
     S_DEMN_PAIN1,              // painstate
     50,                        // painchance
     SFX_DEMON_PAIN,            // painsound
     S_DEMN_ATK1_1,             // meleestate
     S_DEMN_ATK2_1,             // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_DEMON_DEATH,           // deathsound
     13,                        // speed
     32 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     220,                       // mass
     0,                         // damage
     SFX_DEMON_ACTIVE,          // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_ALTSHADOW | MF_NOBLOOD,        // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_MCROSS | MF2_PUSHWALL | MF2_BLASTED     // flags2
     },

    {                           // MT_DEMON2_MASH
     -1,                        // doomednum
     S_DEMN2_LOOK1,             // spawnstate
     250,                       // spawnhealth
     S_DEMN2_CHASE1,            // seestate
     SFX_DEMON_SIGHT,           // seesound
     8,                         // reactiontime
     SFX_DEMON_ATTACK,          // attacksound
     S_DEMN2_PAIN1,             // painstate
     50,                        // painchance
     SFX_DEMON_PAIN,            // painsound
     S_DEMN2_ATK1_1,            // meleestate
     S_DEMN2_ATK2_1,            // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_DEMON_DEATH,           // deathsound
     13,                        // speed
     32 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     220,                       // mass
     0,                         // damage
     SFX_DEMON_ACTIVE,          // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_ALTSHADOW | MF_NOBLOOD,        // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_MCROSS | MF2_PUSHWALL | MF2_BLASTED     // flags2
     },

    {                           // MT_ETTIN_MASH
     -1,                        // doomednum
     S_ETTIN_LOOK1,             // spawnstate
     175,                       // spawnhealth
     S_ETTIN_CHASE1,            // seestate
     SFX_ETTIN_SIGHT,           // seesound
     8,                         // reactiontime
     SFX_ETTIN_ATTACK,          // attacksound
     S_ETTIN_PAIN1,             // painstate
     60,                        // painchance
     SFX_ETTIN_PAIN,            // painsound
     S_ETTIN_ATK1_1,            // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_ETTIN_DEATH,           // deathsound
     13,                        // speed
     25 * FRACUNIT,             // radius
     68 * FRACUNIT,             // height
     175,                       // mass
     3,                         // damage
     SFX_ETTIN_ACTIVE,          // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_ALTSHADOW | MF_NOBLOOD,        // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_MCROSS | MF2_PUSHWALL | MF2_BLASTED     // flags2
     },

    {                           // MT_CENTAUR_MASH
     -1,                        // doomednum
     S_CENTAUR_LOOK1,           // spawnstate
     200,                       // spawnhealth
     S_CENTAUR_WALK1,           // seestate
     SFX_CENTAUR_SIGHT,         // seesound
     8,                         // reactiontime
     SFX_CENTAUR_ATTACK,        // attacksound
     S_CENTAUR_PAIN1,           // painstate
     135,                       // painchance
     SFX_CENTAUR_PAIN,          // painsound
     S_CENTAUR_ATK1,            // meleestate
     0,                         // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_CENTAUR_DEATH,         // deathsound
     13,                        // speed
     20 * FRACUNIT,             // radius
     64 * FRACUNIT,             // height
     120,                       // mass
     0,                         // damage
     SFX_CENTAUR_ACTIVE,        // activesound
     MF_SOLID | MF_SHOOTABLE | MF_COUNTKILL | MF_ALTSHADOW | MF_NOBLOOD,        // flags
     MF2_FLOORCLIP | MF2_PASSMOBJ | MF2_MCROSS | MF2_PUSHWALL | MF2_BLASTED     // flags2
     },

    {                           // MT_KORAX_BOLT
     -1,                        // doomednum
     S_KBOLT1,                  // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     15 * FRACUNIT,             // radius
     35 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_MISSILE | MF_DROPOFF | MF_TRANSLUCENT,    // flags
     MF2_NOTELEPORT             // flags2
     },

    {                           // MT_BAT_SPAWNER
     10225,                     // doomednum
     S_SPAWNBATS1,              // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_NULL,                    // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     0,                         // speed
     20 * FRACUNIT,             // radius
     16 * FRACUNIT,             // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOSECTOR | MF_NOGRAVITY,        // flags
     MF2_DONTDRAW               // flags2
     },

    {                           // MT_BAT
     -1,                        // doomednum
     S_BAT1,                    // spawnstate
     1000,                      // spawnhealth
     S_NULL,                    // seestate
     SFX_NONE,                  // seesound
     8,                         // reactiontime
     SFX_NONE,                  // attacksound
     S_NULL,                    // painstate
     0,                         // painchance
     SFX_NONE,                  // painsound
     S_NULL,                    // meleestate
     S_NULL,                    // missilestate
     S_NULL,                    // crashstate
     S_BAT_DEATH,               // deathstate
     S_NULL,                    // xdeathstate
     SFX_NONE,                  // deathsound
     5 * FRACUNIT,              // speed
     3 * FRACUNIT,              // radius
     3 * FRACUNIT,              // height
     100,                       // mass
     0,                         // damage
     SFX_NONE,                  // activesound
     MF_NOBLOCKMAP | MF_NOGRAVITY | MF_MISSILE, // flags
     MF2_PASSMOBJ | MF2_NOTELEPORT      // flags2
     }
};
